/*
 *  linux/arch/ba/kernel/process.c
 *
 *  BA version
 *    author(s): Matjaz Breskvar (phoenix@beyondsemi.com)
 *
 *  For more information about BA processors, licensing and
 *  design services you may contact Beyond Semiconductor at
 *  sales@beyondsemi.com or visit website http://www.beyondsemi.com.
 *
 *  derived from cris, i386, m68k, ppc, sh ports.
 *
 *  changes:
 *  18. 11. 2003: Matjaz Breskvar (phoenix@beyondsemi.com)
 *    initial port to BA architecture
 *
 * This file handles the architecture-dependent parts of process handling..
 * Based on m86k.
 */

#define __KERNEL_SYSCALLS__
#include <stdarg.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/user.h>
#include <linux/elfcore.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/init_task.h>
#include <linux/mqueue.h>
#include <linux/reboot.h>

#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/processor.h>
#include <asm/spr_defs.h>
#include <asm/aeonR2-hf.h>
#include <asm/asm-offsets.h>
#include <asm/abi.h>

#include <linux/smp.h>

/*
 * Initial task structure. Make this a per-architecture thing,
 * because different architectures tend to have different
 * alignment requirements and potentially different initial
 * setup.
 */

static struct fs_struct init_fs = INIT_FS;
//static struct files_struct init_files = INIT_FILES;
static struct signal_struct init_signals = INIT_SIGNALS(init_signals);
static struct sighand_struct init_sighand = INIT_SIGHAND(init_sighand);
struct mm_struct init_mm = INIT_MM(init_mm);

EXPORT_SYMBOL(init_mm);

/*
 * Power off function, if any
 */
void (*pm_power_off)(void);
EXPORT_SYMBOL(pm_power_off);

/*
 * Initial thread structure.
 *
 * We need to make sure that this is 8192-byte aligned due to the
 * way process stacks are handled. This is done by having a special
 * "init_task" linker map entry..
 */
union thread_union init_thread_union 
	__attribute__((__section__(".data.init_task"))) =
		{ INIT_THREAD_INFO(init_task) };

/*
 * Pointer to Current thread info structure.
 *
 * Used at user space -> kernel transitions. 
 */
struct thread_info *current_thread_info_set[NR_CPUS] = {&init_thread_info, };

/*
 * Initial task structure.
 *
 * All other task structs will be allocated on slabs in fork.c
 */
struct task_struct init_task = INIT_TASK(init_task);

EXPORT_SYMBOL(init_task);

#if 0 // not used for NOW
/*
 * The hlt_counter, disable_hlt and enable_hlt is just here as a hook if
 * there would ever be a halt sequence (for power save when idle) with
 * some largish delay when halting or resuming *and* a driver that can't
 * afford that delay.  The hlt_counter would then be checked before
 * executing the halt sequence, and the driver marks the unhaltable
 * region by enable_hlt/disable_hlt.
 */

static int hlt_counter=0;

void disable_hlt(void)
{
	hlt_counter++;
}

EXPORT_SYMBOL(disable_hlt);

void enable_hlt(void)
{
	hlt_counter--;
}

EXPORT_SYMBOL(enable_hlt);

#endif

void machine_restart(char *command)
{
	printk("*** MACHINE RESTART ***\n");
	__asm__("b.nop 0x0");
}

EXPORT_SYMBOL(machine_restart);

/*
 * Similar to machine_power_off, but don't shut off power.  Add code
 * here to freeze the system for e.g. post-mortem debug purpose when
 * possible.  This halt has nothing to do with the idle halt.
 */

void machine_halt(void)
{
	printk("*** MACHINE HALT ***\n");
	__asm__("b.nop 0x0");
}

EXPORT_SYMBOL(machine_halt);

/* If or when software power-off is implemented, add code here.  */

void machine_power_off(void)
{
	printk("*** MACHINE POWER OFF ***\n");
	__asm__("b.nop 0x0");
}

EXPORT_SYMBOL(machine_power_off);

/*
 * When a process does an "exec", machine state like FPU and debug
 * registers need to be reset.  This is a hook function for that.
 * Currently we don't have any such state to reset, so this is empty.
 */

void flush_thread(void)
{
}

#ifdef CONFIG_AEONR2
// flush the whold d-cache lines and sync write buffer
static void thread_memory_consistent(void)
{
	unsigned long flags;
	unsigned long addr = 0;
	unsigned long size = 0;

	local_irq_save(flags);

#ifdef CONFIG_AEONR2_DCACHE_ENABLED
	asm volatile(
		"b.ori		%0,r0,0x0\n\t"
		"b.ori		%1,r0,%2\n\t"
		"99:\n\t"
		"b.flush_line	0(%0),0\n\t"
		"b.flush_line	0(%0),1\n\t"
		"b.flush_line	0(%0),2\n\t"
		"b.flush_line	0(%0),3\n\t"
		"b.addi		%0,%0,%3\n\t"
		"b.sfleu	%0,%1\n\t"
		"b.bf		99b\n\t"

		"b.syncwritebuffer\n\t"

		: "+r"(addr), "+r"(size)
		: "i"(CONFIG_AEONR2_DC_SIZE/CONFIG_AEONR2_DC_WAY), "i"(CONFIG_AEONR2_DC_LINE)
	);
#endif

#ifdef CONFIG_AEONR2_ICACHE_ENABLED
	asm volatile(
		// invalidate all IC lines
		"b.ori		%0,r0,0x0\n\t"
		"b.ori		%1,r0,%2\n\t"
		"99:\n\t"
		"b.mtspr	r0,%0,%5\n\t"
		"b.addi		%0,%0,%3\n\t"
		"b.sfleu	%0,%1\n\t"
		"b.bf		99b\n\t"
		: "+r"(addr), "+r"(size)
		: "i"(CONFIG_AEONR2_IC_SIZE), "i"(CONFIG_AEONR2_IC_LINE),
		  "i"(SPR_SR), "i"(SPR_ICBIR), "i"(SPR_SR_ICE)
	);
#endif

	local_irq_restore(flags);
}
#endif // CONFIG_AEONR2

void show_regs(struct pt_regs *regs)
{
	extern void show_registers(struct pt_regs *regs);

	/* __PHX__ cleanup this mess */
	show_registers(regs);
}

static void kernel_thread_helper(long dummy1, long dummy2, int (*fn)(void *),
                                 void *arg)
{
	do_exit(fn(arg));
}

/*
 * Create a kernel thread
 */
int kernel_thread(int (*fn)(void *), void * arg, unsigned long flags)
{
	struct pt_regs regs;

	memset(&regs, 0, sizeof(regs));

	regs.gprs[3] = (unsigned long) fn;
	regs.gprs[4] = (unsigned long) arg;

	regs.pc = (unsigned long) kernel_thread_helper;
	regs.sr = KERNEL_SR | SPR_SR_TEE | SPR_SR_IEE;

	/* Ok, create the new process.. */
	return do_fork(flags | CLONE_VM | CLONE_UNTRACED, 0, &regs, 0, NULL, NULL);
}
EXPORT_SYMBOL(kernel_thread);


asmlinkage int sys_fork(int r3, int r4, int r5, int r6, int r7, struct pt_regs *regs)
{
	return do_fork(SIGCHLD, regs->sp, regs, 0, NULL, NULL);
}

asmlinkage int sys_clone(int r3, int r4, int r5, int r6, int r7, struct pt_regs *regs)
{
	unsigned long clone_flags = (unsigned long)r3;
	return do_fork(clone_flags, regs->sp, regs, 0, NULL, NULL);
}

asmlinkage int sys_vfork(int r3, int r4, int r5, int r6, int r7, struct pt_regs *regs)
{
	return do_fork(CLONE_VFORK | CLONE_VM | SIGCHLD, regs->sp, regs, 0, NULL, NULL);
}

unsigned long thread_saved_pc(struct task_struct *t)
{
	return (unsigned long)user_regs(task_thread_info(t))->pc;
}

void release_thread(struct task_struct *dead_task)
{
}

int copy_thread(int nr, unsigned long clone_flags, unsigned long usp,
	unsigned long unused, struct task_struct *p, struct pt_regs *regs)
{
        struct pt_regs *childregs, *kregs;
        extern void ret_from_fork(void);
	unsigned long sp = (unsigned long)task_thread_info(p) + THREAD_SIZE;

#ifdef CONFIG_AEONR2_PROC_TLB_STAT
	p->thread_info->k_itlb_miss = 0;
	p->thread_info->k_dtlb_miss = 0;
	p->thread_info->u_itlb_miss = 0;
	p->thread_info->u_dtlb_miss = 0;
#endif
	
	p->set_child_tid = p->clear_child_tid = NULL;

        /* Copy registers */
	sp -= INT_FRAME_SIZE;
        childregs = (struct pt_regs *) sp;

        *childregs = *regs;
        
	if ((childregs->sr & SPR_SR_SM) == 1) {
                /* for kernel thread, set `current' and stackptr in new task */
                childregs->sp = sp + INT_FRAME_SIZE;
                childregs->gprs[27] = (unsigned long)task_thread_info(p);
        } else {
		childregs->sp = usp;
	}

	/* Return value here should be in sycall return value register. Right
	 * now this is the same as function return value. If you try and change this
	 * the arch/ba/kernel/signal.c code will get broken in sublte way.
	 */
	childregs->gprs[REG_RETVAL_PT] = 0;  /* Result from fork() */

        /*
         * The way this works is that at some point in the future
         * some task will call _switch to switch to the new task.
         * That will pop off the stack frame created below and start
         * the new task running at ret_from_fork.  The new task will
         * do some house keeping and then return from the fork or clone
         * system call, using the stack frame created above.
         */
	sp -= INT_FRAME_SIZE;
        kregs = (struct pt_regs *) sp;

#ifdef CONFIG_AEONR2_RTLSIM
	/*
	 * the __switch will eventualy load registers from this structure
	 * and in case we don't clear it some undefined values will be loaded.
	 * this is not really a problem (since these should never be used)
	 * unless you are running RTL simulation where  XXXX_XXXX will
	 * kill everything.
	 */
	memset(kregs, 0, sizeof(*kregs));
#endif /* CONFIG_AEONR2_RTLSIM */

        task_thread_info(p)->ksp = sp;

	kregs->sr = (regs->sr & ~(SPR_SR_TEE | SPR_SR_IEE)) | SPR_SR_SM;
	kregs->sp = sp + INT_FRAME_SIZE;
	kregs->gprs[27] = (unsigned long)task_thread_info(p); /* current           */
	kregs->pc = (unsigned long)ret_from_fork;
	return 0;
}

/*
 * Set up a thread for executing a new program
 */
void start_thread(struct pt_regs *regs, unsigned long pc, unsigned long sp)
{
	phx_warn("NIP: %lx, SP: %lx", pc, sp);

#ifdef CONFIG_AEONR2
	//thread_memory_consistent();
#endif

	set_fs(USER_DS);
	memset(regs->gprs, 0, sizeof(regs->gprs));

	regs->pc = pc;
	regs->sr = regs->sr & ~SPR_SR_SM;
	regs->sp = sp;
}

/* Fill in the fpu structure for a core dump.  */
int dump_fpu(struct pt_regs *regs, elf_fpregset_t *fpu)
{
	phx_warn("FPU :: TODO");
	return 0;
} 

struct task_struct *_switch_to(struct task_struct *old, struct task_struct *new)
{
	extern struct task_struct *_switch(struct thread_info *old_ti,
					   struct thread_info *new_ti);
	struct thread_info *new_ti, *old_ti;
	long flags;
	struct task_struct *last;

	local_irq_save(flags);
	check_stack(NULL, __FILE__, __FUNCTION__, __LINE__);

	/* current_set is an array of saved current pointers
	 * (one for each cpu). we need them at user->kernel transition,
	 * while we save them at kernel->user transition
	 */
	new_ti = task_thread_info(new);
	old_ti = task_thread_info(old);

	current_thread_info_set[smp_processor_id()] = new_ti;
	last = _switch(old_ti, new_ti);

	check_stack(NULL, __FILE__, __FUNCTION__, __LINE__);
	local_irq_restore(flags);

	return last;
}

/*
 * fill in the user structure for a core dump..
 */
static void dump_thread(struct pt_regs *regs, struct user *dump)
{
	phx_warn("TODO");
}

/*
 * sys_execve() executes a new program.
 */
asmlinkage int sys_execve(char *name, char **argv, char **envp,
			  int r6, int r7, struct pt_regs *regs)
{
	int error;
	char * filename;

	filename = getname(name);
	error = PTR_ERR(filename);

	if (IS_ERR(filename))
	  goto out;

	error = do_execve(filename, argv, envp, regs);
	putname(filename);
 
out:	
	return error;
}

unsigned long get_wchan(struct task_struct *p)
{
	phx_warn("TODO");

	return 0;
}
