/*
 *  linux/arch/ba/kernel/traps.c
 *
 *  BA version
 *    author(s): Matjaz Breskvar (phoenix@beyondsemi.com)
 *
 *  For more information about BA processors, licensing and
 *  design services you may contact Beyond Semiconductor at
 *  sales@beyondsemi.com or visit website http://www.beyondsemi.com.
 *
 *  derived from cris, i386, m68k, ppc, sh ports.
 *
 *  changes:
 *  18. 11. 2003: Matjaz Breskvar (phoenix@beyondsemi.com)
 *    initial port to BA architecture
 *
 *  based on: linux/arch/cris/kernel/traps.c
 *  Copyright (C) 2000,2001 Axis Communications AB
 *
 *  Authors:   Bjorn Wesen
 *  	       Hans-Peter Nilsson
 *
 *  Here we handle the break vectors not used by the system call 
 *  mechanism, as well as some general stack/register dumping 
 *  things.
 *
 */


#include <linux/init.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/ptrace.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/kallsyms.h>
#include <asm/uaccess.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/asm-offsets.h>
#include <asm/abi.h>
#include <asm/aeonR2-hf.h>

extern char _etext, _stext;

static int kstack_depth_to_print = 0x180;

static inline int valid_stack_ptr(struct thread_info *tinfo, void *p)
{
	return	p > (void *)tinfo &&
		p < (void *)tinfo + THREAD_SIZE - 3;
}

static void show_trace(struct task_struct *task, unsigned long * stack)
{
	struct thread_info *context;
	unsigned long addr;
#ifdef	CONFIG_FRAME_POINTER
	unsigned long fp;
#endif

	context = (struct thread_info *)
		((unsigned long)stack & (~(THREAD_SIZE - 1)));

#ifdef	CONFIG_FRAME_POINTER
	if (!task || task == current)
		/* Grab ebp right from our regs */
		asm ("b.ori %0," REG_FP_STR ",0" : "=r" (fp) : );
	else
		fp = ((struct pt_regs *)task->thread_info->ksp)->gprs[REG_FP_PT];


	while (valid_stack_ptr(context, (void *)fp)) {
		addr = *(unsigned long *)(fp - 4);
		print_ip_sym(addr);
		if(fp > *(unsigned long *)(fp - 8)) {
			printk("Next frame inner to this one -- Currupt stack?\n");
			break;
		}
		fp = *(unsigned long *)(fp - 8);
	}
#else
	while (valid_stack_ptr(context, stack)) {
		addr = *stack++;
		if (__kernel_text_address(addr))
			print_ip_sym(addr);
	}
#endif
	printk(" =======================\n");
}

/* displays a short stack trace */
void show_stack(struct task_struct *task, unsigned long *esp)
{
	unsigned long addr, *stack;
	int i;

	// debugging aid: "show_stack(NULL);" prints the
	// back trace for this cpu.

	if(esp==NULL)
		esp=(unsigned long*)&esp;

	/* Also dump the redzone */
	stack = (unsigned long *)((unsigned long)esp - 128);

	if(PAGE_ALIGN((unsigned long)stack) != PAGE_ALIGN((unsigned long)esp))
		stack = (unsigned long *)PAGE_ALIGN((unsigned long)esp);

	printk("Stack dump [0x%08lx]:\n", (unsigned long)esp);
	for(i = -128 / 4; i < kstack_depth_to_print; i++) {
		if (kstack_end(stack))
			break;
		if (__get_user (addr, stack)) {
			/* This message matches "failing address" marked
			   s390 in ksymoops, so lines containing it will
			   not be filtered out by ksymoops.  */
			printk ("Failing address 0x%lx\n", (unsigned long)stack);
			break;
		}
		if(i < 0)
			printk("<%p> sp - %02d: 0x%08lx\n", stack, -i*4, addr);
		else
			printk("<%p> sp + %02d: 0x%08lx\n", stack, i*4, addr);

		stack++;
	}
	printk("\n");

	show_trace(task, esp);

	return;
}

static void show_trace_task(struct task_struct *tsk)
{
	/* 
	 * __PHX__: TODO: SysRq-T trace dump... 
	 */
}

/*
 * The architecture-independent backtrace generator
 */
void dump_stack(void)
{
	unsigned long stack;

	show_stack(current, &stack);
}

#ifdef CONFIG_AEONR2_EXCEPTION_TRACE
static void show_exception_trace(struct thread_info *ti)
{
	print_except_trace_atomic("Thread", __pa(ti->except_trace),
	                          ti->except_trace_pos);
}
#endif

void show_registers(struct pt_regs *regs)
{
	int i;
	int in_kernel = 1;
	unsigned long esp;

	esp = (unsigned long)regs->sp;
	if (user_mode(regs))
		in_kernel = 0;

	printk("CPU #: %d\n"
	       "   PC: %08lx    SR: %08lx    SP: %08lx\n",
	       smp_processor_id(), regs->pc, regs->sr, regs->sp);
	printk("GPR00: %08lx GPR01: %08lx GPR02: %08lx GPR03: %08lx\n",
	       0L, regs->sp, regs->gprs[0], regs->gprs[1]);
	printk("GPR04: %08lx GPR05: %08lx GPR06: %08lx GPR07: %08lx\n",
	       regs->gprs[2], regs->gprs[3], regs->gprs[4], regs->gprs[5]);
	printk("GPR08: %08lx GPR09: %08lx GPR10: %08lx GPR11: %08lx\n",
	       regs->gprs[6], regs->gprs[7], regs->gprs[8], regs->gprs[9]);
	printk("GPR12: %08lx GPR13: %08lx GPR14: %08lx GPR15: %08lx\n",
	       regs->gprs[10], regs->gprs[11], regs->gprs[12], regs->gprs[13]);
	printk("GPR16: %08lx GPR17: %08lx GPR18: %08lx GPR19: %08lx\n",
	       regs->gprs[14], regs->gprs[15], regs->gprs[16], regs->gprs[17]);
	printk("GPR20: %08lx GPR21: %08lx GPR22: %08lx GPR23: %08lx\n",
	       regs->gprs[18], regs->gprs[19], regs->gprs[20], regs->gprs[21]);
	printk("GPR24: %08lx GPR25: %08lx GPR26: %08lx GPR27: %08lx\n",
	       regs->gprs[22], regs->gprs[23], regs->gprs[24], regs->gprs[25]);
	printk("GPR28: %08lx GPR29: %08lx GPR30: %08lx GPR31: %08lx\n",
	       regs->gprs[26], regs->gprs[27], regs->gprs[28], regs->gprs[29]);
	printk("  RES: %08lx oGPR3: %08lx syscallno: %08lx\n",
	       regs->result, regs->orig_gpr3, regs->syscallno);

	printk("Process %s (pid: %d, stackpage=%08lx)\n",
	       current->comm, current->pid, (unsigned long)current);
	/*
	 * When in-kernel, we also print out the stack and code at the
	 * time of the fault..
	 */
	if (in_kernel) {

		printk("\nStack: ");
		show_stack(((struct thread_info *)(regs->sp & PAGE_MASK))->task, (unsigned long*)esp);

		printk("\nCode: ");
		if(regs->pc < PAGE_OFFSET)
			goto bad;

		for(i=-24;i<24;i++)
		{
			unsigned char c;
			if(__get_user(c, &((unsigned char*)regs->pc)[i])) {
bad:
				printk(" Bad PC value.");
				break;
			}

			if (i == 0)
			  printk("(%02x) ", c);
			else
			  printk("%02x ", c);
 		}
	}
	printk("\n");
}	

void nommu_dump_state(struct pt_regs *regs, 
		      unsigned long ea, unsigned long vector)
{
	int i;
	unsigned long addr, stack = regs->sp;

	printk("\n\r[nommu_dump_state] :: ea %lx, vector %lx\n\r",
	       ea, vector);

	printk("CPU #: %d\n"
	       "   PC: %08lx    SR: %08lx    SP: %08lx\n",
	       0, regs->pc, regs->sr, regs->sp);
	printk("GPR00: %08lx GPR01: %08lx GPR02: %08lx GPR03: %08lx\n",
	       0L, regs->sp, regs->gprs[0], regs->gprs[1]);
	printk("GPR04: %08lx GPR05: %08lx GPR06: %08lx GPR07: %08lx\n",
	       regs->gprs[2], regs->gprs[3], regs->gprs[4], regs->gprs[5]);
	printk("GPR08: %08lx GPR09: %08lx GPR10: %08lx GPR11: %08lx\n",
	       regs->gprs[6], regs->gprs[7], regs->gprs[8], regs->gprs[9]);
	printk("GPR12: %08lx GPR13: %08lx GPR14: %08lx GPR15: %08lx\n",
	       regs->gprs[10], regs->gprs[11], regs->gprs[12], regs->gprs[13]);
	printk("GPR16: %08lx GPR17: %08lx GPR18: %08lx GPR19: %08lx\n",
	       regs->gprs[14], regs->gprs[15], regs->gprs[16], regs->gprs[17]);
	printk("GPR20: %08lx GPR21: %08lx GPR22: %08lx GPR23: %08lx\n",
	       regs->gprs[18], regs->gprs[19], regs->gprs[20], regs->gprs[21]);
	printk("GPR24: %08lx GPR25: %08lx GPR26: %08lx GPR27: %08lx\n",
	       regs->gprs[22], regs->gprs[23], regs->gprs[24], regs->gprs[25]);
	printk("GPR28: %08lx GPR29: %08lx GPR30: %08lx GPR31: %08lx\n",
	       regs->gprs[26], regs->gprs[27], regs->gprs[28], regs->gprs[29]);
	printk("  RES: %08lx oGPR3: %08lx syscallno: %08lx\n",
	       regs->result, regs->orig_gpr3, regs->syscallno);

	printk("Process %s (pid: %d, stackpage=%08lx)\n",
	       ((struct task_struct*)(__pa(current)))->comm, 
	       ((struct task_struct*)(__pa(current)))->pid,
	       (unsigned long)current);
	
	printk("\nStack: ");
	printk("Stack dump [0x%08lx]:\n", (unsigned long)stack);
	for(i = 0; i < kstack_depth_to_print; i++) {
		if (((long) stack & (THREAD_SIZE-1)) == 0)
			break;
		stack++;

		printk("%lx :: sp + %02d: 0x%08lx\n", stack, i*4, 
		       *((unsigned long*)(__pa(stack))));
	}
	printk("\n");

	printk("Call Trace:   ");
	i = 1;
	while (((long) stack & (THREAD_SIZE-1)) != 0) {
		addr = *((unsigned long*)__pa(stack));
		stack++;

		if (kernel_text_address(addr)) {
			if (i && ((i % 6) == 0))
				printk("\n ");
			printk(" [<%08lx>]", addr);
			i++;
		}
	}
	printk("\n");
	
	printk("\nCode: ");

	for(i=-24;i<24;i++)
	{
		unsigned char c;
		c = ((unsigned char*)(__pa(regs->pc)))[i];
		
		if (i == 0)
			printk("(%02x) ", c);
		else
			printk("%02x ", c);
	}
	printk("\n");
}


/* This is normally the 'Oops' routine */
void die(const char * str, struct pt_regs * regs, long err)
{
	/* disable tracebuffer as soon as possible 
	 * in case it was not disabled already
	 */

	console_verbose();

	printk("\n%s#: %04lx\n", str, err & 0xffff);

	if (regs) {
#ifdef CONFIG_AEONR2_EXCEPTION_TRACE
		show_exception_trace((struct thread_info *)(regs->sp & PAGE_MASK));
#endif
		show_registers(regs);
	} else {
		printk("regs == NULL, not dumping state\n");
	}

#ifdef CONFIG_JUMP_UPON_UNHANDLED_EXCEPTION
	printk("\n\nUNHANDLED_EXCEPTION: entering infinite loop\n");
	
	/* shut down interrupts */
	local_irq_disable();
	
	__asm__ __volatile__(
		"b.nop 0x0");
	for (;;)
		;
#endif
	do_exit(SIGSEGV);
}

/* This is normally the 'Oops' routine */
void die_if_kernel(const char * str, struct pt_regs * regs, long err)
{
	if(user_mode(regs))
		return;

	die(str, regs, err);
}

void unhandled_exception(struct pt_regs *regs, int ea, int vector)
{
	printk("Unable to handle exception at EA =0x%x, vector 0x%x",
	       	ea, vector);
	die("Oops", regs, 9);
}

void __init trap_init(void)
{
	/* Nothing needs to be done */
}

/* ------------------ functions running without MMU ----------------- */

#ifdef CONFIG_AEONR2_EXCEPTION_TRACE
struct except_stack except_stack[CONFIG_AEONR2_EXCEPTION_TRACE_DEPTH];
unsigned long except_stack_indx = 0;
#endif

extern void emergency_print(const char *str);
extern void emergency_print_nr(unsigned long x);
extern void emergency_print_nr_full(unsigned long x);
extern void print_tracebuffer(void);

#ifdef CONFIG_AEONR2_EXCEPTION_TRACE
void print_except_trace(const char *which, struct except_stack *es,
                        unsigned long pos)
{
	int i;

	safe_print("\n\n\r%x %x %x %x %x %x %x %x\n\n\r", 1, 2, 3, 4, 5, 6, 7, 8);

	emergency_print("\n\r\n");
	emergency_print(which);
	emergency_print(" exception stack (most recent at the top):\n\r");
	for(i = pos; i >= 0; i--) {
		if(es[i].vector == (unsigned long)-1) {
			emergency_print("Switched threads to 0x");
			emergency_print_nr_full(es[i].epcr);
			emergency_print("\n\r");
			continue;
		}
		emergency_print("Exception vector: 0x");
		emergency_print_nr(es[i].vector);
		emergency_print(" EPCR: 0x");
		emergency_print_nr_full(es[i].epcr);
		emergency_print(", ESR: 0x");
		emergency_print_nr_full(es[i].esr);
		emergency_print(", EEAR: 0x");
		emergency_print_nr_full(es[i].eear);
		emergency_print("\n\r");
	}
	for(i = CONFIG_AEONR2_EXCEPTION_TRACE_DEPTH - 1; i > pos; i--) {
		if(es[i].vector == (unsigned long)-1) {
			emergency_print("Switched threads to 0x");
			emergency_print_nr_full(es[i].epcr);
			emergency_print("\n\r");
			continue;
		}
		emergency_print("Exception vector: 0x");
		emergency_print_nr(es[i].vector);
		emergency_print(" EPCR: 0x");
		emergency_print_nr_full(es[i].epcr);
		emergency_print(", ESR: 0x");
		emergency_print_nr_full(es[i].esr);
		emergency_print(", EEAR: 0x");
		emergency_print_nr_full(es[i].eear);
		emergency_print("\n\r");
	}
}
#endif

#define NUM_TLB_ENTRIES 64

int unhandled_exception_in_progress = 0;

void unhandled_exception_handler(struct pt_regs *regs, 
				 unsigned long ea, unsigned long epcr,
				 unsigned long except)
{
	unsigned long *sp;
	/* emergency_print() is function implemented in head.S
	 * and it will transform it's first argument address into
	 * physical address if it is in the 0xcXXX_XXXX range
	 */

	int i, n;
	int *ueip;

	/* make sure that we don't get continuous exceptions as
	 * usually only the first exception is relevant
	 */
	ueip = &unhandled_exception_in_progress;
	if (*((int *)__pa(ueip))) {
		emergency_print("\n\n\runhandled_exception_handler()"
				" triggered unhandled exception\n\r");
		
		goto freeze;
	}

	/* mark that we entered here */
	ueip = &unhandled_exception_in_progress;
	*((int *)__pa(ueip)) = 1;

	emergency_print("Unable to handle exception at EA =0x");
	emergency_print_nr_full(ea);
	emergency_print(", vector 0x");
	emergency_print_nr_full(except);
	emergency_print("\n\rOops#: 0009");
	emergency_print("\n\rCPU #: 0\n\r");
	
	emergency_print("   PC: 0x");
	emergency_print_nr_full(epcr);
	emergency_print(",   SR: 0x");
	emergency_print_nr_full(regs->sr);
	emergency_print("    SP: 0x");
	emergency_print_nr_full(regs->sp);
	
	emergency_print("\n\rGPR00: 0x");
	emergency_print_nr_full(0);
	emergency_print(" GPR01: 0x");
	emergency_print_nr_full(regs->sp);
	emergency_print(" GPR02: 0x");
	emergency_print_nr_full(regs->gprs[0]);
	emergency_print(" GPR03: 0x");
	emergency_print_nr_full(regs->gprs[1]);
	
	emergency_print("\n\rGPR04: 0x");
	emergency_print_nr_full(regs->gprs[2]);
	emergency_print(" GPR05: 0x");
	emergency_print_nr_full(regs->gprs[3]);
	emergency_print(" GPR06: 0x");
	emergency_print_nr_full(regs->gprs[4]);
	emergency_print(" GPR07: 0x");
	emergency_print_nr_full(regs->gprs[5]);
	
	emergency_print("\n\rGPR08: 0x");
	emergency_print_nr_full(regs->gprs[6]);
	emergency_print(" GPR09: 0x");
	emergency_print_nr_full(regs->gprs[7]);
	emergency_print(" GPR10: 0x");
	emergency_print_nr_full(regs->gprs[8]);
	emergency_print(" GPR11: 0x");
	emergency_print_nr_full(regs->gprs[9]);
	
	emergency_print("\n\rGPR12: 0x");
	emergency_print_nr_full(regs->gprs[10]);
	emergency_print(" GPR13: 0x");
	emergency_print_nr_full(regs->gprs[11]);
	emergency_print(" GPR14: 0x");
	emergency_print_nr_full(regs->gprs[12]);
	emergency_print(" GPR15: 0x");
	emergency_print_nr_full(regs->gprs[13]);
	
	emergency_print("\n\rGPR16: 0x");
	emergency_print_nr_full(regs->gprs[14]);
	emergency_print(" GPR17: 0x");
	emergency_print_nr_full(regs->gprs[15]);
	emergency_print(" GPR18: 0x");
	emergency_print_nr_full(regs->gprs[16]);
	emergency_print(" GPR19: 0x");
	emergency_print_nr_full(regs->gprs[17]);
	
	emergency_print("\n\rGPR20: 0x");
	emergency_print_nr_full(regs->gprs[18]);
	emergency_print(" GPR21: 0x");
	emergency_print_nr_full(regs->gprs[19]);
	emergency_print(" GPR22: 0x");
	emergency_print_nr_full(regs->gprs[20]);
	emergency_print(" GPR23: 0x");
	emergency_print_nr_full(regs->gprs[21]);
	
	emergency_print("\n\rGPR24: 0x");
	emergency_print_nr_full(regs->gprs[22]);
	emergency_print(" GPR25: 0x");
	emergency_print_nr_full(regs->gprs[23]);
	emergency_print(" GPR26: 0x");
	emergency_print_nr_full(regs->gprs[24]);
	emergency_print(" GPR27: 0x");
	emergency_print_nr_full(regs->gprs[25]);
	
	emergency_print("\n\rGPR28: 0x");
	emergency_print_nr_full(regs->gprs[26]);
	emergency_print(" GPR29: 0x");
	emergency_print_nr_full(regs->gprs[27]);
	emergency_print(" GPR30: 0x");
	emergency_print_nr_full(regs->gprs[28]);
	emergency_print(" GPR31: 0x");
	emergency_print_nr_full(regs->gprs[29]);

	//print_tracebuffer();


#ifdef CONFIG_AEONR2_EXCEPTION_TRACE
	print_except_trace("Global", (struct except_stack *)__pa(except_stack),
	                   *(unsigned long *)__pa(&except_stack_indx));
	print_except_trace("Thread",
	                   ((struct thread_info *)__pa(current_thread_info()))->except_trace,
	                   ((struct thread_info *)__pa(current_thread_info()))->except_trace_pos);
#endif
	
	emergency_print("\n\n\rDumping iTLB entries\n\r");
	
	for (i = 0; i < NUM_TLB_ENTRIES ; i++) {
		emergency_print("\r\n0x");
		emergency_print_nr(i);
		emergency_print("\tiTLBmr 0x");
		emergency_print_nr_full(mfspr(SPR_ITLBMR_BASE(0)+i));
		emergency_print(" :: iTLBtr 0x");
		emergency_print_nr_full(mfspr(SPR_ITLBTR_BASE(0)+i));
		emergency_print("\t0x");
		emergency_print_nr(i);
	}
	
	emergency_print("\n\n\rDumping dTLB entries\n\r");
	
	for (i = 0; i < NUM_TLB_ENTRIES ; i++) {
		emergency_print("\r\n0x");
		emergency_print_nr(i);
		emergency_print("\tdTLBmr 0x");
		emergency_print_nr_full(mfspr(SPR_DTLBMR_BASE(0)+i));
		emergency_print(" :: dTLBtr 0x");
		emergency_print_nr_full(mfspr(SPR_DTLBTR_BASE(0)+i));
		emergency_print("\t0x");
		emergency_print_nr(i);
	}

	sp = (unsigned long *)(__pa(regs->sp) & ~3ul);

	emergency_print("\n\r\nStack dump [0x");
	emergency_print_nr_full((unsigned long)sp);
	emergency_print("]:\n\r");

	i = 0;
	while((unsigned long)sp & (PAGE_SIZE - 1)) {
		emergency_print("<");
		emergency_print_nr_full((unsigned long)sp);
		emergency_print("> sp + ");
		emergency_print_nr(i);
		emergency_print(": 0x");
		emergency_print_nr_full(*sp);
		emergency_print("\n\r");
		sp++;
		i += 4;
	}
	
	emergency_print("\n\n\rDumping code\n\n\r\t");

	n = 64;
	for(i=0 ; i < 2*n ; i++)
	{
		unsigned char c;
		c = ((unsigned char*)(__pa(regs->pc)))[i-n];

		if (i == n)
			emergency_print("\r    --> ");

		emergency_print_nr(c);
		emergency_print(" ");
				
		if ((i % 16) == 15)
			emergency_print("\n\r\t");
	}
	

freeze:
	emergency_print("\n\r\n\rgoing into infinite loop\n\r");
	
	/* stop the simulator */
	__asm__ __volatile__("b.nop 0x0");
	
	while (1)
		;
}



