#include <stdarg.h>
#include <linux/kernel.h>
#include <asm/system.h>
#include <asm/page.h>

#define PRINTF_MAX_BASE 16
#define PRINTF_MAX_BUFF 32

#define PRINTF_MINUS    PRINTF_MAX_BASE
#define PRINTF_SPACE    (PRINTF_MAX_BASE+1)

/* implemented in head.S, runs with or without MMU's */
extern void emergency_print_ch(char c);

static char numbers_effective[]="0123456789abcdef- ";

static void print_char(char c)
{
	emergency_print_ch(c);
	return;
}

static unsigned long convert_to_physical(unsigned long addr)
{
	/* do not do the conversion in case of running with MMU */
	if (mfspr(SPR_SR) & (SPR_SR_DME|SPR_SR_IME))
		return(addr);

	if (addr > PAGE_OFFSET) {
		addr = addr - PAGE_OFFSET;
	}

	return(addr);
}

void output_number(unsigned int n, int base, int sign_bit, 
		   int padding, int zero_padding)
{
	char *numbers;
	char buff[PRINTF_MAX_BUFF];
	int i = 0;

	numbers = (char *)convert_to_physical((unsigned long)numbers_effective);

	if (base > PRINTF_MAX_BASE)
		return;

	if (sign_bit &&
	    (signed int)(n << ((sizeof(unsigned int)*4) - sign_bit)) < 0) {
		print_char(numbers[PRINTF_MAX_BASE]);
		n = n << ((sizeof(unsigned int)*4) - sign_bit);
		n = -n;
		n = n >> ((sizeof(unsigned int)*4) - sign_bit); 
	}

	do {
		buff[i++] = numbers[n%base];
		n = n / base;
	} while (n);

	if ((signed int)(n << ((sizeof(unsigned int)*4) - sign_bit)) < 0)
		print_char(numbers[PRINTF_MINUS]);

	while (padding-- > i) {
		if (zero_padding)
			print_char(numbers[0]);
		else
			print_char(numbers[PRINTF_SPACE]);
	}

	while (i--)
		print_char(buff[i]);
}

/* in case parameters need to be passed on stack this
 * function is not MMU safe anymore
 */
int safe_print(const char *fmt, ...)
{  
	int zero_padding, padding;
	int args_count = 0;
	int i = 0;
	int n;
	va_list va;

	fmt = (char *)convert_to_physical((unsigned long)fmt);

	va_start(va, fmt);

	while (1) {

		while (fmt[i] != '%' && fmt[i] != '\0')
			print_char(fmt[i++]);
		
		
		if (fmt[i] == '\0')
			goto done;

		/* just trap in case of too many arguments */
		if (++args_count >= 5)
			asm volatile ("bt.trap 0");
		
		i++; /* next character is conversion specifier */
		
		padding = 0;
		zero_padding = 0;
		/* handle (zero) padding */
		if ((fmt[i] >= '0') && (fmt[i] <= '9')) {
			if (fmt[i] == '0') {
				zero_padding = 1;
				i++;
			}
			
			padding = fmt[i++] - '0';
		}

		switch (fmt[i++]) {
			
		case 'i':
		case 'd':
			n = va_arg(va, int);
			output_number(n, 10, 32, padding, zero_padding);
			break;

		case 'u':
			n = va_arg(va, unsigned int);
			output_number(n, 10, 0, padding, zero_padding);
			break;
			
		case 'x':
			n = va_arg(va, unsigned int);
			output_number(n, 16, 0, padding, zero_padding);
			break;
		}

	}
	
done:
	va_end(va);
	return(i);
}
