/*
 *  linux/arch/ba/mm/consistent.c
 *
 *  BA version
 *    author(s): Gyorgy Jeney (nog@beyondsemi.com)
 *
 *  For more information about BA processors, licensing and
 *  design services you may contact Beyond Semiconductor at
 *  sales@beyondsemi.com or visit website http://www.beyondsemi.com.
 *    
 *  changes:
 *  23. 04. 2007: Gyorgy Jeney (nog@beyondsemi.com)
 *    Add dcache physical address range invalidating function
 *  27. 04. 2007: Gyorgy Jeney (nog@beyondsemi.com)
 *    Add coherent DMA allocation routines
 *
 */

#include <linux/types.h>
#include <linux/dma-mapping.h>
#include <asm/tlbflush.h>
#include <asm/cacheflush.h>
#include <asm/spr_defs.h>
#include <linux/vmalloc.h>

/* Invalidates a certain range of the DC */
void inv_dcache_phys_range(unsigned long phys, size_t size)
{
	unsigned long flags_irq;
	unsigned long addr;
	unsigned long addr_end;

	if(!size)
		return;

printk("[KC_DEBUG] inv_dcache_phys_range()\n");

	local_irq_save(flags_irq);

	// cache line size boundary
	phys &= -CONFIG_AEONR2_DC_LINE;

	// cache line size alignment
	size += CONFIG_AEONR2_DC_LINE - 1;
	size &= -CONFIG_AEONR2_DC_LINE;
	if (size > CONFIG_AEONR2_DC_SIZE)
		size = CONFIG_AEONR2_DC_SIZE;

	addr = phys;
	addr_end = addr + size;

	for (; addr < addr_end; addr += CONFIG_AEONR2_DC_LINE) {
		asm volatile(
			"b.invalidate	0(%0)\n\t"
			:
			: "r"(addr)
		);
	}

	local_irq_restore(flags_irq);
}

/* flush a certain range of the DC */
void flush_dcache_phys_range(unsigned long phys, size_t size)
{
	unsigned long flags_irq;
	unsigned long addr;
	unsigned long addr_end;

	if(!size)
		return;

printk("[KC_DEBUG] flush_dcache_phys_range()\n");

	local_irq_save(flags_irq);

	// cache line size boundary
	phys &= -CONFIG_AEONR2_DC_LINE;

	// cache line size alignment
	size += CONFIG_AEONR2_DC_LINE - 1;
	size &= -CONFIG_AEONR2_DC_LINE;
	if (size > CONFIG_AEONR2_DC_SIZE)
		size = CONFIG_AEONR2_DC_SIZE;

	addr = phys;
	addr_end = addr + size;

	for (addr; addr < addr_end; addr += CONFIG_AEONR2_DC_LINE) {
		asm volatile(
			"b.flush	0(%0)\n\t"
			:
			: "r"(addr)
		);
	}

	local_irq_restore(flags_irq);
}

void flush_inv_dcache_phys_range(unsigned long phys, size_t size)
{
	unsigned long flags_irq;
	unsigned long addr;
	unsigned long addr_end;

	if(!size)
		return;

printk("[KC_DEBUG] flush_inv_dcache_phys_range()\n");

	local_irq_save(flags_irq);

	// cache line size boundary
	phys &= -CONFIG_AEONR2_DC_LINE;

	// cache line size alignment
	size += CONFIG_AEONR2_DC_LINE - 1;
	size &= -CONFIG_AEONR2_DC_LINE;
	if (size > CONFIG_AEONR2_DC_SIZE)
		size = CONFIG_AEONR2_DC_SIZE;

	addr = phys;
	addr_end = addr + size;

	for (; addr < addr_end; addr += CONFIG_AEONR2_DC_LINE) {
		asm volatile(
			"b.flush_invalidate	0(%0)\n\t"
			:
			: "r"(addr)
		);
	}

	local_irq_restore(flags_irq);
}

/* Allocate memory that is coherent with the CPU no matter what happens with
 * the DMA (for our case, just set the _CI bit) */
void *dma_alloc_coherent(struct device *dev, size_t size,
                         dma_addr_t *dma_handle, gfp_t flag)
{
	struct page *page;
	unsigned int order = get_order(size);
	unsigned int i;
	struct page **pages;
	void *cpu_addr;

	pages = kmalloc((1 << order) * sizeof(struct page *), GFP_KERNEL);

	if(!pages)
		goto pages_alloc_error;

	page = alloc_pages(flag, order);

	if(!page)
		goto alloc_error;

	split_page(page, order);

	for(i = 0; i < (1 << order); i++)
		pages[i] = page + i;

	cpu_addr = vmap(pages, 1 << order, VM_MAP, PAGE_KERNEL_NOCACHE);

	if(!cpu_addr)
		goto vmap_error;

	kfree(pages);

	*dma_handle = virt_to_phys(page_address(page));
	return cpu_addr;

vmap_error:
	for(i = 0; i < (1 << order); i++)
		__free_page(page + i);

alloc_error:
	kfree(pages);

pages_alloc_error:
	*dma_handle = -1ul;
	return NULL;
}

void dma_free_coherent(struct device *dev, size_t size, void *cpu_addr,
                       dma_addr_t dma_handle)
{
	unsigned int order = get_order(size);
	struct page *page;
	unsigned int i;

	page = vmalloc_to_page(cpu_addr);

	if(!page)
		BUG();

	vunmap(cpu_addr);

	for(i = 0; i < (1 << order); i++)
		__free_page(page + i);
}

