/******************************************************************************
 *-----------------------------------------------------------------------------
 *
 *  Copyright (c) 2008 MStar Semiconductor, Inc.  All rights reserved.
 *
 *-----------------------------------------------------------------------------
 * FILE NAME      columbus_mci.c
 *
 * DESCRIPTION    Columbus LinuxBSP MCI driver.
 *                Implement SD/MMC Host Driver for Columbus FCIE (Flash Card
 *                Interface Engine).
 *
 * AUTHOR         Ray Hsu          Initial Version
 *                      Evan Chang     Update MStar code convention
 *                                            Resolved mounting problem
                         William Wang   Resovled the read/write data inconsistent
                                               modify the timing for high speed CPU
 *
 * HISTORY      2008/05/06     Initial Version
 *                      2008/07/22     Update IRQ assignment & code convention
 *                      2008/07/22     Resolved mounting problem
 *                      2008/08/05     Resovled the read/write data inconsistent
 *                                               modify the timing for high speed CPU
 *                      2008/09/01     modify the check card busy function
 *                      2008/10/02     update the pioneer configuration
 *****************************************************************************/

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/dma-mapping.h>
#include <linux/clk.h>
#include <linux/mmc/host.h>
#include <asm/atomic.h>

#include <linux/list.h>
#include <linux/cdev.h>
#include <asm/uaccess.h>        /* copy_*_user */

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mach/mmc.h>

#include <linux/scatterlist.h>
//#include <mach/irqs.h>
#include "chip_int.h"

#include "mstar_mci.h"
#include "mstar_mci_ioctl.h"

#include <dma-coherence.h>

/******************************************************************************
 * Defines
 ******************************************************************************/

#define DRIVER_NAME "mstar_mci"
#define FL_SENT_COMMAND (1 << 0)
#define FL_SENT_STOP    (1 << 1)
#define LOOP_DELAY_WAIT_COMPLETE   0x1000000

#define fcie_reg_read(host, reg)        __raw_readl((host)->baseaddr + (reg))
#define fcie_reg_write(host, reg, val)  __raw_writel((val), (host)->baseaddr + (reg))

/* For debugging purpose */
#define COLUMBUS_MCI_DEBUG 1
#define FCIE_DYNAMIC_MALLOC 1
#if (COLUMBUS_MCI_DEBUG)
#undef pr_debug
#define pr_debug(fmt, arg...) printk(KERN_INFO fmt, ##arg)
#else
#undef pr_debug
#define pr_debug(fmt, arg...)
#endif
#define simple_debug(fmt, arg...) printk(KERN_INFO fmt, ##arg)
/******************************************************************************
 * Function Prototypes
 ******************************************************************************/

static inline s32 get_data_buffer(struct mstar_mci_host *pHost_st,
                                  volatile u32 *pWords, volatile u32 **pPgAddr);
static inline void do_pio_read(struct mstar_mci_host *pHost_st);
static inline void do_pio_write(struct mstar_mci_host *pHost_st);
static inline void mstar_mci_sg_to_dma(struct mstar_mci_host *pHost_st,
        struct mmc_data *pData_st);
static void mstar_mci_card_power_en(s32 PwrOn, struct mstar_mci_host *pHost_st);
static void mstar_mci_pre_dma_read(struct mstar_mci_host *pHost_st);
static void mstar_mci_post_dma_read(struct mstar_mci_host *pHost_st);
static void mstar_mci_handle_transmitted(struct mstar_mci_host *pHost_st);
static s32 mstar_mci_handle_cmdrdy(struct mstar_mci_host *pHost_st);
static void mstar_mci_reset(struct mstar_mci_host *pHost_st);
static void mstar_mci_switch_sd(struct mstar_mci_host *pHost_st);
static void mstar_mci_enable(struct mstar_mci_host *pHost_st);
static void mstar_mci_disable(struct mstar_mci_host *pHost_st);
static void mstar_mci_send_command(struct mstar_mci_host *pHost_st,
                                      struct mmc_command *pCmd_st);
static void mstar_mci_process_next(struct mstar_mci_host *pHost_st);

#ifdef FCIE_TASKLET /* Defer tasks in IRQ, so IRQ can be finished faster */
static void mstar_mci_completed_command(unsigned long data);
#else   /* Do NOT enable Tasklet */
static void mstar_mci_completed_command(struct mstar_mci_host *pHost_st);
#endif  /* End ifdef FCIE_TASKLET */

static void mstar_mci_request(struct mmc_host *pMMC_st, struct mmc_request *pMRQ_st);
static void mstar_mci_set_ios(struct mmc_host *pMMC_st, struct mmc_ios *pIOS_st);
static irqreturn_t mstar_mci_irq(s32 irq, void *pDevID);
static s32 mstar_mci_get_ro(struct mmc_host *pMMC_st);
static s32 __init mstar_mci_probe(struct platform_device *pDev_st);
static s32 __exit mstar_mci_remove(struct platform_device *pDev_st);
static s32 __init mstar_mci_init(void);
static void __exit mstar_mci_exit(void);

static void mstar_mci_reinit(struct mstar_mci_host *pHost_st);
s32 mstar_mci_clearall_req(struct mmc_host *pMMC_st);

#if defined(CONFIG_MEMSTICK_COLUMBUS_MS)
extern irqreturn_t columbus_ms_interrupt(int irq, void *pDevID);
extern struct memstick_host * pass_fcie_ms_host(void);
#endif
extern unsigned int MDrv_SYS_GetMIU0Length(void);
/******************************************************************************
 * Define Static Global Variables
 ******************************************************************************/

/* Columbus Multimedia Card Interface Operations */
static const struct mmc_host_ops sg_mstar_mci_ops =
{
    .request = mstar_mci_request,
    .set_ios = mstar_mci_set_ios,
    .get_ro = mstar_mci_get_ro,
};

static u64 mmc_dmamask = 0xffffffffUL;

static struct resource sg_mmc_resources_st[] =
{
    [0] =
    {
        .start = FCIE0_BASE,
        .end = FCIE0_BASE + (0x00000400*2) - 1,
        .flags = IORESOURCE_MEM,
    },
    [1] =
    {
        .start = E_IRQ_NFIE,
        .end = E_IRQ_NFIE,
        .flags = IORESOURCE_IRQ,
    },
};

static struct platform_device sg_mstar_mmc_device_st =
{
    .name = "mstar_mci",
    .id = 0,
    .resource = sg_mmc_resources_st,
    .num_resources = ARRAY_SIZE(sg_mmc_resources_st),
    .dev =
    {
        .dma_mask = &mmc_dmamask,
        .coherent_dma_mask = 0xffffffffUL,
    },
};
#if (FCIE_DYNAMIC_MALLOC)
static u32 saved_tx_total_length = 0;
#define sd_txbuf_size (0x200*0xFF)
#endif
static atomic_t is_sd_mmc = ATOMIC_INIT(0);
static int d_initialized = 0;
//static u8 u8SDBuf[512*255] __attribute((aligned(64))); for MIU 0 issue
/******************************************************************************
 * Functions
 ******************************************************************************/
void mstar_mci_disable_stschg(struct mmc_host *pMMC_st)
{
    unsigned long flags;

    local_irq_save(flags);
    *(volatile unsigned short *)(FCIE0_BASE+FCIE_CARD_INT_EN) =
                *(volatile unsigned short *)(FCIE0_BASE+FCIE_CARD_INT_EN) & (~(SDIO2_STS_EN));
    local_irq_restore(flags);

}

void mstar_mci_enable_stschg(struct mmc_host *pMMC_st)
{
    unsigned long flags;

    local_irq_save(flags);
    *(volatile unsigned short *)(FCIE0_BASE+FCIE_CARD_INT_EN) =
                *(volatile unsigned short *)(FCIE0_BASE+FCIE_CARD_INT_EN) | (SDIO2_STS_EN);
    local_irq_restore(flags);
}

void fcie_dma_unmap_page(struct device *dev, dma_addr_t dma_address, size_t size,
	enum dma_data_direction direction)
{
	BUG_ON(direction == DMA_NONE);

	if (!plat_device_is_coherent(dev) && direction != DMA_TO_DEVICE) {
		unsigned long addr;
        //Modify from dma_unmap_page()
        //Benson modify for CPU can't handle physiscal address bug in dma_cache_wback_inv().
        #if 1
        addr = (unsigned long)phys_to_virt(dma_address);
        #else
		addr = plat_dma_addr_to_phys(dma_address);
        #endif
		dma_cache_wback_inv(addr, size);
	}

	plat_unmap_dma_mem(dma_address);
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : checkD0Low
+------------------------------------------------------------------------------
| DESCRIPTION : check the SD data line0
|
| RETURN      : 1 is low; 0 is high
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
static u8 checkD0Low(struct mstar_mci_host *pHost_st)
{
    u8 D0_low = 0;
    if ((fcie_reg_read(pHost_st, FCIE_SDIO_STS) & SDIO2_D0) == 0)
        D0_low = 1;
    return D0_low;
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : waitD0High
+------------------------------------------------------------------------------
| DESCRIPTION : wait the SD data line0
|
| RETURN      : 1 is D0 low means card busy ; 0 is high
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st               | x |      | Columbus MCI Host structure
|--------------------+---+---+-------------------------------------------------
|u32LoopCnt             | x |      | wating time
+------------------------------------------------------------------------------
*/
static u8 waitD0High(struct mstar_mci_host *pHost_st, u32 u32LoopCnt)
{
    u32 u32Cnt=0;

    while (checkD0Low(pHost_st)) /* card busy */
    {
        if ((u32Cnt++) == u32LoopCnt)
        {
            return 1;
        }
        udelay(1);
    }
    return 0;
}

static u8 clear_fcie_status(struct mstar_mci_host *pHost_st, u16 reg, u16 value)
{
   u8 i=0;
   do
   {
       fcie_reg_write(pHost_st, reg, value);
       if (++i == 0)
       {
           printk("Error, clear FCIE status in RIU time out!\n");
           return 1;
       }
   } while(fcie_reg_read(pHost_st, reg)&value);

   return 0;
}

static u8 wait_fifoclk_ready(struct mstar_mci_host *pHost_st)
{
   u16 wait_cnt=0;
   do
   {
       if (++wait_cnt == 0)
       {
           printk("Error, wait FIFO clock ready time out!\n");
           return 1;
       }
   } while((fcie_reg_read(pHost_st, FCIE_MMA_PRI_REG)&FIFO_CLKRDY)==0);

   return 0;
}

static void fcie_reg_dump(struct mstar_mci_host *pHost_st)
{
   u16 i=0;
   u16 reg_v;

   printk("\r\nreg data ===============================================r\n");
   for (i=0;i<0x50;i++)
   {
        reg_v = fcie_reg_read(pHost_st, i*4);
        printk("%04x ", reg_v);
        //if (i%4==3)
        //    printk(" ");
        if (i%8==7)
            printk("||  %X\r\n", i);
   }
   printk("========================================================\r\n");
}

#if 0
static void fcie_dump_dma_data(struct mstar_mci_host *pHost_st, dma_addr_t dma_address, u16 length)
{
    int viraddr;
    u16 u16ByteNo;
    u8 *ptr=NULL;

    __u32 * vaddr;

    ptr = (u8 *)phys_to_virt(dma_address);

    //viraddr = (int)dma_to_virt(&pHost_st->mmc->class_dev, pHost_st->physical_address);
    //ptr = benson_t;//viraddr;

    printk("dma_data ==============================================================\r\n");
    for (u16ByteNo=0; u16ByteNo<length; u16ByteNo++)
    {
        printk("%02x", ptr[u16ByteNo]);
        if (u16ByteNo%4==3)
            printk(" ");
        if (u16ByteNo%32==31)
            printk("||  %X\r\n", u16ByteNo);
    }
    printk("=======================================================================\r\n");

}
#endif


/*
+------------------------------------------------------------------------------
| FUNCTION    : get_data_buffer
+------------------------------------------------------------------------------
| DESCRIPTION : Get RAM buffer address
|
| RETURN      : Error Status; Return 0 if no error.
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
| *pWords            | x |   | Word length
+------------------------------------------------------------------------------
| **pPgAddr          | x |   | Page address
+------------------------------------------------------------------------------
*/
static inline s32 get_data_buffer(struct mstar_mci_host *pHost_st,
                                  volatile u32 *pWords, volatile u32 **pPgAddr)
{
    /* Define Local Variables */
    struct scatterlist *pSG_st = 0;

    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "get_data_buffer parameter pHost_st is NULL\n");
        return -EINVAL;
    }

    if (!pWords)
    {
        printk(KERN_WARNING "get_data_buffer parameter pWords is NULL\n");
        return -EINVAL;
    }

    if (!pPgAddr)
    {
        printk(KERN_WARNING "get_data_buffer parameter pPgAddr is NULL\n");
        return -EINVAL;
    }
    /* End Parameter out-of-bound check */

    if (pHost_st->pio_sgptr >= pHost_st->cmd->data->sg_len)
    {
        pr_debug( "no more buffers (%i/%i)\n",
                  pHost_st->pio_sgptr, pHost_st->cmd->data->sg_len);
        return -EBUSY;
    }

    pSG_st = &pHost_st->cmd->data->sg[pHost_st->pio_sgptr];

    *pWords = pSG_st->length >> 2;     /* 1 word = 32bit = 4 byte, div 4 */
    *pPgAddr = page_address(sg_page(pSG_st)) + pSG_st->offset;

    pHost_st->pio_sgptr++;

    pr_debug( "new buffer (%i/%i)\n",
              pHost_st->pio_sgptr, pHost_st->cmd->data->sg_len);

    return 0;
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : do_pio_read
+------------------------------------------------------------------------------
| DESCRIPTION : Copy Data from RAM into FCIE buffer
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
static inline void do_pio_read(struct mstar_mci_host *pHost_st)
{
    /* Define Local Variables */
    s32 res = 0, i = 0;

    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "do_pio_read parameter pHost_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    if (!pHost_st->pio_words)
    {
        res = get_data_buffer(pHost_st, &pHost_st->pio_words, &pHost_st->pio_ptr);

        if (res)
        {
            pr_debug( "pio_read(): complete (no more data).\n");
            return;
        }

        pr_debug( "pio_read(): new target: [%i]@[%p]\n",
                  pHost_st->pio_words, pHost_st->pio_ptr);
    }

    for (i=0; i < pHost_st->pio_words; i++)
    {
        *(pHost_st->pio_ptr++) = (fcie_readw(FCIE2_BASE + ((i*2) * 4)) & 0xFFFF)
                                 | ((fcie_readw(FCIE2_BASE + (((i*2)+1) * 4)) & 0xFFFF)<<16 );
    }
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : do_pio_write
+------------------------------------------------------------------------------
| DESCRIPTION : Copy Data from FCIE buffer into RAM
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
static inline void do_pio_write(struct mstar_mci_host *pHost_st)
{
    /* Define Local Variables */
    s32 res = 0, i = 0;

    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "do_pio_write parameter pHost_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    if (!pHost_st->pio_words)
    {
        res = get_data_buffer(pHost_st, &pHost_st->pio_words, &pHost_st->pio_ptr);

        if (res)
        {
            pr_debug( "pio_write():complete (no more data).\n");
            return;
        }

        pr_debug( "pio_write(): new source: [%i]@[%p]\n",
                  pHost_st->pio_words, pHost_st->pio_ptr);
    }

    for (i=0; i < pHost_st->pio_words; i++)
    {
        fcie_writew(*(pHost_st->pio_ptr++), FCIE2_BASE + (i * 4));
    }
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_sg_to_dma
+------------------------------------------------------------------------------
| DESCRIPTION : Copy from sg to a dma block - used for transfers
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
| *pData_st          | x |   | MMC data
+------------------------------------------------------------------------------
*/
static inline void mstar_mci_sg_to_dma(struct mstar_mci_host *pHost_st,
        struct mmc_data *pData_st)
{
    /* Define Local Variables */
    u32 i = 0;
    s32 amount = 0;
    u32 *pSGbuffer = 0;
    u32 len = pData_st->sg_len;
    u32 size = pHost_st->total_length;
    unsigned *pDMAbuf = pHost_st->tx_buffer;
    struct scatterlist *pSG_st = 0;

    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "mstars_mci_sg_to_dma parameter pHost_st is NULL\n");
        return;
    }

    if (!pData_st)
    {
        printk(KERN_WARNING "mstar_mci_sg_to_dma parameter pData_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    /* Loop through all entries. Size might not be the entire list though, */
    /* so make sure that we do not transfer too much. */
    for (i = 0; i < len; i++)
    {
        pSG_st = &pData_st->sg[i];

        pSGbuffer = kmap_atomic(sg_page(pSG_st), KM_BIO_SRC_IRQ) + pSG_st->offset;
        amount   = min(size, pSG_st->length);
        memcpy(pDMAbuf, pSGbuffer, amount);
        size -= amount;
        pDMAbuf += (amount >> 2) ;
        kunmap_atomic(pSGbuffer, KM_BIO_SRC_IRQ);

        pr_debug(KERN_WARNING "len=%d, pSG_st->length=%d, size=%d amount=%d, \n", len, pSG_st->length, size, amount);

        if (size == 0)
        {
            break;
        }
    }

    /*
    * Check that we didn't get a request to transfer
    * more data than can fit into the SG list.
    */
    BUG_ON(size != 0);

}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_card_power_en
+------------------------------------------------------------------------------
| DESCRIPTION : Columbus card power on/off
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| PwrOn              | x |   | MCI Card Power On
+------------------------------------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
static void mstar_mci_card_power_en(s32 PwrOn, struct mstar_mci_host *pHost_st)
{
    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "mstar_mci_card_power_en parameter pHost_st is NULL\n");
        return;
    }

    #if 0
    #if 0
    /* if MS card is in socket, do not power cycle */
    if(fcie_reg_read(pHost_st, FCIE_CARD_DET) & MS_DET_N)
    {
        pr_debug("columbus_mci_card_power %s\n", PwrOn ? "ON" : "OFF");
        // mdelay(50);
        set_current_state(TASK_UNINTERRUPTIBLE);
        schedule_timeout(msecs_to_jiffies(50));

        if (PwrOn)
        {
            fcie_reg_write(pHost_st, FCIE_CARD_PWR,
                               (fcie_reg_read(pHost_st, FCIE_CARD_PWR) & (~(SD_PWR_ON | SD_PWR_PAD_OEN))));
            /*
            CLRREG32(REG_GPIO_C_PD, BIT_FCIE_PAD_WPZCLKCMDDAT);// disable pull-down: WPZ, CLK, CMD & D0~D3
            // enable pull-up, low active
            CLRREG32(REG_GPIO_C_PU, BIT_FCIE_PAD_WPZCMDDAT);// enable pull-up: WPZ, CMD & D0~D3
            // enable strong-driving
            // SETREG32(REG_GPIO_C_DRV, BIT_FCIE_PAD_CLKCMDDAT);
            // enable strong-pull up, low active
            CLRREG32(REG_GPIO_C_PU,BIT_FCIE_PAD_STRONGUP);
            */
        }
        else
        {
            /*
            // disable strong-pull up, low active
            SETREG32(REG_GPIO_C_PU,BIT_FCIE_PAD_STRONGUP);
            // disable strong-driving
            CLRREG32(REG_GPIO_C_DRV, BIT_FCIE_PAD_CLKCMDDAT);

            SETREG32(REG_GPIO_C_PD, BIT_FCIE_PAD_WPZCLKCMDDAT);// disable pull-down: WPZ, CMD & D0~D3
            // disable pull-up, low active
            SETREG32(REG_GPIO_C_PU, BIT_FCIE_PAD_WPZCLKCMDDAT);// enable pull-up: WPZ, CMD & D0~D3
            */
            fcie_reg_write(pHost_st, FCIE_CARD_PWR,
                               (fcie_reg_read(pHost_st, FCIE_CARD_PWR) | (SD_PWR_ON | SD_PWR_PAD_OEN)));
        }
    }
    #else
    pr_debug("columbus_mci_card_power %s\n", PwrOn ? "ON" : "OFF");
    // mdelay(50);
    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout(msecs_to_jiffies(50));

    if (PwrOn)
    {
            CLRREG32(PADTOP1_DM_GPIO, DM_GPIO_OEN);
            CLRREG32(PADTOP1_DM_GPIO, DM_GPIO_OUT);
            /*
            CLRREG32(REG_GPIO_C_PD, BIT_FCIE_PAD_WPZCLKCMDDAT);// disable pull-down: WPZ, CLK, CMD & D0~D3
            // enable pull-up, low active
            CLRREG32(REG_GPIO_C_PU, BIT_FCIE_PAD_WPZCMDDAT);// enable pull-up: WPZ, CMD & D0~D3
            // enable strong-driving
            // SETREG32(REG_GPIO_C_DRV, BIT_FCIE_PAD_CLKCMDDAT);
            // enable strong-pull up, low active
            CLRREG32(REG_GPIO_C_PU,BIT_FCIE_PAD_STRONGUP);
            */
    }
    else
    {
            /*
            // disable strong-pull up, low active
            SETREG32(REG_GPIO_C_PU,BIT_FCIE_PAD_STRONGUP);
            // disable strong-driving
            CLRREG32(REG_GPIO_C_DRV, BIT_FCIE_PAD_CLKCMDDAT);

            SETREG32(REG_GPIO_C_PD, BIT_FCIE_PAD_WPZCLKCMDDAT);// disable pull-down: WPZ, CMD & D0~D3
            // disable pull-up, low active
            SETREG32(REG_GPIO_C_PU, BIT_FCIE_PAD_WPZCLKCMDDAT);// enable pull-up: WPZ, CMD & D0~D3
            */
            CLRREG32(PADTOP1_DM_GPIO, DM_GPIO_OEN);
            SETREG32(PADTOP1_DM_GPIO, DM_GPIO_OUT);
    }
    pr_debug("[PADTOP1_DM_GPIO]:%d\n", fcie_readw(PADTOP1_DM_GPIO));
    #endif
    #endif
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_pre_dma_read
+------------------------------------------------------------------------------
| DESCRIPTION : Prepare a dma read
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
static void mstar_mci_pre_dma_read(struct mstar_mci_host *pHost_st)
{
    /* Define Local Variables */
    struct scatterlist *pSG_st = 0;
    struct mmc_command *pCmd_st = 0;
    struct mmc_data *pData_st = 0;
    unsigned int _MIU0_LEN = MDrv_SYS_GetMIU0Length();

    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "mstar_mci_pre_dma_read parameter pHost_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    pr_debug("pre dma read\n");

    pCmd_st = pHost_st->cmd;
    if (!pCmd_st)
    {
        pr_debug("no command\n");
        return;
    }

    pData_st = pCmd_st->data;
    if (!pData_st)
    {
        pr_debug("no data\n");
        return;
    }

    /* nothing left to transfer */
    if (pHost_st->transfer_index >= pData_st->sg_len)
    {
        pr_debug("Nothing left to transfer (index = %d)\n",
                 pHost_st->transfer_index);
    }

    /* Setup the next transfer */
    pSG_st = &pData_st->sg[pHost_st->transfer_index++];

    pSG_st->dma_address = dma_map_page(NULL, sg_page(pSG_st), pSG_st->offset,
                                       pSG_st->length, DMA_FROM_DEVICE);
    #if 0
    if ((u32)pSG_st->dma_address & 0x4)
    {
        /* for 32-bits alignment */
        fcie_reg_write(pHost_st,FCIE_DMA_OFFSET,0xF0);
    }
    else
    {
        fcie_reg_write(pHost_st,FCIE_DMA_OFFSET,0xFF);
    }
    #endif

    if ( ((u32)pSG_st->dma_address) >= _MIU0_LEN )
        printk("\033[35mError, %s dma address is MIU1!\033[0m\n", __FUNCTION__);
    pr_debug("[R]Dma address : [0x%X]\n", pSG_st->dma_address);

    /*
    If MIU 1, do memcpy...
    (u32)pu8Buf) = (u32)pSG_st->dma_address;
    if( ((u32)pu8Buf) >= 0xC0000000 )
		pu8Buf = u8SDBuf;
    */
    fcie_reg_write(pHost_st,FCIE_JOB_BL_CNT,(pSG_st->length/512));
    #if 0
    fcie_reg_write(pHost_st,FCIE_MIU_DMA0,(((u32)pSG_st->dma_address) >> 3));
    fcie_reg_write(pHost_st,FCIE_MIU_DMA1,(((u32)pSG_st->dma_address) >> 19));
    #else
    fcie_reg_write(pHost_st,SDIO_MIU_DMA0,(((u32)pSG_st->dma_address) & 0xFFFF));
    fcie_reg_write(pHost_st,SDIO_MIU_DMA1,(((u32)pSG_st->dma_address) >> 16));
    #endif
    pr_debug("pre dma read done\n");
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_post_dma_read
+------------------------------------------------------------------------------
| DESCRIPTION : Handle after a dma read
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
static void mstar_mci_post_dma_read(struct mstar_mci_host *pHost_st)
{
    /* Define Local Variables */
    struct mmc_command *pCmd_st = 0;
    struct mmc_data *pData_st = 0;
    struct scatterlist *pSG_st = 0;

    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "mstar_mci_post_dma_read parameter pHost_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    pr_debug("post dma read\n");

    pCmd_st = pHost_st->cmd;

    if (!pCmd_st)
    {
        pr_debug("no command\n");
        return;
    }

    pData_st = pCmd_st->data;
    if (!pData_st)
    {
        pr_debug("no data\n");
        return;
    }

    pr_debug("in_use_index=%d, transfer_index=%d \n", pHost_st->in_use_index, pHost_st->transfer_index);

    while (pHost_st->in_use_index < pHost_st->transfer_index)
    {
        pr_debug("finishing index %d\n", pHost_st->in_use_index);
        pSG_st = &pData_st->sg[pHost_st->in_use_index++];
        //Patch dma_unmap_page bug in U4
        //dma_unmap_page(NULL, pSG_st->dma_address, pSG_st->length, DMA_FROM_DEVICE);
        fcie_dma_unmap_page(NULL, pSG_st->dma_address, pSG_st->length, DMA_FROM_DEVICE);
        pData_st->bytes_xfered += pSG_st->length;
        if (sg_page(pSG_st) != NULL)
        {
            flush_dcache_page(sg_page(pSG_st));
        }
        else
        {
            printk("Not flush_dcache_page\n");
        }
    }

    //fcie_dump_dma_data(pHost_st,pSG_st->dma_address,512);
    /* Is there another transfer to trigger? */
    if (pHost_st->transfer_index < pData_st->sg_len)
    {
        pr_debug("another transfer to trigger sg_len=%d transfer_index=%d\n", pData_st->sg_len, pHost_st->transfer_index);
        mstar_mci_pre_dma_read(pHost_st);
    }

    pr_debug("post dma read done\n");
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_handle_transmitted
+------------------------------------------------------------------------------
| DESCRIPTION : Handle transmitted data
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
static void mstar_mci_handle_transmitted(struct mstar_mci_host *pHost_st)
{
    /* Define Local Variables */
    struct mmc_command *pCmd_st  = 0;
    struct mmc_data    *pData_st = 0;

    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "mstar_mci_handle_transmitted parameter pHost_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    pCmd_st = pHost_st->cmd;
    if (!pCmd_st)
    {
        return;
    }

    pData_st = pCmd_st->data;
    if (!pData_st)
    {
        return;
    }

    if (pHost_st->tx_buffer)
    {
        dma_free_coherent(NULL, pHost_st->total_length,
                          pHost_st->tx_buffer, pHost_st->physical_address);
        pHost_st->tx_buffer = NULL;
        pData_st->bytes_xfered = pHost_st->total_length;
    }

    pr_debug("bytes_xfered=%d\n", pData_st->bytes_xfered);
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_handle_cmdrdy
+------------------------------------------------------------------------------
| DESCRIPTION : Handle after command sent ready
|
| RETURN      : Error Status; Return 0 if no error.
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
static s32 mstar_mci_handle_cmdrdy(struct mstar_mci_host *pHost_st)
{
    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "mstar_mci_handle_cmdrdy parameter pHost_st is NULL\n");
        return -EINVAL;
    }
    /* End Parameter out-of-bound check */

    if (!pHost_st->cmd)
    {
        return 1;
    }
    else
    {
        if (!pHost_st->cmd->data)
        {
            return 1;
        }
        else
        {
            if (pHost_st->cmd->data->flags & MMC_DATA_WRITE)
            {
                /*After sendding multi-block-write command, start DMA transfer*/
                if (pHost_st->dodma)
                {
                    if (pHost_st->tx_buffer)
                    {
                        printk("Irq do dma write\n");
                        fcie_reg_write(pHost_st, FCIE_MMA_PRI_REG,
                                           (fcie_reg_read(pHost_st, FCIE_MMA_PRI_REG) | JOB_RW_DIR));
                        fcie_reg_write(pHost_st, FCIE_MIE_PATH_CTL,
                                           (fcie_reg_read(pHost_st, FCIE_MIE_PATH_CTL) | MMA_ENABLE));
                        fcie_reg_write(pHost_st, FCIE_SD_CTL,
                                           (fcie_reg_read(pHost_st, FCIE_SD_CTL) | (SD_DTRX_DIR | SD_DTRX_EN)));
                    }
                }
                else
                {
                    do_pio_write(pHost_st);
                }
            }
        }
    }

    return 0;
}

static void mstar_mci_reset(struct mstar_mci_host *pHost_st)
{
    /* reset IP */
    fcie_reg_write(pHost_st, FCIE_TEST_MODE,
                       (fcie_reg_read(pHost_st, FCIE_TEST_MODE) & (~FCIE_SOFT_RST)));
    mdelay(1);
    fcie_reg_write(pHost_st, FCIE_TEST_MODE,
                       (fcie_reg_read(pHost_st, FCIE_TEST_MODE) | FCIE_SOFT_RST));
}

static void mstar_mci_switch_sd(struct mstar_mci_host *pHost_st)
{
    /* 1. Switch interface to SD */
    fcie_reg_write(pHost_st, FCIE_MIE_PATH_CTL, SD_EN);
    //columbus_mci_reset(pHost_st);
    /* 2. Set SD clock */
    fcie_writew(pHost_st->sd_clk, REG_CLK_NFIE_FCIE);
    /* 3. Set SD mode */
    fcie_reg_write(pHost_st, FCIE_SD_MODE, pHost_st->sd_mod);
    /* 3. Disable SDIO register */
    fcie_reg_write(pHost_st, SDIO_MIU_DMA0, 0);
    fcie_reg_write(pHost_st, SDIO_MIU_DMA1, 0);

    pr_debug("mstar_mci_switch_sd\n");
    pr_debug("1.INTF:0x%X\n", fcie_reg_read(pHost_st, FCIE_MIE_PATH_CTL));
    pr_debug("2.CLK:0x%X\n", fcie_readw(REG_CLK_NFIE_FCIE));
    pr_debug("3.MODE:0x%X\n", fcie_reg_read(pHost_st, FCIE_SD_MODE));
    //printk("4.DET:0x%X\n", fcie_reg_read(pHost_st, FCIE_CARD_DET));
    //printk("4.TOP:0x%X\n", fcie_readw(CHIPTOP_SDIO_MODE));

}


/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_enable
+------------------------------------------------------------------------------
| DESCRIPTION : Enable the controller
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
static void mstar_mci_enable(struct mstar_mci_host *pHost_st)
{
    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "mstar_mci_enable parameter pHost_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    //columbus_mci_reset(pHost_st);
    /* init registers */
    fcie_reg_write(pHost_st, FCIE_MIE_INT_EN, 0);
    fcie_reg_write(pHost_st, FCIE_SD_CTL, 0);
    fcie_reg_write(pHost_st, FCIE_SD_MODE, 0);
    //fcie_reg_write(pHost_st, FCIE_MIE_PATH_CTL, SD_EN);
    /* clear events */
    fcie_reg_write(pHost_st, FCIE_SD_STS, SD_STS_ERRORS);
    //fcie_reg_write(pHost_st, FCIE_MIE_EVENT, SD_ALL_INT);
    fcie_reg_write(pHost_st, FCIE_CARD_EVENT, SD_CARD_EVENTS);
    pHost_st->present = 0;
    /*
     * Winder: some boards need this delay,
     * otherwise the MS_DET_N will be low at boot time.
     * Not sure why.
     */

    // Switch pad to SDIO mode
    CLRREG32(CHIPTOP_SDIO_MODE, PAD_SDIO_MODE_CLR);
    SETREG32(CHIPTOP_SDIO_MODE, PAD_SDIO_MODE_ACT);

    if (!d_initialized)
    {
	    printk("Enable MCI controller.\n");
        mdelay(50);
        d_initialized = 1;
    }
#if defined(CONFIG_MEMSTICK_COLUMBUS_MS)
    if (!(fcie_reg_read(pHost_st, FCIE_CARD_DET) & MS_DET_N))	//jesse.su
    {
        pHost_st->present = 1;
        pr_debug("MS Card present!\n");
    }
#endif
    if (!(fcie_reg_read(pHost_st, FCIE_CARD_DET) & SDIO2_DET_N))
    {
        pHost_st->present = 1;
        pr_debug("SD Card present!\n");
    }
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_disable
+------------------------------------------------------------------------------
| DESCRIPTION : Disable the controller
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
static void mstar_mci_disable(struct mstar_mci_host *pHost_st)
{
    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "mstar_mci_disable parameter pHost_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    fcie_reg_write(pHost_st, FCIE_MIE_PATH_CTL, ~(MMA_ENABLE | SD_EN));
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_send_command
+------------------------------------------------------------------------------
| DESCRIPTION : Send a command
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
| *pCmd_st           | x |   | MCI command structure
+------------------------------------------------------------------------------
*/
static void mstar_mci_send_command(struct mstar_mci_host *pHost_st,
                                      struct mmc_command *pCmd_st)
{
    /* Define Local Variables */
    struct mmc_data *pData_st;
    u32 mie_int = 0, sd_ctl = 0;

    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "mstar_mci_send_command parameter pHost_st is NULL\n");
        return;
    }

    if (!pCmd_st)
    {
        printk(KERN_WARNING "mstar_mci_send_command parameter pCmd_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    pr_debug("mstar_mci_send_command\n");
    mstar_mci_switch_sd(pHost_st);

    pData_st = pCmd_st->data;

    pHost_st->cmd = pCmd_st;
    fcie_reg_write(pHost_st, FCIE_SD_CTL, sd_ctl);

    if (pData_st)
    {
        //u32 unknown_fix;

        pr_debug("pData_st!=0\n");
        #if 0
        /* It is impossible error..but, if it happened, it would be a nightmare. */
        unknown_fix = ((u32)pData_st) & 0xf0000000;
        if (unknown_fix != 0xc0000000)
        {
            printk("pData_st wrong %p.\n", pData_st);
            pCmd_st->error = -EINVAL;
            if ((pHost_st->mmc != NULL) && (pHost_st->request != NULL))
            {
                mmc_request_done(pHost_st->mmc, pHost_st->request);
            }
            SD_UNLOCK_FCIE();
            return;
        }
        #endif

        if (pData_st->blksz & 0x3)
        {
            pr_debug("Unsupported block size\n");
            pCmd_st->error = -EINVAL;
            mmc_request_done(pHost_st->mmc, pHost_st->request);
            return;
        }

        if (pData_st->flags & MMC_DATA_READ)
        {
            sd_ctl |= SD_DTRX_EN;
        }
    }

    sd_ctl |= SD_CMD_EN;
    mie_int |= SD_CMD_ENDE;

    pHost_st->dodma = 0;
    if (pData_st)
    {
        pr_debug("blocks = %d, length = %d \n", pData_st->blocks, pData_st->blksz);
        wait_fifoclk_ready(pHost_st);
        if ((pData_st->blksz * pData_st->blocks) < 512)
        {
            pHost_st->pio_sgptr = 0;
            pHost_st->pio_words = 0;
            fcie_reg_write(pHost_st, FCIE_SD_MODE,
                               fcie_reg_read(pHost_st, FCIE_SD_MODE) | SD_DEST);
            fcie_reg_write(pHost_st, FCIE_CARD_WD_CNT,
                               ((pData_st->blksz * pData_st->blocks)>>1));
            pr_debug("2__FCIE_SD_MODE=%x, FCIE_CARD_WD_CNT=%x\n",
                     fcie_reg_read(pHost_st, FCIE_SD_MODE),
                     fcie_reg_read(pHost_st, FCIE_CARD_WD_CNT));
            mie_int |= SD_DATA_ENDE;
            fcie_reg_write(pHost_st, SDIO_BLK_SIZE, (u16)pData_st->blksz);
        }
        else
        {
            pHost_st->dodma = 1;
            fcie_reg_write(pHost_st, FCIE_SD_MODE,
                               fcie_reg_read(pHost_st, FCIE_SD_MODE) & ~SD_DEST);
            //mie_int |= MMA_DATA_ENDE;
            pData_st->bytes_xfered = 0;
            pHost_st->transfer_index = 0;
            pHost_st->in_use_index = 0;
            fcie_reg_write(pHost_st, SDIO_BLK_SIZE, (SDIO_BLK_MOD_EN|(u16)pData_st->blksz));
            pr_debug("[SDIO_BLK_SIZE]:%X\n", fcie_reg_read(pHost_st, SDIO_BLK_SIZE));
            if (pData_st->flags & MMC_DATA_WRITE)
            {
                /* Handle a write */
                //mie_int |= SD_DATA_ENDE;
                mie_int |= CARD_DMA_END_EN;
                if (!pHost_st->tx_buffer)
                {
                    #if (FCIE_DYNAMIC_MALLOC)
                    pHost_st->total_length = pData_st->blksz * pData_st->blocks;
                    saved_tx_total_length = pHost_st->total_length;
                    pHost_st->tx_buffer = dma_alloc_coherent(NULL,
                                          pHost_st->total_length,
                                          &pHost_st->physical_address, GFP_KERNEL);
                    #endif
                    if (pHost_st->tx_buffer)
                    {
                        mstar_mci_sg_to_dma(pHost_st, pData_st);
                        if ((u32)pHost_st->physical_address & 0x4)
                        {
                            fcie_reg_write(pHost_st, FCIE_DMA_OFFSET, 0xF0); // for 32-bits alignment
                        }
                        else
                        {
                            fcie_reg_write(pHost_st, FCIE_DMA_OFFSET, 0xFF);
                        }
                        if ( ((u32)pHost_st->physical_address) >= 0xC0000000 )
                               printk("\033[34mError, %s dma address is MIU1!\033[0m\n", __FUNCTION__);
                        pr_debug("[W]Dma address : [0x%X]\n", pHost_st->physical_address);
                        fcie_reg_write(pHost_st, FCIE_JOB_BL_CNT,
                                           (pHost_st->total_length / 512));
                        #if 0
                        fcie_reg_write(pHost_st, FCIE_MIU_DMA0,
                                           (((u32)pHost_st->physical_address) >> 3));
                        fcie_reg_write(pHost_st, FCIE_MIU_DMA1,
                                           (((u32)pHost_st->physical_address) >> 19));
                        #else
                        fcie_reg_write(pHost_st, SDIO_MIU_DMA0,
                                           (((u32)pHost_st->physical_address) & 0xFFFF));
                        fcie_reg_write(pHost_st, SDIO_MIU_DMA1,
                                           (((u32)pHost_st->physical_address) >> 16));
                        #endif
                    }
                    else
                    {
                        pCmd_st->error = -ENOMEM;
                        return;
                    }
                }
                else
                {
                    printk("tx_buff not NULL?!\n");
                }
            }
            else
            {
                /* Handle a read */
                pHost_st->rx_buffer = NULL;
                pHost_st->total_length = 0;
                mie_int |= MMA_DATA_ENDE;

                mstar_mci_pre_dma_read(pHost_st);
                fcie_reg_write(pHost_st, FCIE_MMA_PRI_REG,
                                   (fcie_reg_read(pHost_st, FCIE_MMA_PRI_REG) & (~JOB_RW_DIR)));
                fcie_reg_write(pHost_st, FCIE_MIE_PATH_CTL,
                                   (fcie_reg_read(pHost_st, FCIE_MIE_PATH_CTL) | MMA_ENABLE));
            }
        }
    }

    pr_debug("FCIE_SD_STS = %08X\n",fcie_reg_read(pHost_st, FCIE_SD_STS));

    if (waitD0High(pHost_st, 0x200000))
    {
        printk(KERN_WARNING "Card busy!!\n");
        pCmd_st->error = -ETIMEDOUT;
        return;
    }

    /* Send the command and then enable the PDC - not the other way round as */
    /* the data sheet says */
    fcie_writew((((pCmd_st->arg >> 24) << 8) | (0x40 | pCmd_st->opcode)), FCIE1_BASE);
    fcie_writew(((pCmd_st->arg & 0xFF00) | ((pCmd_st->arg >> 16) & 0xFF)), FCIE1_BASE + 4);
    fcie_writew(((fcie_readw(FCIE1_BASE + 8) & 0xFF00) + (pCmd_st->arg & 0xFF)), FCIE1_BASE + 8);

    #if 0
    pr_debug("cifc+0 = %08X\n", fcie_readw(FCIE1_BASE + 0));
    pr_debug("cifc+4 = %08X\n", fcie_readw(FCIE1_BASE + 4));
    pr_debug("cifc+8 = %08X\n", fcie_readw(FCIE1_BASE + 8));
    #endif
    fcie_reg_write(pHost_st, FCIE_CMD_SIZE, 5);

    if (mmc_resp_type(pCmd_st) == MMC_RSP_NONE)
    {
        sd_ctl &= ~SD_RSP_EN;
        fcie_reg_write(pHost_st, FCIE_RSP_SIZE, 0);
    }
    else
    {
        if (mmc_resp_type(pCmd_st) == MMC_RSP_R2)
        {
            sd_ctl |= SD_RSPR2_EN;
            fcie_reg_write(pHost_st, FCIE_RSP_SIZE, 16); /* (136-8)/8 */
        }
        else
        {
            fcie_reg_write(pHost_st, FCIE_RSP_SIZE, 5); /*(48-8)/8 */
        }

        sd_ctl |= SD_RSP_EN;
    }

    simple_debug("Sending command %d as %08X, mod %08X, arg = %08X\n",
             pCmd_st->opcode, sd_ctl, fcie_reg_read(pHost_st, FCIE_SD_MODE), pCmd_st->arg);
    simple_debug("Clock:0x%X\n", fcie_readw(REG_CLK_NFIE_FCIE));

    //Jonas_990326: switch the sequence of FCIE_MIE_INT_EN and FCIE_SD_CTL
    /* Enable selected interrupts */
    fcie_reg_write(pHost_st, FCIE_MIE_INT_EN, mie_int);
    //fcie_reg_write(pHost_st, FCIE_MIE_INT_EN, 0x9FFF);
    pr_debug("[FCIE_MIE_INT_EN]:%X\n", fcie_reg_read(pHost_st, FCIE_MIE_INT_EN));
    pr_debug("[FCIE_MIE_EVENT]:%X\n", fcie_reg_read(pHost_st, FCIE_MIE_EVENT));
    fcie_reg_write(pHost_st, FCIE_SD_CTL, sd_ctl);

}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_process_next
+------------------------------------------------------------------------------
| DESCRIPTION : Process the next step in the request
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
static void mstar_mci_process_next(struct mstar_mci_host *pHost_st)
{
    unsigned long flags;

    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "mstar_mci_process_next parameter pHost_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    if (!(pHost_st->flags & FL_SENT_COMMAND))
    {
        pr_debug("request->cmd\n");
        local_irq_save(flags);
        pHost_st->flags |= FL_SENT_COMMAND;
        mstar_mci_send_command(pHost_st, pHost_st->request->cmd);
        local_irq_restore(flags);
    }
    else if ((!(pHost_st->flags & FL_SENT_STOP)) && pHost_st->request->stop)
    {
        pr_debug("request->stop\n");
        local_irq_save(flags);
        pHost_st->flags |= FL_SENT_STOP;
        mstar_mci_send_command(pHost_st, pHost_st->request->stop);
        local_irq_restore(flags);
    }
    else
    {
        pr_debug("mmc_request_done\n");
        local_irq_save(flags);
        mmc_request_done(pHost_st->mmc, pHost_st->request);
        local_irq_restore(flags);
    }
}


/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_completed_command
+------------------------------------------------------------------------------
| DESCRIPTION : Handle a command that has been completed
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| data               | x |   | Columbus MCI Host Data
+------------------------------------------------------------------------------
| *pHost_st          | x |   | Columbus MCI Host structure
+------------------------------------------------------------------------------
*/
#ifdef FCIE_TASKLET /* Defer tasks in IRQ, so IRQ can be finished faster */

static void mstar_mci_completed_command(unsigned long data)
{
    /* Define Local Variables */
    struct mstar_mci_host *pHost_st = (struct mstar_mci_host *) data;

#else               /* Do NOT enable Tasklet */

static void mstar_mci_completed_command(struct mstar_mci_host *pHost_st)
{

#endif  /* End ifdef FCIE_TASKLET */

    /* Define Local Variables */
    u8 *pTemp = 0;
    u32 status = 0, i = 0;
    struct mmc_command *pCmd_st = pHost_st->cmd;

    /* Parameter out-of-bound check */
    if (!pHost_st)
    {
        printk(KERN_WARNING "mstar_mci_completed_command parameter pHost_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    pTemp = (u8*)&(pCmd_st->resp[0]);

    #if 0
    pr_debug("response:%08X,%08X,%08X,%08X\n",
             fcie_readw(FCIE1_BASE), fcie_readw(FCIE1_BASE + 4),
             fcie_readw(FCIE1_BASE + 8), fcie_readw(FCIE1_BASE + 0x0c));
    #endif
    //if(waitD0High(pHost_st, 150))
    //{
    //    printk(KERN_WARNING "Card busy 2!!\n");
    //}

    for (i=0; i < 15; i++)
    {
        pTemp[(3 - (i % 4)) + (4 * (i / 4))] =
            (u8)(fcie_readw(FCIE1_BASE + (((i + 1) / 2) * 4)) >> (8 * ((i + 1) % 2)));
    }

    status = fcie_reg_read(pHost_st, FCIE_SD_STS);

    #if 1
    simple_debug( "Status = %08X [%08X %08X %08X %08X]\n",
              status, pCmd_st->resp[0], pCmd_st->resp[1], pCmd_st->resp[2], pCmd_st->resp[3]);
    #endif
    if (status & SD_STS_ERRORS)
    {
        if ((status & SD_CMDRSP_CERR) && !(mmc_resp_type(pCmd_st) & MMC_RSP_CRC))
        {
            pCmd_st->error = 0;
        }
        else
        {
            //printk(KERN_WARNING "r = %x\n", status);
            if (status & (SD_DAT_NORSP | SD_DAT_STSERR))
            {
                pCmd_st->error = -ETIMEDOUT;
            }
            else if (status & (SD_CMDRSP_CERR | SD_DAT_CERR | SD_DAT_STSNEG))
            {
                pCmd_st->error = -EILSEQ;
            }
            else
            {
                pCmd_st->error = -EIO;
            }

            printk(KERN_WARNING "Error detected and set to %d (cmd = %d, retries = %d)\n",
                     pCmd_st->error, pCmd_st->opcode, pCmd_st->retries);
        }
    }
    else
    {
        pCmd_st->error = 0;
    }

    //printk(KERN_WARNING "pCmd_st->error =%x\n",pCmd_st->error);
    //fcie_reg_write(pHost_st, FCIE_SD_STS, SD_STS_ERRORS);
    clear_fcie_status(pHost_st, FCIE_SD_STS, SD_STS_ERRORS);
    mstar_mci_process_next(pHost_st);
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_request
+------------------------------------------------------------------------------
| DESCRIPTION : Handle an MMC request
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pMMC_st           | x |   | MMC Host structure
+------------------------------------------------------------------------------
| *pMRQ_st           | x |   | MMC Request structure
+------------------------------------------------------------------------------
*/
//extern void columbus_disable_workaround_timer(void);
static void mstar_mci_request(struct mmc_host *pMMC_st, struct mmc_request *pMRQ_st)
{
    /* Define Local Variables */
    struct mstar_mci_host *pHost_st = mmc_priv(pMMC_st);
    // s32 present;

    /* Parameter out-of-bound check */
    if (!pMMC_st)
    {
        printk(KERN_WARNING "mstar_mci_request parameter pMMC_st is NULL\n");
        return;
    }

    if (!pMRQ_st)
    {
        printk(KERN_WARNING "mstar_mci_request parameter pMRQ_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    //columbus_disable_workaround_timer();

    pHost_st->request = pMRQ_st;
    pHost_st->flags = 0;

#if 0
    present = !(fcie_reg_read(pHost_st, FCIE_CARD_DET) & SD_DET_N);

    if (!present)
    {
        printk("Card Not Present..\n");
        pMRQ_st->cmd->error = -ENOMEDIUM;
        mmc_request_done(pMMC_st, pMRQ_st);
        return;
    }
#endif
    mstar_mci_process_next(pHost_st);
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_set_ios
+------------------------------------------------------------------------------
| DESCRIPTION : Set MMC IOs
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pMMC_st           | x |   | MMC Host structure
+------------------------------------------------------------------------------
| *pIOS_st           | x |   | MMC Inputs/Outputs
+------------------------------------------------------------------------------
*/
static void mstar_mci_set_ios(struct mmc_host *pMMC_st, struct mmc_ios *pIOS_st)
{
    /* Define Local Variables */
    s32 clkdiv = 0;
    u32 sd_mod = 0;
    unsigned long flags;
    u16 reg_mask = ~((u16)0x3F);

    struct mstar_mci_host *pHost_st = mmc_priv(pMMC_st);

    /* Parameter out-of-bound check */
    if (!pMMC_st)
    {
        printk(KERN_WARNING "mstar_mci_set_ios parameter pMMC_st is NULL\n");
        return;
    }

    if (!pIOS_st)
    {
        printk(KERN_WARNING "mstar_mci_set_ios parameter pIOS_st is NULL\n");
        return;
    }
    /* End Parameter out-of-bound check */

    //SD_LOCK_FCIE();
    pr_debug("MMC: mstar_mci_set_ios\n");
    // mdelay(50);
    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout(msecs_to_jiffies(50));

    local_irq_save(flags);

    sd_mod = 0;
    pHost_st->bus_mode = pIOS_st->bus_mode;
    clkdiv =fcie_readw(REG_CLK_NFIE_FCIE); /* High triggle level */

    pr_debug("clkdiv:%d\n", clkdiv);
    pr_debug("clock:%d\n", pIOS_st->clock);
    if (pIOS_st->clock == 0)
    {
        /* Disable the MCI controller */
        sd_mod &= ~(SD_CLK_EN | SD_CS_EN);
        clkdiv |= 0x01; /* disable clock */
    }
    else
    {
        /* Enable the MCI controller */
        sd_mod |= (SD_CLK_EN | SD_CS_EN | SD_DATSYNC | SDIO_SD_BUS_SW | SDIO_PORT_SEL);

        pr_debug("pIOS_st->clock=%d\n", pIOS_st->clock);

        if (pIOS_st->clock == 300000)
        {
    #if defined(CONFIG_MSTAR_URANUS4)
            clkdiv &= reg_mask; /* 300KHz */
            clkdiv |= (BIT_FCIE_CLK_375K);
    #else
            clkdiv &= reg_mask; /* 300KHz */
    #endif
        }
        else if (pIOS_st->clock == 50000000)    /* Chun for High Speed SD */
        {
            clkdiv &= reg_mask; /* 300KHz */
            //clkdiv |= (BIT_FCIE_CLK_43M << reg_shift); /* 43MHz */
            //clkdiv |= (BIT_FCIE_CLK_32M); /* 32MHz */
            clkdiv |= (BIT_FCIE_CLK_22_8M);
        }
        else if (pIOS_st->clock == 36000000)    /* 36MHz */
        {
            clkdiv &= reg_mask; /* 300KHz */
            //clkdiv |= (BIT_FCIE_CLK_32M); /* 32MHz */
            clkdiv |= (BIT_FCIE_CLK_22_8M);
        }
        else
        {
            clkdiv &= reg_mask; /* 300KHz */
            clkdiv |= BIT_FCIE_CLK_22_8M;
        }

        //fcie_writew(clkdiv, REG_CLK_NFIE_FCIE);
        pHost_st->sd_clk = (u16)clkdiv;
        pr_debug("[clkdiv]=%d\n", clkdiv);
        pr_debug("[pHost_st->sd_clk]=%d\n", pHost_st->sd_clk);

    }

    if (pIOS_st->bus_width == MMC_BUS_WIDTH_4)
    {
        pr_debug("MMC: Setting controller bus width to 4\n");
        sd_mod = (sd_mod & ~SD_DAT_LINE) | SD_DAT_LINE0;
    }
    else
    {
        pr_debug("MMC: Setting controller bus width to 1\n");
        sd_mod = (sd_mod & ~SD_DAT_LINE);
    }

    pr_debug("[FCIE_SD_MODE]=%x\n", sd_mod);
    pHost_st->sd_mod = sd_mod;
    //fcie_reg_write(pHost_st, FCIE_SD_MODE, sd_mod);
    //fcie_reg_write(pHost_st, FCIE_CARD_WD_CNT, 0);
    /* maybe switch power to the card */
    switch (pIOS_st->power_mode)
    {
        case MMC_POWER_OFF:
            mstar_mci_card_power_en(0, pHost_st);
            break;

        case MMC_POWER_UP:
            /* Fall through */
        case MMC_POWER_ON:
            mstar_mci_card_power_en(1, pHost_st);
            break;

        default:
            /* Need to do error handler here */
            break;
    }
    local_irq_restore(flags);
}


/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mic_SetClockRate
+------------------------------------------------------------------------------
| DESCRIPTION : Set SD card Clock Rate
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pMMC_st           | x |   | MMC Host structure
+------------------------------------------------------------------------------
| u32ClockRate       | x |   | SD Clock Rate
+------------------------------------------------------------------------------
*/
static void columbus_mci_SetClockRate(struct mmc_host *pHost_st, u32 u32ClockRate)
{
    /* Define Local Variables */
    u16 clkdiv = 0;
    struct mmc_ios *ios;

    #if defined(CONFIG_ARCH_PIONEER)
    u8  reg_shift = 8;
    u16 reg_mask = ~((u16)0x7F << 8);
    #elif defined(CONFIG_MSTAR_URANUS4)
    u8  reg_shift = 0;
    u16 reg_mask = ~((u16)0x3F);
    #endif

    ios = &pHost_st->ios;
    clkdiv =fcie_readw(REG_CLK_NFIE_FCIE); /* High triggle level */
    clkdiv &= reg_mask; /* 300KHz */
    printk("[columbus_mci_SetClockRate]clk:%d\n", u32ClockRate);

    if(u32ClockRate >= CLK_43MHz)
    {
        //clkdiv |= (BIT_FCIE_CLK_43M << reg_shift); /* 43MHz */
        clkdiv |= (BIT_FCIE_CLK_32M << reg_shift); /* 32MHz */
        ios->clock = CLK_43MHz;
    }
    else if(u32ClockRate >= CLK_36MHz)
    {
        clkdiv |= (BIT_FCIE_CLK_32M << reg_shift); /* 32MHz */
        ios->clock = CLK_36MHz;
    }
    else if(u32ClockRate >= CLK_27MHz)
    {
        clkdiv |= (BIT_FCIE_CLK_27M << reg_shift); /* 36MHz */
        ios->clock = CLK_27MHz;
    }
    else /*default 22.8Mhz*/
    {
        clkdiv |= (BIT_FCIE_CLK_22_8M << reg_shift); /* 22.8MHz */
        ios->clock = CLK_21_5MHz;
    }

    fcie_writew(clkdiv, REG_CLK_NFIE_FCIE);

    printk("MMC: Set Clock Rate to %uHz \n", ios->clock);
}
/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_irq
+------------------------------------------------------------------------------
| DESCRIPTION : Handle an interrupt
|
| RETURN      : Interrupt Process Status. 1 = IRQ_HANDLED, 0 = IRQ_NONE.
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| irq                | x |   | Interrupt Request
+------------------------------------------------------------------------------
| *pDevID            | x |   | Device ID
+------------------------------------------------------------------------------
*/
static irqreturn_t mstar_mci_irq(s32 irq, void *pDevID)
{
    /* Define Local Variables */
    s32 completed = 0, present = 0;
    u32 sd_sts = 0, card_evt = 0, mie_evt = 0, mie_int = 0, mie_int_sts = 0,i;
    u32 card_int_en;
    struct mstar_mci_host *pHost_st = pDevID;
    struct memstick_host    *msh;
    struct mmc_host *pMMC_st;
    struct mmc_ios *pIOS_st;
    u32 u32Cmd = 0;

    card_int_en = 0;
    msh = NULL;

    pr_debug("mstar_mci_irq\n");

    /* Get ios structure and SD command number*/
    pMMC_st = pHost_st->mmc;
    pIOS_st = &pMMC_st->ios;
    if(pHost_st->cmd == NULL)
        u32Cmd = 0;
    else
        u32Cmd = pHost_st->cmd->opcode;
    /* Parameter out-of-bound check */
    if (!pDevID)
    {
        printk(KERN_WARNING "mstar_mci_irq parameter pDevID is NULL\n");
        return -EINVAL;
    }
    /* End Parameter out-of-bound check */

    card_evt = fcie_reg_read(pHost_st, FCIE_CARD_EVENT);
    card_int_en = fcie_reg_read(pHost_st, FCIE_CARD_INT_EN);  /* mask don't care int  */
    //printk("1:%x 2:%x 3:%x 4:%x\n", card_evt, card_int_en, fcie_reg_read(pHost_st, FCIE_CARD_DET), fcie_readw(CHIPTOP_SDIO_MODE));
    //printk("5:%x 6:%x\n", fcie_reg_read(pHost_st, FCIE_JOB_BL_CNT), fcie_reg_read(pHost_st, FCIE_TR_BL_CNT));


#if defined(CONFIG_MEMSTICK_COLUMBUS_MS)
    /* share IRQ line with SD and MS driver, for passing MS interrupt */
    if ( ((card_evt & card_int_en) & MS_STS_CHG) || !(fcie_reg_read(pHost_st, FCIE_CARD_DET) & MS_DET_N))
    {
        pr_debug("columbus_MS_STS_CHG occurred \n");
        present = !(fcie_reg_read(pHost_st, FCIE_CARD_DET) & MS_DET_N);		//jesse.su
        if (present != pHost_st->present)
        {
            pHost_st->present = present;
            pr_debug("[MMC]%s: card %s\n", mmc_hostname(pHost_st->mmc),
                     present ? "insert" : "remove");

            mstar_mci_reinit(pHost_st);

            /* Winder add for card status change. */
            pHost_st->card_chg = 1;
            wake_up_interruptible(&pHost_st->event_q);
        }

        msh = pass_fcie_ms_host();
        atomic_set(&is_sd_mmc, 0);
        return columbus_ms_interrupt(irq, (void *)msh); //Pass to memory stick driver
    }
#endif

    if ((card_evt & SD_SDIO2_CHG)&&
        (card_int_en & SDIO2_STS_EN))
    {
        mdelay(10);
        present = !(fcie_reg_read(pHost_st, FCIE_CARD_DET) & SDIO2_DET_N);
        if (present != pHost_st->present)
        {
            pHost_st->present = present;
            printk("%s: card %s\n", mmc_hostname(pHost_st->mmc),
                     present ? "insert" : "remove");

            /* Winder add for card status change. */
            pHost_st->card_chg = 1;
            atomic_set(&is_sd_mmc, 1);
            wake_up_interruptible(&pHost_st->event_q);

            /* use write to avoid both SD and MS enable */
            fcie_reg_write(pHost_st, FCIE_MIE_PATH_CTL, SD_EN);
            //fcie_reg_write(pHost_st, FCIE_MIE_EVENT, SD_ALL_INT);
            clear_fcie_status(pHost_st, FCIE_MIE_EVENT, SD_ALL_INT);
            /* force to release host if card is removed.. */
            // printk("host release..\n");
            mmc_release_host(pHost_st->mmc);
            mmc_detect_change(pHost_st->mmc, msecs_to_jiffies(1));
            // mmc_detect_change(pHost_st->mmc, 0);
            //columbus_mci_disable_stschg(pHost_st->mmc);
            // if (!present)
            {
                mstar_mci_reinit(pHost_st);
                mstar_mci_clearall_req(pMMC_st);
            }
            /* Winder add for dynamic removal/insertion. */
#if (FCIE_DYNAMIC_MALLOC)
            if (pHost_st->tx_buffer)
            {
                if (!pHost_st->total_length)
                {
                    pHost_st->total_length = saved_tx_total_length;
                }
                dma_free_coherent(NULL, pHost_st->total_length,
                                  pHost_st->tx_buffer, pHost_st->physical_address);
                pHost_st->tx_buffer = NULL;
            }
#else
            if (pHost_st->present)
            {
                pHost_st->tx_buffer = dma_alloc_coherent(NULL, pHost_st->total_length,
                                          &pHost_st->physical_address, GFP_KERNEL);
	            if (!pHost_st->tx_buffer) {
			         printk(KERN_WARNING "dma_alloc_coherent %ld failed\n",
					 pHost_st->total_lengthe);
			         return -ENOMEM;
                }

	            pr_debug("dma_alloc_coherent data is at %p (%ld)\n",
		   	    pHost_st->tx_buffer, pHost_st->total_length);
            }
            else
            {
                if (pHost_st->tx_buffer) {
                dma_free_coherent(NULL, pHost_st->total_length,
                                  pHost_st->tx_buffer, pHost_st->physical_address);
                                  pHost_st->tx_buffer = NULL;
                }
            }
#endif
        }

        //fcie_reg_write(pHost_st, FCIE_CARD_EVENT, card_evt);
        //fcie_reg_write(pHost_st, FCIE_SD_STS, SD_STS_ERRORS);
        clear_fcie_status(pHost_st, FCIE_CARD_EVENT, card_evt);
        clear_fcie_status(pHost_st, FCIE_SD_STS, SD_STS_ERRORS);
    }

    mie_evt = fcie_reg_read(pHost_st, FCIE_MIE_EVENT);
    mie_int = fcie_reg_read(pHost_st, FCIE_MIE_INT_EN);
    mie_int_sts = mie_evt & mie_int;
    sd_sts  = fcie_reg_read(pHost_st, FCIE_SD_STS);

    pr_debug("KERN_WARNING MCI irq: mie_evt=%08X, sd_sts=%08X, card_evt=%08X\n",
             mie_evt, sd_sts, card_evt);

    if (sd_sts & SD_STS_ERRORS)
    {
        completed = 1;

        if (sd_sts & SD_DAT_CERR)
        {
            printk(KERN_WARNING "MMC: Received data phase CRC error event\n");
            fcie_reg_dump(pHost_st);
        }
        if (sd_sts & SD_DAT_STSERR)
        {
            printk(KERN_WARNING "MMC: Transmitted data phase:Flash program error\n");
        }
        if (sd_sts & SD_DAT_STSNEG)
        {
            printk(KERN_WARNING "MMC: Transmitted data phase:Transmission error\n");
        }
        if (sd_sts & SD_DAT_NORSP)
        {
            printk(KERN_WARNING "MMC: Transmitted CMD phase: No response\n");
        }
        if (sd_sts & SD_CMDRSP_CERR)
        {
            printk(KERN_WARNING "MMC: Response CRC error\n");
        }

        /* set Driving Strength if status error with high speed card(43MHz), except command 1, 5, 8 or no command(0)*/
        if(!(u32Cmd==0x0 || u32Cmd == 0x1 || u32Cmd == 0x5 || u32Cmd == 0x8) && (pIOS_st->clock >= CLK_43MHz))
        {
            //SETREG32(REG_GPIO_C_DRV, BIT_FCIE_PAD_CLKCMDDAT);/* Driving Strength */
            //columbus_mci_SetClockRate(pMMC_st, CLK_36MHz); /* Set Clock Rate to 36MHz*/
            printk(KERN_INFO "MMC: Enable Driving Strenth, Command:0x%x err\n", u32Cmd);
        }
    }
    else
    {
        if (mie_int_sts & SD_CMD_END)
        {
            completed = mstar_mci_handle_cmdrdy(pHost_st);
            //fcie_reg_write(pHost_st, FCIE_MIE_EVENT, SD_CMD_END);
        }

        if (mie_int_sts & (MMA_DATA_END|SD_DATA_END|CARD_DMA_END))
        {
            if (pHost_st->cmd->data->flags & MMC_DATA_WRITE)
            {
                for (i=0; i < LOOP_DELAY_WAIT_COMPLETE; i++)
                {

                    if (0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT) &&
                            0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT) &&
                            0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT) &&
                            0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT))
                    {
                        break;
                    }
                    udelay(1);
                }
                //completed =1;
                if (i==LOOP_DELAY_WAIT_COMPLETE)
                    printk("Error loop\n");

                mstar_mci_handle_transmitted(pHost_st);
                //udelay(200);
                completed =1;
            }
            else
            {
                if (pHost_st->dodma)
                {
                    for (i=0; i < LOOP_DELAY_WAIT_COMPLETE; i++)
                    {

                        if (0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT) &&
                                0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT) &&
                                0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT) &&
                                0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT))
                        {
                            break;
                        }
                        udelay(1);
                    }
                    //completed =1;
                    if (i==LOOP_DELAY_WAIT_COMPLETE)
                        printk("Error loop\n");
                    mstar_mci_post_dma_read(pHost_st);
                    completed =1;
                }
            }
            fcie_reg_write(pHost_st, FCIE_MIE_PATH_CTL,
                               (fcie_reg_read(pHost_st, FCIE_MIE_PATH_CTL) & ~MMA_ENABLE));
        }

        /* Only continue processing if no errors */
        if (mie_int_sts & SD_DATA_END)
        {
            pr_debug("SD_DATA_END,FCIE_TR_BL_CNT=%d\n",
                     fcie_reg_read(pHost_st, FCIE_TR_BL_CNT));

            if (pHost_st->dodma)
            {
                if (pHost_st->cmd->data->flags & MMC_DATA_WRITE)
                {
#if 0
                    for (i=0; i < LOOP_DELAY_WAIT_COMPLETE; i++)
                    {

                        if (0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT) &&
                                0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT) &&
                                0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT) &&
                                0 == fcie_reg_read(pHost_st, FCIE_TR_BL_CNT))
                        {
                            //completed =1;
                            break;
                        }
                        udelay(1);
                    }
                    completed =1;
                    if (i==LOOP_DELAY_WAIT_COMPLETE)
                        printk("Error loop\n");
#endif
                }
            }
            else
            {
                if (pHost_st->cmd->data->flags & MMC_DATA_READ)
                {
                    do_pio_read(pHost_st);
                }
                completed =1;
            }
        }
    }
    //printk("[FCIE_MIE_EVENT]:%X\n", fcie_reg_read(pHost_st, FCIE_MIE_EVENT));
    //printk("[FCIE_MIE_INT_EN]:%X\n", fcie_reg_read(pHost_st, FCIE_MIE_INT_EN));
    if (completed)
    {
        pr_debug("Completed command\n");
        //fcie_reg_write(pHost_st, FCIE_MIE_EVENT, SD_ALL_INT);
        clear_fcie_status(pHost_st, FCIE_MIE_EVENT, SD_ALL_INT);

#ifdef FCIE_TASKLET /* Defer tasks in IRQ, so IRQ can be finished faster */
        tasklet_schedule(&pHost_st->pio_tasklet);
#else /* Do NOT enable Tasklet */
        mstar_mci_completed_command(pHost_st);
#endif /* End ifdef FCIE_TASKLET */
    }
    else
    {
        //printk("\033[34mIrq no complete!\033[0m\n");
        //fcie_reg_write(pHost_st, FCIE_MIE_EVENT, mie_evt);
        clear_fcie_status(pHost_st, FCIE_MIE_EVENT, mie_evt);
    }

    return IRQ_HANDLED;
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_get_ro
+------------------------------------------------------------------------------
| DESCRIPTION : Detect SD card read/write permission
|
| RETURN      : Card access permission. 1 = read-only, 0 = read-write.
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pMMC_st           | x |   | MMC Host structure
+------------------------------------------------------------------------------
*/
static s32 mstar_mci_get_ro(struct mmc_host *pMMC_st)
{
    /* Define Local Variables */
    s32 read_only = 0;
    struct mstar_mci_host *pHost_st = mmc_priv(pMMC_st);

    /* Parameter out-of-bound check */
    if (!pMMC_st)
    {
        printk(KERN_WARNING "mstar_mci_get_ro parameter pMMC_st is NULL\n");
        return -EINVAL;
    }
    /* End Parameter out-of-bound check */

    if (fcie_reg_read(pHost_st, FCIE_SD_STS) & SD_WR_PRO_N)
    {
        read_only = 1;
        pr_debug("%s: card is %s\n", mmc_hostname(pMMC_st),
                 (read_only ? "read-only" : "read-write"));
    }
    else
    {
        pr_debug("%s: host does not support reading read-only "
                 "switch.  Assuming write-enable.\n", mmc_hostname(pMMC_st));
    }

    return read_only;
}

static int mstar_mci_major =   MSTAR_MCI_MAJOR;
static int mstar_mci_minor =   0;
static struct class *mstar_mci_class;

/*
 * Open and close
 */
int mstar_mci_open(struct inode *inode, struct file *filp)
{
    struct mstar_mci_host *dev; /* device information */

    dev = container_of(inode->i_cdev, struct mstar_mci_host, cdev);
    filp->private_data = dev; /* for other methods */

    if (!dev)
    {
        printk("Driver not registered yet!\n");

        return -EFAULT;
    }
    if (down_interruptible(&dev->sem))
        return -ERESTARTSYS;

    // printk("columbus_mci opened.\n");

    up(&dev->sem);

    return 0;          /* success */
}


int mstar_mci_release(struct inode *inode, struct file *filp)
{
    // printk("release columbus mci.\n");

    return 0;
}

/*
 * The ioctl() implementation
 */
int mstar_mci_ioctl(struct inode *inode, struct file *filp,
                       unsigned int cmd, unsigned long arg)
{
    int retval = 0;
    int err = 0;
    struct mstar_mci_host *dev = (struct mstar_mci_host *)filp->private_data;

    /*
     * extract the type and number bitfields, and don't decode
     * wrong cmds: return ENOTTY (inappropriate ioctl) before access_ok()
     */
    if (_IOC_TYPE(cmd) != MSTAR_MCI_IOC_MAGIC)
        return -ENOTTY;
    if (_IOC_NR(cmd) > MSTAR_MCI_IOC_MAXNR)
        return -ENOTTY;

    /*
     * the direction is a bitmask, and VERIFY_WRITE catches R/W
     * transfers. `Type' is user-oriented, while
     * access_ok is kernel-oriented, so the concept of "read" and
     * "write" is reversed
     */
    if (_IOC_DIR(cmd) & _IOC_READ)
        err = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
    else if (_IOC_DIR(cmd) & _IOC_WRITE)
        err =  !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
    if (err)
        return -EFAULT;

    switch (cmd)
    {
        case MSTAR_MCI_CARD_CHG:
            {
                if (wait_event_interruptible(dev->event_q, (dev->card_chg != 0)))
                    return -ERESTARTSYS; /* signal: tell the fs layer to handle it */

                dev->card_chg = 0;

                retval = __put_user(dev->present, (int __user *)arg);

                if (atomic_read(&is_sd_mmc))
                {
                    mmc_detect_change(dev->mmc, msecs_to_jiffies(1));
                }

                break;
            }

        case MSTAR_MCI_CARD_TYPE:
            {
                int card_type;
		int ms_det, sd_det;

		if (dev->present)
		{
			ms_det = !(fcie_reg_read(dev, FCIE_CARD_DET) & MS_DET_N);
			if (ms_det)
			{
				card_type = FCIE_CARD_DET_MS;
			}
			else
			{
				sd_det = !(fcie_reg_read(dev, FCIE_CARD_DET) & SDIO2_DET_N);
				if (sd_det)
				{
					card_type = FCIE_CARD_DET_SD;
				}
				else
				{
					/* Should not be here. */
					card_type = FCIE_CARD_DET_NONE;
				}
			}
		}
		else
		{
			card_type = FCIE_CARD_DET_NONE;
		}

		retval = __put_user(card_type, (int __user *)arg);

                break;
            }

        default:  /* redundant, as cmd was checked against MAXNR */

            return -ENOTTY;
    }

    return retval;

}

struct file_operations mstar_mci_fops =
{
    .owner =    THIS_MODULE,
    .ioctl =    mstar_mci_ioctl,
    .open =     mstar_mci_open,
    .release =  mstar_mci_release,
};

s32 mstar_mci_get_present(struct mmc_host *pMMC_st)
{
    s32 present;
    struct mstar_mci_host *pHost_st = mmc_priv(pMMC_st);

    present = !(fcie_reg_read(pHost_st, FCIE_CARD_DET) & SDIO2_DET_N);

    return present;
}
EXPORT_SYMBOL(mstar_mci_get_present);

s32 mstar_mci_queue_req(struct mmc_host *pMMC_st, struct mmc_request *mrq)
{
    struct mstar_mci_host *pHost_st = mmc_priv(pMMC_st);
    unsigned long flags;

    if (!pMMC_st)
    {
        return (s32)(-1);
    }

    if (!mrq)
    {
        return (s32)(-1);
    }

    local_irq_save(flags);
    list_add_tail(&mrq->queue_node, &pHost_st->queue);
    local_irq_restore(flags);

    return 1;
}
EXPORT_SYMBOL(mstar_mci_queue_req);

s32 mstar_mci_dequeue_req(struct mmc_host *pMMC_st, struct mmc_request *mrq)
{
    struct mstar_mci_host *pHost_st = mmc_priv(pMMC_st);
    unsigned long flags;

    if (!pMMC_st)
    {
        return (s32)(-1);
    }

    if (!mrq)
    {
        return (s32)(-1);
    }

    local_irq_save(flags);
    if (!list_empty(&pHost_st->queue))
    {
        list_del_init(&mrq->queue_node);
    }
    local_irq_restore(flags);

    return 1;
}
EXPORT_SYMBOL(mstar_mci_dequeue_req);

s32 mstar_mci_clearall_req(struct mmc_host *pMMC_st)
{
    struct mstar_mci_host *pHost_st = mmc_priv(pMMC_st);
    unsigned long flags;
    struct mmc_request *mrq = NULL;

    if (!pMMC_st)
    {
        return (s32)(-1);
    }

    local_irq_save(flags);
    while (!list_empty(&pHost_st->queue))
    {
        mrq = list_entry (pHost_st->queue.next, struct mmc_request, queue_node);
        if (mrq)
        {
            list_del_init (&mrq->queue_node);
            printk("%s: clearall req\n", __FUNCTION__);
            mmc_request_done(pMMC_st, mrq);
        }
    }
    local_irq_restore(flags);

    return 1;
}
EXPORT_SYMBOL(mstar_mci_clearall_req);


/*
 * Set up the char_dev structure for this device.
 */
static int mstar_mci_setup_cdev(struct mstar_mci_host *dev)
{
    int result = 0;
    dev_t devno;
    int i;

    /*
     * Get a range of minor numbers to work with, asking for a dynamic
     * major unless directed otherwise at load time.
     */
    if (mstar_mci_major)
    {
        devno = MKDEV(mstar_mci_major, mstar_mci_minor);
        result = register_chrdev_region(devno, MSTAR_MCI_NR_DEVS, "columbus_mci");
    }
    else
    {
        result = alloc_chrdev_region(&devno, mstar_mci_minor, MSTAR_MCI_NR_DEVS,
                                     "columbus_mci");
        mstar_mci_major = MAJOR(devno);
    }

    if (result < 0)
    {
        printk(KERN_WARNING "scull: can't get major %d\n", mstar_mci_major);
        return result;
    }

    init_MUTEX(&dev->sem);

    cdev_init(&dev->cdev, &mstar_mci_fops);
    dev->cdev.owner = THIS_MODULE;
    dev->cdev.ops = &mstar_mci_fops;
    result = cdev_add (&dev->cdev, devno, 1);
    /* Fail gracefully if need be */
    if (result)
        printk(KERN_NOTICE "Error %d adding mstar_mci0", result);

    mstar_mci_class = class_create(THIS_MODULE, "mstar_mci");
    if (IS_ERR(mstar_mci_class))
    {
        cdev_del(&dev->cdev);
        /* cleanup_module is never called if registering failed */
        unregister_chrdev_region(devno, MSTAR_MCI_NR_DEVS);

        return PTR_ERR(mstar_mci_class);
    }

    for (i=0; i<MSTAR_MCI_NR_DEVS; i++)
    {
        device_create(mstar_mci_class, NULL, devno,
                      NULL, "columbus_mci%d", i);
    }

    return result;
}


/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_probe
+------------------------------------------------------------------------------
| DESCRIPTION : Probe for the device
|
| RETURN      : Error Status; Return 0 if no error.
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pDev_st           | x |   | Columbus MCI device structure
+------------------------------------------------------------------------------
*/
static s32 __init mstar_mci_probe(struct platform_device *pDev_st)
{
    /* Define Local Variables */
    struct mmc_host         *pMMC_st = 0;
    struct mstar_mci_host *pHost_st = 0;
    struct resource         *pRes_st = 0;
    s32 ret = 0;
    // u32 temp = 0;

    printk("mstar_mci_probe\n");
    /* Parameter out-of-bound check */
    if (!pDev_st)
    {
        printk(KERN_WARNING "mstar_mci_probe parameter pDev_st is NULL\n");
        return -EINVAL;
    }
    /* End Parameter out-of-bound check */

    printk("Probe MCI devices\n");

    pRes_st = platform_get_resource(pDev_st, IORESOURCE_MEM, 0);
    if (!pRes_st)
    {
        return -ENXIO;
    }

    if (!request_mem_region(pRes_st->start, pRes_st->end - pRes_st->start + 1, DRIVER_NAME))
    {
        return -EBUSY;
    }

    pMMC_st = mmc_alloc_host(sizeof(struct mstar_mci_host), &pDev_st->dev);
    if (!pMMC_st)
    {
        pr_debug("Failed to allocate mmc host\n");
        release_mem_region(pRes_st->start, pRes_st->end - pRes_st->start + 1);
        return -ENOMEM;
    }

    pMMC_st->ops = &sg_mstar_mci_ops;
    pMMC_st->f_min = 300000;
    pMMC_st->f_max = 54000000;
    pMMC_st->ocr_avail = MMC_VDD_32_33 | MMC_VDD_33_34;

    pMMC_st->max_blk_count  = 255; /* 8bit */
    pMMC_st->max_blk_size   = 512; /* sector */
    pMMC_st->max_req_size   = pMMC_st->max_blk_count  * pMMC_st->max_blk_size;
    pMMC_st->max_seg_size   = pMMC_st->max_req_size;

    pMMC_st->max_phys_segs  = 128;
    pMMC_st->max_hw_segs    = 128;

    pHost_st           = mmc_priv(pMMC_st);
    pHost_st->mmc      = pMMC_st;
    pHost_st->tx_buffer = NULL;
    pHost_st->rx_buffer = NULL;
    pHost_st->bus_mode = 0;
    pHost_st->total_length = sd_txbuf_size;
    pMMC_st->caps = pMMC_st->caps
                    | MMC_CAP_4_BIT_DATA     /* CHUN 20090518 for High speed support */
                    | MMC_CAP_MMC_HIGHSPEED
                    | MMC_CAP_SD_HIGHSPEED;

#ifdef FCIE_TASKLET /* Defer tasks in IRQ, so IRQ can be finished faster */
    tasklet_init(&pHost_st->pio_tasklet, mstar_mci_completed_command,
                 (unsigned long) pHost_st);
#endif  /* End ifdef FCIE_TASKLET */
    /***** Map I/O region *****/
    #if 1
    pHost_st->baseaddr = (void __iomem *)FCIE0_BASE;
    #else
    pHost_st->baseaddr = ioremap(pRes_st->start, pRes_st->end - pRes_st->start + 1);
    if (!pHost_st->baseaddr)
    {
        printk("ioremap fail\n");
        mmc_free_host(pMMC_st);
        release_mem_region(pRes_st->start, pRes_st->end - pRes_st->start + 1);
        return -ENOMEM;
    }
    #endif

    //fcie_writew(fcie_readw(REG_PM_IRQ_MASK)&(~(BIT10|BIT11)), REG_PM_IRQ_MASK);
    /***** Reset hardware *****/
    mstar_mci_enable(pHost_st);

    // disable strong-driving
    //CLRREG32(REG_GPIO_C_DRV, BIT_FCIE_PAD_CLKCMDDAT);

    /***** Allocate the MCI interrupt *****/
    pHost_st->irq = platform_get_irq(pDev_st, 0);
    ret = request_irq(pHost_st->irq, mstar_mci_irq,
                      IRQF_SHARED, DRIVER_NAME, pHost_st);
    if (ret)
    {
        printk(KERN_ERR "Columbus MMC: Failed to request MCI interrupt\n");
        mmc_free_host(pMMC_st);
        iounmap(pHost_st->baseaddr);
        release_mem_region(pRes_st->start, pRes_st->end - pRes_st->start + 1);
        return ret;
    }

    pr_debug("Columbus MMC: request MCI interrupt = %d\n", pHost_st->irq);
    platform_set_drvdata(pDev_st, pMMC_st);

    /***** Add host to MMC layer *****/
    device_init_wakeup(&pDev_st->dev, 1);

    // winder add.
    // fcie_writew(0x0800|fcie_readw(0xA0007E10), 0xA0007E10); // Open the USB clock source, bit_11 reg_force_pll_on
    init_waitqueue_head(&(pHost_st->event_q));
    pHost_st->card_chg = 0;
    mstar_mci_setup_cdev(pHost_st);

    //chun 20090611 enable SD_STS_EN
    //clear_fcie_status(pHost_st, FCIE_CARD_EVENT, fcie_reg_read(pHost_st, FCIE_CARD_EVENT));
    //fcie_reg_write(pHost_st, FCIE_CARD_EVENT, SD_SDIO2_CHG);
    fcie_reg_write(pHost_st, FCIE_CARD_INT_EN, fcie_reg_read(pHost_st, FCIE_CARD_INT_EN)|SDIO2_STS_EN);
    pr_debug("[FCIE_CARD_INT_EN]:%x\n", fcie_reg_read(pHost_st, FCIE_CARD_INT_EN));


    mmc_add_host(pMMC_st);

    pr_debug("Added MCI driver\n");
    INIT_LIST_HEAD(&pHost_st->queue);

    return 0;
}

static void mstar_mci_reinit(struct mstar_mci_host *pHost_st)
{
#if 0
    /* Reset IP */
    mstar_mci_reset(pHost_st);
#endif
    /***** Reset hardware *****/
    mstar_mci_enable(pHost_st);
#if 0
    // disable strong-driving
    CLRREG32(REG_GPIO_C_DRV, BIT_FCIE_PAD_CLKCMDDAT);
    // winder add.
    fcie_writew(0x0800|fcie_readw(0xA0007E10), 0xA0007E10); // Open the USB clock source, bit_11 reg_force_pll_on
    //chun 20090611 enable SD_STS_EN
    fcie_reg_write(pHost_st, FCIE_CARD_EVENT, SD_STS_CHG);
    fcie_reg_write(pHost_st, FCIE_CARD_INT_EN, fcie_reg_read(pHost_st, FCIE_CARD_INT_EN)|SD_STS_EN);
#endif
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_remove
+------------------------------------------------------------------------------
| DESCRIPTION : Remove a device
|
| RETURN      : Error Status; Return 0 if no error.
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pDev_st           | x |   | Columbus MCI device structure
+------------------------------------------------------------------------------
*/
static s32 __exit mstar_mci_remove(struct platform_device *pDev_st)
{
    /* Define Local Variables */
    struct mmc_host          *pMMC_st   = platform_get_drvdata(pDev_st);
    struct mstar_mci_host  *pHost_st  = mmc_priv(pMMC_st);
    struct resource          *pRes_st   = 0;

    /* Parameter out-of-bound check */
    if (!pDev_st)
    {
        printk(KERN_WARNING "mstar_mci_remove parameter pDev_st is NULL\n");
        return -EINVAL;
    }
    /* End Parameter out-of-bound check */

    if (!pMMC_st)
    {
        return -1;
    }

    mstar_mci_disable(pHost_st);
    mmc_remove_host(pMMC_st);
    free_irq(pHost_st->irq, pHost_st);

    //clk_disable(pHost_st->mci_clk);         /* Disable the peripheral clock */
    //clk_put(pHost_st->mci_clk);

    iounmap(pHost_st->baseaddr);
    pRes_st = platform_get_resource(pDev_st, IORESOURCE_MEM, 0);
    release_mem_region(pRes_st->start, pRes_st->end - pRes_st->start + 1);

    mmc_free_host(pMMC_st);
    platform_set_drvdata(pDev_st, NULL);
    pr_debug("MCI Removed\n");

    return 0;
}


#ifdef CONFIG_PM
/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_suspend
+------------------------------------------------------------------------------
| DESCRIPTION : MCI Suspend
|
| RETURN      : Error Status; Return 0 if no error.
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pDev_st           | x |   | Columbus MCI device structure
+------------------------------------------------------------------------------
| state              | x |   | Power Management Transition State
+------------------------------------------------------------------------------
*/
static s32 mstar_mci_suspend(struct platform_device *pDev_st, pm_message_t state)
{
    /* Define Local Variables */
    struct mmc_host         *pMMC_st    = platform_get_drvdata(pDev_st);
    s32 ret = 0;

    if (pMMC_st)
    {
        ret = mmc_suspend_host(pMMC_st, state);
    }

    pr_debug("MCI suspend ret=%d\n", ret);
    return ret;
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_resume
+------------------------------------------------------------------------------
| DESCRIPTION : MCI Resume
|
| RETURN      : Error Status; Return 0 if no error.
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| *pDev_st           | x |   | Columbus MCI device structure
+------------------------------------------------------------------------------
*/
static s32 mstar_mci_resume(struct platform_device *pDev_st)
{
    /* Define Local Variables */
    struct mmc_host          *pMMC_st   = platform_get_drvdata(pDev_st);
    s32 ret = 0;

    if (pMMC_st)
    {
        ret = mmc_resume_host(pMMC_st);
    }

    pr_debug("MCI resume ret=%d\n", ret);
    return ret;
}

#else
/*
 * Current driver does not support following two functions,
 * therefore set them to NULL.
 */

#define mstar_mci_suspend    NULL
#define mstar_mci_resume     NULL

#endif  /* End ifdef CONFIG_PM */


/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_add_device_mmc
+------------------------------------------------------------------------------
| DESCRIPTION : Register MMC device to kernel
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
| mmc_id             | x |   | MMC ID number
+------------------------------------------------------------------------------
*/
void __init mstar_add_device_mmc(s16 mmc_id)
{
    printk(KERN_INFO "%s entered\n",__FUNCTION__);
    platform_device_register(&sg_mstar_mmc_device_st);
}



/******************************************************************************
 * Define Static Global Variables
 ******************************************************************************/

/* Columbus MCI Driver (Include Remove, Suspend, and Resume) */
static struct platform_driver sg_mstar_mci_driver =
{
    .remove = __exit_p(mstar_mci_remove),
    .suspend = mstar_mci_suspend,
    .resume = mstar_mci_resume,
    .probe = mstar_mci_probe,
    .driver  =
    {
        .name = DRIVER_NAME,
        .owner = THIS_MODULE,
    },
};

/******************************************************************************
 * Init & Exit Modules
 ******************************************************************************/

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_init
+------------------------------------------------------------------------------
| DESCRIPTION : MCI module init.
|
| RETURN      : Error Status; Return 0 if no error.
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
|                    |   |   |
+------------------------------------------------------------------------------
*/
static s32 __init mstar_mci_init(void)
{
    // int err = 0;

    printk(KERN_INFO "Columbus SD Card Init: %s.\n",__FUNCTION__);

    platform_device_register(&sg_mstar_mmc_device_st);
    return platform_driver_register(&sg_mstar_mci_driver);
}

/*
+------------------------------------------------------------------------------
| FUNCTION    : columbus_mci_exit
+------------------------------------------------------------------------------
| DESCRIPTION : MCI module exit.
|
| RETURN      : No return value
|
+------------------------------------------------------------------------------
| Parameter Name     |IN |OUT|                   Usage
|--------------------+---+---+-------------------------------------------------
|                    |   |   |
+------------------------------------------------------------------------------
*/
static void __exit mstar_mci_exit(void)
{
    platform_driver_unregister(&sg_mstar_mci_driver);
}

module_init(mstar_mci_init);
module_exit(mstar_mci_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Mstar Multimedia Card Interface driver");
MODULE_AUTHOR("Benson.Hsiao");
