////////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2008-2009 MStar Semiconductor, Inc.
// All rights reserved.
//
// Unless otherwise stipulated in writing, any and all information contained
// herein regardless in any format shall remain the sole proprietary of
// MStar Semiconductor Inc. and be kept in strict confidence
// ("MStar Confidential Information") by the recipient.
// Any unauthorized act including without limitation unauthorized disclosure,
// copying, use, reproduction, sale, distribution, modification, disassembling,
// reverse engineering and compiling of the contents of MStar Confidential
// Information is unlawful and strictly prohibited. MStar hereby reserves the
// rights to any and all damages, losses, costs and expenses resulting therefrom.
//
////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/// @file   mdrv_gflip.c
/// @brief  MStar gflip Interface
/// @author MStar Semiconductor Inc.
/// @attention
/// <b><em></em></b>
///////////////////////////////////////////////////////////////////////////////////////////////////

#define _MDRV_GFLIP_C

//=============================================================================
// Include Files
//=============================================================================
#include <linux/kernel.h>
#include <linux/jiffies.h>
#include <linux/sched.h>

#include "mdrv_mstypes.h"
#include "mdrv_gflip.h"
#include "mhal_gflip.h"

//=============================================================================
// Compile options
//=============================================================================


//=============================================================================
// Local Defines
//=============================================================================
#define MAX_FLIP_ADDR_FIFO 	(0x10)
#define FLIP_INTERRUPT_TIMEOUT   (100)  //In MS

//=============================================================================
// Debug Macros
//=============================================================================
#define GFLIP_DEBUG
#ifdef GFLIP_DEBUG
    #define GFLIP_PRINT(fmt, args...)      printk("[GFlip (Driver)][%05d] " fmt, __LINE__, ## args)
    #define GFLIP_ASSERT(_cnd, _fmt, _args...)                   \
                                    if (!(_cnd)) {              \
                                        GFLIP_PRINT(_fmt, ##_args);  \
                                    }
#else
    #define GFLIP_PRINT(_fmt, _args...)
    #define GFLIP_ASSERT(_cnd, _fmt, _args...)
#endif

//=============================================================================
// Macros
//=============================================================================

//=============================================================================
// Local Variables
//=============================================================================
static volatile GFLIP_INFO _GFlipInfo[MAX_GOP_SUPPORT][MAX_GOP_GWIN][MAX_FLIP_ADDR_FIFO];
static volatile MS_U32 _u32GFlipInfoReadPtr[MAX_GOP_SUPPORT][MAX_GOP_GWIN];
static volatile MS_U32 _u32GFlipInfoWritePtr[MAX_GOP_SUPPORT][MAX_GOP_GWIN];
static volatile MS_BOOL _bGFlipInVsyncLimitation[MAX_GOP_SUPPORT] = { FALSE };

static GFLIP_DWININT_INFO _GFlipDWinIntInfo = { 0x0 };
static MS_U32 _u32GFlipLatestIntTicket; //The jiffies when Interrupt happen
//=============================================================================
// Global Variables
//=============================================================================

//=============================================================================
// Local Function Prototypes
//=============================================================================
static void _MDrv_GFLIP_ResetFlipInfoPtr(MS_U32 u32GopIdx);

//=============================================================================
// Local Function
//=============================================================================
//-------------------------------------------------------------------------------------------------
/// Reset Flip Info Reader/Writer ptr.
/// @param  u32GopIdx    \b IN: the idx of ptr's owner(gop)
/// @return void
/// @attention
/// <b>[OBAMA] <em></em></b>
//-------------------------------------------------------------------------------------------------
void _MDrv_GFLIP_ResetFlipInfoPtr(MS_U32 u32GopIdx)
{
    MS_U16 u16Idx;

    for(u16Idx=0; u16Idx<MAX_GOP_GWIN; u16Idx++)
    {
        _u32GFlipInfoReadPtr[u32GopIdx][u16Idx] = 0;
        _u32GFlipInfoWritePtr[u32GopIdx][u16Idx] = 0;
    }
}

//=============================================================================
// GFLIP Driver Function
//=============================================================================

//-------------------------------------------------------------------------------------------------
/// GFlip Init(reset flip info ptr, interrupt enable, etc..)
/// @param  u32GopIdx                  \b IN:the idx of gop which need be initialized.
/// @return TRUE: success
/// @attention
/// <b>[OBAMA] <em></em></b>
//-------------------------------------------------------------------------------------------------
MS_BOOL MDrv_GFLIP_Init(MS_U32 u32GopIdx)
{
    //reset flip Info:
    _MDrv_GFLIP_ResetFlipInfoPtr(u32GopIdx);
        //enable interrupt:
    MHal_GFLIP_IntEnable(u32GopIdx,TRUE);

    return TRUE;
}

//-------------------------------------------------------------------------------------------------
/// GFlip DeInit(interrupt disable)
/// @param  u32GopIdx                  \b IN:the idx of gop which need be de-initialized.
/// @return TRUE: success
/// @attention
/// <b>[OBAMA] <em></em></b>
//-------------------------------------------------------------------------------------------------
MS_BOOL MDrv_GFLIP_DeInit(MS_U32 u32GopIdx)
{
    //disable interrupt:
    MHal_GFLIP_IntEnable(u32GopIdx,FALSE);

    return TRUE;
}

//-------------------------------------------------------------------------------------------------
/// Clear Irq
/// @param  void
/// @return TRUE: success
/// @attention
/// <b>[OBAMA] <em></em></b>
//-------------------------------------------------------------------------------------------------
MS_BOOL MDrv_GFLIP_ClearIRQ(void)
{
    MS_U32 u32GopIdx;

    MHal_GFLIP_ClearDWINIRQ(&_GFlipDWinIntInfo);

    for(u32GopIdx=0; u32GopIdx<MAX_GOP_SUPPORT; u32GopIdx++)
    {
        if(MHal_GFLIP_IsVSyncInt(u32GopIdx) == FALSE)
        {
            continue;
        }

        MHal_GFLIP_IntEnable(u32GopIdx, FALSE);

        if(MHal_GFLIP_IsVSyncInt(u32GopIdx) == TRUE)
        {   //Wow...The Vsync Issue Happened...
            MHal_GFLIP_HandleVsyncLimitation(u32GopIdx); //different chip maybe has different handle.
            _bGFlipInVsyncLimitation[u32GopIdx] = TRUE;
            continue;
        }

        MHal_GFLIP_IntEnable(u32GopIdx, TRUE);
    }

    //Record current Time Ticket:
    _u32GFlipLatestIntTicket = jiffies;
    return TRUE;
}

//-------------------------------------------------------------------------------------------------
/// Process Irq
/// @param  void
/// @return TRUE: success kickoff flip to gop
/// @return FALSE: no flip kickoff to gop
/// @attention
/// <b>[OBAMA] <em>
///         step1: check src of vsync. interrupt
///         step2: check if any flip request in queue with this Gop
///         step3: check if TAG back
///         step4: set flip to GOP
///         step5: set ReadPtr to next, this entry consumed!
/// </em></b>
//-------------------------------------------------------------------------------------------------
MS_BOOL MDrv_GFLIP_ProcessIRQ(void)
{
    MS_U32 u32GopIdx, u32GwinIdx;

    for(u32GopIdx=0; u32GopIdx<MAX_GOP_SUPPORT; u32GopIdx++)
    {
        //step 1: check src of vsync. interrupt
        if(MHal_GFLIP_IsVSyncInt(u32GopIdx) == FALSE)
        {
            continue;
        }

        for(u32GwinIdx=0; u32GwinIdx<MAX_GOP_GWIN; u32GwinIdx++)
        {
            //step 2:check if any flip request in queue with this Gop:
            if(_u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx] == _u32GFlipInfoWritePtr[u32GopIdx][u32GwinIdx]) //no any flip request in queue
            {
                continue;
            }

            //step 3: if get queue, check if TAG back.
            if(MHal_GFLIP_IsTagIDBack(_GFlipInfo[u32GopIdx][u32GwinIdx][_u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx]].u32TagId) == FALSE)
            {
                continue;
            }

            //Has Flip Request, and Tag Returned! we need programming flip address:
            //step 4: if Tag Back: set flip to GOP.
            MHal_GFLIP_SetFlipToGop(u32GopIdx, u32GwinIdx, _GFlipInfo[u32GopIdx][u32GwinIdx][_u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx]].u32Addr);

            //Step 5: set ReadPtr to next, this entry consumed!
            _u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx] = (_u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx]+1)%MAX_FLIP_ADDR_FIFO;
        }
    }

    return TRUE;
}

//-------------------------------------------------------------------------------------------------
/// Restore HW Limitation -- Vsync Limitation.
/// @return TRUE: DDI call success
/// @attention
/// <b>[OBAMA] <em>The Signal of GOP Dst from IP/VE restored and we re-enable GFlip. </em></b>
//-------------------------------------------------------------------------------------------------
MS_BOOL MDrv_GFLIP_RestoreFromVsyncLimitation(void)
{
    MS_U32 u32GopIdx;

    for(u32GopIdx=0; u32GopIdx<MAX_GOP_SUPPORT; u32GopIdx++)
    {
        if(TRUE == _bGFlipInVsyncLimitation[u32GopIdx])
        {
            MHal_GFLIP_RestoreFromVsyncLimitation(u32GopIdx);
            _bGFlipInVsyncLimitation[u32GopIdx] = FALSE;
        }
    }

    return TRUE;
}

//-------------------------------------------------------------------------------------------------
/// Get DWin Interrupt Info
/// @param  pGFlipDWinIntInfo       \b OUT: the dwin interrupt info
/// @param  bResetDWinIntInfo       \b IN: reset dwin interrupt infor or not
/// @return TRUE: DDI call success
/// @attention
/// <b>[OBAMA] <em>Get DWin Interrupt Info and reset it if requested. </em></b>
//-------------------------------------------------------------------------------------------------
MS_BOOL MDrv_GFLIP_GetDWinIntInfo(GFLIP_DWININT_INFO *pGFlipDWinIntInfo, MS_BOOL bResetDWinIntInfo)
{
    pGFlipDWinIntInfo->u8DWinIntInfo = _GFlipDWinIntInfo.u8DWinIntInfo;

    if(bResetDWinIntInfo)
    {
        _GFlipDWinIntInfo.u8DWinIntInfo = 0x0;
    }

    return TRUE;
}

//-------------------------------------------------------------------------------------------------
/// Set Flip Info(flip request) to GFlip
/// @param  u32GopIdx       \b IN: the gop which flip kickoff to
/// @param  u32GwinIdx       \b IN: the gwin which flip kickoff to
/// @param  u32Addr             \b IN: the flip address which will be kickoff
/// @param  u32TagId            \b IN: the TagId need be waited before flip
/// @param  u32QEntry           \b IN: the queued flips, if current queued flips >= u32QEntry, won't set this flip request to gop
/// @param  u32QEntry           \b out: the current queued flips in GFLIP
/// @param  u32Result   \b out: kickoff flip info success(TRUE) or fail(FALSE)
/// @return TRUE: DDI call success
/// @attention
/// <b>[OBAMA] <em>Flip when vsync interrupt, and the DDI used for set flip info to GFlip. </em></b>
//-------------------------------------------------------------------------------------------------
MS_BOOL MDrv_GFLIP_SetFlipInfo(MS_U32 u32GopIdx, MS_U32 u32GwinIdx, MS_U32 u32Addr, MS_U32 u32TagId, MS_U32 * u32QEntry, MS_U32 *u32Result)
{
    MS_U32 u32NextWritePtr;
    MS_U32 u32QCnt = 0;

    if ((u32GopIdx >= MAX_GOP_SUPPORT) || (u32GwinIdx >= MAX_GOP_GWIN)) //overflow
    {
        *u32Result = FALSE;
        return TRUE;
    }

    if(TRUE == _bGFlipInVsyncLimitation[u32GopIdx])
    { //in handling vsync limitation status, just print the error and return true to avoid block app:
        printk("\n Error in MDrv_GFLIP_SetFlipInfo. No Signal for GOP Clock Source and GFlip Vsync Interrupt disabled!!!");
        *u32Result = TRUE;
        return TRUE;
    }

    u32NextWritePtr = (_u32GFlipInfoWritePtr[u32GopIdx][u32GwinIdx]+1) % MAX_FLIP_ADDR_FIFO;
    if(u32NextWritePtr != _u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx]) //can write
    {
        if(_u32GFlipInfoWritePtr[u32GopIdx][u32GwinIdx] >= _u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx])
        {
            u32QCnt = _u32GFlipInfoWritePtr[u32GopIdx][u32GwinIdx] - _u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx];
        }
        else
        {
            u32QCnt = MAX_FLIP_ADDR_FIFO + _u32GFlipInfoWritePtr[u32GopIdx][u32GwinIdx] - _u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx];
        }

        if( u32QCnt >= *u32QEntry )
        {
            *u32QEntry = u32QCnt;
            *u32Result = FALSE;
            return TRUE;
        }

        _GFlipInfo[u32GopIdx][u32GwinIdx][_u32GFlipInfoWritePtr[u32GopIdx][u32GwinIdx]].u32Addr = (u32Addr >> GFLIP_ADDR_ALIGN_RSHIFT);
        _GFlipInfo[u32GopIdx][u32GwinIdx][_u32GFlipInfoWritePtr[u32GopIdx][u32GwinIdx]].u32TagId= u32TagId;
        _u32GFlipInfoWritePtr[u32GopIdx][u32GwinIdx] = u32NextWritePtr;

        //Hold on when New QCnt can't match requested QCnt:
        //Fix issue if we set Flip Info and back which leads to App write to on-show window.
        while(TRUE)
        {
            if(_u32GFlipInfoWritePtr[u32GopIdx][u32GwinIdx] >= _u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx])
            {
                u32QCnt = _u32GFlipInfoWritePtr[u32GopIdx][u32GwinIdx] - _u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx];
            }
            else
            {
                u32QCnt = MAX_FLIP_ADDR_FIFO + _u32GFlipInfoWritePtr[u32GopIdx][u32GwinIdx] - _u32GFlipInfoReadPtr[u32GopIdx][u32GwinIdx];
            }

            if( u32QCnt < *u32QEntry )
            {
                break;
            }

            if(time_after(jiffies, (msecs_to_jiffies(FLIP_INTERRUPT_TIMEOUT)+_u32GFlipLatestIntTicket)))
            {
                break;
            }

            schedule();
        }

        //since do not do the sync. with mutex/spinlock, the return of avaiable queue number maybe not accurate.
        *u32QEntry = u32QCnt + 1;
        *u32Result = TRUE;

        return TRUE;
    }
    else
    {
        *u32QEntry = MAX_FLIP_ADDR_FIFO-1;
        *u32Result = FALSE;
        return TRUE;
    }
}


