/*
 * Copyright (C) 2008, OGAWA Hirofumi
 * Released under GPL v2.
 */

#include <linux/buffer_head.h>
#include <linux/thread_info.h>
#include "exfat.h"
#include "bitmap.h"

/* Carry-Save Adder */
#define CSA(h, l, a, b, c)	do {		\
	unsigned long u = a ^ b;		\
	unsigned long v = c;			\
	h = (a & b) | (u & v);			\
	l = u ^ v;				\
} while (0)

/* TODO: maybe, should merge to bitmap_weight(). */
static int hweight_array(const unsigned long *bitmap, int n)
{
	unsigned long ones, twos, twosA, twosB, fours, foursA, foursB, eights;
	int counts, i;

	counts = 0;
	ones = twos = fours = 0;
	for (i = 0; i <= n - 8; i += 8) {
		CSA(twosA, ones, ones, bitmap[i], bitmap[i + 1]);
		CSA(twosB, ones, ones, bitmap[i + 2], bitmap[i + 3]);
		CSA(foursA, twos, twos, twosA, twosB);
		CSA(twosA, ones, ones, bitmap[i + 4], bitmap[i + 5]);
		CSA(twosB, ones, ones, bitmap[i + 6], bitmap[i + 7]);
		CSA(foursB, twos, twos, twosA, twosB);
		CSA(eights, fours, fours, foursA, foursB);
		counts += hweight_long(eights);
	}
	counts = 8 * counts + 4 * hweight_long(fours)
		+ 2 * hweight_long(twos) + hweight_long(ones);
	for (; i < n; i++)
		counts += hweight_long(bitmap[i]);
	return counts;
}

#if BITS_PER_LONG == 64
typedef __le64			__lelong;
#define lelong_to_cpu(x)	le64_to_cpu(x);
#else
typedef __le32			__lelong;
#define lelong_to_cpu(x)	le32_to_cpu(x);
#endif

/* Hack: want bitmap_weight_le() */
static int exfat_hweight_le(const void *buf, int bits)
{
	const unsigned long *bitmap = buf;
	int counts;

	/* counts = bitmap_weight(bitmap, bits & ~(BITS_PER_LONG - 1)); */
	counts = hweight_array(bitmap, BIT_WORD(bits & ~(BITS_PER_LONG - 1)));
	if (bits & (BITS_PER_LONG - 1)) {
		unsigned long word;
		word = lelong_to_cpu(*(__lelong *)&bitmap[BIT_WORD(bits) + 1]);
		counts += hweight_long(word & BITMAP_LAST_WORD_MASK(bits));
	}
	return counts;
}

static int exfat_count_free_clusters(struct inode *inode)
{
	struct super_block *sb = inode->i_sb;
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct address_space *mapping = inode->i_mapping;
	pgoff_t index;
	u32 bits, bits_left, used_bits;

	/* FIXME: readahead all data of bitmap */

	index = 0;
	used_bits = 0;
	bits_left = sbi->total_clusters;
	while (bits_left > 0) {
		struct page *page;
		void *kaddr;

		page = read_mapping_page(mapping, index, NULL);
		if (IS_ERR(page))
			return PTR_ERR(page);

		bits = min_t(u32, bits_left, PAGE_CACHE_SIZE * 8);
		kaddr = kmap_atomic(page, KM_USER0);

		used_bits += exfat_hweight_le(kaddr, bits);

		kunmap_atomic(kaddr, KM_USER0);
		page_cache_release(page);

		bits_left -= bits;
		index++;
	}

	sbi->free_clusters = sbi->total_clusters - used_bits;
	PRINTK("%s: free_clusters %u\n", __func__, sbi->free_clusters);

	return 0;
}

int exfat_setup_bitmap(struct super_block *sb, u32 clusnr, u64 i_size)
{
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct inode *inode;
	int err;

	if (!clusnr) {
		PRINTK(KERN_ERR "exFAT: free space bitmap isn't available\n");
		return -EINVAL;
	}
	if (i_size < DIV_ROUND_UP(sbi->total_clusters, 8)) {
		PRINTK(KERN_ERR "exFAT: free space bitmap is too small\n");
		return -EINVAL;
	}

	inode = exfat_new_internal_inode(sb, EXFAT_BITMAP_INO, 0, clusnr,
					 i_size);
	if (!inode)
		return -ENOMEM;

	err = exfat_count_free_clusters(inode);
	if (err)
		return err;

	sbi->bitmap_inode = inode;

	return 0;
}

void exfat_free_bitmap(struct exfat_sb_info *sbi)
{
	if (sbi->bitmap_inode) {
		iput(sbi->bitmap_inode);
		sbi->bitmap_inode = NULL;
	}
}

/*
	@sb:super block 
	@dclusnr:set cluster
	@clus_len:cluster length
	@set: set or clear(TRUE for set)

*/
int exfat_bits_set(struct super_block *sb,u32 dclusnr,u32 clus_len,int set)
{
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct exfat_inode_info *bitmap_exi = EXFAT_I(sbi->bitmap_inode);

	struct buffer_head *bh;	
	u8 *kaddr = NULL;
	unsigned long index;
	u32 offset,available_bits,blocksize,blocks = 0;
	int isize = sbi->bitmap_inode->i_size;		//bitmap inode size
	int idisplay = 0;

	if((dclusnr+clus_len) > isize << 3){
		printk("bits_set overflow \n");
		return -1;
	}

	blocksize = 1<<sbi->bitmap_inode->i_blkbits;

	dclusnr -= EXFAT_START_ENT;

	index = (unsigned long)(dclusnr / (blocksize<<3));

	offset = dclusnr & ((blocksize<<3) - 1);

	MY_DEBUG(idisplay,"index=%d,offset=%d,dclusnr=%d \n",index,offset,dclusnr);

	while(clus_len){
		//get the block
		sector_t blocknr;		//current blocknr
		blocknr = exfat_clus_to_blknr(sbi, bitmap_exi->clusnr) + index;

		bh = sb_bread(sb,blocknr);
		if (!bh) 
			return -EIO;
		
		available_bits = blocksize*8 -offset;
		if(available_bits>clus_len)
			available_bits = clus_len;
		clus_len -= available_bits;

		kaddr = (u8 *)bh->b_data;
		MY_DEBUG(idisplay,"available_bits=%d,kaddr[0]=%x \n",available_bits,kaddr[0]);

		while((offset&7) && available_bits)
		{
			MY_DEBUG(idisplay,"set=%d,offset>>3 = %d, 1<<(offset&7) = %x \n",set,offset>>3,1<<(offset&7));
			if(set)
				kaddr[offset>>3] |= 1<<(offset&7);
			else
				kaddr[offset>>3] &=  ~(1<<(offset&7));
			offset++;
			available_bits--;
			MY_DEBUG(idisplay,"after set  kaddr[offset>>3] = %x \n",kaddr[offset>>3]);
			
		}
		
		while(available_bits>=8)
		{
			if(set)
				kaddr[offset>>3] = 0xff;
			else
				kaddr[offset>>3]&= 0x00;
			offset += 8;
			available_bits -= 8;
		}
		
		while(available_bits)
		{
			if(set)
				kaddr[offset>>3] |= 1<<(offset&7);
			else
				kaddr[offset>>3] &=  ~(1<<(offset&7));
			offset++;
			available_bits--;
		} 


		offset = 0;
		index ++;

		mark_buffer_dirty(bh);
//		sync_dirty_buffer(bh);
		brelse(bh);
	}

	return 0;
}

static inline void exfat_unmap_page(struct page *page)
{
	kunmap(page);
	page_cache_release(page);
}

static inline struct page *exfat_map_page(struct address_space *mapping,
		unsigned long index)
{
	struct page *page = read_mapping_page(mapping, index, NULL);

	if (!IS_ERR(page)) {
		kmap(page);
		if (!PageError(page))
			return page;
		exfat_unmap_page(page);
		return ERR_PTR(-EIO);
	}
	return page;
}

//exfat_bits_set_with_page
int exfat_bits_set_with_page(struct super_block *sb,u32 dclusnr,u32 clus_len,int set)
{
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct address_space *mapping;
	struct exfat_inode_info *bitmap_exi = EXFAT_I(sbi->bitmap_inode);
	
	struct page *page;
	u8 *kaddr;

	int idisplay = 0;
	
	unsigned long index;
	u32 offset,available_bits;
	int isize = sbi->bitmap_inode->i_size;		//bitmap inode size

	if((dclusnr+clus_len) > isize << 3){
		printk("bits_set overflow \n");
		return -1;
	}

	dclusnr -= EXFAT_START_ENT;

	mapping = sbi->bitmap_inode->i_mapping;	//bitmap address space

	index = (unsigned long)(dclusnr / (PAGE_CACHE_SIZE<<3));

	offset = dclusnr & ((PAGE_CACHE_SIZE<<3) - 1);

	MY_DEBUG(idisplay,"index=%d,offset=%d,dclusnr=%d \n",index,offset,dclusnr);
	while(clus_len){
//		page = read_mapping_page(mapping, index, NULL);
		page = exfat_map_page(mapping,index);
		if (IS_ERR(page))
			return -1;

 		kaddr = page_address(page);
//		kaddr = kmap_atomic(page, KM_USER0);
		
		available_bits = PAGE_CACHE_SIZE*8 -offset;
		if(available_bits>clus_len)
			available_bits = clus_len;
		clus_len -= available_bits;

//		lock_page(page);
		if (unlikely(page_has_buffers(page))) 
		{
			struct buffer_head *bh, *head;
			unsigned int  bh_ofs;
			unsigned bh_size = 1<<sbi->bitmap_inode->i_blkbits;
			   
			bh = head = page_buffers(page);
			do {
				bh_ofs = bh_offset(bh);
				if (bh_ofs + bh_size <= (offset>>3))
					continue;
				if (bh_ofs >= ((offset+available_bits+7)>>3))
					break;
				set_buffer_dirty(bh);
			} while ((bh = bh->b_this_page) != head);
		}

		MY_DEBUG(idisplay,"available_bits=%d,kaddr[0]=%x \n",available_bits,kaddr[0]);
		
		set_page_dirty(page);

		while((offset&7) && available_bits)
		{
			MY_DEBUG(idisplay,"set=%d,offset>>3 = %d, 1<<(offset&7) = %x \n",set,offset>>3,1<<(offset&7));
			if(set)
				kaddr[offset>>3] |= 1<<(offset&7);
			else
				kaddr[offset>>3] &=  ~(1<<(offset&7));
			offset++;
			available_bits--;
			MY_DEBUG(idisplay,"after set  kaddr[offset>>3] = %x \n",kaddr[offset>>3]);
			
		}
		
		while(available_bits>=8)
		{
			if(set)
				kaddr[offset>>3] = 0xff;
			else
				kaddr[offset>>3]&= 0x00;
			offset += 8;
			available_bits -= 8;
		}
		
		while(available_bits)
		{
			if(set)
				kaddr[offset>>3] |= 1<<(offset&7);
			else
				kaddr[offset>>3] &=  ~(1<<(offset&7));
			offset++;
			available_bits--;
		} 
		

		offset = 0;
		index ++;
		flush_dcache_page(page);
		exfat_unmap_page(page);

//		kunmap_atomic(kaddr, KM_USER0);
//		unlock_page(page);
//		page_cache_release(page);

	}

	return 0;
}


/**
 * ffz - Find the first 0 bit in an byte
 * @x:
 *
 * Description...
 *
 * Returns:
 */
static u32 myffz(unsigned char x)
{
	u32 r = 0x01,temp;
	int i;
	
	if (!x)
		return 0;
	for(i = 0 ; i < 8 ; i ++){
//		temp = r << i;
//		printk("i=%d: x=%x,temp=%x,x & temp = %x\n",i,x,temp,x&temp);
		if(!(x & (r << i)))
			return i;
	}

	return 0xFFFF;
}


/*
	@sb:super block
	@srch_beg:search begin cluster
	@found_clus:return found clusnr

	return 0 to success, or -1 to fail
*/
int exfat_find_1free_cluster(struct super_block *sb,u32 srch_beg,u32 *found_clus)
{
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct exfat_inode_info *bitmap_exi = EXFAT_I(sbi->bitmap_inode);
	struct buffer_head *bh;	
	u32 index,i,blocksize;
	u32 bit,pos,total_bits,cur_srch,temppos;
	u8* byte,*kaddr;

	static int idisplay = 0;
	static int aaa = 140930;
	
	int isize = sbi->bitmap_inode->i_size;		//bitmap inode size	
	total_bits = sbi->total_clusters;				//total bits
	MY_DEBUG(idisplay,"isize=%d,total_size=%d \n",isize,total_bits);
//	DEBUG_ON((isize<<3)!= total_bits,"bug");
	blocksize = 1<<sbi->bitmap_inode->i_blkbits;
	
	if(srch_beg< EXFAT_START_ENT || srch_beg >= total_bits)
		return -1;	
		
	if(srch_beg < sbi->last_free_clus)
		srch_beg = sbi->last_free_clus + 1; 

	srch_beg -= EXFAT_START_ENT;	
	
	index = (unsigned long)(srch_beg / (blocksize<<3));	
	pos = srch_beg - index*(blocksize<<3);	//start offset in this index
	temppos = pos & 7;		//offset in one byte				
	cur_srch = pos & ~7;		//current search begin position in this block (8 alignment)
	pos = srch_beg;

	MY_DEBUG(idisplay,"index=%d,pos=%d,cur_srch=%d,PAGE_CACHE_SIZE = %d \n",index,pos,cur_srch,PAGE_CACHE_SIZE);
	
	while(pos < total_bits){
		//get the block
		sector_t blocknr;		//current blocknr
		blocknr = exfat_clus_to_blknr(sbi, bitmap_exi->clusnr) + index;

		bh = sb_bread(sb,blocknr);
		if (!bh) 
			return -EIO;	

		kaddr = (u8 *)bh->b_data;

		byte = kaddr + cur_srch/8;

		//unaligned bits
		if(temppos > 0){
			for(i = temppos ; i < 8 ; i ++){
				if(pos >= total_bits)
					goto not_found;
				if(!(*byte & (0x01 << i))){
					cur_srch += i;
					goto clus_found;
				}
				pos ++;
			}

			cur_srch += 8;
			byte++;
		}
		pos = (pos + 7) & 7;		//pos to aligned

		temppos = total_bits - pos;
		
		//aligned bytes search
		while((cur_srch < blocksize<<3) && (temppos >= 8) ){
			bit = myffz(*byte);
			
			MY_DEBUG(idisplay,"*byte=%x,bit=%d ,pos=%d \n",*byte,bit,cur_srch);

			if(bit != 0xFFFF){
				DEBUG_ON(bit >= 8,"bug");
				cur_srch += bit;
				goto clus_found;
			}
				
			cur_srch+= 8; 
			pos += 8;
			byte ++;

			temppos = total_bits - pos;
		}

		if(temppos < 8){// the  last end bits(less than 8)
			pos -= 8;
			cur_srch -= 8;
			
			for(i = 0 ; i < temppos ; i ++){
				if(pos >= total_bits)
					goto not_found;
				if(!(*byte & (0x01 << i))){
					cur_srch += i;
					goto clus_found;
				}
				pos ++;
			}

			goto not_found;
		}
		
		cur_srch = 0;
			
		index ++;
		brelse(bh);
	}

	MY_DEBUG(idisplay,"cannot find from %d to end \n",srch_beg);

clus_found:
	*found_clus = cur_srch + index*(blocksize<<3) + EXFAT_START_ENT;
	sbi->last_free_clus = *found_clus;
//	if(*found_clus & 0x10000)
//		idisplay = 1;		

//	printk("[PID %d]  *found_clus=%d,index = %d \n",current->pid,*found_clus,index);	

	MY_DEBUG(idisplay,"out  *found_clus=%d,cur_srch=%d,pos=%d,index = %d \n",*found_clus,cur_srch,pos,index);
	return 0;

not_found:
	*found_clus = -1;
	return -1;
}




int exfat_find_1free_cluster1111(struct super_block *sb,u32 srch_beg,u32 *found_clus)
{
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct exfat_inode_info *bitmap_exi = EXFAT_I(sbi->bitmap_inode);
	
	unsigned long index;
	u32 bit,pos,total_bits,cur_srch,blocksize,i,temppos,left_length;
	u8* byte,*kaddr;
	struct buffer_head *bh;	
	

	int isize = sbi->bitmap_inode->i_size;		//bitmap inode size
	total_bits = sbi->total_clusters;				//total bits
	DEBUG_ON((isize<<3)!= total_bits,"bug");
	
	blocksize = 1<<sbi->bitmap_inode->i_blkbits;
	
	if(srch_beg< EXFAT_START_ENT || srch_beg >= total_bits)
		return -1;

	srch_beg -= EXFAT_START_ENT;	
	
	index = (unsigned long)(srch_beg / (blocksize<<3));
	//offset = srch_beg 	
	pos = srch_beg - index*(blocksize<<3);	//start offset in this index
	temppos = pos & 7;		//offset in one byte	
	pos &= ~7;					
	cur_srch = pos;		//current search begin position	(8 alignment)
	pos = index * (blocksize<<3);	//srch beg , 8 alignment


	MY_DEBUG(0,"index=%d,pos=%d,cur_srch=%d,PAGE_CACHE_SIZE = %d \n",index,pos,cur_srch,PAGE_CACHE_SIZE);
	
	while(pos < total_bits){
		//get the block
		sector_t blocknr;		//current blocknr
		blocknr = exfat_clus_to_blknr(sbi, bitmap_exi->clusnr) + index;

		bh = sb_bread(sb,blocknr);
		if (!bh) 
			return -EIO;

		kaddr = (u8 *)bh->b_data;

		byte = kaddr + cur_srch/8;

		//unaligned bits
		if(temppos > 0){
			for(i = temppos ; i < 8 ; i ++){
				if(!(*byte & (0x01 << i))){
					cur_srch += i;
					goto clus_found;
				}					
			}

			cur_srch += 8;
			byte++;
		}

		//aligned bytes search
		while(cur_srch < blocksize<<3){
			bit = myffz(*byte);
			
			MY_DEBUG(0,"*byte=%x,bit=%d ,pos=%d \n",*byte,bit,cur_srch);

			if(bit != 0xFFFF){
				DEBUG_ON(bit >= 8,"bug");
				cur_srch += bit;

				goto clus_found;
			}
				
			cur_srch+= 8; 
			byte ++;
		}
		
		cur_srch = 0;
		pos += blocksize<<3;		
		
		index ++;
		brelse(bh);
	}

	MY_DEBUG(0,"cannot find from %d to end \n",srch_beg);


	//if don't find in [srch_beg, end], search it from (2, srch_beg)
	index = 0;
	cur_srch = 0;
	pos = 0;
	
	while(pos < srch_beg){
		//get the block
		sector_t blocknr;		//current blocknr
		blocknr = exfat_clus_to_blknr(sbi, bitmap_exi->clusnr) + index;

		bh = sb_bread(sb,blocknr);
		if (!bh) 
			return -EIO;

		kaddr = (u8 *)bh->b_data;

		byte = kaddr + cur_srch/8;
		
		while(cur_srch < blocksize<<3){
			bit = ffz(*byte);
			if(bit != 0xFFFF){
				DEBUG_ON(bit >= 8,"bug");
				cur_srch += bit;

				goto clus_found;
			}
				
			cur_srch+= 8; 
			byte ++;
		}
		
		cur_srch = 0;
		pos += blocksize<<3;
		index ++;
	}

clus_found:
	*found_clus = cur_srch + pos + EXFAT_START_ENT;
	MY_DEBUG(0,"out  *found_clus=%d,cur_srch=%d,pos=%d \n",*found_clus,cur_srch,pos);
	return 0;
	
}



int exfat_find_1free_cluster_with_page(struct super_block *sb,u32 srch_beg,u32 *found_clus)
{
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct address_space *mapping;
	struct exfat_inode_info *bitmap_exi = EXFAT_I(sbi->bitmap_inode);
	
	unsigned long index;
	u32 bit,pos,total_bits,cur_srch;
	u8* byte;
	
	int isize = sbi->bitmap_inode->i_size;		//bitmap inode size
	total_bits = sbi->total_clusters;				//total bits
	DEBUG_ON((isize<<3)!= total_bits,"bug");
	
	mapping = sbi->bitmap_inode->i_mapping;	//bitmap address space

	MY_DEBUG(0,"isize=%d,total_bits=%d,srch_beg=%d \n",isize,total_bits,srch_beg);

	if(srch_beg< EXFAT_START_ENT || srch_beg >= total_bits)
		return -1;

	srch_beg -= EXFAT_START_ENT;	

	index = (unsigned long)(srch_beg / (PAGE_CACHE_SIZE<<3));
	//offset = srch_beg 	
	pos = srch_beg - index*(PAGE_CACHE_SIZE<<3);	//start offset in this index
	pos &= ~7;	
	cur_srch = srch_beg - pos;		//current search begin position	

	MY_DEBUG(0,"index=%d,pos=%d,cur_srch=%d,PAGE_CACHE_SIZE = %d \n",index,pos,cur_srch,PAGE_CACHE_SIZE);
	
	while(cur_srch < total_bits){
		struct page *page;
		void *kaddr;
		page = read_mapping_page(mapping, index, NULL);
		if (IS_ERR(page))
			return -1;
		
		kaddr = kmap_atomic(page, KM_USER0);

		byte = kaddr + pos/8;
		while(pos < PAGE_CACHE_SIZE<<3){
			bit = myffz(*byte);
			
			MY_DEBUG(0,"*byte=%x,bit=%d ,pos=%d \n",*byte,bit,pos);

			if(bit != 0xFFFF){
				DEBUG_ON(bit >= 8,"bug");
				pos += bit;

				kunmap_atomic(kaddr, KM_USER0);
				page_cache_release(page);

				goto clus_found;
			}
				
			pos+= 8; 
			byte ++;
		}
		
		pos = 0;
		cur_srch += PAGE_CACHE_SIZE<<3;
		index ++;
		kunmap_atomic(kaddr, KM_USER0);
		page_cache_release(page);
	}

	MY_DEBUG(0,"cannot find from %d to end \n",srch_beg);


	//if don't find in [srch_beg, end], search it from (2, srch_beg)
	index = 0;
	cur_srch = 0;
	pos = 2;
	
	while(cur_srch < srch_beg){
		struct page *page;
		void *kaddr;
		page = read_mapping_page(mapping, index, NULL);
		if (IS_ERR(page))
			return -1;
		
		kaddr = kmap_atomic(page, KM_USER0);

		byte = kaddr + pos*8;
		while(pos < PAGE_CACHE_SIZE<<3){
			bit = ffz(*byte);
			if(bit != 0xFFFF){
				DEBUG_ON(bit >= 8,"bug");
				pos += bit;

				kunmap_atomic(kaddr, KM_USER0);
				page_cache_release(page);

				goto clus_found;
			}
				
			pos+= 8; 
			byte ++;
		}
		
		pos = 0;
		cur_srch += PAGE_CACHE_SIZE<<3;
		index ++;
		kunmap_atomic(kaddr, KM_USER0);
		page_cache_release(page);
	}

clus_found:
	*found_clus = cur_srch + pos + EXFAT_START_ENT;
	MY_DEBUG(0,"out  *found_clus=%d,cur_srch=%d,pos=%d \n",*found_clus,cur_srch,pos);
	return 0;
	
}
