/*
 * Copyright (C) 2008, OGAWA Hirofumi
 * Released under GPL v2.
 */

#include "exfat.h"
#include "cache.h"
#include "fatent.h"
#include "cluster.h"
#include "bitmap.h"
#include "inode.h"

static void calc_cmap(struct exfat_clus_map *cmap, struct exfat_cache_id *cid,
		      u32 iclusnr)
{
	u32 offset = iclusnr - cid->iclusnr;
	cmap->iclusnr = iclusnr;
	cmap->clusnr = cid->clusnr + offset;
	cmap->len = cid->len - offset;
}

/*
 * Caller must be guarantee it have no race with truncate().  I.e. the
 * range of "iclusnr" and "iclusnr + len" must not be truncated.
 */
 extern int g_flag;
int exfat_get_cluster(struct inode *inode, u32 iclusnr, u32 clus_len,
		      struct exfat_clus_map *cmap)
{
	struct super_block *sb = inode->i_sb;
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct exfat_inode_info *exi = EXFAT_I(inode);
	struct exfat_cache_id cid;
	struct exfat_ent fatent;
	u32 clusnr;
	int err = 0, found_target;

	int idisplay = 0;
	if(g_flag == 1)
		idisplay = 1;

	if (clus_len && (exi->data_flag == EXFAT_DATA_CONTIGUOUS)) {
		//printk("exfat_get_cluster: EXFAT_DATA_CONTIGUOUS\n");
		DEBUG_ON(clus_len >
		  (/*i_size_read(inode)*/EXFAT_I(inode)->phys_size+ (sbi->clus_size - 1)) >> sbi->clus_bits,
			 "clus_len %u, i_size %lld, physize %lld,clus_size %lu\n",
			 clus_len, i_size_read(inode), EXFAT_I(inode)->phys_size,sbi->clus_size);
		/* FIXME: should check clus_len limit from i_size? */
		cmap->iclusnr = iclusnr;
		cmap->clusnr = exi->clusnr + iclusnr;
		cmap->len = clus_len;
		MY_DEBUG(idisplay,"EXFAT_DATA_CONTIGUOUS: contig: iclusnr %u, clusnr %u, len %u\n",
		       cmap->iclusnr, cmap->clusnr, cmap->len);
		return 0;
	}

	DEBUG0_ON(!exi->clusnr);

	/*
	 * FIXME: use rw_semaphore (r: here, w: truncate), then we can
	 * do readahead of contiguous clusters and cache it. (This
	 * should optimize disk-seek very much if clusters is not
	 * fragmented heavily)
	 */

	/* Setup the start cluster to walk */
	if (!exfat_cache_lookup(inode, iclusnr, &cid)){
		MY_DEBUG(idisplay,"cache_init \n");
		cache_init(&cid, 0, exi->clusnr);
	}
	clusnr = cid.clusnr + cid.len - 1;

	/* Walk the cluster-chain */
	found_target = 0;
	exfat_ent_init(&fatent);
	MY_DEBUG(idisplay,"%s: iclusnr %u, clus_len %u: cid.iclusnr %u, cid.clusnr %u,"
	       " cid.len %u, search start clusnr %u\n",
	       __func__, iclusnr, clus_len, cid.iclusnr, cid.clusnr, cid.len,
	       clusnr);
	while (cid.iclusnr + cid.len < iclusnr + clus_len) {
		/* Target iclusnr was found, so find contiguous clusters */
		if (!found_target && (iclusnr < cid.iclusnr + cid.len))
			found_target = 1;

		/* FIXME: should be readahead FAT blocks */
		u32 prev = clusnr;
		err = exfat_ent_read(inode, &fatent, prev, &clusnr);
		MY_DEBUG(idisplay,"%s: ent_read: clusnr %u, next clusnr %u\n",
		       __func__, prev, clusnr);
		
		if (err < 0)
			goto out;

		if (exfat_ent_eof(sbi, clusnr)) {
			exfat_cache_add(inode, &cid);

			/* Return special cmap */
			cmap->iclusnr = cid.iclusnr + cid.len;
			cmap->clusnr = clusnr;
			cmap->len = 0;
			goto out;
		} else if (exfat_ent_bad(sbi, clusnr)) {
			PRINTK(KERN_ERR "exFAT: found bad cluster entry"
			       " (start cluster 0x%08x)\n", exi->clusnr);
			err = -EIO;
			goto out;
		} else if (!exfat_ent_valid(sbi, clusnr)) {
			exfat_fs_panic(sb, "found invalid cluster chain"
				       " (start cluster 0x%08x, entry 0x%08x)",
				       exi->clusnr, clusnr);
			err = -EIO;
			goto out;
		}

		if (cache_contiguous(&cid, clusnr))
			cid.len++;
		else {
			if (found_target) {
				/* Found a discontiguous cluster */
				calc_cmap(cmap, &cid, iclusnr);
				exfat_cache_add(inode, &cid);

				cache_init(&cid, cid.iclusnr + cid.len, clusnr);
				exfat_cache_add(inode, &cid);
				goto out;
			}
			cache_init(&cid, cid.iclusnr + cid.len, clusnr);
		}

		/* Prevent the infinite loop of cluster chain */
		if (cid.iclusnr + cid.len > sbi->total_clusters) {
			exfat_fs_panic(sb, "detected the cluster chain loop"
				       " (start cluster 0x%08x)", exi->clusnr);
			err = -EIO;
			goto out;
		}
	}

	calc_cmap(cmap, &cid, iclusnr);
	exfat_cache_add(inode, &cid);
out:
	exfat_ent_release(&fatent);
	MY_DEBUG(idisplay,"%s: result cmap: iclusnr %u, clusnr %u, len %u\n", __func__,
	       cmap->iclusnr, cmap->clusnr, cmap->len);

	return err;
}

/*
	@inode : inode the chain to add
	@new_dclus : added cluster
	@nr_cluster : number of cluster(s)
*/
int exfat_chain_add(struct inode *inode, int new_dclus, int nr_cluster)
{
	struct super_block *sb = inode->i_sb;
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct exfat_inode_info *exi = EXFAT_I(inode);
	int err = 0,new_fclus, last,next_dclus;
	int idisplay = 0;
//	static flag = 0;
//	if(flag)
//		return 0;
	

	last = new_fclus = 0;
	MY_DEBUG(idisplay,"exi->clusnr=%d \n",exi->clusnr);
	if(exi->clusnr){
		struct exfat_clus_map cmap;
		err = exfat_get_cluster(inode,EXFAT_ENT_EOF,0,&cmap);
		MY_DEBUG(idisplay,"cmap.clusnr=%u,sbi->total_clusters=%d \n",cmap.clusnr,sbi->total_clusters);

		new_fclus = cmap.iclusnr - 1;
		if(new_fclus < 0)
			new_fclus = 0;
		err = exfat_get_cluster(inode,new_fclus,0,&cmap);
		new_fclus = cmap.iclusnr + 1;
		last = cmap.clusnr;
	}

	MY_DEBUG(idisplay,"last=%d, new_fclus = %d\n",last,new_fclus);

	if(last){	//add new one to the cluster chain
		struct exfat_ent exfatent;

		exfat_ent_init(&exfatent);	
		err = exfat_ent_prepare(inode, &exfatent, last, &next_dclus);
		if(err)
			return -EIO;
		else{		
			if(new_dclus != last + 1)	//if the cluster chain is not contiguous, set off the contiguous flag
				exi->data_flag &= ~(EXFAT_DATA_CONTIGUOUS - 1);
			exfat_ent_write(inode, &exfatent, new_dclus);
			
			mark_buffer_dirty(exfatent.bh);
//			sync_dirty_buffer(exfatent.bh);

			exfat_ent_release(&exfatent);

			//Todo: if the cluster chain is contiguous, don't write the FAT; and then if find the cluster
			//chain is not contiguous, write all the FAT
		}	
	}
	else{	// new cluster chain
		exi->clusnr = new_dclus;
		if (S_ISDIR(inode->i_mode) && IS_DIRSYNC(inode)) {
			err = exfat_sync_inode(inode);
			if (err)
				return err;
		} else
			mark_inode_dirty(inode);

//		flag = 1;	
	}

	inode->i_blocks += nr_cluster << (sbi->clus_bits - inode->i_blkbits);
	
	return 0;
}


/*
	@inode: inode entry write to
	@clusnr:write cluster number
	@valuetoset:write data
	
*/
int exfat_write_entry(struct inode *inode, u32 clusnr,u32 valuetoset)
{
	struct exfat_ent exfatent = {0};
	u32 next_clus;

	MY_DEBUG(0,"clusnr=%d,valuetoset=%x \n",clusnr,valuetoset);

	exfat_ent_init(&exfatent);	
	exfat_ent_prepare(inode,&exfatent,clusnr,&next_clus);
	exfat_ent_write(inode,&exfatent,valuetoset);

	mark_buffer_dirty(exfatent.bh);
	
	//no need to sync it right now	
//	sync_dirty_buffer(exfatent.bh);	

	exfat_ent_release(&exfatent);

	return 0;
}

/*
	@inode : inode to free clusters
	@clusnr : clusnr to begin free
*/
int exfat_free_clusters(struct inode *inode, u32 clusnr,int clus_len)
{
	struct super_block *sb = inode->i_sb;
//	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct exfat_inode_info *exi = EXFAT_I(inode);
	int err = 0;
	struct exfat_ent exfatent;
	u32 next_clus,write = 1;
	int idisplay = 0;

	MY_DEBUG(idisplay,"clusnr=%d, clus_len=%d \n",clusnr,clus_len);

	//should we lock fs?
	exfat_ent_init(&exfatent);	
	do{
		err = exfat_ent_read(inode, &exfatent, clusnr, &next_clus);
		if(err)
			break;
		MY_DEBUG(idisplay,"clusnr=%d,next_clus=%u\n",clusnr,next_clus);
		
		if(clusnr < 0){
			DEBUG_ON(1,"bug");
			break;
		}	
		else /*if(next_clus == EXFAT_ENT_FREE)*/{
			if(exi->data_flag == EXFAT_DATA_CONTIGUOUS){
				next_clus = clusnr + 1;
				write = 0;
				MY_DEBUG(idisplay,"found contiguous cluster file.next_clus=%d \n",next_clus);
			}	
		}

		if(write)
			exfat_write_entry(inode, clusnr, EXFAT_ENT_FREE);
		exfat_bits_set(sb, clusnr, 1, 0);
		clus_len --;
		write = 1;

		MY_DEBUG(idisplay,"clusnr=%d,next_clus=%d \n",clusnr,next_clus);
		
		clusnr = next_clus;
		
	}while(clusnr != EXFAT_ENT_EOF && clus_len);

	exfat_ent_release(&exfatent);
	
	return err;
}


int exfat_clear_cluster(struct inode *dir,int clus)
{
	struct super_block *sb = dir->i_sb;
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct buffer_head *bh = NULL;
	sector_t blocknr;
	int i,blocksize;

	blocksize = 1 << dir->i_blkbits;
	blocknr = exfat_clus_to_blknr(sbi,clus);

	for(i = 0 ; i  < sbi->bpc ; i ++){
		bh = sb_bread(sb, blocknr);
		if (!bh) 
			return -EIO;

		memset(bh->b_data,0,blocksize);
		mark_buffer_dirty(bh);
//		sync_dirty_buffer(bh);
		brelse(bh);

		blocknr ++;
	}	

	return 0;
	
}


/*
	@inode: file inode attatched to 
	@iclusnr:file cluster num extended to

	extend cluster chain
*/
int exfat_add_cluster(struct inode *inode, u32 *iclusnr)
{
	struct super_block *sb = inode->i_sb;
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct exfat_inode_info *exi = EXFAT_I(inode);
	int err = 0;
	unsigned long blocksize,cur_block;
	u32 cur_iclusnr,dclusnr;

	int idisplay = 0;
	static int aaa = 140930;

	blocksize = 1 << inode->i_blkbits;
	cur_block = (i_size_read(inode) + (blocksize - 1)) >> inode->i_blkbits;
	cur_iclusnr = (cur_block+ (sbi->bpc - 1)) >> sbi->bpc_bits;
	MY_DEBUG(idisplay,"cur_block=%d,cur_iclusnr=%d,sbi->bpc=%d,sbi->bpc_bits=%d\n",cur_block,cur_iclusnr,sbi->bpc,sbi->bpc_bits);
	MY_DEBUG(idisplay,"inode size=%lld,*iclusnr=%d,cur_iclusnr=%d\n",inode->i_size,*iclusnr,cur_iclusnr);
	DEBUG_ON(*iclusnr < cur_iclusnr,"bug");

	while(cur_iclusnr <= *iclusnr){
		mutex_lock(&sbi->bitmap_mutex);
		/*step1: find free cluster from bitmap*/
//		if(aaa == 0){
			err = exfat_find_1free_cluster(sb,EXFAT_START_ENT,&dclusnr);
//			aaa = dclusnr;
//		}
//		else
//			dclusnr = aaa ++;
		MY_DEBUG(idisplay,"dclusnr=%d,err=%d\n",dclusnr,err);
		if(err < 0)
			break;

		/*step2: set the bitmap*/	
		exfat_bits_set(sb,dclusnr,1,1);

		mutex_unlock(&sbi->bitmap_mutex);

		/*step3: mark the cluster as EOF cluster*/	
		exfat_write_entry(inode,dclusnr,EXFAT_ENT_EOF);

		MY_DEBUG(idisplay,"before exfat_chain_add \n");

		/*step4: chain the cluster*/
		err = exfat_chain_add(inode,dclusnr,1);
		MY_DEBUG(idisplay,"after exfat_chain_add, err=%d,inode->i_size = %lld,physize=%d \n",err,inode->i_size,exi->phys_size);
		if(err){
			exfat_bits_set(sb,dclusnr,1,0);
			exfat_free_clusters(inode,dclusnr,1);
			
			break;
		}	

//		exfat_clear_cluster(inode,dclusnr);		
 
		cur_iclusnr ++;
		exi->phys_size += sbi->clus_size;
		err = 0;
	}

	*iclusnr = dclusnr;
	
	return err;
	
}

int exfat_allocate_clusters(struct inode *inode, int *cluster, int nr_cluster)
{
	struct super_block *sb = inode->i_sb;
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
//	struct exfat_inode_info *exi = EXFAT_I(inode);
	int err = 0,index = 0, found_clus,i ;
	
	while(index < nr_cluster){
		err = exfat_find_1free_cluster(sb, EXFAT_START_ENT, &found_clus);
		if(err)
			break;
		
		cluster[index ++] = found_clus;

		if(index == nr_cluster)
			break;	
	}

	if(index == nr_cluster){
		for(i = 0 ; i < index ; i ++)
			exfat_bits_set(sb,cluster[i],1,1);
	}	

	return err;
}
