/*
 * Copyright (C) 2008, OGAWA Hirofumi
 * Released under GPL v2.
 */

#include <linux/buffer_head.h>
#include "exfat.h"
#include "bitmap.h"
#include "upcase.h"
#include "inode.h"
#include "cluster.h"

static void pd_init(struct exfat_parse_data *pd)
{
	pd->nr_bhs = 0;
#ifdef DEBUG
	pd->bh_cnt = 0;
#endif
}

static void pd_brelse(struct exfat_parse_data *pd)
{
	while (pd->nr_bhs) {
		pd->nr_bhs--;
		brelse(pd->bhs[pd->nr_bhs]);
#ifdef DEBUG
		pd->bh_cnt--;
#endif
	}
	pd_init(pd);
}

static void *pd_get_chunk(struct exfat_parse_data *pd)
{
	return pd->bhs[0]->b_data + pd->bh_offset;
}

struct exfat_parse_iter_data {
	unsigned long	left;
	unsigned long	bh_offset;
	unsigned int	bh_index;
};

static void *pd_iter_first_de(struct exfat_parse_data *pd,
			      struct exfat_parse_iter_data *pd_iter)
{
	DEBUG_ON(pd->size < EXFAT_CHUNK_SIZE, "pd->size %lu\n", pd->size);
	pd_iter->left = pd->size - EXFAT_CHUNK_SIZE;
	pd_iter->bh_offset = pd->bh_offset;
	pd_iter->bh_index = 0;
	return pd_get_chunk(pd);
}

static void *pd_iter_next_de(struct exfat_parse_data *pd,
			     struct exfat_parse_iter_data *pd_iter)
{
	if (pd_iter->left < EXFAT_CHUNK_SIZE)
		return NULL;
	pd_iter->left -= EXFAT_CHUNK_SIZE;
	pd_iter->bh_offset += EXFAT_CHUNK_SIZE;
	if (pd_iter->bh_offset >= pd->bhs[pd_iter->bh_index]->b_size) {
		pd_iter->bh_index++;
		pd_iter->bh_offset = 0;
	}
	return pd->bhs[pd_iter->bh_index]->b_data + pd_iter->bh_offset;
}

enum { PARSE_NEXT, PARSE_STOP, };

struct exfat_parse_callback {
	int		(*parse)(struct inode *, loff_t,
				 struct exfat_parse_data *,
				 struct exfat_parse_callback *);
	void		*data;
};

static int exfat_parse_dir(struct file *filp, struct inode *dir, loff_t *ppos,
			   struct exfat_parse_callback *pcb)
{
	struct super_block *sb = dir->i_sb;
	struct buffer_head map_bh, *bh;
	sector_t uninitialized_var(blocknr), last_iblock;
	unsigned long blocksize, blocks, offset, de_left;
	struct exfat_parse_data pd;
	loff_t pos;
	int ret, err;
	int idisplay = 0;

	blocksize = 1 << dir->i_blkbits;
	last_iblock = dir->i_size >> dir->i_blkbits;

	/* FIXME: there is this limitation? Or limitation is cluster size? */
	if (dir->i_size & (blocksize - 1)) {
		exfat_fs_panic(sb, "invalid directory size (size %lld)",
			       dir->i_size);
		return -EIO;
	}

	pos = *ppos;

	/* FIXME: can cache the free space for next create()/mkdir()/etc? */
	pd_init(&pd);
	err = ret = 0;
	blocks = de_left = 0;
	while (!ret && pos < dir->i_size) {
		if (!blocks) {
			/* FIXME: readahead, probably last_iblock is too big */
			sector_t iblock;

			iblock = pos >> dir->i_blkbits;

			map_bh.b_state = 0;
			map_bh.b_blocknr = 0;
			map_bh.b_size = (last_iblock-iblock) << dir->i_blkbits;

			err = exfat_get_block(dir, iblock, &map_bh, 0);
			if (err)
				break;
			DEBUG0_ON(!buffer_mapped(&map_bh));

			blocknr = map_bh.b_blocknr;
			blocks = map_bh.b_size >> dir->i_blkbits;
		}

		offset = pos & (blocksize - 1);
		bh = sb_bread(sb, blocknr);
		if (!bh) {
			err = -EIO;
			break;
		}
#ifdef DEBUG
		pd.bh_cnt++;
#endif
		if (de_left) {
			unsigned long len;
#ifdef DEBUG
			pd.bh_cnt++;
#endif
			get_bh(bh);
			pd.bhs[pd.nr_bhs] = bh;
			pd.nr_bhs++;

			len = min(de_left, blocksize);
			de_left -= len;
			offset += len;
			pos += len;

			MY_DEBUG(idisplay,"22%s: blocknr %llu, de_left %lu, offset %lu,"
			       " pos %llu, len %lu\n", __func__,
			       (llu)bh->b_blocknr, de_left, offset, pos, len);
			if (!de_left)
				goto dirent_complete;

			goto next_block;
		}

		MY_DEBUG(idisplay,"offset=%d,bh->b_size=%d,dir->i_blkbits=%d\n",offset,bh->b_size,dir->i_blkbits);
		
		while (offset < bh->b_size) {
			struct exfat_chunk_dirent *dirent;
			unsigned long len, chunks;

			
			dirent = (void *)bh->b_data + offset;

			MY_DEBUG(idisplay,"dirent->type = %x, offset = %d\n",dirent->type,offset);
			if (dirent->type == EXFAT_TYPE_EOD) {
				pos = dir->i_size;
				break;
			} else if (!(dirent->type & EXFAT_TYPE_VALID)) {
				offset += EXFAT_CHUNK_SIZE;
				pos += EXFAT_CHUNK_SIZE;
				continue;
			}

			if (dirent->type == EXFAT_TYPE_DIRENT)
				chunks = 1 + dirent->sub_chunks;
			else
				chunks = 1;

			de_left = chunks << EXFAT_CHUNK_BITS;

			pd.size = de_left;
#ifdef DEBUG
			pd.bh_cnt++;
#endif
			get_bh(bh);
			pd.bhs[0] = bh;
			pd.nr_bhs = 1;
			pd.bh_offset = offset;

			len = min(de_left, blocksize - offset);
			de_left -= len;
			offset += len;
			pos += len;

			MY_DEBUG(idisplay,"2222%s: blocknr %llu, de_left %lu, offset %lu,"
			       " pos %llu, len %lu\n", __func__,
			       (llu)bh->b_blocknr, de_left, offset, pos, len);
			if (!de_left) {
dirent_complete:
				ret = pcb->parse(dir, pos - pd.size, &pd, pcb);
				pd_brelse(&pd);
				if (ret != PARSE_NEXT) {
					if (ret < 0)
						err = ret;
					pos -= pd.size;
					break;
				}
			}
		}
next_block:
		*ppos = pos;
		brelse(bh);
#ifdef DEBUG
		pd.bh_cnt--;
#endif
		blocknr++;
		blocks--;
	}
	pd_brelse(&pd);
#ifdef DEBUG
	DEBUG_ON(pd.bh_cnt, "pd.bh_cnt %d\n", pd.bh_cnt);
#endif

	return err;
}

/* FIXME: unicode string should be cached? */
static int exfat_d_revalidate(struct dentry *dentry, struct nameidata *nd)
{
	return 1;
}

static int exfat_d_hash(struct dentry *dentry, struct qstr *name)
{
	/* FIXME: can't cache converted name/hash? */
	return 0;
}

static int exfat_d_compare(struct dentry *dentry, struct qstr *a,
			   struct qstr *b)
{
	/* FIXME: can't use converted name/hash? */
	if (a->len != b->len)
		return 1;
	return memcmp(a->name, b->name, a->len);
}

static int exfat_d_delete(struct dentry *dentry)
{
	return 0;
}

static void exfat_d_release(struct dentry *dentry)
{
}

static void exfat_d_iput(struct dentry *dentry, struct inode *inode)
{
	iput(inode);
}

struct dentry_operations exfat_dentry_ops = {
	.d_revalidate	= exfat_d_revalidate,
	.d_hash		= exfat_d_hash,
	.d_compare	= exfat_d_compare,
	.d_delete	= exfat_d_delete,
	.d_release	= exfat_d_release,
	.d_iput		= exfat_d_iput,
};

static int exfat_match(struct exfat_sb_info *sbi, wchar_t *a, __le16 *b, u8 len)
{
	u8 i;
	for (i = 0; i < len; i++) {
		/* FIXME: assume a[i] is upper-case */
		if (a[i] != exfat_towupper(sbi, le16_to_cpu(b[i])))
			return 0;
	}
	return -1;
}

struct lookup_parse_data {
	struct inode *inode;
	wchar_t *wstr;
	u8 wlen;
	u16 hash;
	u8	is_dir;
	u8	delete_flag;
};

static int lookup_parse(struct inode *dir, loff_t pos,
			struct exfat_parse_data *pd,
			struct exfat_parse_callback *pcb)
{
	struct super_block *sb = dir->i_sb;
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct lookup_parse_data *pcb_data = pcb->data;
	struct exfat_parse_iter_data pd_iter;
	struct exfat_chunk_dirent *dirent;
	struct exfat_chunk_data *data;
	struct exfat_chunk_name *name;
	wchar_t *wstr;
	u16 csum;
	u8 len, de_namelen;
	int idisplay = 0;

	MY_DEBUG(idisplay,"In:pd->size=%d,pos=%lld\n",pd->size,pos);

	dirent = pd_iter_first_de(pd, &pd_iter);
	MY_DEBUG(idisplay,"dirent->type=%x \n",dirent->type);
	if (dirent->type != EXFAT_TYPE_DIRENT)
		return PARSE_NEXT;

	MY_DEBUG(idisplay,"1\n");	
	/* ->sub_chunk was already checked by caller. */
	DEBUG_ON((1 + dirent->sub_chunks) << EXFAT_CHUNK_BITS != pd->size,
		 "sub_chunks %u, pd->size %lu\n", dirent->sub_chunks, pd->size);
	PRINTK("exFAT: 0x%02x found\n", dirent->type);
	/* Skip dirent->checksum field */
	csum = exfat_checksum16(0, dirent, 2);
	csum = exfat_checksum16(csum, (u8 *)dirent + 4, EXFAT_CHUNK_SIZE - 4);

	if(dirent->attrib & EXFAT_ATTR_DIR)
		pcb_data->is_dir = 1;
	else
		pcb_data->is_dir = 0;

	data = pd_iter_next_de(pd, &pd_iter);
	if (!data)
		return PARSE_NEXT;
	if (data->type != EXFAT_TYPE_DATA)
		return PARSE_NEXT;

	MY_DEBUG(idisplay," 2\n");
	
	PRINTK("%s: wlen %u, name_len %u\n", __func__,
	       pcb_data->wlen, data->name_len);
	if (pcb_data->wlen != data->name_len)
		return PARSE_NEXT;
	MY_DEBUG(idisplay," 3\n");	
	PRINTK("%s: hash 0x%04x, checksum 0x%04x\n", __func__,
	       pcb_data->hash, le16_to_cpu(data->hash));
	if (pcb_data->hash != le16_to_cpu(data->hash))
		return PARSE_NEXT;
	de_namelen = data->name_len;
	csum = exfat_checksum16(csum, data, EXFAT_CHUNK_SIZE);
	MY_DEBUG(idisplay," 4\n");
	wstr = pcb_data->wstr;
	while (de_namelen) {
		name = pd_iter_next_de(pd, &pd_iter);
		if (!name)
			return PARSE_NEXT;
		if (name->type != EXFAT_TYPE_NAME)
			return PARSE_NEXT;

		len = min_t(u8, de_namelen, EXFAT_CHUNK_NAME_SIZE);
		if (!exfat_match(sbi, wstr, name->name, len))
			return PARSE_NEXT;
		wstr += len;
		de_namelen -= len;

		csum = exfat_checksum16(csum, name, EXFAT_CHUNK_SIZE);
	}
	MY_DEBUG(idisplay,"5\n ");
	/* Checksum of the remaining entries (may not be exfat_chunk_name) */
	while ((name = pd_iter_next_de(pd, &pd_iter)) != NULL) {
		if (!(name->type & EXFAT_TYPE_SUBCHUNK))
			return PARSE_NEXT;
		csum = exfat_checksum16(csum, name, EXFAT_CHUNK_SIZE);
	}
	MY_DEBUG(idisplay," 6\n");
	if (le16_to_cpu(dirent->checksum) != csum) {
		exfat_fs_panic(sb, "checksum failed for directory entry"
			       " in lookup (0x%04x != 0x%04x)",
			       le16_to_cpu(dirent->checksum), csum);
		return -EIO;
	}
	MY_DEBUG(idisplay," 7\n");

	MY_DEBUG(idisplay,"delete_flag = %d \n",pcb_data->delete_flag);

	if(pcb_data->delete_flag){	//found & delete it
		int i ;
		dirent = pd_iter_first_de(pd, &pd_iter);
		dirent->type &= ~EXFAT_TYPE_VALID;

		data = pd_iter_next_de(pd, &pd_iter);
		data->type &= ~EXFAT_TYPE_VALID;
		
		de_namelen = data->name_len;
		wstr = pcb_data->wstr;
		while (de_namelen) {
			name = pd_iter_next_de(pd, &pd_iter);
			name->type &= ~EXFAT_TYPE_VALID;
			len = min_t(u8, de_namelen, EXFAT_CHUNK_NAME_SIZE);
			wstr += len;
			de_namelen -= len;

		}
		/* Checksum of the remaining entries (may not be exfat_chunk_name) 
		while ((name = pd_iter_next_de(pd, &pd_iter)) != NULL) {
			if (!(name->type & EXFAT_TYPE_SUBCHUNK))
				return PARSE_NEXT;
			csum = exfat_checksum16(csum, name, EXFAT_CHUNK_SIZE);
		}*/
		for(i = 0 ; i < pd->nr_bhs ; i ++){
			mark_buffer_dirty(pd->bhs[i]);
			sync_dirty_buffer(pd->bhs[i]);
		}
		
	}
	else{
		pcb_data->inode = exfat_iget(sb, pd, dirent, data);
		if (IS_ERR(pcb_data->inode))
			return PTR_ERR(pcb_data->inode);
	}

	MY_DEBUG(idisplay," out\n");
	return PARSE_STOP;
}

static struct dentry *exfat_lookup(struct inode *dir, struct dentry *dentry,
				   struct nameidata *nd)
{
	struct exfat_sb_info *sbi = EXFAT_SB(dir->i_sb);
	struct nls_table *nls = sbi->opts.nls;
	struct exfat_parse_callback pcb;
	struct lookup_parse_data pcb_data;
	const unsigned char *name;
	wchar_t *wstr;
	unsigned int i, name_left, wlen, wleft;
	loff_t pos;
	int err;

	int idisplay = 0;

	MY_DEBUG(idisplay,"INnnnnn,name = %s  \n",dentry->d_name.name);
	
	/* FIXME: can't use converted name/hash? */

	wlen = wleft = min(dentry->d_name.len, EXFAT_MAX_NAMELEN);
	pcb_data.wstr = wstr = kmalloc(wlen * sizeof(*wstr), GFP_KERNEL);
	if (!wstr) {
		err = -ENOMEM;
		goto error;
	}

	name = dentry->d_name.name;
	name_left = dentry->d_name.len;

//	printk("1@@@@@ name=%s,name_left=%d\n",name,name_left);
	
	/* FIXME: maybe toupper and hash should be computed at same time */
//	err = exfat_conv_c2u_for_cmp(nls, &name, &name_left, &wstr, &wleft);
	err = exfat_conv_c2u(nls, &name, &name_left, &wstr, &wleft);
	/* FIXME: shouldn't return -ENAMETOOLONG? */
	if (err)
		goto error;
	pcb_data.wlen = wlen - wleft;

//	printk("@@@@@ wlen=%d,pcb_data.wlen=%d\n",wlen,pcb_data.wlen);

	pcb_data.hash = 0;
	pcb_data.delete_flag = 0;
	for (i = 0; i < pcb_data.wlen; i++) {
		__le16 uc;
		MY_DEBUG(idisplay,"uc %d= %x\n",i,pcb_data.wstr[i]);
		pcb_data.wstr[i] = exfat_towupper(sbi, pcb_data.wstr[i]);
		uc = cpu_to_le16(pcb_data.wstr[i]);
		pcb_data.hash = exfat_checksum16(pcb_data.hash, &uc, sizeof(uc));
	}

	pcb_data.inode = NULL;
	pcb.parse = lookup_parse;
	pcb.data = &pcb_data;

	/* FIXME: pos can be previous lookuped position. */
	pos = 0;
	err = exfat_parse_dir(NULL, dir, &pos, &pcb);
	if (err)
		goto error;

	kfree(pcb_data.wstr);

	/* FIXME: can cache the position for next lookup()?
	 * Assume sequential locality (stat() for all files after readdir()) */

	if(pcb_data.inode)
		MY_DEBUG(idisplay,"inode_nlink=%d , icount = %d \n",pcb_data.inode->i_nlink,pcb_data.inode->i_count.counter);


	dentry->d_op = &exfat_dentry_ops;
	dentry = d_splice_alias(pcb_data.inode, dentry);
	if (dentry)
		dentry->d_op = &exfat_dentry_ops;

	MY_DEBUG(idisplay,"Outtttttt  \n");
	
	return dentry;

error:
	MY_DEBUG(idisplay,"Errrrrr out \n");
	kfree(pcb_data.wstr);
	return ERR_PTR(err);
}


struct dentry_data {
	struct inode *inode;
	wchar_t *wstr;
	u8 wlen;
	u16 hash;
};

/* returns the length of a struct qstr, ignoring trailing dots */
static unsigned int exfat_striptail_len(struct qstr *qstr)
{
	unsigned int len = qstr->len;

	while (len && qstr->name[len - 1] == '.')
		len--;
	return len;
}

static inline wchar_t exfat_invalid_char(wchar_t w)
{
	return ((w < 0x0020) || (w == '*') 	|| (w == '?')	|| 
		   	(w == '<')   	|| (w == '>')	|| (w == '|')	|| 
		   	(w == '"')	|| (w == ':')	|| (w == '/')	|| 
		   	(w == '\\'));
}

static inline int extfat_name_valid(const wchar_t*name,int len)
{
	int i = 0;
	printk("len = %d\n",len);
	for(i = 0 ; i < len ; i ++){
		printk("char = %x\n",name[i]);
		if(exfat_invalid_char(name[i])){
			return 1;
		}
	}

	if (name[i - 1] == ' ') /* last character cannot be space */
		return 1;

	return 0;
}

static int exfat_fill_dirent(	struct exfat_chunk_dirent *dirent_slot,
							int checksum,struct timespec *ts,
							int nr_sub_slot,int is_dir)
{
	 __le16 time;
	 __le16 date;
	 u8 time_cs;

	memset(dirent_slot,0,sizeof(struct exfat_chunk_dirent));
	 
	dirent_slot->type = EXFAT_TYPE_DIRENT;
	dirent_slot->sub_chunks = nr_sub_slot;
	dirent_slot->checksum = checksum;

	exfat_time_unix2exfat(ts, &time, &date, &time_cs);
	dirent_slot->atime = dirent_slot->mtime = dirent_slot->crtime = time;//0x8e39;//time;
	dirent_slot->adate = dirent_slot->mdate = dirent_slot->crdate = date;//0x3c95;//date;
//	dirent_slot->mtime =0x889b;
//	dirent_slot->mdate = 0x3c8e;
	
//	dirent_slot->crtime_cs = 0xa2;
	dirent_slot->mtime_cs = time_cs;//0x00;//time_cs;
	memset(dirent_slot->xxxx02,0xA0,3);
	if(is_dir)
		dirent_slot->attrib |= EXFAT_ATTR_DIR;
	else{
			dirent_slot->attrib &= ~EXFAT_ATTR_DIR;
			dirent_slot->attrib |= EXFAT_ATTR_ARCH;
	}

	return 0;
}

static int exfat_fill_chunkdata(	struct exfat_sb_info *sbi, 
								struct exfat_chunk_data *chkdata_slot,
								u16 hash,int cluster,int namelen,int is_dir)
{
	memset(chkdata_slot,0,sizeof(struct exfat_chunk_data));

	if(is_dir){
		chkdata_slot->flag = EXFAT_DATA_CONTIGUOUS;
		chkdata_slot->size = sbi->clus_size;
		chkdata_slot->size2 = sbi->clus_size;	
	}
	else{
		chkdata_slot->flag = EXFAT_DATA_CONTIGUOUS;
		chkdata_slot->size = 0;
		chkdata_slot->size2 = 0;
	}
	chkdata_slot->type = EXFAT_TYPE_DATA;
	chkdata_slot->name_len = namelen;
	chkdata_slot->hash = hash;
	chkdata_slot->clusnr = cluster;
	
	return 0;
}

static int exfat_fill_chunkname(	struct exfat_chunk_name *chkname_slot,
									wchar_t* name,int namelen,
									int nr_slots)
{
	int i = 0,len;
	unsigned char *pbyte;
	for(i = 0 ; i < nr_slots ; i ++){
		memset(chkname_slot,0,sizeof(struct exfat_chunk_name));
	 
		chkname_slot->type = EXFAT_TYPE_NAME;
		pbyte = (unsigned char *)chkname_slot;
		len = min(EXFAT_CHUNK_NAME_SIZE,namelen);
		memcpy(pbyte + 2,name,len*sizeof(wchar_t));

		name+= len;
		namelen -= len;
		chkname_slot ++;
	}

	DEBUG_ON(namelen != 0,"bug");
		
	return 0;
}

static int exfat_build_slots(struct inode *dir, struct qstr *qname, 
							int is_dir, int cluster,
							struct timespec *ts,
							struct exfat_chunk_dirent *slots, int *nr_slots)
{
	
	struct exfat_sb_info *sbi = EXFAT_SB(dir->i_sb);
	struct nls_table *nls = sbi->opts.nls;
	wchar_t *wstr,*pbwstr=NULL;
	unsigned int i, name_left, wlen, wleft;
	const unsigned char *name;
	int err;
	u16 hash,checksum;
	struct exfat_chunk_data* chunk_data;
	struct exfat_chunk_name* chunk_name;

	int idisplay = 0;

	wlen = wleft = min(qname->len, EXFAT_MAX_NAMELEN);
	wstr = kmalloc(wlen * sizeof(*wstr), GFP_KERNEL);
	if (!wstr) {
		err = -ENOMEM;
		goto error;
	}
	pbwstr = wstr;

	name = qname->name;
	name_left = qname->len;

	MY_DEBUG(idisplay,"name=%s,len=%d\n",name,name_left);
	
	/* FIXME: maybe toupper and hash should be computed at same time */
	err = exfat_conv_c2u(nls, &name, &name_left, &wstr, &wleft);
	/* FIXME: shouldn't return -ENAMETOOLONG? */
	if (err)
		goto error;
		
	wlen -= wleft;		//actual length

//	for(i = 0 ; i < wlen ; i ++)
//		printk(" %x ",pbwstr[i]);
	MY_DEBUG(idisplay,"wlen=%d\n",wlen);

//	if(err = extfat_name_valid(wstr,wlen))
//		goto error;

	//0x85 + 0xC0 + n*0XC1	
	*nr_slots = 2 + (wlen + EXFAT_CHUNK_NAME_SIZE - 1) / EXFAT_CHUNK_NAME_SIZE;

	MY_DEBUG(idisplay,"nr_slots=%d\n",*nr_slots);
	
	//hash
	for (i = 0; i < wlen; i++) {
		__le16 uc;
		uc = exfat_towupper(sbi, pbwstr[i]);
		uc = cpu_to_le16(uc);
		hash = exfat_checksum16(hash, &uc, sizeof(uc));
	}


	MY_DEBUG(idisplay,"hash=%d",hash);

	//fill dirent
	exfat_fill_dirent(slots, 0,  ts, *nr_slots - 1,is_dir);
	checksum = exfat_checksum16(0, slots, 2);
	checksum = exfat_checksum16(checksum, (u8 *)slots + 4, EXFAT_CHUNK_SIZE - 4);

	//fill chunk data
	chunk_data = (struct exfat_chunk_data*)(slots + 1);
	exfat_fill_chunkdata(sbi,chunk_data, hash,cluster, wlen,is_dir);
	checksum = exfat_checksum16(checksum, chunk_data, EXFAT_CHUNK_SIZE);

	//fill chunk name
	chunk_name = (struct exfat_chunk_name*)(chunk_data + 1);
	exfat_fill_chunkname(chunk_name, pbwstr, wlen, *nr_slots - 2);
	for(i = 0 ; i < slots->sub_chunks - 1 ; i ++){
		chunk_name = (struct exfat_chunk_name*)(chunk_data + 1 + i);	
		checksum = exfat_checksum16(checksum, chunk_name, EXFAT_CHUNK_SIZE);
	}

	//fill the chksum(because the chksum use all the slots' data)
	slots->checksum = checksum;
	
error:
	MY_DEBUG(idisplay,"out err=%d\n",err);
	kfree(pbwstr);
	return err;
	
}

extern int g_index;
int g_flag = 0;

static int exfat_find_free_dentries(struct inode *dir,loff_t *ppos,int nr_slots,
									struct exfat_parse_data *pd)
{
	struct super_block *sb = dir->i_sb;
	struct exfat_sb_info *sbi = EXFAT_SB(sb);
//	struct exfat_chunk_dirent *dirent;
	struct buffer_head map_bh, *bh;
	sector_t blocknr, last_iblock,mappedblocks;
	unsigned long blocksize, blocks, offset, de_left = 0;
	loff_t pos,size;
	int ret = 0, err,found_slots = 0,flag = 0;	

//	printk("[exfat_find_free_dentries] g_index=%d\n",g_index);

	int idisplay = 0;
//	if(g_index >= 300)
//		idisplay = 1;
//	if(g_index >= 690)
//		idisplay = 0;

	blocksize = 1 << dir->i_blkbits;
	last_iblock = EXFAT_I(dir)->phys_size >> dir->i_blkbits;
//	last_iblock = sbi->clus_size >> dir->i_blkbits;


	MY_DEBUG(idisplay,"dir->i_size=%lld,blocksize=%d,dir->i_blkbits=%d\n",dir->i_size,blocksize,dir->i_blkbits);	

	/* FIXME: there is this limitation? Or limitation is cluster size? */
	if (EXFAT_I(dir)->phys_size & (blocksize - 1)) {
		exfat_fs_panic(sb, "invalid directory size (size %lld)",
			       EXFAT_I(dir)->phys_size);
		return -EIO;
	}

	size = EXFAT_I(dir)->phys_size;
	pos = *ppos;

	//Todo: max dentry limitation

	pd_init(pd);
	blocks = 0;
retry:	
	while (!ret && pos < size) {
		MY_DEBUG(idisplay,"ret=%d,pos=%lld,blocks=%d \n",ret,pos,blocks);	
		if (!blocks) {
			/* FIXME: readahead, probably last_iblock is too big */
			sector_t iblock;

			iblock = pos >> dir->i_blkbits;

			map_bh.b_state = 0;
			map_bh.b_blocknr = 0;
			mappedblocks = min(last_iblock-iblock,sbi->bpc);
//			map_bh.b_size = (last_iblock-iblock) << dir->i_blkbits;
			map_bh.b_size = mappedblocks << dir->i_blkbits;
			
			MY_DEBUG(idisplay,"last_iblock=%d,iblock=%d,pos=%lld \n",last_iblock,iblock,pos);	
			
			err = exfat_get_block(dir, iblock, &map_bh, 0);
			if (err)
				break;
			DEBUG0_ON(!buffer_mapped(&map_bh));

			blocknr = map_bh.b_blocknr;
			blocks = map_bh.b_size >> dir->i_blkbits;
		}
		offset = pos & (blocksize - 1);
		MY_DEBUG(idisplay,"flag=%d,blocknr=%d,offset=%d,de_left=%d,blocks=%d\n",flag,blocknr,offset,de_left,blocks);		
		bh = sb_bread(sb, blocknr);
		if (!bh) {
			err = -EIO;
			break;
		}

		if (flag) {
			//save the next block
			get_bh(bh);
			pd->bhs[pd->nr_bhs] = bh;
			pd->nr_bhs++;
			
			MY_DEBUG(idisplay,"pd->nr_bhs=%d,blocknr=%d \n",pd->nr_bhs,bh->b_blocknr);
		}

next_free_dentry:
		while (offset < bh->b_size) {
			struct exfat_chunk_dirent *dirent;
			unsigned long len, chunks;

			
			dirent = (void *)bh->b_data + offset;

//			MY_DEBUG(idisplay,"dirent->type = %x, offset = %d\n",dirent->type,offset);
			 if (!(dirent->type & EXFAT_TYPE_VALID)) {
				if(flag == 0){	//record the first one
					flag = 1;
					pd->size = nr_slots*EXFAT_CHUNK_SIZE;
					get_bh(bh);
					pd->bhs[0] = bh;
					pd->nr_bhs = 1;
					pd->bh_offset = offset;
					MY_DEBUG(idisplay,"found_slots = %d, offset = %d,org_pos=%d\n",found_slots,offset,pd->bh_offset);
	
				}
				offset += EXFAT_CHUNK_SIZE;
				pos += EXFAT_CHUNK_SIZE;
				found_slots ++;
				
//				MY_DEBUG(idisplay,"found_slots = %d, offset = %d,org_pos=%d\n",found_slots,offset,pd->bh_offset);
				
				if(found_slots == nr_slots){	//found all the free slots
					ret = 1;	
					err = 0;
					break;
				}

				continue;
			}

			//clean up
			flag = 0;
			found_slots = 0;		
			pd_brelse(pd);

			//forward to next slot  !!! can speedup here
			offset += EXFAT_CHUNK_SIZE;
			pos += EXFAT_CHUNK_SIZE;

//			MY_DEBUG(idisplay,"1111%s: blocknr %llu, de_left %lu, offset %lu,"
//			       " pos %llu, len %lu\n", __func__,
//			       (llu)bh->b_blocknr, de_left, offset, pos, len);
		}
next_block:
		*ppos = pos;
		brelse(bh);
		blocknr++;
		blocks--;
		MY_DEBUG(idisplay,"pos = %lld, blocknr = %d,blocks=%d\n",pos,blocknr,blocks);
	}

	if(!ret){	//pos >= dir->i_size
		u32 clus;
		
		//allocate new block
		clus = last_iblock >> sbi->bpc_bits;	
		err = exfat_add_cluster(dir, &clus);
		blocks = 0;
		last_iblock += sbi->bpc;
		
		
/*		exfat_get_block(dir, last_iblock + 1, &map_bh, 1);
		blocknr = map_bh.b_blocknr;
		blocks = map_bh.b_size >> dir->i_blkbits;
*/
		size += sbi->clus_size;
		MY_DEBUG(idisplay,"clus=%d,size=%lld \n",clus,size);

//		g_flag = 1;

//		MY_DEBUG(idisplay,"blocknr=%d,last_iblock=%d,size=%lld,blocks=%d \n",blocknr,last_iblock,size,blocks);

		goto retry;
	}
	MY_DEBUG(idisplay,"out err =%d\n",err);	
	
	return err;
}

static int exfat_write_entries(	struct inode *dir,struct exfat_chunk_dirent * slots,
								loff_t pos,int nr_slots,struct exfat_parse_data *pd)
{
	struct super_block *sb = dir->i_sb;
	struct buffer_head  *bh;
	sector_t blocknr;
	unsigned long blocksize, offset, de_left,index;
	loff_t cur_pos;
	int err = 0,len;
	int idisplay = 0;

//	if(g_index >= 300)
//		idisplay = 1;
//	if(g_index >= 690)
//		idisplay = 0;


	blocksize = 1 << dir->i_blkbits;

	/* FIXME: there is this limitation? Or limitation is cluster size? */
	if (EXFAT_I(dir)->phys_size & (blocksize - 1)) {
		exfat_fs_panic(sb, "invalid directory size (size %lld)",
			       EXFAT_I(dir)->phys_size);
		return -EIO;
	}

	cur_pos = pos - nr_slots*EXFAT_CHUNK_SIZE;
	MY_DEBUG(idisplay,"cur_pos=%lld,pos=%lld,nr_slots=%d\n",cur_pos,pos,nr_slots);
	DEBUG_ON(cur_pos < 0,"bug");
	
	de_left = nr_slots * EXFAT_CHUNK_SIZE;
	index = 0;
	offset = cur_pos & (blocksize - 1);		//offset in block
	while(cur_pos < pos){
		blocknr = pd->bhs[index]->b_blocknr;
		bh = sb_bread(sb, blocknr);
		if (!bh) {
			err = -EIO;
			break;
		}

		len = min(de_left,blocksize - offset);
		memcpy(bh->b_data + offset,slots,len);
		slots += len/EXFAT_CHUNK_SIZE;

		MY_DEBUG(idisplay,"[index=%d,blocknr=%d]len=%d,de_left=%d,offset=%d \n",index,blocknr,len,de_left,offset);	
	
		de_left -= len;
		cur_pos += len;
		offset += len;

		mark_buffer_dirty(bh);
//		sync_dirty_buffer(bh);
		brelse(bh);

		if(offset >= blocksize){
			offset -= blocksize;
			index ++;
			DEBUG_ON(pd->nr_bhs < index,"bug");
		}
	}

	MY_DEBUG(idisplay,"err=%d \n",err);
	
	return err;
}

static int exfat_write_entries_1(	struct inode *dir,struct exfat_chunk_dirent * slots,
								loff_t pos,int nr_slots,struct exfat_parse_data *pd)
{
	struct super_block *sb = dir->i_sb;
//	struct exfat_chunk_dirent *dirent;
//	struct exfat_chunk_data *data;
//	struct exfat_chunk_name *name;
	struct buffer_head map_bh, *bh;
	sector_t blocknr, last_iblock;
	unsigned long blocksize, blocks, offset, de_left;
	loff_t cur_pos;
	int err = 0,len;
	int idisplay = 0;

	blocksize = 1 << dir->i_blkbits;
	last_iblock = dir->i_size >> dir->i_blkbits;

	/* FIXME: there is this limitation? Or limitation is cluster size? */
	if (dir->i_size & (blocksize - 1)) {
		exfat_fs_panic(sb, "invalid directory size (size %lld)",
			       dir->i_size);
		return -EIO;
	}

	cur_pos = pos - nr_slots*EXFAT_CHUNK_SIZE;
	MY_DEBUG(idisplay,"cur_pos=%lld,pos=%lld,nr_slots=%d\n",cur_pos,pos,nr_slots);
	DEBUG_ON(cur_pos < 0,"bug");
	

	de_left = nr_slots * EXFAT_CHUNK_SIZE;
	blocks = 0;
	while(cur_pos < pos){
		if (!blocks) {
			/* FIXME: readahead, probably last_iblock is too big */
			sector_t iblock;

			iblock = cur_pos >> dir->i_blkbits;
//			if(pos & (blocksize - 1))
//				;
//			else
//				iblock -= 1;
MY_DEBUG(idisplay,"iblock=%d,last_iblock=%d \n",iblock,last_iblock);
			if(iblock < 0)
				iblock = 0;

			map_bh.b_state = 0;
			map_bh.b_blocknr = 0;
			map_bh.b_size = (last_iblock-iblock) << dir->i_blkbits;

			err = exfat_get_block(dir, iblock, &map_bh, 0);
			if (err)
				break;
			DEBUG0_ON(!buffer_mapped(&map_bh));

			blocknr = map_bh.b_blocknr;
			blocks = map_bh.b_size >> dir->i_blkbits;
		}

		MY_DEBUG(idisplay,"blocknr=%d,blocks=%d,nr_slots=%d \n",blocknr,blocks,nr_slots);	

		offset = cur_pos & (blocksize - 1);		//offset in block
		bh = sb_bread(sb, blocknr);
		if (!bh) {
			err = -EIO;
			break;
		}

		if(offset < blocksize){
	//		exfat_dump_bh(bh);
			len = min(de_left,blocksize - offset);
			memcpy(bh->b_data + offset,slots,len);
			slots += len/EXFAT_CHUNK_SIZE;

			MY_DEBUG(idisplay,"len=%d,de_left=%d,offset=%d \n",len,de_left,offset);	
			
			de_left -= len;
			cur_pos += len;
			offset += len;

	//		exfat_dump_bh(bh);

			mark_buffer_dirty(bh);

			sync_dirty_buffer(bh);
			
			brelse(bh);
		}
		blocknr++;
		blocks--;
	}

	MY_DEBUG(idisplay,"err=%d \n",err);
	
	return err;
}

static int exfat_add_entries(	struct inode *dir, void *slots, int nr_slots,
								struct exfat_parse_data *pd)
{
	struct super_block *sb = dir->i_sb;
//	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	struct exfat_inode_info *exi = EXFAT_I(dir);
	int err = 0;
	loff_t pos = 0;

	int idisplay = 0;

	MY_DEBUG(idisplay,"In,nr_slots=%d\n",nr_slots);

	//find free entries
	err = exfat_find_free_dentries(dir, &pos, nr_slots,pd);
	if(err)
		goto error;

	MY_DEBUG(idisplay,"1\n");	

	//write dentry data to disk
	err = exfat_write_entries(dir, slots, pos, nr_slots,pd);
	if(err)
		goto error;
	
error:
	MY_DEBUG(idisplay,"out err=%d",err);
	return err;
}

static int exfat_name_exist(struct inode *dir,struct qstr *qname, int is_dir)
{
	struct exfat_sb_info *sbi = EXFAT_SB(dir->i_sb);
	struct nls_table *nls = sbi->opts.nls;
	struct exfat_parse_callback pcb;
	struct lookup_parse_data pcb_data = {0};
	const unsigned char *name;
	wchar_t *wstr;
	unsigned int i, name_left, wlen, wleft;
	loff_t pos;
	int err;

	/* FIXME: can't use converted name/hash? */

	wlen = wleft = min(qname->len, EXFAT_MAX_NAMELEN);
	pcb_data.wstr = wstr = kmalloc(wlen * sizeof(*wstr), GFP_KERNEL);
	if (!wstr) {
		err = -ENOMEM;
		goto error;
	}

	name = qname->name;
	name_left = qname->len;
	
	/* FIXME: maybe toupper and hash should be computed at same time */
	err = exfat_conv_c2u(nls, &name, &name_left, &wstr, &wleft);
	/* FIXME: shouldn't return -ENAMETOOLONG? */
	if (err)
		goto error;
	pcb_data.wlen = wlen - wleft;

	pcb_data.hash = 0;
	for (i = 0; i < pcb_data.wlen; i++) {
		__le16 uc;
		pcb_data.wstr[i] = exfat_towupper(sbi, pcb_data.wstr[i]);
		uc = cpu_to_le16(pcb_data.wstr[i]);
		pcb_data.hash = exfat_checksum16(pcb_data.hash, &uc, sizeof(uc));
	}

	pcb_data.inode = NULL;
	pcb.parse = lookup_parse;
	pcb.data = &pcb_data;

	/* FIXME: pos can be previous lookuped position. */
	pos = 0;
	err = exfat_parse_dir(NULL, dir, &pos, &pcb);
	if (err)
		goto error;

	kfree(pcb_data.wstr);

	/* FIXME: can cache the position for next lookup()?
	 * Assume sequential locality (stat() for all files after readdir()) */
	if(is_dir == pcb_data.is_dir)
		return EEXIST;

	return 0;

error:
	kfree(pcb_data.wstr);
	return 0;

}

static int exfat_create_entries(	struct inode *dir,struct qstr *qname, int is_dir,
			  					int cluster, struct timespec *ts,
			  					struct inode **inode)
{
	struct super_block *sb = dir->i_sb;
	struct exfat_chunk_dirent *slots = NULL;
	unsigned int len;
	int err = 0, nr_slots=0;
	struct exfat_parse_data pd;
	int idisplay = 0;

	MY_DEBUG(idisplay,"[exfat_create_entries] In,name=%s\n",qname->name);
	len = exfat_striptail_len(qname);
	if (len == 0)
		return -ENOENT;

	slots = kmalloc(sizeof(*slots) * EXFAT_MAX_CHUNK, GFP_NOFS);
	if (slots == NULL)
		goto cleanup;
//	printk("[exfat_create_entries] 1,is_dir = %d\n",is_dir);
	
//	err = exfat_name_exist(dir, qname, is_dir);
//	if(err)
//		goto cleanup;
		
//	printk("[exfat_create_entries] 2\n");
	err = exfat_build_slots(dir,qname,
						is_dir,cluster, ts, slots, &nr_slots);
	if(err)
		goto cleanup;
		
//	printk("[exfat_create_entries] 3\n");
	pd_init(&pd);
	err = exfat_add_entries(dir, slots, nr_slots,&pd);
	if(err)
		goto cleanup;
//	printk("[exfat_create_entries] 4\n");
	*inode = exfat_iget(sb, &pd, slots, (struct exfat_chunk_data *)(slots+1));
	pd_brelse(&pd);

	MY_DEBUG(idisplay,"created ino : %d",(*inode)->i_ino);
	
	/* update timestamp */
	dir->i_ctime = dir->i_mtime = dir->i_atime = *ts;
	if (IS_DIRSYNC(dir))
		exfat_sync_inode(dir);
	else
		mark_inode_dirty(dir);
		
cleanup:
//	printk("[exfat_create_entries] out, err=%d\n",err);
	kfree(slots);
	return err;
}

static int exfat_create(struct inode *dir, struct dentry *dentry, int mode,
		       struct nameidata *nd)
{
	struct super_block *sb = dir->i_sb;
	struct timespec ts;
	struct inode *inode = NULL;
	int err;

	int idisplay = 0;

//	printk("\n[exfat_create]PID=%d, In,name=%s\n",current->pid,dentry->d_name.name);

	//should we lock fs?
	lock_super(sb);
	
	//step1:find avaiable slots for this entry,write the entry
	ts = CURRENT_TIME_SEC;
	err = exfat_create_entries(dir,&dentry->d_name,0,0,&ts,&inode);
	if(err)
		goto error;

	MY_DEBUG(idisplay,"1 \n");

	if (IS_ERR(inode)) {
		err = PTR_ERR(inode);
		goto error;
	}

	MY_DEBUG(idisplay,"2 \n");

	inode->i_version++;
	inode->i_mtime = inode->i_atime = inode->i_ctime = ts;
	/* timestamp is already written, so mark_inode_dirty() is unneeded. */

	dentry->d_time = dentry->d_parent->d_inode->i_version;
	d_instantiate(dentry, inode);

error:
	unlock_super(sb);

	MY_DEBUG(idisplay,"out err=%d \n",err);
	
	return err;
	
}

static int exfat_alloc_new_dir(struct inode *dir,int *clus)
{
	struct super_block *sb = dir->i_sb;
//	struct exfat_sb_info *sbi = EXFAT_SB(sb);
	int err;

	err = exfat_allocate_clusters(dir,clus, 1);
	if(!err)
		err = exfat_clear_cluster(dir, *clus);
		
	return err;
}

int g_index = 0;
static int exfat_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
	struct super_block *sb = dir->i_sb;
	struct timespec ts;
	struct inode *inode;
	int err,clusnr;
	
	g_index ++;
	int idisplay = 0;

//	if(g_index >= 650)
		printk("\n[exfat_mkdir]PID=%d, In,name=%s\n",current->pid,dentry->d_name.name);
	//should we lock fs?
	lock_super(sb);

	MY_DEBUG(idisplay,"1\n");
	
	err = exfat_alloc_new_dir(dir,&clusnr);
	if(err)
		goto error;
	
	//step1:find avaiable slots for this entry,write the entry
	ts = CURRENT_TIME_SEC;

	MY_DEBUG(idisplay,"2,err=%d,clusnr=%d\n",err,clusnr);
	
	err = exfat_create_entries(dir,&dentry->d_name,1,clusnr,&ts,&inode);
	if(err)
		goto error;
//	printk("[exfat_mkdir] 3\n");
	if (IS_ERR(inode)) {
		err = PTR_ERR(inode);
		goto error;
	}
	err = exfat_write_entry(inode,clusnr,EXFAT_ENT_EOF);
	if(err)
		goto error;

	err = exfat_bits_set(sb,clusnr,1,1);
	if(err)
		goto error;

	MY_DEBUG(idisplay,"4\n");
		
	inode->i_version++;
	inode->i_mtime = inode->i_atime = inode->i_ctime = ts;
	/* timestamp is already written, so mark_inode_dirty() is unneeded. */

	dentry->d_time = dentry->d_parent->d_inode->i_version;
	d_instantiate(dentry, inode);

	unlock_super(sb);
//	printk("[exfat_mkdir] ok out err=%d\n",err);	
	return 0;

error:
	exfat_bits_set(sb,clusnr,1,0);
	unlock_super(sb);	
	return err;
}

static int exfat_find_rm_entries(struct inode *dir, struct qstr* qname,int delete)
{
	struct exfat_sb_info *sbi = EXFAT_SB(dir->i_sb);
	struct nls_table *nls = sbi->opts.nls;
	struct exfat_parse_callback pcb;
	struct lookup_parse_data pcb_data;
	const unsigned char *name;
	wchar_t *wstr;
	unsigned int i, name_left, wlen, wleft;
	loff_t pos;
	int err;

	int idisplay = 0;
	MY_DEBUG(idisplay,"remove entries! \n");

	/* FIXME: can't use converted name/hash? */

	wlen = wleft = min(qname->len, EXFAT_MAX_NAMELEN);
	pcb_data.wstr = wstr = kmalloc(wlen * sizeof(*wstr), GFP_KERNEL);
	if (!wstr) {
		err = -ENOMEM;
		goto error;
	}

	name = qname->name;
	name_left = qname->len;

	
	/* FIXME: maybe toupper and hash should be computed at same time */
	err = exfat_conv_c2u(nls, &name, &name_left, &wstr, &wleft);
	/* FIXME: shouldn't return -ENAMETOOLONG? */
	if (err)
		goto error;
	pcb_data.wlen = wlen - wleft;

	pcb_data.hash = 0;
	for (i = 0; i < pcb_data.wlen; i++) {
		__le16 uc;
		pcb_data.wstr[i] = exfat_towupper(sbi, pcb_data.wstr[i]);
		uc = cpu_to_le16(pcb_data.wstr[i]);
		pcb_data.hash = exfat_checksum16(pcb_data.hash, &uc, sizeof(uc));
	}

	pcb_data.inode = NULL;
	pcb_data.delete_flag = delete;
	pcb.parse = lookup_parse;
	pcb.data = &pcb_data;
	
	pos = 0;

	err = exfat_parse_dir(NULL, dir, &pos, &pcb);
	if (err)
		goto error;

	kfree(pcb_data.wstr);

	/* FIXME: can cache the position for next lookup()?
	 * Assume sequential locality (stat() for all files after readdir()) */

	return 0;
	
error:
	kfree(pcb_data.wstr);
	return ERR_PTR(err);

}


static int exfat_unlink(struct inode *dir, struct dentry *dentry)
{
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = dir->i_sb;
	int err;

	int idisplay = 0;

	MY_DEBUG(idisplay,"inode_nlink=%d , icount = %d \n",inode->i_nlink,inode->i_count.counter);

	lock_super(sb);

	err = exfat_find_rm_entries(dir, &dentry->d_name,1);
	if(err)
		goto out;

	MY_DEBUG(idisplay,"inode_nlink=%d , icount = %d \n",inode->i_nlink,inode->i_count.counter);
	clear_nlink(inode);
	inode->i_mtime = inode->i_atime = CURRENT_TIME_SEC;	
//	mark_inode_dirty(inode);

	exfat_detach(inode);	
out:
	unlock_super(sb);

	return err;
}


static int exfat_dir_empty(struct inode *dir)
{
	struct super_block *sb = dir->i_sb;
	struct buffer_head map_bh, *bh;
	sector_t blocknr, last_iblock;
	unsigned long blocksize, blocks, offset, de_left = 0;
	loff_t pos;
	int idisplay = 0,err;

	blocksize = 1 << dir->i_blkbits;
	last_iblock = dir->i_size >> dir->i_blkbits;

	MY_DEBUG(idisplay,"dir->i_size=%lld,blocksize=%d,dir->i_blkbits=%d\n",dir->i_size,blocksize,dir->i_blkbits);	

	/* FIXME: there is this limitation? Or limitation is cluster size? */
	if (dir->i_size & (blocksize - 1)) {
		exfat_fs_panic(sb, "invalid directory size (size %lld)",
			       dir->i_size);
		return -EIO;
	}

	//Todo: max dentry limitation


	pos = 0;
	blocks = 0;
retry:	
	while (pos < dir->i_size) {
		MY_DEBUG(idisplay,"pos=%lld,blocks=%d \n",pos,blocks);	
		if (!blocks) {
			/* FIXME: readahead, probably last_iblock is too big */
			sector_t iblock;
			iblock = pos >> dir->i_blkbits;

			map_bh.b_state = 0;
			map_bh.b_blocknr = 0;
			map_bh.b_size = (last_iblock-iblock) << dir->i_blkbits;
			
			MY_DEBUG(idisplay,"last_iblock=%d,iblock=%d,pos=%lld \n",last_iblock,iblock,pos);	
			
			err = exfat_get_block(dir, iblock, &map_bh, 0);
			if (err)
				break;
			DEBUG0_ON(!buffer_mapped(&map_bh));

			blocknr = map_bh.b_blocknr;
			blocks = map_bh.b_size >> dir->i_blkbits;
		}
		offset = pos & (blocksize - 1);
		MY_DEBUG(idisplay,"blocknr=%d,offset=%d,de_left=%d,blocks=%d\n",blocknr,offset,de_left,blocks);		
		bh = sb_bread(sb, blocknr);
		if (!bh) {
			err = -EIO;
			break;
		}

		if (de_left) {
			unsigned long len;

			get_bh(bh);

			len = min(de_left, blocksize);
			de_left -= len;
			offset += len;
			pos += len;

			MY_DEBUG(idisplay,"%s: blocknr %llu, de_left %lu, offset %lu,"
			       " pos %llu, len %lu\n", __func__,
			       (llu)bh->b_blocknr, de_left, offset, pos, len);
			if (!de_left)
				goto next_free_dentry;

			goto next_block;
		}

next_free_dentry:
		while (offset < bh->b_size) {
			struct exfat_chunk_dirent *dirent;
			unsigned long len, chunks;

			dirent = (void *)bh->b_data + offset;

			MY_DEBUG(idisplay,"dirent->type = %x, offset = %d\n",dirent->type,offset);
			if (dirent->type == 0x03) {
				return -EEXIST;
			} else if (!(dirent->type & EXFAT_TYPE_VALID)) {
				offset += EXFAT_CHUNK_SIZE;
				pos += EXFAT_CHUNK_SIZE;
				continue;
			}
			else 
				return -EEXIST;

		}
next_block:
		brelse(bh);
		blocknr++;
		blocks--;
		MY_DEBUG(idisplay,"pos = %lld, blocknr = %d,blocks=%d\n",pos,blocknr,blocks);
	}

	return 0;
}


static int exfat_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = dir->i_sb;
	int err;

	int idisplay = 0;

	MY_DEBUG(idisplay,"In .... \n");
	lock_super(sb);

	//find if the folder empty?
	err = exfat_dir_empty(inode);
	if(err)
		goto out;

	err = exfat_find_rm_entries(dir, &dentry->d_name,1);
	if(err)
		goto out;

	MY_DEBUG(idisplay,"B[dir%d:nlink=%d , icount = %d],[inode%d:nlink=%d , icount = %d]\n",dir->i_ino,dir->i_nlink,dir->i_count.counter,inode->i_ino,inode->i_nlink,inode->i_count.counter);
	drop_nlink(dir);
	clear_nlink(inode);

	MY_DEBUG(idisplay,"A[dir%d:nlink=%d , icount = %d],[inode%d:nlink=%d , icount = %d]\n",dir->i_ino,dir->i_nlink,dir->i_count.counter,inode->i_ino,inode->i_nlink,inode->i_count.counter);

	inode->i_mtime = inode->i_atime = CURRENT_TIME_SEC;	
//	mark_inode_dirty(inode);

	exfat_detach(inode);
	
out:
	unlock_super(sb);
	return err;
}

static int exfat_store_pd(struct inode *inode,struct exfat_parse_data *pd)
{
	struct super_block * sb = inode->i_sb;
	struct exfat_inode_info *exi = EXFAT_I(inode);
	struct buffer_head *bh;
	int i ;
	int idisplay = 0;
	
	pd->nr_bhs = exi->nr_bhs;
	for( i = 0 ; i < pd->nr_bhs && i < EXFAT_DE_MAX_BH; i ++ ){
		bh = sb_bread(sb, exi->de_blocknr[i]);
		if(!bh)
			return -EIO;			
		pd->bhs[i] = bh;
	}
	pd->bh_offset = exi->de_offset;
	pd->size = exi->de_size;

	MY_DEBUG(idisplay,"[inode %d]nr_bhs=%d,bh_offset=%d,size=%d \n",inode->i_ino,pd->nr_bhs,pd->bh_offset,pd->size);

	return 0;
}

static int exfat_rename2(struct inode *old_dir, struct dentry *old_dentry,
		       struct inode *new_dir, struct dentry *new_dentry)
{
	int err,is_dir;
	struct super_block *sb = old_dir->i_sb;
	struct inode* old_inode, *new_inode,*created_inode;
	struct exfat_parse_data pd,old_pd;
	int clus,backup_clus;
	struct timespec ts;
	int idisplay = 0;

	MY_DEBUG(idisplay,"oldname=%s,newname=%s \n",old_dentry->d_name.name,new_dentry->d_name.name);

	lock_super(sb);
	
	ts = CURRENT_TIME_SEC;
	err = exfat_find_rm_entries(old_dir, &old_dentry->d_name,0);
	if(err)	//the old folder doesn't exist
		goto out;

	MY_DEBUG(idisplay,"find old \n");	

	old_inode = old_dentry->d_inode;
	is_dir = S_ISDIR(old_inode->i_mode);

	new_inode = new_dentry->d_inode;
	MY_DEBUG(idisplay,"new_inode=%x \n",new_inode);
	if(new_inode){
		MY_DEBUG(idisplay,"new inode \n");
		if(is_dir){
			err = exfat_dir_empty(new_inode);
			if(err)
				goto out;
		}
		exfat_store_pd(new_inode,&pd);
		exfat_detach(new_inode);
		clus = EXFAT_I(new_inode)->clusnr;
	}
	else{
		err = exfat_create_entries(new_dir,&new_dentry->d_name,is_dir,0,&ts,&created_inode);
		if(err)
			goto out;
		exfat_store_pd(created_inode,&pd);
		clus = EXFAT_I(created_inode)->clusnr;
	}

	exfat_store_pd(old_inode,&old_pd);
	backup_clus = EXFAT_I(old_inode)->clusnr;
	
	exfat_detach(old_inode);
	exfat_attach(old_inode,&pd);

//	MY_DEBUG(idisplay,"old ino:%d,old clus:%d   new ino:%d,new clus:%d \n",old_inode->i_ino,backup_clus,new_inode->i_ino,clus);

	if (IS_DIRSYNC(new_dir)) {
		MY_DEBUG(idisplay,"sync \n");
		err = exfat_sync_inode(old_inode);
		if (err)
			goto error_inode;
	} 
	else{
		MY_DEBUG(idisplay,"mark dirty \n");
		mark_inode_dirty(old_inode);
	}	

	MY_DEBUG(idisplay,"add new \n");
	
	err = exfat_find_rm_entries(old_dir,&old_dentry->d_name,1);
	if(err)
		goto out;
		
	drop_nlink(old_dir);
	inc_nlink(new_dir);

	if (new_inode) {
		drop_nlink(new_inode);
		new_inode->i_ctime = ts;
	}
	
	MY_DEBUG(idisplay," rm old \n");

	goto out;
	
error_inode:
	exfat_detach(old_inode);
	exfat_attach(old_inode,&old_pd);

	if(new_inode){
		exfat_attach(new_inode,&pd);
	}
	else{
		err = exfat_find_rm_entries(new_dir,&new_dentry->d_name,1);
		if(err)
			goto out;
	}

out:
	MY_DEBUG(idisplay,"out err=%d",err);

	pd_brelse(&pd);
	pd_brelse(&old_pd);

	unlock_super(sb);
	return err;
}


static int exfat_rename(struct inode *old_dir, struct dentry *old_dentry,
		       struct inode *new_dir, struct dentry *new_dentry)
{
	int err,is_dir;
	struct super_block *sb = old_dir->i_sb;
	struct inode* old_inode, *new_inode,*created_inode;
	struct exfat_parse_data pd,old_pd;
	int backup_clus = 0;
	struct timespec ts;
	int idisplay = 0;

	MY_DEBUG(idisplay,"oldname=%s,newname=%s \n",old_dentry->d_name.name,new_dentry->d_name.name);

	lock_super(sb);
	
	ts = CURRENT_TIME_SEC;
	err = exfat_find_rm_entries(old_dir, &old_dentry->d_name,0);
	if(err)	//the old folder doesn't exist
		goto out;

	MY_DEBUG(idisplay,"find old \n");	

	old_inode = old_dentry->d_inode;
	is_dir = S_ISDIR(old_inode->i_mode);

	new_inode = new_dentry->d_inode;
	MY_DEBUG(idisplay,"new_inode=%x,is_dir=%d \n",new_inode,is_dir);
	pd_init(&pd);
	if(new_inode){
		MY_DEBUG(idisplay,"new inode \n");
		if(is_dir){
			err = exfat_dir_empty(new_inode);
			if(err)
				goto out;
		}
		exfat_store_pd(new_inode,&pd);
		exfat_detach(new_inode);
		backup_clus = EXFAT_I(new_inode)->clusnr;

	}
	else{
		err = exfat_create_entries(new_dir,&new_dentry->d_name,is_dir,0,&ts,&created_inode);
		if(err)
			goto out;
		exfat_store_pd(created_inode,&pd);
	}

	pd_init(&old_pd);
	exfat_store_pd(old_inode,&old_pd);
	exfat_detach(old_inode);

	MY_DEBUG(idisplay,"backup_clus=%d \n",backup_clus);

	if(new_inode)
		EXFAT_I(new_inode)->clusnr = EXFAT_I(old_inode)->clusnr;
	else
		EXFAT_I(created_inode)->clusnr = EXFAT_I(old_inode)->clusnr;
		
	if (IS_DIRSYNC(new_dir)) {
		MY_DEBUG(idisplay,"sync \n");
		if(new_inode)
			err = exfat_sync_inode(new_inode);
		else
			err = exfat_sync_inode(created_inode);

		if (err)
			goto error_inode;
	} 
	else{
		MY_DEBUG(idisplay,"mark dirty \n");
		if(new_inode)
			mark_inode_dirty(new_inode);
		else
			mark_inode_dirty(created_inode);
	}	
	
	err = exfat_find_rm_entries(old_dir,&old_dentry->d_name,1);
	if(err)
		goto out;
		
	drop_nlink(old_dir);
	inc_nlink(new_dir);

	if (new_inode) {
		drop_nlink(new_inode);
		new_inode->i_ctime = ts;
	}
	
	MY_DEBUG(idisplay," rm old \n");

	goto out;
	
error_inode:

	MY_DEBUG(idisplay,"err_inode\n");
	
	exfat_attach(old_inode,&old_pd);
	mark_inode_dirty(old_inode);

	if(new_inode){
		exfat_attach(new_inode,&pd);
		mark_inode_dirty(new_inode);
	}
	else
		err = exfat_find_rm_entries(new_dir,&new_dentry->d_name,1);

out:
	pd_brelse(&pd);
	pd_brelse(&old_pd);
	
	MY_DEBUG(idisplay,"out err=%d \n",err);

	unlock_super(sb);
	
	return err;
}


struct readdir_parse_data {
	void *dirbuf;
	filldir_t filldir;
	void *str;
};

#define EXFAT_MAX_NAME_BUFSIZE	(EXFAT_MAX_NAMELEN * NLS_MAX_CHARSET_SIZE)
#define DO_READDIR_CHECKSUM

static int readdir_parse(struct inode *dir, loff_t pos,
			 struct exfat_parse_data *pd,
			 struct exfat_parse_callback *pcb)
{
	struct super_block *sb = dir->i_sb;
	struct nls_table *nls = EXFAT_SB(sb)->opts.nls;
	struct readdir_parse_data *pcb_data = pcb->data;
	struct exfat_parse_iter_data pd_iter;
	struct exfat_chunk_dirent *dirent;
	struct exfat_chunk_data *data;
	struct exfat_chunk_name *name;
	struct inode *inode;
	unsigned long ino;
	unsigned int d_type;
#ifdef DO_READDIR_CHECKSUM
	u16 csum;
#endif
	u8 de_namelen;
	const __le16 *ustr;
	unsigned char *str;
	unsigned int uleft, str_left, str_len;
	int err;

	int idisplay = 0;
	unsigned char *strback;

	/* FIXME: should cache whether checksum was done */

	dirent = pd_iter_first_de(pd, &pd_iter);
	if (dirent->type != EXFAT_TYPE_DIRENT)
		return PARSE_NEXT;
	/* ->sub_chunk was already checked by caller. */
	DEBUG_ON((1 + dirent->sub_chunks) << EXFAT_CHUNK_BITS != pd->size,
		 "sub_chunks %u, pd->size %lu\n", dirent->sub_chunks, pd->size);
	MY_DEBUG(idisplay,"exFAT dirent: 0x%02x found\n", dirent->type);
#ifdef DO_READDIR_CHECKSUM
	/* Skip dirent->checksum field */
	csum = exfat_checksum16(0, dirent, 2);
	csum = exfat_checksum16(csum, (u8 *)dirent + 4, EXFAT_CHUNK_SIZE - 4);
#endif

	data = pd_iter_next_de(pd, &pd_iter);
	if (!data)
		return PARSE_NEXT;
	if (data->type != EXFAT_TYPE_DATA)
		return PARSE_NEXT;
	
	de_namelen = data->name_len;
	
	MY_DEBUG(idisplay,"exFAT data: 0x%02x found,de_namelen=%d\n", data->type,de_namelen);
	
#ifdef DO_READDIR_CHECKSUM
	csum = exfat_checksum16(csum, data, EXFAT_CHUNK_SIZE);
#endif

	if (!pcb_data->str) {
		/* FIXME: too big, should be own kmem_cachep */
		pcb_data->str = kmalloc(EXFAT_MAX_NAME_BUFSIZE, GFP_KERNEL);
		if (!pcb_data->str)
			return -ENOMEM;
	}

	str = pcb_data->str;
	strback = str;
	str_left = EXFAT_MAX_NAME_BUFSIZE;
	while (de_namelen) {
		name = pd_iter_next_de(pd, &pd_iter);
		if (!name)
			return PARSE_NEXT;
		if (name->type != EXFAT_TYPE_NAME)
			return PARSE_NEXT;

		ustr = name->name;
		uleft = min_t(u8, de_namelen, EXFAT_CHUNK_NAME_SIZE);
		de_namelen -= uleft;

		err = exfat_conv_u2c(nls, &ustr, &uleft, &str, &str_left);
		//printk("after u2c: str=%s\n",pcb_data->str);
		if (err) {
			/* FIXME: if char was invalid, what to do */
			if (err == -EINVAL) {
				printk(KERN_ERR "exFAT: invalid char 0x%04x in"
				       " file name\n", le16_to_cpu(*ustr));
			}
			/* FIXME: shouldn't return -ENAMETOOLONG */
			return err;
		}
#ifdef DO_READDIR_CHECKSUM
		csum = exfat_checksum16(csum, name, EXFAT_CHUNK_SIZE);
#endif
	}

	MY_DEBUG(idisplay,"str=%s \n",strback);

	/* Checksum of the remaining entries (may not be exfat_chunk_name) */
	while ((name = pd_iter_next_de(pd, &pd_iter)) != NULL) {
		if (!(name->type & EXFAT_TYPE_SUBCHUNK))
			return PARSE_NEXT;
#ifdef DO_READDIR_CHECKSUM
		csum = exfat_checksum16(csum, name, EXFAT_CHUNK_SIZE);
#endif
	}
#ifdef DO_READDIR_CHECKSUM
	if (le16_to_cpu(dirent->checksum) != csum) {
		exfat_fs_panic(sb, "checksum failed for directory entry"
			       " in readdir (0x%04x != 0x%04x)",
			       le16_to_cpu(dirent->checksum), csum);
		return -EIO;
	}
#endif

	inode = exfat_ilookup(sb, pd->bhs[0]->b_blocknr, pd->bh_offset);
	if (inode) {
		ino = inode->i_ino;
		iput(inode);
	} else
		ino = iunique(sb, EXFAT_RESERVED_INO);

	if (dirent->attrib & cpu_to_le16(EXFAT_ATTR_DIR))
		d_type = DT_DIR;
	else
		d_type = DT_REG;

	str_len = EXFAT_MAX_NAME_BUFSIZE - str_left;

	str[0] = '\0';
	MY_DEBUG(idisplay,"%s: ent %p, str %s, len %u, pos %lld, ino %lu, type 0x%x\n",
	       __func__, pcb_data->dirbuf, (char *)pcb_data->str, str_len,
	       pos, ino, d_type);

	if (pcb_data->filldir(pcb_data->dirbuf, pcb_data->str, str_len, pos,
			      ino, d_type))
		return PARSE_STOP;

	return PARSE_NEXT;
}

static int exfat_readdir(struct file *filp, void *dirbuf, filldir_t filldir)
{
	struct inode *dir = filp->f_path.dentry->d_inode;
	struct exfat_parse_callback pcb;
	struct readdir_parse_data pcb_data;
	loff_t pos;
	int err;

	int idisplay = 0;

	MY_DEBUG(idisplay,"f_pos=%lld \n",filp->f_pos);
	/* exFAT doesn't have "." and "..", so returns fake entries */
	while (filp->f_pos < 2) {
		unsigned long ino;
		int len = filp->f_pos + 1;

		if (filp->f_pos == 0)
			ino = dir->i_ino;
		else
			ino = parent_ino(filp->f_path.dentry);

		MY_DEBUG(idisplay,"%s: ent %p, .., len %u, pos %lld, ino %lu, type 0x%x\n",
		       __func__, dirbuf, len, filp->f_pos, ino, DT_DIR);
		if (filldir(dirbuf, "..", len, filp->f_pos, ino, DT_DIR))
			return 0;

		filp->f_pos++;
	}
	if (filp->f_pos > 2 && (filp->f_pos & (EXFAT_CHUNK_SIZE - 1)))
		return -ENOENT;

	pcb_data.dirbuf = dirbuf;
	pcb_data.filldir = filldir;
	pcb_data.str = NULL;
	pcb.parse = readdir_parse;
	pcb.data = &pcb_data;

	pos = filp->f_pos & ~(loff_t)(EXFAT_CHUNK_SIZE - 1);
	err = exfat_parse_dir(filp, dir, &pos, &pcb);
	if (pos < 2)
		filp->f_pos = 2;
	else
		filp->f_pos = pos;

	kfree(pcb_data.str);

	MY_DEBUG(idisplay,"pos=%lld \n",pos);

	return err;
}

struct rootdir_parse_data {
	u32 bitmap_clusnr;
	u64 bitmap_size;
	u32 upcase_checksum;
	u32 upcase_clusnr;
	u64 upcase_size;
};

static int rootdir_parse(struct inode *dir, loff_t pos,
			 struct exfat_parse_data *pd,
			 struct exfat_parse_callback *pcb)
{
	struct exfat_chunk_bitmap *bitmap;
	struct exfat_chunk_upcase *upcase;
	struct rootdir_parse_data *pcb_data = pcb->data;

	bitmap = pd_get_chunk(pd);
	switch (bitmap->type) {
	case EXFAT_TYPE_BITMAP:
		if (pcb_data->bitmap_clusnr) {
			/* FIXME: for this, what do we do */
			PRINTK(KERN_WARNING "exFAT: found another"
			       " free space bitmap, ignored\n");
			break;
		}
		pcb_data->bitmap_clusnr = le32_to_cpu(bitmap->clusnr);
		pcb_data->bitmap_size = le64_to_cpu(bitmap->size);
		break;
	case EXFAT_TYPE_UPCASE:
		if (pcb_data->upcase_clusnr) {
			/* FIXME: for this, what do we do */
			PRINTK(KERN_WARNING "exFAT: found another"
			       " upper-case table, ignored\n");
			break;
		}
		upcase = pd_get_chunk(pd);
		pcb_data->upcase_checksum = le32_to_cpu(upcase->checksum);
		pcb_data->upcase_clusnr = le32_to_cpu(upcase->clusnr);
		pcb_data->upcase_size = le64_to_cpu(upcase->size);
		break;
	}

	return PARSE_NEXT;
}

int exfat_read_rootdir(struct inode *dir)
{
	struct super_block *sb = dir->i_sb;
	struct exfat_parse_callback pcb;
	struct rootdir_parse_data pcb_data;
	loff_t pos;
	int err;

	pcb_data.bitmap_clusnr = 0;
	pcb_data.upcase_clusnr = 0;
	pcb.parse = rootdir_parse;
	pcb.data = &pcb_data;

	pos = 0;
	err = exfat_parse_dir(NULL, dir, &pos, &pcb);
	if (err)
		return err;

	err = exfat_setup_bitmap(sb, pcb_data.bitmap_clusnr,
				 pcb_data.bitmap_size);
	if (err)
		return err;

	err = exfat_setup_upcase(sb, pcb_data.upcase_checksum,
				 pcb_data.upcase_clusnr,
				 pcb_data.upcase_size);
	if (err) {
		exfat_free_bitmap(EXFAT_SB(sb));
		return err;
	}

	return 0;
}

const struct inode_operations exfat_dir_inode_ops = {
	.create		= exfat_create,
	.lookup		= exfat_lookup,
//	.link		= ext3_link,
	.unlink		= exfat_unlink,
//	.symlink	= ext3_symlink,
	.mkdir		= exfat_mkdir,
	.rmdir		= exfat_rmdir,
//	.mknod		= ext3_mknod,
	.rename		= exfat_rename,
//	.setattr	= ext3_setattr,
	.getattr	= exfat_getattr
#ifdef CONFIG_EXT3_FS_XATTR
//	.setxattr	= generic_setxattr,
//	.getxattr	= generic_getxattr,
//	.listxattr	= ext3_listxattr,
//	.removexattr	= generic_removexattr,
#endif
//	.permission	= ext3_permission,
	/* FIXME: why doesn't ext4 support this for directory? */
//	.fallocate	= ext4_fallocate,
//	.fiemap		= ext4_fiemap,
};

const struct file_operations exfat_dir_ops = {
	.llseek		= generic_file_llseek,
	.read		= generic_read_dir,
	.readdir	= exfat_readdir,
//	.unlocked_ioctl	= exfat_ioctl,
#ifdef CONFIG_COMPAT
//	.compat_ioctl	= exfat_compat_ioctl,
#endif
//	.fsync		= exfat_sync_file,
//	.release	= exfat_release_dir,
};
