/**
 * aops.c - NTFS kernel address space operations and page cache handling.
 *	    Part of the Linux-NTFS project.
 *
 * Copyright (c) 2001-2007 Anton Altaparmakov
 * Copyright (c) 2002 Richard Russon
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/kernel.h>
#include <linux/syscalls.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/percpu.h>
#include <linux/slab.h>
#include <linux/capability.h>
#include <linux/blkdev.h>
#include <linux/file.h>
#include <linux/quotaops.h>
#include <linux/highmem.h>
#include <linux/module.h>
#include <linux/writeback.h>
#include <linux/hash.h>
#include <linux/suspend.h>
#include <linux/buffer_head.h>
#include <linux/task_io_accounting_ops.h>
#include <linux/bio.h>
#include <linux/notifier.h>
#include <linux/cpu.h>
#include <linux/bitops.h>
#include <linux/mpage.h>
#include <linux/bit_spinlock.h>
#include <linux/pagevec.h>
#include <linux/wait.h>
#include <linux/jiffies.h>
#include <linux/delay.h>

#include "aops.h"
#include "volume.h"
#include "attrib.h"
#include "debug.h"
#include "inode.h"
#include "mst.h"
#include "mft.h"
#include "runlist.h"
#include "types.h"
#include "compress.h"

static inline void flush_dcache_mft_record_page(ntfs_inode *ni)
{
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
	flush_dcache_page(ni->base_ni_attr.mft_page);
}

static inline int ntfs3g_need_remap_bh(struct buffer_head *bh)
{
    return !buffer_mapped(bh) ||(bh->b_blocknr==LCN_RL_NOT_MAPPED);
}

/**
 * ntfs_end_buffer_async_read - async io completion for reading attributes
 * @bh:		buffer head on which io is completed
 * @uptodate:	whether @bh is now uptodate or not
 *
 * Asynchronous I/O completion handler for reading pages belonging to the
 * attribute address space of an inode.  The inodes can either be files or
 * directories or they can be fake inodes describing some attribute.
 *
 * If NInoMstProtected(), perform the post read mst fixups when all IO on the
 * page has been completed and mark the page uptodate or set the error bit on
 * the page.  To determine the size of the records that need fixing up, we
 * cheat a little bit by setting the index_block_size in ntfs_inode to the ntfs
 * record size, and index_block_size_bits, to the log(base 2) of the ntfs
 * record size.
 */
static void ntfs_end_buffer_async_read(struct buffer_head *bh, int uptodate)
{
	unsigned long flags;
	struct buffer_head *first, *tmp;
	struct page *page;
	struct inode *vi;
	ntfs_inode *ni;
	int page_uptodate = 1;
       int errcode;

	page = bh->b_page;
	vi = page->mapping->host;
	ni = NTFS_I(vi);

	if (likely(uptodate)) {
		loff_t i_size;
		s64 file_ofs, init_size;

		set_buffer_uptodate(bh);

		file_ofs = ((s64)page->index << PAGE_CACHE_SHIFT) +
				bh_offset(bh);
//@tmp		read_lock_irqsave(&ni->size_lock, flags);
		init_size = ni->map_na->initialized_size;
		i_size = i_size_read(vi);
//@tmp		read_unlock_irqrestore(&ni->size_lock, flags);
		if (unlikely(init_size > i_size)) {
			/* Race with shrinking truncate. */
			init_size = i_size;
		}
		/* Check for the current buffer head overflowing. */
		if (unlikely(file_ofs + bh->b_size > init_size)) {
			int ofs;
			void *kaddr;

			ofs = 0;
			if (file_ofs < init_size)
				ofs = init_size - file_ofs;
			local_irq_save(flags);
			kaddr = kmap_atomic(page, KM_BIO_SRC_IRQ);
			memset(kaddr + bh_offset(bh) + ofs, 0,
					bh->b_size - ofs);
			flush_dcache_page(page);
			kunmap_atomic(kaddr, KM_BIO_SRC_IRQ);
			local_irq_restore(flags);
		}
	} else {
		clear_buffer_uptodate(bh);
		SetPageError(page);
		ntfs_log_perror(ni->vol->sb, "Buffer I/O error, logical block "
				"0x%llx.", (unsigned long long)bh->b_blocknr);
	}
	first = page_buffers(page);
	local_irq_save(flags);
	bit_spin_lock(BH_Uptodate_Lock, &first->b_state);
	clear_buffer_async_read(bh);
	unlock_buffer(bh);
	tmp = bh;
	do {
		if (!buffer_uptodate(tmp))
			page_uptodate = 0;
		if (buffer_async_read(tmp)) {
			if (likely(buffer_locked(tmp)))
				goto still_busy;
			/* Async buffers must be locked. */
			BUG();
		}
		tmp = tmp->b_this_page;
	} while (tmp != bh);
	bit_spin_unlock(BH_Uptodate_Lock, &first->b_state);
	local_irq_restore(flags);
	/*
	 * If none of the buffers had errors then we can set the page uptodate,
	 * but we first have to perform the post read mst fixups, if the
	 * attribute is mst protected, i.e. if NInoMstProteced(ni) is true.
	 * Note we ignore fixup errors as those are detected when
	 * map_mft_record() is called which gives us per record granularity
	 * rather than per page granularity.
	 */
	if (!NInoAttrMstProtected(ni)) {
		if (likely(page_uptodate && !PageError(page)))
			SetPageUptodate(page);
	} else {
		u8 *kaddr;
		unsigned int i, recs;
		u32 rec_size;

		rec_size = ni->fixup_block_size;
		recs = PAGE_CACHE_SIZE / rec_size;
		/* Should have been verified before we got here... */
		NTFS3G_BUG_TRACE(!recs);
		local_irq_save(flags);
		kaddr = kmap_atomic(page, KM_BIO_SRC_IRQ);
		for (i = 0; i < recs; i++)
			ntfs_mst_post_read_fixup(&errcode, (NTFS_RECORD*)(kaddr +
					i * rec_size), rec_size);
		kunmap_atomic(kaddr, KM_BIO_SRC_IRQ);
		local_irq_restore(flags);
		flush_dcache_page(page);
		if (likely(page_uptodate && !PageError(page)))
			SetPageUptodate(page);
	}
	unlock_page(page);
	return;
still_busy:
	bit_spin_unlock(BH_Uptodate_Lock, &first->b_state);
	local_irq_restore(flags);
	return;
}

static void ntfs_end_read(struct buffer_head *bh, int uptodate)
{
	unsigned long flags;
	struct buffer_head *first, *tmp;
	struct page *page;
	struct inode *vi;
	ntfs_inode *ni;
	int page_uptodate = 1;
       int errcode;

	page = bh->b_page;
	vi = page->mapping->host;
	ni = NTFS_I(vi);

	if (likely(uptodate)) {
		loff_t i_size;
		s64 file_ofs, init_size;

		set_buffer_uptodate(bh);

		file_ofs = ((s64)page->index << PAGE_CACHE_SHIFT) +
				bh_offset(bh);
//@tmp		read_lock_irqsave(&ni->size_lock, flags);
		init_size = ni->map_na->initialized_size;
		i_size = i_size_read(vi);
//@tmp		read_unlock_irqrestore(&ni->size_lock, flags);
		if (unlikely(init_size > i_size)) {
			/* Race with shrinking truncate. */
			init_size = i_size;
		}
  #if 0
		/* Check for the current buffer head overflowing. */
		if (unlikely(file_ofs + bh->b_size > init_size)) {
			int ofs;
			void *kaddr;

			ofs = 0;
			if (file_ofs < init_size)
				ofs = init_size - file_ofs;
			local_irq_save(flags);
			kaddr = kmap_atomic(page, KM_BIO_SRC_IRQ);
			memset(kaddr + bh_offset(bh) + ofs, 0,
					bh->b_size - ofs);
			flush_dcache_page(page);
			kunmap_atomic(kaddr, KM_BIO_SRC_IRQ);
			local_irq_restore(flags);
		}
  #endif
	} else {
		clear_buffer_uptodate(bh);
		SetPageError(page);
	}
	first = page_buffers(page);
	local_irq_save(flags);
	bit_spin_lock(BH_Uptodate_Lock, &first->b_state);
	clear_buffer_async_read(bh);
	unlock_buffer(bh);
	tmp = bh;
	do {
		if (!buffer_uptodate(tmp))
			page_uptodate = 0;
		if (buffer_async_read(tmp)) {
			if (likely(buffer_locked(tmp)))
				goto still_busy;
			/* Async buffers must be locked. */
			BUG();
		}
		tmp = tmp->b_this_page;
	} while (tmp != bh);
	bit_spin_unlock(BH_Uptodate_Lock, &first->b_state);
	local_irq_restore(flags);
	/*
	 * If none of the buffers had errors then we can set the page uptodate,
	 * but we first have to perform the post read mst fixups, if the
	 * attribute is mst protected, i.e. if NInoMstProteced(ni) is true.
	 * Note we ignore fixup errors as those are detected when
	 * map_mft_record() is called which gives us per record granularity
	 * rather than per page granularity.
	 */
	if (!NInoAttrMstProtected(ni)) {
		if (likely(page_uptodate && !PageError(page)))
			SetPageUptodate(page);
	} else {
		u8 *kaddr;
		unsigned int i, recs;
		u32 rec_size;

              if(likely(page_uptodate && !PageError(page)))
              {
		     rec_size = ni->fixup_block_size;
		     recs = PAGE_CACHE_SIZE / rec_size;
		     /* Should have been verified before we got here... */
		     NTFS3G_BUG_TRACE(!recs);
		     local_irq_save(flags);
		     kaddr = kmap_atomic(page, KM_BIO_SRC_IRQ);
		     for (i = 0; i < recs; i++)
			      ntfs_mst_post_read_fixup(&errcode, (NTFS_RECORD*)(kaddr +
					      i * rec_size), rec_size);
		     kunmap_atomic(kaddr, KM_BIO_SRC_IRQ);
		     local_irq_restore(flags);
		     flush_dcache_page(page);
		     SetPageUptodate(page);
                }
	}
	unlock_page(page);
	return;
still_busy:
	bit_spin_unlock(BH_Uptodate_Lock, &first->b_state);
	local_irq_restore(flags);
	return;
}

static void ntfs_end_write(struct buffer_head *bh, int uptodate)
{
	unsigned long flags;
	struct buffer_head *page_first, *tmp;
	struct page *page;
	struct inode *vi;



	page = bh->b_page;
	vi = page->mapping->host;

       NTFS3G_BUG_TRACE(NInoAttrMstProtected(NTFS_I(vi)));

	if (uptodate) {
		set_buffer_uptodate(bh);
	} else {
		set_bit(AS_EIO, &page->mapping->flags);
		set_buffer_write_io_error(bh);
		clear_buffer_uptodate(bh);
		SetPageError(page);
	}

    	page_first = page_buffers(page);

	local_irq_save(flags);
	bit_spin_lock(BH_Uptodate_Lock, &page_first->b_state);
	clear_buffer_async_write(bh);
       unlock_buffer(bh);

	tmp = page_first;
	do {
		if (buffer_async_write(tmp)) {
			if (likely(buffer_locked(tmp)))
				goto still_busy;
			/* Async buffers must be locked. */
			BUG();
		}
		tmp = tmp->b_this_page;
	} while (tmp != page_first);
	bit_spin_unlock(BH_Uptodate_Lock, &page_first->b_state);
	local_irq_restore(flags);
	end_page_writeback(page);
	return;
still_busy:
	bit_spin_unlock(BH_Uptodate_Lock, &page_first->b_state);
	local_irq_restore(flags);
	return;
};



static void ntfs_end_bio_io_sync(struct bio *bio, int err)
{

    struct submit_rw_io_data *sdata;
    struct buffer_head *bh;
    const int uptodate = test_bit(BIO_UPTODATE, &bio->bi_flags);
    int i, j;
    sdata=bio->bi_private;

    for( i=0; i<sdata->bh_page_count; i++)
    {
        bh=sdata->bh[i];
        for(j=0;j<sdata->bh_count[i];j++)
       {
             if(sdata->b_read)
                   ntfs_end_read( bh, uptodate);
             else
                   ntfs_end_write(bh, uptodate);
             bh = bh->b_this_page;
       }
    }
    bio_put(bio);
    kmem_cache_free(ntfs3g_bio_data_cache, sdata);


}


static struct bio *
ntfs_bio_alloc(struct block_device *bdev,
		sector_t first_sector, int nr_vecs,
		gfp_t gfp_flags)
{
	struct bio *bio;

	bio = bio_alloc(gfp_flags, nr_vecs);

	if (bio == NULL && (current->flags & PF_MEMALLOC)) {
		while (!bio && (nr_vecs /= 2))
			bio = bio_alloc(gfp_flags, nr_vecs);
	}

	if (bio) {
		bio->bi_bdev = bdev;
		bio->bi_sector = first_sector;
	}
	return bio;
}

static void ntfs_submit_rw_block(struct buffer_head *start_bh, int b_read, unsigned long length,unsigned int bh_count, struct bio **pp_bio)
{

     struct submit_rw_io_data *sdata;


	if(*pp_bio==NULL || ((struct submit_rw_io_data *)(*pp_bio)->bi_private)->bh_page_count>=MAX_SUBMIT_PAGE ||
           (*pp_bio)->bi_sector+bio_sectors(*pp_bio) != start_bh->b_blocknr ||
		bio_add_page(*pp_bio, start_bh->b_page, length, bh_offset(start_bh))<length)
	{
	    if(*pp_bio)
	    {

	    	 bio_get(*pp_bio);
	        submit_bio(b_read?READ:WRITE, *pp_bio);
	        bio_put(*pp_bio);
	    }
           sdata =  kmem_cache_alloc(ntfs3g_bio_data_cache, GFP_NOFS);
	    if(sdata == NULL)
	    {
	        printk("NTFS:Fatal error, not enough memory in ntfs_submit_read_block\n");
		 return;
	    }
	    *pp_bio = ntfs_bio_alloc(start_bh->b_bdev, start_bh->b_blocknr* (start_bh->b_size >> 9),
			min_t(int, MAX_SUBMIT_PAGE, bio_get_nr_vecs(start_bh->b_bdev)),
			GFP_KERNEL);
	    if(*pp_bio == NULL)
	    {
	        kfree(sdata);
	        printk("NTFS:Fatal error, not enough memory in ntfs_submit_read_block\n");
		 return;
	    }

	    (*pp_bio)->bi_end_io = ntfs_end_bio_io_sync;
           sdata->bh[0]=start_bh;
           sdata->bh_count[0]=bh_count;
	    sdata->bh_page_count = 1;
           sdata->b_read = b_read;
	    (*pp_bio)->bi_private = sdata;

	    bio_add_page(*pp_bio, start_bh->b_page, length, bh_offset(start_bh));
	}
	else
	{
	    sdata = ((struct submit_rw_io_data *)(*pp_bio)->bi_private);
           NTFS3G_BUG_TRACE(sdata->b_read != b_read);
	    sdata->bh[sdata->bh_page_count]=start_bh;
           sdata->bh_count[sdata->bh_page_count]=bh_count;
	    sdata->bh_page_count++;
	}

#if 0
       bio->bi_bdev = start_bh->b_bdev;
	bio->bi_sector = start_bh->b_blocknr* (start_bh->b_size >> 9);
       bio->bi_size =length;

       bio->bi_io_vec[0].bv_page = start_bh->b_page;
	bio->bi_io_vec[0].bv_len = length;
	bio->bi_io_vec[0].bv_offset = bh_offset(start_bh);

	bio->bi_vcnt = 1;
	bio->bi_idx = 0;



	bio->bi_end_io = ntfs_end_bio_io_sync;

       sdata->bh=start_bh;
       sdata->bh_count=bh_count;
	bio->bi_private = sdata;

	bio_get(bio);
	submit_bio(READ, bio);

	bio_put(bio);
#endif
}

static int ntfs_read_page_block(struct page *page, ntfs_attr *na, struct bio **pp_bio)
{
	loff_t i_size;
	VCN vcn;
	LCN lcn;
	s64 init_size;
	struct inode *vi;
	ntfs_inode *ni;
	ntfs_volume *vol;
	runlist_element *rl;
	struct buffer_head *bh, *head, *arr[MAX_BUF_PER_PAGE];
	sector_t iblock, lblock, zblock;
//	unsigned long flags;
	unsigned int blocksize, vcn_ofs;
	int i, nr;
	unsigned char blocksize_bits;
       int errcode;

	vi = page->mapping->host;
	ni = NTFS_I(vi);
	vol = ni->vol;

	/* $MFT/$DATA must have its complete runlist in memory at all times. */
	//NTFS3G_BUG_TRACE(!ni->runlist.rl && !ni->mft_no && !NInoAttr(ni));

	blocksize = vol->sb->s_blocksize;
	blocksize_bits = vol->sb->s_blocksize_bits;

       NTFS3G_BUG_TRACE(!page_has_buffers(page));

	bh = head = page_buffers(page);
	NTFS3G_BUG_TRACE(!bh);

	/*
	 * We may be racing with truncate.  To avoid some of the problems we
	 * now take a snapshot of the various sizes and use those for the whole
	 * of the function.  In case of an extending truncate it just means we
	 * may leave some buffers unmapped which are now allocated.  This is
	 * not a problem since these buffers will just get mapped when a write
	 * occurs.  In case of a shrinking truncate, we will detect this later
	 * on due to the runlist being incomplete and if the page is being
	 * fully truncated, truncate will throw it away as soon as we unlock
	 * it so no need to worry what we do with it.
	 */
	iblock = (s64)page->index << (PAGE_CACHE_SHIFT - blocksize_bits);
//@tmp	read_lock_irqsave(&ni->size_lock, flags);
	lblock = (ni->map_na->allocated_size + blocksize - 1) >> blocksize_bits;
	init_size = ni->map_na->initialized_size;
	i_size = i_size_read(vi);
//@tmp	read_unlock_irqrestore(&ni->size_lock, flags);
	if (unlikely(init_size > i_size)) {
		/* Race with shrinking truncate. */
		init_size = i_size;
	}
	zblock = (init_size + blocksize - 1) >> blocksize_bits;

	/* Loop through all the buffers in the page. */
	rl = NULL;
	nr = i = 0;
	do {
		errcode = 0;

		if (unlikely(buffer_uptodate(bh)))
			continue;
		if (unlikely(!ntfs3g_need_remap_bh(bh))) {
			arr[nr++] = bh;
			continue;
		}
		bh->b_bdev = vol->sb->s_bdev;
		/* Is the block within the allowed limits? */
		if (iblock < lblock) {
			bool is_retry = false;

			/* Convert iblock into corresponding vcn and offset. */
			vcn = (VCN)iblock << blocksize_bits >>
					vol->cluster_size_bits;
			vcn_ofs = ((VCN)iblock << blocksize_bits) &
					((1UL<<vol->cluster_size_bits)-1);
			if (!rl) {
//@tmp				down_read(&ni->runlist.lock);
				rl = ntfs_attr_find_vcn(&errcode, na,  vcn);
				if(!rl)
				{
				       if(errcode == ENOENT)
                                       lcn = LCN_ENOENT;
                                   else
                                       lcn = LCN_EIO;
				 }
			}
 lock_retry_vcn_2_lcn:
			if (likely(rl != NULL)) {
				/* Seek to element containing target vcn. */
				while (rl->length && rl[1].vcn <= vcn)
					rl++;
				lcn = ntfs_rl_vcn_to_lcn(&errcode, rl, vcn);
			}
			/* Successful remap. */
			if (lcn >= 0) {
				/* Setup buffer head to correct block. */
				bh->b_blocknr = ((lcn << vol->cluster_size_bits)
						+ vcn_ofs) >> blocksize_bits;
				set_buffer_mapped(bh);
				/* Only read initialized data blocks. */
				if (iblock < zblock) {
					arr[nr++] = bh;
					continue;
				}
				/* Fully non-initialized data block, zero it. */
				goto handle_zblock;
			}
			/* It is a hole, need to zero it. */
			if (lcn == LCN_HOLE)
				goto handle_hole;
			/* If first try and runlist unmapped, map and retry. */
			if (!is_retry && lcn == LCN_RL_NOT_MAPPED) {
				is_retry = true;
				/*
				 * Attempt to map runlist, dropping lock for
				 * the duration.
				 */
//@tmp				up_read(&ni->runlist.lock);
                           errcode = 0;
				rl = ntfs_attr_find_vcn(&errcode, na, vcn);
				if (rl)
					goto lock_retry_vcn_2_lcn;
				rl = NULL;
			} else if (!rl)
			{
//@tmp				up_read(&ni->runlist.lock);
			 }
			/*
			 * If buffer is outside the runlist, treat it as a
			 * hole.  This can happen due to concurrent truncate
			 * for example.
			 */
			if (errcode == ENOENT || lcn == LCN_ENOENT) {
				errcode = 0;
				goto handle_hole;
			}
			/* Hard error, zero out region. */
			if (!errcode)
				errcode = EIO;
			bh->b_blocknr = -1;
			SetPageError(page);
			ntfs_log_perror("Failed to read from inode 0x%lx, "
					"attribute type 0x%x, vcn 0x%llx, "
					"offset 0x%x because its location on "
					"disk could not be determined%s "
					"(error code %i).", ni->mft_no,
					ni->type, (unsigned long long)vcn,
					vcn_ofs, is_retry ? " even after "
					"retrying" : "", -errcode);
		}
		/*
		 * Either iblock was outside lblock limits or
		 * ntfs_rl_vcn_to_lcn() returned error.  Just zero that portion
		 * of the page and set the buffer uptodate.
		 */
handle_hole:
		bh->b_blocknr = -1UL;
		clear_buffer_mapped(bh);
handle_zblock:
		zero_user(page, i * blocksize, blocksize);
		set_buffer_uptodate(bh);
	} while (i++, iblock++, (bh = bh->b_this_page) != head);

	/* Release the lock if we took it. */
//@tmp	if (rl)
//@tmp		up_read(&ni->runlist.lock);

	/* Check we have at least one buffer ready for i/o. */
	if (nr) {
		struct buffer_head *tbh;
        struct buffer_head *sbh=NULL;

        unsigned long io_size=0;
        unsigned int  bh_count=0;


		/* Lock the buffers. */
		for (i = 0; i < nr; i++) {
			tbh = arr[i];
			lock_buffer(tbh);
			tbh->b_end_io = ntfs_end_buffer_async_read;
			set_buffer_async_read(tbh);
		}
		/* Finally, start i/o on the buffers. */
		for (i = 0; i < nr; i++) {

			tbh = arr[i];

			if (likely(!buffer_uptodate(tbh)))
			{
                           if(sbh==NULL)
                           {
                                  io_size=tbh->b_size;
                                  sbh=tbh;
                                  bh_count=1;
                            }
                            else if((sbh->b_blocknr +(io_size>>blocksize_bits))!=tbh->b_blocknr)
                            {
                                   ntfs_submit_rw_block(sbh,1, io_size,bh_count, pp_bio);
                                   io_size=tbh->b_size;
                                   sbh=tbh;
                                   bh_count=1;
                            }
                            else
                           {
                                   io_size+=tbh->b_size;
                                   bh_count++;
                           }

			}
			else
			{
                        if(sbh!=NULL)
                        {
                             ntfs_submit_rw_block(sbh, 1, io_size,bh_count, pp_bio);
                             bh_count=0;
                             sbh=NULL;
                        }
			   ntfs_end_buffer_async_read(tbh, 1);
			}
		}

        if(sbh!=NULL)
        {
            ntfs_submit_rw_block(sbh,1, io_size,bh_count, pp_bio);
            bh_count=0;
            sbh=NULL;
        }
		return 0;
	}
	/* No i/o was scheduled on any of the buffers. */
	if (likely(!PageError(page)))
		SetPageUptodate(page);
	else /* Signal synchronous i/o error. */
		nr = -EIO;
	unlock_page(page);
	return nr;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * ntfs_intern_readpage - fill a @page of a @file with data from the device
 * @file:	open file to which the page @page belongs or NULL
 * @page:	page cache page to fill with data
 *
 * For non-resident attributes, ntfs_readpage() fills the @page of the open
 * file @file by calling the ntfs version of the generic block_read_full_page()
 * function, ntfs_read_block(), which in turn creates and reads in the buffers
 * associated with the page asynchronously.
 *
 * For resident attributes, OTOH, ntfs_readpage() fills @page by copying the
 * data from the mft record (which at this stage is most likely in memory) and
 * fills the remainder with zeroes. Thus, in this case, I/O is synchronous, as
 * even if the mft record is not cached at this point in time, we need to wait
 * for it to be read in before we can do the copy.
 *
 * Return 0 on success and -errno on error.
 */
static int ntfs_intern_readpage(struct file *filp, struct page *page, ntfs_attr *na, struct bio **pp_bio)
{
	loff_t i_size;
	struct inode *vi;
	u8 *addr;
	ntfs_attr_search_ctx *ctx;
	MFT_RECORD *mrec;
//	unsigned long flags;
	u32 attr_len;
	int err = 0;
       ntfs_inode *ni;
       ntfs_inode *vfs_base_ni;
       struct buffer_head *bh, *head;

//retry_readpage:
	NTFS3G_BUG_TRACE(!PageLocked(page));
	vi = page->mapping->host;
	i_size = i_size_read(vi);
	/* Is the page fully outside i_size? (truncate in progress) */
	if (unlikely(page->index >= (i_size + PAGE_CACHE_SIZE - 1) >>
			PAGE_CACHE_SHIFT)) {
              if(unlikely(*pp_bio!=0))
              {
	             submit_bio(READ, *pp_bio);
	             *pp_bio = NULL;
              }
		zero_user(page, 0, PAGE_CACHE_SIZE);
		ntfs_log_debug("Read outside i_size - truncated?");

		goto done;
	}
	/*
	 * This can potentially happen because we clear PageUptodate() during
	 * ntfs_writepage() of MstProtected() attributes.
	 */
	if (PageUptodate(page)) {
		unlock_page(page);
              if(unlikely(*pp_bio!=0))
              {
	             submit_bio(READ, *pp_bio);
	             *pp_bio = NULL;
              }
		return 0;
	}

	ni = NTFS_I(vi);


    	if (!page_has_buffers(page)) {
		create_empty_buffers(page,  ni->vol->sb->s_blocksize, 0);
		if (unlikely(!page_has_buffers(page))) {
			unlock_page(page);
			return -ENOMEM;
		}
	}

	/*
	 * Only $DATA attributes can be encrypted and only unnamed $DATA
	 * attributes can be compressed.  Index root can have the flags set but
	 * this means to create compressed/encrypted files, not that the
	 * attribute is compressed/encrypted.  Note we need to check for
	 * AT_INDEX_ALLOCATION since this is the type of both directory and
	 * index inodes.
	 */
	if (na->type != AT_INDEX_ALLOCATION) {
		/* If attribute is encrypted, deny access, just like NT4. */
		if (NAttrEncrypted(na)) {
			NTFS3G_BUG_TRACE(na->type != AT_DATA);
			err = -EACCES;

                     if(unlikely(*pp_bio!=0))
                     {
	                  submit_bio(READ, *pp_bio);
	                  *pp_bio = NULL;
                     }
			goto err_out;
		}
		/* Compressed data streams are handled in compress.c. */
		if (NAttrNonResident(na) && NAttrCompressed(na)) {
			NTFS3G_BUG_TRACE(na->type != AT_DATA);
			NTFS3G_BUG_TRACE(na->name_len);
                     if(unlikely(*pp_bio!=0))
                     {
	                  submit_bio(READ, *pp_bio);
	                  *pp_bio = NULL;
                     }
			return ntfs_read_compressed_block(page, na);
		}
	}
	/* NInoNonResident() == NInoIndexAllocPresent() */
	if (NAttrNonResident(na)) {
		/* Normal, non-resident data stream. */

        err=ntfs_read_page_block(page, na, pp_bio);
	 return err;
    }
      if(unlikely(*pp_bio!=0))
      {
	    submit_bio(READ, *pp_bio);
	    *pp_bio = NULL;
      }

      if(NInoNTFSAttrNI(ni))
      {
            NTFS3G_BUG_TRACE(!ni->ntfs_mft_base_ni);
            vfs_base_ni = ni->ntfs_mft_base_ni;
            if(vfs_base_ni->base_ni_attr.nr_extents < 0)
                vfs_base_ni = vfs_base_ni->base_ni_attr.base_ni;
      }
      else
            vfs_base_ni = ni;
	/*
	 * Attribute is resident, implying it is not compressed or encrypted.
	 * This also means the attribute is smaller than an mft record and
	 * hence smaller than a page, so can simply zero out any pages with
	 * index above 0.  Note the attribute can actually be marked compressed
	 * but if it is resident the actual data is not compressed so we are
	 * ok to ignore the compressed flag here.
	 */
	if (unlikely(page->index > 0)) {
		zero_user(page, 0, PAGE_CACHE_SIZE);
		goto done;
	}

	/* Map, pin, and lock the mft record. */
	mrec = map_mft_record(vfs_base_ni);
	if (!mrec) {
		err = -EIO;
		goto err_out;
	}
	/*
	 * If a parallel write made the attribute non-resident, drop the mft
	 * record and retry the readpage.
	 */
	ctx = ntfs_attr_get_search_ctx(&err, vfs_base_ni, mrec);
	if (unlikely(!ctx)) {
		err = -ENOMEM;
		goto unm_err_out;
	}
	err = ntfs_attr_lookup(&err, na->type, na->name, na->name_len,
			CASE_SENSITIVE, 0, NULL, 0, ctx);
	if (unlikely(err))
		goto put_unm_err_out;
	attr_len = le32_to_cpu(ctx->attr->value_length);
//@tmp	read_lock_irqsave(&ni->size_lock, flags);
	if (unlikely(attr_len > ni->map_na->initialized_size))
		attr_len = ni->map_na->initialized_size;
	i_size = i_size_read(vi);
//@tmp	read_unlock_irqrestore(&ni->size_lock, flags);
	if (unlikely(attr_len > i_size)) {
		/* Race with shrinking truncate. */
		attr_len = i_size;
	}
	addr = kmap_atomic(page, KM_USER0);
	/* Copy the data to the page. */
       bh = head = page_buffers(page);
	NTFS3G_BUG_TRACE(!bh);
	do {

		if (!buffer_uptodate(bh))
		{
		       int cpy = attr_len-bh_offset(bh);
			if(cpy <= 0)
                            memset(addr+bh_offset(bh), 0, ni->vol->sb->s_blocksize);
                      else if(cpy>=ni->vol->sb->s_blocksize)
                            memcpy(addr+bh_offset(bh), (u8*)ctx->attr +le16_to_cpu(ctx->attr->value_offset)+bh_offset(bh),
                                   ni->vol->sb->s_blocksize);
                      else
                      {
                             memcpy(addr+bh_offset(bh), (u8*)ctx->attr +le16_to_cpu(ctx->attr->value_offset)+bh_offset(bh),
                                            cpy);
                             memset(addr+bh_offset(bh)+cpy, 0, ni->vol->sb->s_blocksize-cpy);
                      }
                      set_buffer_uptodate(bh);
		}
		bh = bh->b_this_page;
	} while (bh != head);
	//memcpy(addr, (u8*)ctx->attr +
			//le16_to_cpu(ctx->attr->value_offset), attr_len);
	/* Zero the remainder of the page. */
	//memset(addr + attr_len, 0, PAGE_CACHE_SIZE - attr_len);
	flush_dcache_page(page);
	kunmap_atomic(addr, KM_USER0);
put_unm_err_out:
	ntfs_attr_put_search_ctx(&err, ctx);
unm_err_out:
	unmap_mft_record(vfs_base_ni);
done:
	SetPageUptodate(page);
       if (unlikely(page_has_buffers(page)))
       {
                   struct buffer_head *bh, *head;
		     bh = head = page_buffers(page);
	            NTFS3G_BUG_TRACE(!bh);
	            do {
		             set_buffer_uptodate(bh);
	          } while ((bh = bh->b_this_page) != head);
        }
err_out:
	unlock_page(page);
	return err;
}

static int ntfs3g_readpage(struct file *filp, struct page *page)
{
     struct bio *bio = NULL;
     int err;
     ntfs_inode *ni;
     struct inode *vi;

     vi = page->mapping->host;
     ni = NTFS_I(vi);


     if (!ni->map_na)
     {
	       return -EIO;
     }
     err = ntfs_intern_readpage(filp, page, ni->map_na,  &bio);
     if(bio)
     {
	  bio_get(bio);
	  submit_bio(READ, bio);
	  bio_put(bio);
     }
     return err;
}

static int
ntfs3g_readpages(struct file *filp, struct address_space *mapping,
			struct list_head *pages, unsigned nr_pages)
{
	struct bio *bio = NULL;
	unsigned page_idx;
	int err = 0;
       ntfs_inode *ni;
       struct inode *vi;

       vi = mapping->host;
       ni = NTFS_I(vi);
       if (!ni->map_na)
       {
	       return -EIO;
       }
	for (page_idx = 0; page_idx < nr_pages; page_idx++) {
		struct page *page = list_entry(pages->prev, struct page, lru);

		prefetchw(&page->flags);
		list_del(&page->lru);
		if (!add_to_page_cache_lru(page, mapping,
					page->index, GFP_KERNEL)) {

			err |= ntfs_intern_readpage(filp, page, ni->map_na,  &bio);
			}
		page_cache_release(page);
	}
	NTFS3G_BUG_TRACE(!list_empty(pages));
	if (bio)
       {
	     bio_get(bio);
	     submit_bio(READ, bio);
	     bio_put(bio);
       }
	return err;
}



/**
 * ntfs_write_block - write a @page to the backing store
 * @page:	page cache page to write out
 * @wbc:	writeback control structure
 *
 * This function is for writing pages belonging to non-resident, non-mst
 * protected attributes to their backing store.
 *
 * For a page with buffers, map and write the dirty buffers asynchronously
 * under page writeback. For a page without buffers, create buffers for the
 * page, then proceed as above.
 *
 * If a page doesn't have buffers the page dirty state is definitive. If a page
 * does have buffers, the page dirty state is just a hint, and the buffer dirty
 * state is definitive. (A hint which has rules: dirty buffers against a clean
 * page is illegal. Other combinations are legal and need to be handled. In
 * particular a dirty page containing clean buffers for example.)
 *
 * Return 0 on success and -errno on error.
 *
 * Based on ntfs_read_block() and __block_write_full_page().
 */
static int ntfs_write_block(struct page *page, ntfs_attr *na, struct writeback_control *wbc, struct bio **pp_bio)
{
	VCN vcn;
	LCN lcn = LCN_EIO;
	s64 initialized_size;
	loff_t i_size;
	sector_t block, dblock, iblock;
	struct inode *vi;
	ntfs_inode *ni;
	ntfs_volume *vol;
	runlist_element *rl;
	struct buffer_head *bh, *head;
//	unsigned long flags;
	unsigned int blocksize, vcn_ofs;
	int err;
	bool need_end_writeback;
	unsigned char blocksize_bits;
       struct buffer_head *sbh=NULL;
       unsigned long io_size=0;
       unsigned int  bh_count=0;

	vi = page->mapping->host;
	ni = NTFS_I(vi);
	vol = ni->vol;

	ntfs_log_debug("Entering for inode 0x%lx, attribute type 0x%x, page index "
			"0x%lx.", ni->mft_no, ni->type, page->index);
	NTFS3G_BUG_TRACE(!NAttrNonResident(na));
	NTFS3G_BUG_TRACE(NInoAttrMstProtected(ni));
	blocksize = vol->sb->s_blocksize;
	blocksize_bits = vol->sb->s_blocksize_bits;
       NTFS3G_BUG_TRACE(!page_has_buffers(page));
	bh = head = page_buffers(page);
	NTFS3G_BUG_TRACE(!bh);

	/* NOTE: Different naming scheme to ntfs_read_block()! */

	/* The first block in the page. */
	block = (s64)page->index << (PAGE_CACHE_SHIFT - blocksize_bits);

//@tmp	read_lock_irqsave(&ni->size_lock, flags);
	i_size = i_size_read(vi);
	initialized_size = ni->map_na->initialized_size;
//@tmp	read_unlock_irqrestore(&ni->size_lock, flags);

	/* The first out of bounds block for the data size. */
	dblock = (i_size + blocksize - 1) >> blocksize_bits;

	/* The last (fully or partially) initialized block. */
	iblock = initialized_size >> blocksize_bits;

	/*
	 * Be very careful.  We have no exclusion from __set_page_dirty_buffers
	 * here, and the (potentially unmapped) buffers may become dirty at
	 * any time.  If a buffer becomes dirty here after we've inspected it
	 * then we just miss that fact, and the page stays dirty.
	 *
	 * Buffers outside i_size may be dirtied by __set_page_dirty_buffers;
	 * handle that here by just cleaning them.
	 */

	/*
	 * Loop through all the buffers in the page, mapping all the dirty
	 * buffers to disk addresses and handling any aliases from the
	 * underlying block device's mapping.
	 */
	rl = NULL;
	err = 0;
	do {
		bool is_retry = false;

		if (unlikely(block >= dblock)) {
			/*
			 * Mapped buffers outside i_size will occur, because
			 * this page can be outside i_size when there is a
			 * truncate in progress. The contents of such buffers
			 * were zeroed by ntfs_writepage().
			 *
			 * FIXME: What about the small race window where
			 * ntfs_writepage() has not done any clearing because
			 * the page was within i_size but before we get here,
			 * vmtruncate() modifies i_size?
			 */
			clear_buffer_dirty(bh);
			set_buffer_uptodate(bh);
			continue;
		}

		/* Clean buffers are not written out, so no need to map them. */
		if (!buffer_dirty(bh))
			continue;

		/* Make sure we have enough initialized size. */
		if (unlikely((block >= iblock) &&
				(initialized_size < i_size))) {
			/*
			 * If this page is fully outside initialized size, zero
			 * out all pages between the current initialized size
			 * and the current page. Just use ntfs_readpage() to do
			 * the zeroing transparently.
			 */
			if (block > iblock) {
				// TODO:
				// For each page do:
				// - read_cache_page()
				// Again for each page do:
				// - wait_on_page_locked()
				// - Check (PageUptodate(page) &&
				//			!PageError(page))
				// Update initialized size in the attribute and
				// in the inode.
				// Again, for each page do:
				//	__set_page_dirty_buffers();
				// page_cache_release()
				// We don't need to wait on the writes.
				// Update iblock.
			}
			/*
			 * The current page straddles initialized size. Zero
			 * all non-uptodate buffers and set them uptodate (and
			 * dirty?). Note, there aren't any non-uptodate buffers
			 * if the page is uptodate.
			 * FIXME: For an uptodate page, the buffers may need to
			 * be written out because they were not initialized on
			 * disk before.
			 */
			if (!PageUptodate(page)) {
				// TODO:
				// Zero any non-uptodate buffers up to i_size.
				// Set them uptodate and dirty.
			}
			// TODO:
			// Update initialized size in the attribute and in the
			// inode (up to i_size).
			// Update iblock.
			// FIXME: This is inefficient. Try to batch the two
			// size changes to happen in one go.
			ntfs_log_error(vol->sb, "Writing beyond initialized size "
					"is not supported yet. Sorry.");
			err = -EOPNOTSUPP;
			break;
			// Do NOT set_buffer_new() BUT DO clear buffer range
			// outside write request range.
			// set_buffer_uptodate() on complete buffers as well as
			// set_buffer_dirty().
		}

		/* No need to map buffers that are already mapped. */
		if (likely(!ntfs3g_need_remap_bh(bh)))
			continue;

		/* Unmapped, dirty buffer. Need to map it. */
		bh->b_bdev = vol->sb->s_bdev;

		/* Convert block into corresponding vcn and offset. */
		vcn = (VCN)block << blocksize_bits;
		vcn_ofs = vcn & ((1UL<<vol->cluster_size_bits)-1);
		vcn >>= vol->cluster_size_bits;
		if (!rl) {
//@tmp			down_read(&ni->runlist.lock);
                            err = 0;
				rl = ntfs_attr_find_vcn(&err, na,  vcn);
				if(!rl)
				{
				       if(err == ENOENT)
                                       lcn = LCN_ENOENT;
                                   else
                                       lcn = LCN_EIO;
				 }
		}
lock_retry_vcn_2_lcn:
		if (likely(rl != NULL)) {
			/* Seek to element containing target vcn. */
			while (rl->length && rl[1].vcn <= vcn)
				rl++;
			lcn = ntfs_rl_vcn_to_lcn(&err, rl, vcn);
		}
              if(lcn == LCN_HOLE)
              {
                   VCN update_from = -1;
                   s64 ofs;

                   ofs =vcn_ofs;
                   err = 0;
                  if (ntfs_attr_fill_hole(&err, na, (s64)blocksize, &ofs, &rl, &update_from))
                 {
                         err = -ENOMEM;
	                  break;
                  }
                  if (update_from!=-1 &&
		            ntfs_attr_update_mapping_pairs(&err, na, 0 /*update_from*/))
                 {
                         err = -EIO;
	                  break;
                  }
                  lcn = ntfs_rl_vcn_to_lcn(&err, rl, vcn);
               }

		/* Successful remap. */
		if (lcn >= 0) {
			/* Setup buffer head to point to correct block. */
			bh->b_blocknr = ((lcn << vol->cluster_size_bits) +
					vcn_ofs) >> blocksize_bits;
			set_buffer_mapped(bh);
			continue;
		}

		/* If first try and runlist unmapped, map and retry. */
		if (!is_retry && lcn == LCN_RL_NOT_MAPPED) {
			is_retry = true;
			/*
			 * Attempt to map runlist, dropping lock for
			 * the duration.
			 */
//@tmp           up_read(&ni->runlist.lock);
			rl = ntfs_attr_find_vcn(&err, na,  vcn);

			if (rl)
				goto lock_retry_vcn_2_lcn;
		}
		/*
		 * If buffer is outside the runlist, truncate has cut it out
		 * of the runlist.  Just clean and clear the buffer and set it
		 * uptodate so it can get discarded by the VM.
		 */
		if (err == -ENOENT || lcn == LCN_ENOENT) {
			bh->b_blocknr = -1;
			clear_buffer_dirty(bh);
			zero_user(page, bh_offset(bh), blocksize);
			set_buffer_uptodate(bh);
			err = 0;
			continue;
		}
		/* Failed to map the buffer, even after retrying. */
		if (!err)
			err = -EIO;
		bh->b_blocknr = -1;
		ntfs_log_error(na->ni->vol->sb, "Failed to write to inode 0x%lx, "
				"attribute type 0x%x, vcn 0x%llx, offset 0x%x "
				"because its location on disk could not be "
				"determined%s (error code %i).", ni->mft_no,
				na->type, (unsigned long long)vcn,
				vcn_ofs, is_retry ? " even after "
				"retrying" : "", err);
		break;
	} while (block++, (bh = bh->b_this_page) != head);

	/* Release the lock if we took it. */
//@tmp	if (rl)
//@tmp		up_read(&ni->runlist.lock);

	/* For the error case, need to reset bh to the beginning. */
	bh = head;

	/* Just an optimization, so ->readpage() is not called later. */
	if (unlikely(!PageUptodate(page))) {
		int uptodate = 1;
		do {
			if (!buffer_uptodate(bh)) {
				uptodate = 0;
				bh = head;
				break;
			}
		} while ((bh = bh->b_this_page) != head);
		if (uptodate)
			SetPageUptodate(page);
	}

	/* Setup all mapped, dirty buffers for async write i/o. */
	do {
		if (buffer_mapped(bh) && buffer_dirty(bh) && (!buffer_write_io_error(bh)) ) {
			lock_buffer(bh);
			if (test_clear_buffer_dirty(bh)) {
               if(!buffer_uptodate(bh))
                printk("!!!!!!!!!!!!ntfs_write_block error:%08lu\n\n",bh->b_state);
				NTFS3G_BUG_TRACE(!buffer_uptodate(bh));
				mark_buffer_async_write(bh);
			} else
				unlock_buffer(bh);
		} else if (unlikely(err)) {
			/*
			 * For the error case. The buffer may have been set
			 * dirty during attachment to a dirty page.
			 */
			if (err != -ENOMEM)
				clear_buffer_dirty(bh);
		}
	} while ((bh = bh->b_this_page) != head);

	if (unlikely(err)) {
		// TODO: Remove the -EOPNOTSUPP check later on...
		if (unlikely(err == -EOPNOTSUPP))
			err = 0;
		else if (err == -ENOMEM) {
			ntfs_log_warning(vol->sb, "Error allocating memory. "
					"Redirtying page so we try again "
					"later.");
			/*
			 * Put the page back on mapping->dirty_pages, but
			 * leave its buffer's dirty state as-is.
			 */
			redirty_page_for_writepage(wbc, page);
			err = 0;
		} else
			SetPageError(page);
	}

	NTFS3G_BUG_TRACE(PageWriteback(page));
	set_page_writeback(page);	/* Keeps try_to_free_buffers() away. */

	/* Submit the prepared buffers for i/o. */
	need_end_writeback = true;

       NTFS3G_BUG_TRACE(bh != head);
	do {
              if(buffer_async_write(bh))
             {
                    need_end_writeback = false;
                    if(sbh==NULL)
                    {
                            io_size=bh->b_size;
                            sbh=bh;
                            bh_count=1;
                    }
                    else if((sbh->b_blocknr +(io_size>>blocksize_bits))!=bh->b_blocknr)
                    {
                            ntfs_submit_rw_block(sbh,0, io_size,bh_count, pp_bio);
                            io_size=bh->b_size;
                            sbh=bh;
                            bh_count=1;
                    }
                    else
                    {
                            io_size+=bh->b_size;
                            bh_count++;
                    }

		}
              bh = bh->b_this_page;
	}while(bh != head);
       if(bh_count)
            ntfs_submit_rw_block(sbh,0, io_size,bh_count, pp_bio);
	unlock_page(page);

	/* If no i/o was started, need to end_page_writeback(). */
	if (unlikely(need_end_writeback))
		end_page_writeback(page);

	ntfs_log_debug("Done.");
	return err;
}
struct submit_mst_w_data
{
    //bh_end_io_t *end_submit;
    int       nr_bhs;
    unsigned long    all_buffer_finished;
    unsigned long     bio_start_jiffies;
    struct buffer_head* bhs[MAX_BUF_PER_PAGE];
};


static void ntfs_end_mst_buffer_async_write(struct buffer_head *bh, int uptodate, struct submit_mst_w_data *pmst_write_data)
{
	unsigned long flags;
	struct page *page;
	struct inode *vi;
      struct buffer_head *page_first;
	ntfs_inode *ni;
       unsigned int  i;



       NTFS3G_BUG_TRACE(!buffer_async_write(bh));
	page = bh->b_page;
    	vi = page->mapping->host;
	ni = NTFS_I(vi);
       NTFS3G_BUG_TRACE(!NInoAttrMstProtected(ni));


	if (uptodate) {
		set_buffer_uptodate(bh);
	} else {
		printk(KERN_WARNING "ntfs_end_mst_buffer_async_write failed at page index 0x%08x, ofs 0x%08x",
					(unsigned int)page->index, (unsigned int)bh_offset(bh));
		set_bit(AS_EIO, &page->mapping->flags);
		set_buffer_write_io_error(bh);
		clear_buffer_uptodate(bh);
		SetPageError(page);
	}
       page_first = page_buffers(page);
	local_irq_save(flags);
	bit_spin_lock(BH_Uptodate_Lock, &page_first->b_state);
	clear_buffer_async_write(bh);

       for( i=0; i<pmst_write_data->nr_bhs; i++)
       {
           if(pmst_write_data->bhs[i])
           {
                 NTFS3G_BUG_TRACE(!buffer_locked(pmst_write_data->bhs[i]));
                 if(buffer_async_write(pmst_write_data->bhs[i]))
                         goto still_busy;
           }
       }
	bit_spin_unlock(BH_Uptodate_Lock, &page_first->b_state);
	local_irq_restore(flags);
       clear_bit(0, &pmst_write_data ->all_buffer_finished);
	return;
still_busy:
	bit_spin_unlock(BH_Uptodate_Lock, &page_first->b_state);
	local_irq_restore(flags);
	return;
}

static void mst_end_bio_bh_io_sync(struct bio *bio, int err)
{
	struct buffer_head *bh = bio->bi_private;

	if (err == -EOPNOTSUPP) {
		set_bit(BIO_EOPNOTSUPP, &bio->bi_flags);
		set_bit(BH_Eopnotsupp, &bh->b_state);
	}
       ntfs_end_mst_buffer_async_write(bh, test_bit(BIO_UPTODATE, &bio->bi_flags),
               (struct submit_mst_w_data  *)bh->b_end_io);
	//bh->b_end_io(bh, test_bit(BIO_UPTODATE, &bio->bi_flags));
	bio_put(bio);
}
int submit_mst_bh(int rw, struct buffer_head * bh)
{
	struct bio *bio;
	int ret = 0;

	NTFS3G_BUG_TRACE(!buffer_locked(bh));
	NTFS3G_BUG_TRACE(!buffer_mapped(bh));
	NTFS3G_BUG_TRACE(!bh->b_end_io);

	/*
	 * Mask in barrier bit for a write (could be either a WRITE or a
	 * WRITE_SYNC
	 */
	if (buffer_ordered(bh) && (rw & WRITE))
		rw |= WRITE_BARRIER;

	/*
	 * Only clear out a write error when rewriting
	 */
	if (test_set_buffer_req(bh) && (rw & WRITE))
		clear_buffer_write_io_error(bh);

	/*
	 * from here on down, it's all bio -- do the initial mapping,
	 * submit_bio -> generic_make_request may further map this bio around
	 */
	bio = bio_alloc(GFP_NOIO, 1);

	bio->bi_sector = bh->b_blocknr * (bh->b_size >> 9);
	bio->bi_bdev = bh->b_bdev;
	bio->bi_io_vec[0].bv_page = bh->b_page;
	bio->bi_io_vec[0].bv_len = bh->b_size;
	bio->bi_io_vec[0].bv_offset = bh_offset(bh);

	bio->bi_vcnt = 1;
	bio->bi_idx = 0;
	bio->bi_size = bh->b_size;

	bio->bi_end_io = mst_end_bio_bh_io_sync;
	bio->bi_private = bh;

	bio_get(bio);
	submit_bio(rw, bio);

	if (bio_flagged(bio, BIO_EOPNOTSUPP))
		ret = -EOPNOTSUPP;

	bio_put(bio);
	return ret;
}

/**
 * ntfs_write_mst_block - write a @page to the backing store
 * @page:	page cache page to write out
 * @wbc:	writeback control structure
 *
 * This function is for writing pages belonging to non-resident, mst protected
 * attributes to their backing store.  The only supported attributes are index
 * allocation and $MFT/$DATA.  Both directory inodes and index inodes are
 * supported for the index allocation case.
 *
 * The page must remain locked for the duration of the write because we apply
 * the mst fixups, write, and then undo the fixups, so if we were to unlock the
 * page before undoing the fixups, any other user of the page will see the
 * page contents as corrupt.
 *
 * We clear the page uptodate flag for the duration of the function to ensure
 * exclusion for the $MFT/$DATA case against someone mapping an mft record we
 * are about to apply the mst fixups to.
 *
 * Return 0 on success and -errno on error.
 *
 * Based on ntfs_write_block(), ntfs_mft_writepage(), and
 * write_mft_record_nolock().
 */
 static int ntfs3g_mst_wait_action(void *word)
{
    struct submit_mst_w_data *pmst_w_data = container_of(word, struct submit_mst_w_data, all_buffer_finished);
    if(jiffies-pmst_w_data->bio_start_jiffies >= HZ/2)
    {
            block_sync_page(pmst_w_data->bhs[0]->b_page);
            pmst_w_data->bio_start_jiffies = jiffies;
     }
    msleep(1);

    return 0;
};
static int ntfs_write_mst_block(struct page *page, ntfs_attr *na,
		struct writeback_control *wbc)
{
	sector_t block, dblock, rec_block;
	struct inode *vi = page->mapping->host;
	ntfs_inode *ni = NTFS_I(vi);
       ntfs_inode *vfs_base_ni;
	ntfs_volume *vol = ni->vol;
	u8 *kaddr;
	unsigned int rec_size = ni->fixup_block_size;
//	ntfs_inode *locked_nis[PAGE_CACHE_SIZE / rec_size];
	struct buffer_head *bh, *head, *tbh, *rec_start_bh;
	runlist_element *rl;
	int i,  nr_recs, max_bhs, bhs_per_rec, err, err2;
	unsigned bh_size;
	bool  is_mft, page_is_dirty, rec_is_dirty;
	unsigned char bh_size_bits;
       struct submit_mst_w_data *pmst_w_data;
//	 int retry_count = 0;
       pmst_w_data=kmalloc(sizeof(struct submit_mst_w_data),GFP_ATOMIC);

       if(!pmst_w_data)
       {
           printk("ntfs3g run out of memory\n");
           return -ENOMEM;
       }


	ntfs_log_debug("Entering for inode 0x%lx, attribute type 0x%x, page index "
			"0x%lx.", vi->i_ino, ni->type, page->index);
	NTFS3G_BUG_TRACE(!NAttrNonResident(na));
	NTFS3G_BUG_TRACE(!NInoAttrMstProtected(ni));
       NTFS3G_BUG_TRACE(!page_has_buffers(page));

      if(NInoNTFSAttrNI(ni))
      {
            NTFS3G_BUG_TRACE(!ni->ntfs_mft_base_ni);
            vfs_base_ni = ni->ntfs_mft_base_ni;
            if(vfs_base_ni->base_ni_attr.nr_extents < 0)
                vfs_base_ni = vfs_base_ni->base_ni_attr.base_ni;
      }
      else
            vfs_base_ni = ni;

	is_mft = (S_ISREG(VFS_I(vfs_base_ni)->i_mode) && (FILE_MFT==ni->mft_no||FILE_MFTMirr==ni->mft_no));
	/*
	 * NOTE: ntfs_write_mst_block() would be called for $MFTMirr if a page
	 * in its page cache were to be marked dirty.  However this should
	 * never happen with the current driver and considering we do not
	 * handle this case here we do want to BUG(), at least for now.
	 */
	if(unlikely(!(is_mft || ni->map_na->type==AT_INDEX_ALLOCATION)))
	{
	       printk("inode(%lld:%08x), attr%d should not run into ntfs_write_mst_block\n", ni->mft_no, VFS_I(vfs_base_ni)->i_mode,  ni->map_na->type);
              BUG();
	}
	//NTFS3G_BUG_TRACE(!(is_mft || ni->map_na->type==AT_INDEX_ALLOCATION));
	bh_size = vol->sb->s_blocksize;
	bh_size_bits = vol->sb->s_blocksize_bits;
	max_bhs = PAGE_CACHE_SIZE / bh_size;
	NTFS3G_BUG_TRACE(!max_bhs);
	if(max_bhs > MAX_BUF_PER_PAGE)
	 {
	     printk("max_bhs=%d, bh_size = %d\n", max_bhs, bh_size);
            BUG();
	 }

RETRY_WRITE_MST_PAGE:
	
/*	if (!PageDirty(page)) {
		// someone wrote it for during unlock page period in this function
		printk("!PageDirty(page) %d\n", retry_count);
		  goto done;
	}*/

	if (PageWriteback(page)) 
	{
		 if (wbc->sync_mode != WB_SYNC_NONE)
			 wait_on_page_writeback(page);
		 else
		 	goto done;
	        
	}
       pmst_w_data->nr_bhs = 0;

	/* Make sure we have mapped buffers. */
	bh = head = page_buffers(page);
	NTFS3G_BUG_TRACE(!bh);

	NTFS3G_BUG_TRACE(PAGE_CACHE_SIZE<rec_size);
	bhs_per_rec = rec_size >> bh_size_bits;
	NTFS3G_BUG_TRACE(!bhs_per_rec);

	/* The first block in the page. */
	rec_block = block = (sector_t)page->index <<
			(PAGE_CACHE_SHIFT - bh_size_bits);

	/* The first out of bounds block for the data size. */
	dblock = (i_size_read(vi) + bh_size - 1) >> bh_size_bits;

	rl = NULL;
	err = err2 = pmst_w_data->nr_bhs = nr_recs  = 0;
	page_is_dirty = rec_is_dirty = false;
	rec_start_bh = NULL;
	do {
		bool is_retry = false;

		if (likely(block < rec_block)) {
			if (unlikely(block >= dblock)) {
				clear_buffer_dirty(bh);
				set_buffer_uptodate(bh);
				continue;
			}
			/*
			 * This block is not the first one in the record.  We
			 * ignore the buffer's dirty state because we could
			 * have raced with a parallel mark_ntfs_record_dirty().
			 */
			if (!rec_is_dirty)
				continue;
			if (unlikely(err2)) {
				if (err2 != ENOMEM)
					clear_buffer_dirty(bh);
				continue;
			}
		} else /* if (block == rec_block) */ {
			NTFS3G_BUG_TRACE(block > rec_block);
			/* This block is the first one in the record. */
			rec_block += bhs_per_rec;
			err2 = 0;
			if (unlikely(block >= dblock)) {
				clear_buffer_dirty(bh);
				set_buffer_uptodate(bh);
				continue;
			}
			if (!buffer_dirty(bh)) {
				/* Clean records are not written out. */
				rec_is_dirty = false;
				continue;
			}
			rec_is_dirty = true;
			rec_start_bh = bh;
		}
		/* Need to map the buffer if it is not mapped already. */
		if (unlikely(ntfs3g_need_remap_bh(bh))) {
			VCN vcn;
			LCN lcn;
			unsigned int vcn_ofs;

			bh->b_bdev = vol->sb->s_bdev;
			/* Obtain the vcn and offset of the current block. */
			vcn = (VCN)block << bh_size_bits;
			vcn_ofs = vcn & ((1UL<<vol->cluster_size_bits)-1);
			vcn >>= vol->cluster_size_bits;
			if (!rl) {
                            err = 0;
				rl = ntfs_attr_find_vcn(&err, na,  vcn);
				if(!rl)
				{
				       if(err == ENOENT)
                                       lcn = LCN_ENOENT;
                                   else
                                       lcn = LCN_EIO;
                                   err = -err;
				 }
				//down_read(&ni->runlist.lock);
				//rl = ni->runlist.rl;
			}
 lock_retry_vcn_2_lcn:
			if (likely(rl != NULL)) {
				/* Seek to element containing target vcn. */
				while (rl->length && rl[1].vcn <= vcn)
					rl++;
				lcn = ntfs_rl_vcn_to_lcn(&err, rl, vcn);
			}
			/* Successful remap. */
			if (likely(lcn >= 0)) {
				/* Setup buffer head to correct block. */
				bh->b_blocknr = ((lcn <<
						vol->cluster_size_bits) +
						vcn_ofs) >> bh_size_bits;
				set_buffer_mapped(bh);
			} else {
				/*
				 * Remap failed.  Retry to map the runlist once
				 * unless we are working on $MFT which always
				 * has the whole of its runlist in memory.
				 */
				if (!is_mft && !is_retry &&
						lcn == LCN_RL_NOT_MAPPED) {
					is_retry = true;
					/*
					 * Attempt to map runlist, dropping
					 * lock for the duration.
					 */
//@tmp				up_read(&ni->runlist.lock);
					rl = ntfs_attr_find_vcn(&err2, na,  vcn);
					if (rl)
						goto lock_retry_vcn_2_lcn;
					if (err2 == ENOMEM)
						page_is_dirty = true;
                                   err2 = -err2;
					lcn = err2;
				} else {
					err2 = -EIO;
//@tmp					if (!rl)
//@tmp						up_read(&ni->runlist.lock);
				}
				/* Hard error.  Abort writing this record. */
				if (!err || err == -ENOMEM)
					err = err2;
				bh->b_blocknr = -1;
				ntfs_log_error(na->ni->vol->sb, "Cannot write ntfs record "
						"0x%llx (inode 0x%lx, "
						"attribute type 0x%x) because "
						"its location on disk could "
						"not be determined (error "
						"code %lli).",
						(long long)block <<
						bh_size_bits >>
						vol->mft_record_size_bits,
						ni->mft_no, na->type,
						(long long)lcn);
				/*
				 * If this is not the first buffer, remove the
				 * buffers in this record from the list of
				 * buffers to write and clear their dirty bit
				 * if not error -ENOMEM.
				 */
				if (rec_start_bh != bh) {
					while (pmst_w_data->bhs[--pmst_w_data->nr_bhs] != rec_start_bh)
						;
					if (err2 != -ENOMEM) {
						do {
							clear_buffer_dirty(
								rec_start_bh);
						} while ((rec_start_bh =
								rec_start_bh->
								b_this_page) !=
								bh);
					}
				}
				continue;
			}
		}
        if(buffer_write_io_error(bh))
				clear_buffer_dirty(bh);
        else {
    		NTFS3G_BUG_TRACE(!buffer_uptodate(bh));
    		NTFS3G_BUG_TRACE(pmst_w_data->nr_bhs >= max_bhs);
    		pmst_w_data->bhs[pmst_w_data->nr_bhs++] = bh;
        }
	} while (block++, (bh = bh->b_this_page) != head);
//@tmp	if (unlikely(rl))
//@tmp		up_read(&ni->runlist.lock);
	/* If there were no dirty buffers, we are done. */
	if (!pmst_w_data->nr_bhs)
		goto done;
	/* Map the page so we can access its contents. */

       for( i=0; i<pmst_w_data->nr_bhs; i++)
       {
             /*if trylock_buffer failed, unlock all previous buffer  to avoid following code race competition with map_mft_record*/
             if(!trylock_buffer(pmst_w_data->bhs[i]))
             {
                 int j = i-1;
		   while(j>=0)
		   {
		       unlock_buffer(pmst_w_data->bhs[j]);
		   }
		  
		   unlock_page(page);
		   schedule_timeout_interruptible(1);
		   //wait the other task to finish operations on current buffer and got try again
		   //lock_buffer(pmst_w_data->bhs[i]); 
		   //unlock_buffer(pmst_w_data->bhs[i]);
		   lock_page(page);
		   //page buffer may be de-allocated during kernel cache droping 
    	          if (unlikely(!page_has_buffers(page))) 
		   {
		             NTFS3G_BUG_TRACE(!PageUptodate(page));
		             create_empty_buffers(page, ni->vol->sb->s_blocksize,
				                (1 << BH_Uptodate) | (1 << BH_Dirty));
		             if (unlikely(!page_has_buffers(page))) 
			      {
			              printk("%s:%dError allocating page "
					             "buffers.  Redirtying page so we try "
					              "again later.", __FILE__, __LINE__);
			              /*
			                   * Put the page back on mapping->dirty_pages, but leave
			                  * its buffers' dirty state as-is.
			              */
			              redirty_page_for_writepage(wbc, page);
			              unlock_page(page);
					kfree(pmst_w_data);
			              return 0;
		             }
	             }
		   //   printk("mst buffer lock confliction found, try again!(%d)\n", ++retry_count);
		      goto RETRY_WRITE_MST_PAGE;      
             } 
       }

       for( i=0; i<pmst_w_data->nr_bhs; i++)
	   	set_buffer_async_write(pmst_w_data->bhs[i]);
	/* Clear the page uptodate flag whilst the mst fixups are applied. */
	NTFS3G_BUG_TRACE(!PageUptodate(page));
	
	kaddr = kmap(page); //kemap in advance to avoid page cache droped during unlock--lock produce
	
	for (i = 0; i < pmst_w_data->nr_bhs; i++) {
		unsigned int ofs;

		/* Skip buffers which are not at the beginning of records. */
		if (i % bhs_per_rec)
			continue;
		tbh = pmst_w_data->bhs[i];
		ofs = bh_offset(tbh);

		/* Apply the mst protection fixups. */
              err2 = 0;
		ntfs_mst_pre_write_fixup(&err2, (NTFS_RECORD*)(kaddr + ofs),
				rec_size);
		if (unlikely(err2)) {
			if (!err || err == -ENOMEM)
				err = -EIO;
			printk("Failed to apply mst fixups "
					"(inode 0x%lx, attribute type 0x%x, "
					"page index 0x%lx, page offset 0x%x)!"
					"  Unmount and run chkdsk.", vi->i_ino,
					ni->map_na->type, page->index, ofs);
			/*
			 * Mark all the buffers in this record clean as we do
			 * not want to write corrupt data to disk.
			 */
			do {
				clear_buffer_dirty(pmst_w_data->bhs[i]);
                            clear_buffer_async_write(pmst_w_data->bhs[i]);
                            unlock_buffer(pmst_w_data->bhs[i]);
				pmst_w_data->bhs[i] = NULL;
			} while (++i % bhs_per_rec);
			continue;
		}
		nr_recs++;
	}
	/* If no records are to be written out, we are done. */
       kunmap(page);
	if (!nr_recs)
		goto done;
	flush_dcache_page(page);

 	NTFS3G_BUG_TRACE(PageWriteback(page));
	set_page_writeback(page);
       set_bit(0, &pmst_w_data->all_buffer_finished);


	/* Lock buffers and start synchronous write i/o on them. */
	for (i = 0; i < pmst_w_data->nr_bhs; i++) {
		tbh = pmst_w_data->bhs[i];
		if (!tbh)
			continue;
		/* The buffer dirty state is now irrelevant, just clean it. */
		clear_buffer_dirty(tbh);
		NTFS3G_BUG_TRACE(!buffer_uptodate(tbh));
		NTFS3G_BUG_TRACE(!buffer_mapped(tbh));
              NTFS3G_BUG_TRACE(!buffer_async_write(tbh));
		get_bh(tbh);
		tbh->b_end_io = (bh_end_io_t*)pmst_w_data;
		submit_mst_bh(WRITE, tbh);
	}

     pmst_w_data->bio_start_jiffies = jiffies;

     wait_on_bit(&pmst_w_data->all_buffer_finished, 0, ntfs3g_mst_wait_action, TASK_UNINTERRUPTIBLE);

     kaddr = kmap(page);

	for (i = 0; i < pmst_w_data->nr_bhs; i++) {
		unsigned int ofs;

              if(!pmst_w_data->bhs[i])
                           continue;

              NTFS3G_BUG_TRACE(buffer_async_write(pmst_w_data->bhs[i]));
		/* Skip buffers which are not at the beginning of records. */
		if (i % bhs_per_rec)
		{
		       NTFS3G_BUG_TRACE(!buffer_locked(pmst_w_data->bhs[i]));
                     unlock_buffer(pmst_w_data->bhs[i]);
			continue;
		}
		tbh = pmst_w_data->bhs[i];
		ofs = bh_offset(tbh);
		ntfs_mst_post_write_fixup(&err, (NTFS_RECORD*)(kaddr+ofs));
              unlock_buffer(pmst_w_data->bhs[i]);
	}
	/* If no records are to be written out, we are done. */
      kunmap(page);
      unlock_page(page);
      end_page_writeback(page);
      kfree(pmst_w_data);
      return 0;
done:
       unlock_page(page);
       kfree(pmst_w_data);
       return 0;
}

/**
 * ntfs_writepage - write a @page to the backing store
 * @page:	page cache page to write out
 * @wbc:	writeback control structure
 *
 * This is called from the VM when it wants to have a dirty ntfs page cache
 * page cleaned.  The VM has already locked the page and marked it clean.
 *
 * For non-resident attributes, ntfs_writepage() writes the @page by calling
 * the ntfs version of the generic block_write_full_page() function,
 * ntfs_write_block(), which in turn if necessary creates and writes the
 * buffers associated with the page asynchronously.
 *
 * For resident attributes, OTOH, ntfs_writepage() writes the @page by copying
 * the data to the mft record (which at this stage is most likely in memory).
 * The mft record is then marked dirty and written out asynchronously via the
 * vfs inode dirty code path for the inode the mft record belongs to or via the
 * vm page dirty code path for the page the mft record is in.
 *
 * Based on ntfs_readpage() and fs/buffer.c::block_write_full_page().
 *
 * Return 0 on success and -errno on error.
 */
static int ntfs3g_internal_writepage(struct page *page, struct writeback_control *wbc, struct bio **pp_bio)
{
	loff_t i_size;
	struct inode *vi = page->mapping->host;
	ntfs_inode *ni = NTFS_I(vi);
       ntfs_inode *vfs_base_ni;
	char *addr;
	ntfs_attr_search_ctx *ctx = NULL;
	MFT_RECORD *m = NULL;
	u32 attr_len;
	int err;
       struct buffer_head *bh, *head;

       if (!ni->map_na)
       {
	       return -EIO;
       }
retry_writepage:
	NTFS3G_BUG_TRACE(!PageLocked(page));
	i_size = i_size_read(vi);
	/* Is the page fully outside i_size? (truncate in progress) */
	if (unlikely(page->index >= (i_size + PAGE_CACHE_SIZE - 1) >>
			PAGE_CACHE_SHIFT)) {
		/*
		 * The page may have dirty, unmapped buffers.  Make them
		 * freeable here, so the page does not leak.
		 */
		block_invalidatepage(page, 0);
		unlock_page(page);
		ntfs_log_debug("Write outside i_size - truncated?");
		return 0;
	}
    	if (!page_has_buffers(page)) {
		NTFS3G_BUG_TRACE(!PageUptodate(page));
		create_empty_buffers(page, ni->vol->sb->s_blocksize,
				(1 << BH_Uptodate) | (1 << BH_Dirty));
		if (unlikely(!page_has_buffers(page))) {
			ntfs_log_warning(vol->sb, "Error allocating page "
					"buffers.  Redirtying page so we try "
					"again later.");
			/*
			 * Put the page back on mapping->dirty_pages, but leave
			 * its buffers' dirty state as-is.
			 */
			redirty_page_for_writepage(wbc, page);
			unlock_page(page);
			return 0;
		}
	}
	/*
	 * Only $DATA attributes can be encrypted and only unnamed $DATA
	 * attributes can be compressed.  Index root can have the flags set but
	 * this means to create compressed/encrypted files, not that the
	 * attribute is compressed/encrypted.  Note we need to check for
	 * AT_INDEX_ALLOCATION since this is the type of both directory and
	 * index inodes.
	 */
	if (ni->map_na->type != AT_INDEX_ALLOCATION) {
		/* If file is encrypted, deny access, just like NT4. */
		if (NAttrEncrypted(ni->map_na)) {
			unlock_page(page);
			NTFS3G_BUG_TRACE(ni->map_na->type != AT_DATA);
			ntfs_log_debug("Denying write access to encrypted file.");
			return -EACCES;
		}
		/* Compressed data streams are handled in compress.c. */
		if (NAttrNonResident(ni->map_na) && NAttrCompressed(ni->map_na)) {
			NTFS3G_BUG_TRACE(ni->map_na->type != AT_DATA);
			NTFS3G_BUG_TRACE(ni->map_na->name_len);
			// TODO: Implement and replace this with
			// return ntfs_write_compressed_block(page);
			unlock_page(page);
			ntfs_log_error(vi->i_sb, "Writing to compressed files is "
					"not supported yet.  Sorry.");
			return -EOPNOTSUPP;
		}
              #if 0
		// TODO: Implement and remove this check.
		if (NAttrNonResident(ni->map_na) && NAttrSparse(ni->map_na)) {
			unlock_page(page);
			ntfs_log_error(vi->i_sb, "Writing to sparse files is not "
					"supported yet.  Sorry.");
			return -EOPNOTSUPP;
		}
              #endif
	}
	/* NInoNonResident() == NInoIndexAllocPresent() */
	if (NAttrNonResident(ni->map_na)) {
		/* We have to zero every time due to mmap-at-end-of-file. */
		if (page->index >= (i_size >> PAGE_CACHE_SHIFT)) {
			/* The page straddles i_size. */
			unsigned int ofs = i_size & ~PAGE_CACHE_MASK;
			zero_user_segment(page, ofs, PAGE_CACHE_SIZE);
		}
		/* Handle mst protected attributes. */
		if (NInoAttrMstProtected(ni))
		{
		       NTFS3G_BUG_TRACE(*pp_bio);
			return ntfs_write_mst_block(page, ni->map_na,  wbc);
		}
		/* Normal, non-resident data stream. */
		err =  ntfs_write_block(page, ni->map_na, wbc, pp_bio);
              return err;
	}
       if(unlikely(*pp_bio!=0))
       {
	        submit_bio(WRITE, *pp_bio);
	        *pp_bio = NULL;
       }
	/*
	 * Attribute is resident, implying it is not compressed, encrypted, or
	 * mst protected.  This also means the attribute is smaller than an mft
	 * record and hence smaller than a page, so can simply return error on
	 * any pages with index above 0.  Note the attribute can actually be
	 * marked compressed but if it is resident the actual data is not
	 * compressed so we are ok to ignore the compressed flag here.
	 */

	if (unlikely(page->index > 0)) {
		ntfs_log_error(vi->i_sb, "BUG()! page->index (0x%lx) > 0.  "
				"Aborting write.", page->index);
		NTFS3G_BUG_TRACE(PageWriteback(page));
		set_page_writeback(page);
		unlock_page(page);
		end_page_writeback(page);
		return -EIO;
	}
      if(NInoNTFSAttrNI(ni))
      {
            NTFS3G_BUG_TRACE(!ni->ntfs_mft_base_ni);
            vfs_base_ni = ni->ntfs_mft_base_ni;
            if(vfs_base_ni->base_ni_attr.nr_extents < 0)
                vfs_base_ni = vfs_base_ni->base_ni_attr.base_ni;
      }
      else
            vfs_base_ni = ni;

      	NTFS3G_BUG_TRACE(PageWriteback(page));
	set_page_writeback(page);
	/* Map, pin, and lock the mft record. */
	m = map_mft_record(vfs_base_ni);
	if (!m) {
		err = -EIO;
		m = NULL;
		ctx = NULL;
		goto err_out;
	}
	/*
	 * If a parallel write made the attribute non-resident, drop the mft
	 * record and retry the writepage.
	 */
	if (unlikely(NAttrNonResident(ni->map_na))) {
		unmap_mft_record(vfs_base_ni);
              test_clear_page_writeback(page);
		goto retry_writepage;
	}
 TRY_AGAIN:
       err = 0;
	ctx = ntfs_attr_get_search_ctx(&err, vfs_base_ni, m);
	if (unlikely(!ctx)) {
		err = -ENOMEM;
		goto err_out;
	}
	err = ntfs_attr_lookup(&err, ni->map_na->type, ni->map_na->name, ni->map_na->name_len,
			CASE_SENSITIVE, 0, NULL, 0, ctx);
	if (unlikely(err))
		goto err_out;
	/*
	 * Keep the VM happy.  This must be done otherwise the radix-tree tag
	 * PAGECACHE_TAG_DIRTY remains set even though the page is clean.
	 */
	attr_len = le32_to_cpu(ctx->attr->value_length);
	i_size = i_size_read(vi);
	if (unlikely(attr_len != i_size)) {
              ntfs_attr_put_search_ctx(&err, ctx);
		/* Race with shrinking truncate or a failed truncate. */
		attr_len = i_size;
		/*
		 * If the truncate failed, fix it up now.  If a concurrent
		 * truncate, we do its job, so it does not have to do anything.
		 */
		if(STATUS_OK == ntfs_attr_truncate(&err, ni->map_na, 	attr_len))
		{
                        ntfs_sync_inode_size(ni, ni->map_na);
		          /* Shrinking cannot fail. */
  	                 if (unlikely(NAttrNonResident(ni->map_na)))
                       {
		                     unmap_mft_record(vfs_base_ni);
                                   test_clear_page_writeback(page);
		                     goto retry_writepage;
	                 }
	       }
              else
             {
                   unmap_mft_record(vfs_base_ni);
		     SetPageError(page);
                   unlock_page(page);
                   end_page_writeback(page);
                   return -EIO;
             }
             goto TRY_AGAIN;
	}

	addr = kmap_atomic(page, KM_USER0);

       bh = head = page_buffers(page);
	NTFS3G_BUG_TRACE(!bh);
	do {

		if (buffer_dirty(bh))
		{
		     if(attr_len > bh_offset(bh))
		     {
		           int cpy = attr_len-bh_offset(bh);
			   if(cpy <= ni->vol->sb->s_blocksize)
                            memcpy((u8*)ctx->attr +le16_to_cpu(ctx->attr->value_offset)+bh_offset(bh), addr+bh_offset(bh), cpy);
                        else
                            memcpy((u8*)ctx->attr +le16_to_cpu(ctx->attr->value_offset)+bh_offset(bh), addr+bh_offset(bh),
                                   ni->vol->sb->s_blocksize);
		     }
                   clear_buffer_dirty(bh);
		}
		bh = bh->b_this_page;
	} while (bh != head);

	/* Copy the data from the page to the mft record. */
	//memcpy((u8*)ctx->attr +
            //le16_to_cpu(ctx->attr->value_offset), addr, attr_len);
	/* Zero out of bounds area in the page cache page. */
	//memset(addr + attr_len, 0, PAGE_CACHE_SIZE - attr_len);
	kunmap_atomic(addr, KM_USER0);
	flush_dcache_page(page);
	flush_dcache_mft_record_page(ctx->ntfs_ino);

	/* Finally, mark the mft record dirty, so it gets written back. */
	mark_mft_record_dirty(ctx->ntfs_ino);
	ntfs_attr_put_search_ctx(&err, ctx);
	unmap_mft_record(vfs_base_ni);
	/* We are done with the page. */
       unlock_page(page);
	end_page_writeback(page);
	return 0;
err_out:
	if (err == -ENOMEM) {
		ntfs_log_warning(vi->i_sb, "Error allocating memory. Redirtying "
				"page so we try again later.");
		/*
		 * Put the page back on mapping->dirty_pages, but leave its
		 * buffers' dirty state as-is.
		 */
		redirty_page_for_writepage(wbc, page);
		err = 0;
	} else {
		ntfs_log_error(vi->i_sb, "Resident attribute write failed with "
				"error %i.", err);
		SetPageError(page);
//@tmp		NVolSetErrors(ni->vol);
	}
	unlock_page(page);
       end_page_writeback(page);

	if (ctx)
		ntfs_attr_put_search_ctx(&err, ctx);
	if (m)
		unmap_mft_record(vfs_base_ni);
	return err;
}
static int ntfs3g_writepage(struct page *page, struct writeback_control *wbc)
{
    struct bio *bio = NULL;
     int err;

     err = ntfs3g_internal_writepage(page, wbc,  &bio);
     if(bio)
     {
	  bio_get(bio);
	  submit_bio(WRITE, bio);
	  bio_put(bio);
     }
     return err;
}
static int ntfs3g_writepages(struct address_space *mapping,
			  struct writeback_control *wbc)
{
	struct backing_dev_info *bdi = mapping->backing_dev_info;
	int ret = 0;
	int done = 0;
	struct pagevec pvec;
	int nr_pages;
	pgoff_t uninitialized_var(writeback_index);
	pgoff_t index;
	pgoff_t end;		/* Inclusive */
	pgoff_t done_index;
	int cycled;
	int range_whole = 0;
	long nr_to_write = wbc->nr_to_write;
       struct bio *bio = NULL;

	if (wbc->nonblocking && bdi_write_congested(bdi)) {
		wbc->encountered_congestion = 1;
		return 0;
	}

	pagevec_init(&pvec, 0);
	if (wbc->range_cyclic) {
		writeback_index = mapping->writeback_index; /* prev offset */
		index = writeback_index;
		if (index == 0)
			cycled = 1;
		else
			cycled = 0;
		end = -1;
	} else {
		index = wbc->range_start >> PAGE_CACHE_SHIFT;
		end = wbc->range_end >> PAGE_CACHE_SHIFT;
		if (wbc->range_start == 0 && wbc->range_end == LLONG_MAX)
			range_whole = 1;
		cycled = 1; /* ignore range_cyclic tests */
	}
retry:
	done_index = index;
	while (!done && (index <= end)) {
		int i;

		nr_pages = pagevec_lookup_tag(&pvec, mapping, &index,
			      PAGECACHE_TAG_DIRTY,
			      min(end - index, (pgoff_t)PAGEVEC_SIZE-1) + 1);
		if (nr_pages == 0)
			break;

		for (i = 0; i < nr_pages; i++) {
			struct page *page = pvec.pages[i];

			/*
			 * At this point, the page may be truncated or
			 * invalidated (changing page->mapping to NULL), or
			 * even swizzled back from swapper_space to tmpfs file
			 * mapping. However, page->index will not change
			 * because we have a reference on the page.
			 */
			if (page->index > end) {
				/*
				 * can't be range_cyclic (1st pass) because
				 * end == -1 in that case.
				 */
				done = 1;
				break;
			}

			done_index = page->index + 1;

			lock_page(page);

			/*
			 * Page truncated or invalidated. We can freely skip it
			 * then, even for data integrity operations: the page
			 * has disappeared concurrently, so there could be no
			 * real expectation of this data interity operation
			 * even if there is now a new, dirty page at the same
			 * pagecache address.
			 */
			if (unlikely(page->mapping != mapping)) {
continue_unlock:
				unlock_page(page);
				continue;
			}

			if (!PageDirty(page)) {
				/* someone wrote it for us */
				goto continue_unlock;
			}

			if (PageWriteback(page)) {
				if (wbc->sync_mode != WB_SYNC_NONE)
					wait_on_page_writeback(page);
				else
					goto continue_unlock;
			}

			NTFS3G_BUG_TRACE(PageWriteback(page));
			if (!clear_page_dirty_for_io(page))
				goto continue_unlock;

			ret = ntfs3g_internal_writepage(page, wbc, &bio);
			if (unlikely(ret)) {
				if (ret == AOP_WRITEPAGE_ACTIVATE) {
					unlock_page(page);
					ret = 0;
				} else {
					/*
					 * done_index is set past this page,
					 * so media errors will not choke
					 * background writeout for the entire
					 * file. This has consequences for
					 * range_cyclic semantics (ie. it may
					 * not be suitable for data integrity
					 * writeout).
					 */
					done = 1;
					break;
				}
 			}

			if (nr_to_write > 0) {
				nr_to_write--;
				if (nr_to_write == 0 &&
				    wbc->sync_mode == WB_SYNC_NONE) {
					/*
					 * We stop writing back only if we are
					 * not doing integrity sync. In case of
					 * integrity sync we have to keep going
					 * because someone may be concurrently
					 * dirtying pages, and we might have
					 * synced a lot of newly appeared dirty
					 * pages, but have not synced all of the
					 * old dirty pages.
					 */
					done = 1;
					break;
				}
			}

			if (wbc->nonblocking && bdi_write_congested(bdi)) {
				wbc->encountered_congestion = 1;
				done = 1;
				break;
			}
		}
		pagevec_release(&pvec);
		cond_resched();
	}
	if (!cycled && !done) {
		/*
		 * range_cyclic:
		 * We hit the last page and there is more work to be done: wrap
		 * back to the start of the file
		 */
		cycled = 1;
		index = 0;
		end = writeback_index - 1;
		goto retry;
	}
	if (!wbc->no_nrwrite_index_update) {
		if (wbc->range_cyclic || (range_whole && nr_to_write > 0))
			mapping->writeback_index = done_index;
		wbc->nr_to_write = nr_to_write;
	}
       if(bio)
      {
	  bio_get(bio);
	  submit_bio(WRITE, bio);
	  bio_put(bio);
      }
	return ret;
}
static int ntfs3g_get_block(struct inode *inode, sector_t iblock,
			struct buffer_head *bh_result, int create)
{
     ntfs_inode *ni = NTFS_I(inode);

       runlist *rl = NULL;
       int err;
       VCN vcn;
       LCN lcn = LCN_EIO;
       int is_retry = 0;
       u32 vcn_ofs;
       int blocksize = ni->vol->sb->s_blocksize, blocksize_bits=ni->vol->sb->s_blocksize_bits;

        NTFS3G_BUG_TRACE(iblock> ((ni->map_na->data_size+blocksize-1)>>ni->vol->sb->s_blocksize_bits));

        if(!NAttrNonResident(ni->map_na))
        {
             ntfs_attr_search_ctx *ctx;
             u32 attr_len;
             int ret;
             u8 *addr;
             u16 copy_len;
	      ctx = ntfs_attr_get_search_ctx(&err, ni, NULL);
	      if (unlikely(!ctx)) {
		     return -ENOMEM;
	      }
	     ret = ntfs_attr_lookup(&err, ni->map_na->type, ni->map_na->name, ni->map_na->name_len,
			CASE_SENSITIVE, 0, NULL, 0, ctx);
	      if (unlikely(ret))
	      {
		    ntfs_attr_put_search_ctx(&err, ctx);
                  return -err;
	      }
	      attr_len = le32_to_cpu(ctx->attr->value_length);

              if(attr_len<= bh_offset(bh_result))
                   copy_len = 0;
              else
                   copy_len = attr_len-bh_offset(bh_result);
              if(copy_len>blocksize)
                  copy_len = blocksize;

	      addr = kmap_atomic(bh_result->b_page, KM_USER0);
	      /* Copy the data to the page. */

	       memcpy(addr+bh_offset(bh_result), (u8*)ctx->attr +
			le16_to_cpu(ctx->attr->value_offset)+bh_offset(bh_result),
			copy_len);
	       memset(addr+bh_offset(bh_result)+copy_len, 0, blocksize-copy_len);        ;
	       kunmap_atomic(addr, KM_USER0);
	       ntfs_attr_put_search_ctx(&err, ctx);

	     //set_buffer_new(bh_result);
            //set_buffer_mapped(bh_result);
               set_buffer_uptodate(bh_result);
               mark_buffer_dirty(bh_result);
	        bh_result->b_bdev = ni->vol->sb->s_bdev;
	        bh_result->b_blocknr = LCN_RL_NOT_MAPPED;
	        bh_result->b_size = blocksize;
               return 0;
        }


        err = 0;
        vcn = (VCN)iblock << blocksize_bits;
	 vcn >>= ni->vol->cluster_size_bits;
     	 vcn_ofs = (u32)(((VCN)iblock << blocksize_bits) &
					((1UL<<ni->vol->cluster_size_bits)-1));

         NTFS3G_BUG_TRACE(NInoAttrMstProtected(ni));

	  rl = ntfs_attr_find_vcn(&err, ni->map_na,  vcn);
	  if(!rl)
	  {
		if(err == ENOENT)
                    lcn = LCN_ENOENT;
               else
                    lcn = LCN_EIO;
	  }
try_map_bh:
	if (likely(rl != NULL)) {
			/* Seek to element containing target vcn. */
			while (rl->length && rl[1].vcn <= vcn)
				rl++;
			lcn = ntfs_rl_vcn_to_lcn(&err, rl, vcn);
	 }
        if(lcn == LCN_HOLE)
        {
              VCN update_from = -1;
              s64 ofs;

              ofs =vcn_ofs + ((vcn - rl->vcn) << ni->vol->cluster_size_bits);
              err = 0;
               if (ntfs_attr_fill_hole(&err, ni->map_na, (s64)blocksize, &ofs, &rl, &update_from))
               {
                         return  -ENOMEM;
               }
               if (update_from!=-1 &&
		            ntfs_attr_update_mapping_pairs(&err, ni->map_na, 0 /*update_from*/))
               {
                         return -EIO;
               }
               lcn = ntfs_rl_vcn_to_lcn(&err, rl, vcn);
         }

		/* Successful remap. */
	   if (lcn >= 0) {
			/* Setup buffer head to point to correct block. */
			bh_result->b_blocknr = ((lcn << ni->vol->cluster_size_bits) +
					vcn_ofs) >> blocksize_bits;
	               //set_buffer_new(bh_result);
                      set_buffer_mapped(bh_result);
	               bh_result->b_bdev = ni->vol->sb->s_bdev;
	               bh_result->b_size = blocksize;
                      return 0;
	   }

		/* If first try and runlist unmapped, map and retry. */
	   if (!is_retry && lcn == LCN_RL_NOT_MAPPED) {
			is_retry = true;
			/*
			 * Attempt to map runlist, dropping lock for
			 * the duration.
			 */
//@tmp           up_read(&ni->runlist.lock);
			rl = ntfs_attr_find_vcn(&err, ni->map_na,  vcn);

			if (rl)
				goto try_map_bh;
		}
              return -ENOSPC;
}
static int ntfs3g_write_begin(struct file *file, struct address_space *mapping,
			loff_t pos, unsigned len, unsigned flags,
			struct page **pagep, void **fsdata)
{

	struct inode *vi =mapping->host;
	ntfs_inode *ni = NTFS_I(vi);
       int err = 0;


	*pagep = NULL;


       if(ni->map_na->data_size < pos+len)
       {
              ntfs_attr_search_ctx *ctx;
		if (ntfs_attr_truncate(&err, ni->map_na,pos+len)) {
			ntfs_log_perror("Failed to enlarge attribute");
                     NTFS3G_BUG_TRACE(!err);
			return -EIO;
		}
              NTFS3G_BUG_TRACE(ni->map_na->data_size < pos+len);

              if(NAttrNonResident(ni->map_na))
             {
                    ctx = ntfs_attr_get_search_ctx(&err, ni, NULL);
		      if (!ctx)
			       return -ENOMEM;
		     if (ntfs_attr_lookup(&err, ni->map_na->type, ni->map_na->name, ni->map_na->name_len, 0,
				0, NULL, 0, ctx))
		    {
			      ntfs_attr_put_search_ctx(&err, ctx);
                           return -EIO;
		     }

		     ctx->attr->initialized_size = cpu_to_sle64(pos + len);
                   mark_mft_record_dirty(ctx->ntfs_ino);
                   ni->map_na->initialized_size = pos + len;
                   ntfs_attr_put_search_ctx(&err, ctx);
             }
             else
             {
                 NTFS3G_BUG_TRACE(ni->map_na->data_size != ni->map_na->initialized_size);
             }


        }

        return block_write_begin(file, mapping, pos, len, flags, pagep, fsdata,
							ntfs3g_get_block);
}

static int ntfs3g_write_end(struct file *file, struct address_space *mapping,
			loff_t pos, unsigned len, unsigned copied,
			struct page *pagep, void *fsdata)
{
	struct inode *inode = mapping->host;

	copied = generic_write_end(file, mapping, pos, len, copied, pagep, fsdata);

       if(copied >=0)
       {
    	     inode->i_mtime = inode->i_ctime = CURRENT_TIME_SEC;
    	     mark_inode_dirty(inode);
        }
	return copied;
}


/**
 * ntfs_aops - general address space operations for inodes and attributes
 */
const struct address_space_operations ntfs3g_aops = {
	.readpage	= ntfs3g_readpage,	/* Fill page with data. */
	.readpages     = ntfs3g_readpages,
	.sync_page	= block_sync_page,	/* Currently, just unplugs the
						   disk request queue. */
	.writepage	= ntfs3g_writepage,	/* Write dirty page to disk. */

       .writepages     = ntfs3g_writepages,

	.migratepage	= buffer_migrate_page,	/* Move a page cache page from
						   one physical page to an
						   other. */
	.write_begin	= ntfs3g_write_begin,
	.write_end	= ntfs3g_write_end
};

/**
 * ntfs_mst_aops - general address space operations for mst protecteed inodes
 *		   and attributes
 */
const struct address_space_operations ntfs3g_mst_aops = {
	.readpage	= ntfs3g_readpage,	/* Fill page with data. */
	.sync_page	= block_sync_page,	/* Currently, just unplugs the
						   disk request queue. */
	.writepage	= ntfs3g_writepage,	/* Write dirty page to disk. */
	.set_page_dirty	= __set_page_dirty_nobuffers,	/* Set the page dirty
						   without touching the buffers
						   belonging to the page. */
	.migratepage	= buffer_migrate_page,	/* Move a page cache page from
						   one physical page to an
						   other. */
};
/**
 * mark_ntfs_record_dirty - mark an ntfs record dirty
 * @page:	page containing the ntfs record to mark dirty
 * @ofs:	byte offset within @page at which the ntfs record begins
 *
 * Set the buffers and the page in which the ntfs record is located dirty.
 *
 * The latter also marks the vfs inode the ntfs record belongs to dirty
 * (I_DIRTY_PAGES only).
 *
 * If the page does not have buffers, we create them and set them uptodate.
 * The page may not be locked which is why we need to handle the buffers under
 * the mapping->private_lock.  Once the buffers are marked dirty we no longer
 * need the lock since try_to_free_buffers() does not free dirty buffers.
 */
void mark_ntfs_record_dirty(struct page *page, const unsigned int ofs) {
	struct address_space *mapping;
	ntfs_inode *ni;
	struct buffer_head *bh, *head, *buffers_to_free = NULL;
	unsigned int end, bh_size, bh_ofs;


       mapping = page->mapping;
       ni = NTFS_I(mapping->host);
	NTFS3G_BUG_TRACE(!PageUptodate(page));
	end = ofs + ni->fixup_block_size;
	bh_size = VFS_I(ni)->i_sb->s_blocksize;
	spin_lock(&mapping->private_lock);
	if (unlikely(!page_has_buffers(page))) {
		spin_unlock(&mapping->private_lock);
		bh = head = alloc_page_buffers(page, bh_size, 1);
		spin_lock(&mapping->private_lock);
		if (likely(!page_has_buffers(page))) {
			struct buffer_head *tail;

			do {
				set_buffer_uptodate(bh);
				tail = bh;
				bh = bh->b_this_page;
			} while (bh);
			tail->b_this_page = head;
			attach_page_buffers(page, head);
		} else
			buffers_to_free = bh;
	}
	spin_unlock(&mapping->private_lock);
	lock_page(page);
	bh = head = page_buffers(page);
	NTFS3G_BUG_TRACE(!bh);
	do {
		bh_ofs = bh_offset(bh);
		if (bh_ofs + bh_size <= ofs)
			continue;
		if (unlikely(bh_ofs >= end))
			break;
		set_buffer_dirty(bh);
	} while ((bh = bh->b_this_page) != head);

	__set_page_dirty_nobuffers(page);
	
	unlock_page(page);
	
	if (unlikely(buffers_to_free)) {
		do {
			bh = buffers_to_free->b_this_page;
			free_buffer_head(buffers_to_free);
			buffers_to_free = bh;
		} while (buffers_to_free);
	}
}

