/**
 * bitmap.c - Bitmap handling code. Originated from the Linux-NTFS project.
 *
 * Copyright (c) 2002-2006 Anton Altaparmakov
 * Copyright (c) 2004-2005 Richard Russon
 * Copyright (c) 2004-2008 Szabolcs Szakacsits
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the NTFS-3G
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/stddef.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>	/* For bdev_hardsect_size(). */
#include <linux/backing-dev.h>
#include <linux/buffer_head.h>
#include <linux/vfs.h>
#include <linux/moduleparam.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/mount.h>

#include "config.h"
#include "types.h"
#include "attrib.h"
#include "bitmap.h"
#include "debug.h"
#include "misc.h"
#include "aops.h"

/**
 * ntfs_bitmap_set_bits_in_run - set a run of bits in a bitmap to a value
 * @na:		attribute containing the bitmap
 * @start_bit:	first bit to set
 * @count:	number of bits to set
 * @value:	value to set the bits to (i.e. 0 or 1)
 *
 * Set @count bits starting at bit @start_bit in the bitmap described by the
 * attribute @na to @value, where @value is either 0 or 1.
 *
 * On success return 0 and on error return -1 with errno set to the error code.
 */
static int ntfs_bitmap_set_bits_node_mmap(int *p_errno, struct inode *vi, s64 start_bit,
				       s64 count, int value)
{
       s64 available_bits, offset_bit;

    	u32 index;
	struct address_space *mapping;
       struct page *page;
       u8 *kaddr;
       s64 isize;

	if (!vi  || start_bit < 0 || count < 0) {
		*p_errno = EINVAL;
		ntfs_log_perror("%s:   b=%p  pos=%lld  count=%lld",
				__FUNCTION__,  b, (long long)pos,
				(long long)count);
		return -1;
	}
       isize =  i_size_read(vi);
       if(start_bit+count > (isize<<3))
      {
		*p_errno = EINVAL;
		ntfs_log_perror("%s:   b=%p  pos=%lld  count=%lld",
				__FUNCTION__,  b, (long long)pos,
				(long long)count);
		return -1;
	}
    
      	mapping = vi->i_mapping;

       index = (u32)(start_bit/(PAGE_CACHE_SIZE<<3));
       offset_bit = start_bit&(PAGE_CACHE_SIZE*8-1);
       
       while(count)
        {
		/* Update @index and get the next page. */
		page = ntfs_map_page(mapping, index);
		if (IS_ERR(page))
			break;
        
               kaddr = page_address(page);
               available_bits = PAGE_CACHE_SIZE*8 -offset_bit;
               if(available_bits>count)
                  available_bits = count;
               count -= available_bits;

               lock_page(page);
               if (unlikely(page_has_buffers(page))) 
               {
                       struct buffer_head *bh, *head;
                        unsigned int  bh_ofs;
                        unsigned bh_size = vi->i_sb->s_blocksize;
                               
		          bh = head = page_buffers(page);
	                 NTFS3G_BUG_TRACE(!bh);
	                 do {
		                bh_ofs = bh_offset(bh);
		               if (bh_ofs + bh_size <= (offset_bit>>3))
			             continue;
		             if (bh_ofs >= ((offset_bit+available_bits+7)>>3))
			             break;
		             set_buffer_dirty(bh);
	                 } while ((bh = bh->b_this_page) != head);
               }
               set_page_dirty(page);
               
               while((offset_bit&7) && available_bits)
               {
                     if(value)
                         kaddr[offset_bit>>3] |= 1<<(offset_bit&7);
                     else
                         kaddr[offset_bit>>3] &=  ~(1<<(offset_bit&7));
                     offset_bit++;
                     available_bits--;
               }
               while(available_bits>=8)
               {
                    if(value)
                         kaddr[offset_bit>>3] = 0xff;
                     else
                         kaddr[offset_bit>>3]&= 0x00;
                     offset_bit += 8;
                     available_bits -= 8;
               }
               while(available_bits)
               {
                     if(value)
                         kaddr[offset_bit>>3] |= 1<<(offset_bit&7);
                     else
                         kaddr[offset_bit>>3] &=  ~(1<<(offset_bit&7));
                     offset_bit++;
                     available_bits--;
               }                
		 unlock_page(page);
               ntfs_unmap_page(page);
               index++;
               offset_bit = 0;
	}
       if(count)
           return -1;
	return 0;
}
/**
 * ntfs_bitmap_set_run - set a run of bits in a bitmap
 * @na:		attribute containing the bitmap
 * @start_bit:	first bit to set
 * @count:	number of bits to set
 *
 * Set @count bits starting at bit @start_bit in the bitmap described by the
 * attribute @na.
 *
 * On success return 0 and on error return -1 with errno set to the error code.
 */
int  ntfs_bitmap_set_run(int *p_errno, ntfs_inode *bitmap_ni, s64 start_bit, s64 count)
{
	int ret; 
	
	ntfs_log_enter("Set from bit %lld, count %lld\n",
		       (long long)start_bit, (long long)count);
	ret = ntfs_bitmap_set_bits_node_mmap(p_errno, VFS_I(bitmap_ni), start_bit, count, 1);

	ntfs_log_leave("\n");
	return ret;
}

/**
 * ntfs_bitmap_clear_run - clear a run of bits in a bitmap
 * @na:		attribute containing the bitmap
 * @start_bit:	first bit to clear
 * @count:	number of bits to clear
 *
 * Clear @count bits starting at bit @start_bit in the bitmap described by the
 * attribute @na.
 *
 * On success return 0 and on error return -1 with errno set to the error code.
 */
int  ntfs_bitmap_clear_run(int *p_errno, ntfs_inode *bitmap_ni, s64 start_bit, s64 count)
{
	int ret; 
	
	ntfs_log_enter("Clear from bit %lld, count %lld\n",
		       (long long)start_bit, (long long)count); 
	ret = ntfs_bitmap_set_bits_node_mmap(p_errno, VFS_I(bitmap_ni), start_bit, count, 0);

	ntfs_log_leave("\n");
	return ret;
}

