/**
 * dir.c - Directory handling code. Originated from the Linux-NTFS project.
 *
 * Copyright (c) 2002-2005 Anton Altaparmakov
 * Copyright (c) 2004-2005 Richard Russon
 * Copyright (c) 2004-2008 Szabolcs Szakacsits
 * Copyright (c) 2005-2007 Yura Pakhuchiy
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the NTFS-3G
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/stddef.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>	/* For bdev_hardsect_size(). */
#include <linux/backing-dev.h>
#include <linux/buffer_head.h>
#include <linux/vfs.h>
#include <linux/moduleparam.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/mount.h>

#include "config.h"
#include "types.h"
#include "debug.h"
#include "attrib.h"
#include "inode.h"
#include "dir.h"
#include "volume.h"
#include "mft.h"
#include "index.h"
#include "ntfstime.h"
#include "lcnalloc.h"
#include "misc.h"
#include "security.h"


#define NTFS_VOLUME_NAME_LEN 32
#define NTFS_IOCTL_GET_VOLUME_LABEL	_IOR('r', 0x13, __u32)
/*
 * The little endian Unicode strings "$I30", "$SII", "$SDH", "$O"
 *  and "$Q" as global constants.
 */
ntfschar NTFS_INDEX_I30[5] = { const_cpu_to_le16('$'), const_cpu_to_le16('I'),
		const_cpu_to_le16('3'), const_cpu_to_le16('0'),
		const_cpu_to_le16('\0') };
ntfschar NTFS_INDEX_SII[5] = { const_cpu_to_le16('$'), const_cpu_to_le16('S'),
		const_cpu_to_le16('I'), const_cpu_to_le16('I'),
		const_cpu_to_le16('\0') };
ntfschar NTFS_INDEX_SDH[5] = { const_cpu_to_le16('$'), const_cpu_to_le16('S'),
		const_cpu_to_le16('D'), const_cpu_to_le16('H'),
		const_cpu_to_le16('\0') };
ntfschar NTFS_INDEX_O[3] = { const_cpu_to_le16('$'), const_cpu_to_le16('O'),
		const_cpu_to_le16('\0') };
ntfschar NTFS_INDEX_Q[3] = { const_cpu_to_le16('$'), const_cpu_to_le16('Q'),
		const_cpu_to_le16('\0') };
ntfschar NTFS_INDEX_R[3] = { const_cpu_to_le16('$'), const_cpu_to_le16('R'),
		const_cpu_to_le16('\0') };


typedef struct
{   
   int          pre_uname_len;
   ntfschar  pre_uname[NTFS_MAX_NAME_LEN];
}ntfs3g_dir_private_info;

/**
 * ntfs_inode_lookup_by_name - find an inode in a directory given its name
 * @dir_ni:	ntfs inode of the directory in which to search for the name
 * @uname:	Unicode name for which to search in the directory
 * @uname_len:	length of the name @uname in Unicode characters
 *
 * Look for an inode with name @uname in the directory with inode @dir_ni.
 * ntfs_inode_lookup_by_name() walks the contents of the directory looking for
 * the Unicode name. If the name is found in the directory, the corresponding
 * inode number (>= 0) is returned as a mft reference in cpu format, i.e. it
 * is a 64-bit number containing the sequence number.
 *
 * On error, return -1 with errno set to the error code. If the inode is is not
 * found errno is ENOENT.
 *
 * Note, @uname_len does not include the (optional) terminating NULL character.
 *
 * Note, we look for a case sensitive match first but we also look for a case
 * insensitive match at the same time. If we find a case insensitive match, we
 * save that for the case that we don't find an exact match, where we return
 * the mft reference of the case insensitive match.
 *
 * If the volume is mounted with the case sensitive flag set, then we only
 * allow exact matches.
 */
u64 ntfs_inode_lookup_by_name(int *p_errno, ntfs_inode *dir_ni, const ntfschar *uname,
		const int uname_len)
{
	VCN vcn;
	u64 mref = 0;
	s64 br;
	ntfs_volume *vol = dir_ni->vol;
	ntfs_attr_search_ctx *ctx;
	INDEX_ROOT *ir;
	INDEX_ENTRY *ie;
	INDEX_ALLOCATION *ia;
	u8 *index_end;
	int eo, rc;
	u32 index_block_size, index_vcn_size;
	u8 index_vcn_size_bits;
       ntfs_inode *ia_ni = NULL;

	ntfs_log_trace("Entering\n");

	if (!dir_ni || !uname || uname_len <= 0) {
		*p_errno = EINVAL;
		return -1;
	}
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(dir_ni));
	ctx = ntfs_attr_get_search_ctx(p_errno, dir_ni, NULL);
	if (!ctx)
		return -1;

	/* Find the index root attribute in the mft record. */
	if (ntfs_attr_lookup(p_errno, AT_INDEX_ROOT, NTFS_INDEX_I30, 4, CASE_SENSITIVE, 0, NULL,
			0, ctx)) {
		ntfs_log_perror("Index root attribute missing in directory inode "
				"%lld", (unsigned long long)dir_ni->mft_no);
		goto put_err_out;
	}
	/* Get to the index root value. */
	ir = (INDEX_ROOT*)((u8*)ctx->attr +
			le16_to_cpu(ctx->attr->value_offset));
	index_block_size = le32_to_cpu(ir->index_block_size);
	if (index_block_size < NTFS_BLOCK_SIZE ||
			index_block_size & (index_block_size - 1)) {
		ntfs_log_error(dir_ni->vol->sb, "Index block size %u is invalid.\n",
				(unsigned)index_block_size);
		goto put_err_out;
	}
	index_end = (u8*)&ir->index + le32_to_cpu(ir->index.index_length);
	/* The first index entry. */
	ie = (INDEX_ENTRY*)((u8*)&ir->index +
			le32_to_cpu(ir->index.entries_offset));
	/*
	 * Loop until we exceed valid memory (corruption case) or until we
	 * reach the last entry.
	 */
	for (;; ie = (INDEX_ENTRY*)((u8*)ie + le16_to_cpu(ie->length))) {
		/* Bounds checks. */
		if ((u8*)ie < (u8*)ctx->mrec || (u8*)ie +
				sizeof(INDEX_ENTRY_HEADER) > index_end ||
				(u8*)ie + le16_to_cpu(ie->key_length) >
				index_end) {
			ntfs_log_error(dir_ni->vol->sb, "Index entry out of bounds in inode %lld"
				       "\n", (unsigned long long)dir_ni->mft_no);
			goto put_err_out;
		}
		/*
		 * The last entry cannot contain a name. It can however contain
		 * a pointer to a child node in the B+tree so we just break out.
		 */
		if (ie->ie_flags & INDEX_ENTRY_END)
			break;
		
		if (!le16_to_cpu(ie->length)) {
			ntfs_log_error(dir_ni->vol->sb, "Zero length index entry in inode %lld"
				       "\n", (unsigned long long)dir_ni->mft_no);
			goto put_err_out;
		}
		/*
		 * Not a perfect match, need to do full blown collation so we
		 * know which way in the B+tree we have to go.
		 */
		rc = ntfs_names_collate(uname, uname_len,
				(ntfschar*)&ie->key.file_name.file_name,
				ie->key.file_name.file_name_length, 1,
				IGNORE_CASE, vol->upcase, vol->upcase_len);
		/*
		 * If uname collates before the name of the current entry, there
		 * is definitely no such name in this index but we might need to
		 * descend into the B+tree so we just break out of the loop.
		 */
		if (rc == -1)
			break;
		/* The names are not equal, continue the search. */
		if (rc)
			continue;
		/*
		 * Names match with case insensitive comparison, now try the
		 * case sensitive comparison, which is required for proper
		 * collation.
		 */
		rc = ntfs_names_collate(uname, uname_len,
				(ntfschar*)&ie->key.file_name.file_name,
				ie->key.file_name.file_name_length, 1,
				CASE_SENSITIVE, vol->upcase, vol->upcase_len);
		if (rc == -1)
			break;
		if (rc)
			continue;
		/*
		 * Perfect match, this will never happen as the
		 * ntfs_are_names_equal() call will have gotten a match but we
		 * still treat it correctly.
		 */
		mref = le64_to_cpu(ie->indexed_file);
		ntfs_attr_put_search_ctx(p_errno, ctx);
		return mref;
	}
	/*
	 * We have finished with this index without success. Check for the
	 * presence of a child node and if not present return error code
	 * ENOENT, unless we have got the mft reference of a matching name
	 * cached in mref in which case return mref.
	 */
	if (!(ie->ie_flags & INDEX_ENTRY_NODE)) {
		ntfs_attr_put_search_ctx(p_errno, ctx);
		if (mref)
			return mref;
		ntfs_log_debug("Entry not found.\n");
		*p_errno = ENOENT;
		return -1;
	} /* Child node present, descend into it. */

	/* Open the index allocation attribute. */
       ia_ni = ntfs_attr_inode_open(p_errno, dir_ni, AT_INDEX_ALLOCATION, NTFS_INDEX_I30, 4);
	if (!ia_ni) {
		ntfs_log_perror("Failed to open index allocation (inode %lld)",
				(unsigned long long)dir_ni->mft_no);
             //*p_errno = ENOENT;
		goto put_err_out;
	}

	/* Allocate a buffer for the current index block. */
	ia = ntfs_malloc_nofs(p_errno, index_block_size);
	if (!ia) {
		goto put_err_out;
	}

	/* Determine the size of a vcn in the directory index. */
	if (vol->cluster_size <= index_block_size) {
		index_vcn_size = vol->cluster_size;
		index_vcn_size_bits = vol->cluster_size_bits;
	} else {
		index_vcn_size = vol->sector_size;
		index_vcn_size_bits = vol->sector_size_bits;
	}

	/* Get the starting vcn of the index_block holding the child node. */
	vcn = sle64_to_cpup((sle64*)((u8*)ie + le16_to_cpu(ie->length) - 8));

descend_into_child_node:

	/* Read the index block starting at vcn. */
	br = ntfs_inode_map_read(p_errno, VFS_I(ia_ni), vcn << index_vcn_size_bits, index_block_size, (u8*)ia);
	if (br != index_block_size) {
		if (br != -1)
			*p_errno = EIO;
		ntfs_log_perror("Failed to read vcn 0x%llx",
			       	(unsigned long long)vcn);
		goto close_err_out;
	}

	if (sle64_to_cpu(ia->index_block_vcn) != vcn) {
		ntfs_log_error(dir_ni->vol->sb, "Actual VCN (0x%llx) of index buffer is different "
				"from expected VCN (0x%llx).\n",
				(long long)sle64_to_cpu(ia->index_block_vcn),
				(long long)vcn);
		*p_errno = EIO;
		goto close_err_out;
	}
	if (le32_to_cpu(ia->index.allocated_size) + 0x18 != index_block_size) {
		ntfs_log_error(dir_ni->vol->sb, "Index buffer (VCN 0x%llx) of directory inode 0x%llx "
				"has a size (%u) differing from the directory "
				"specified size (%u).\n", (long long)vcn,
				(unsigned long long)ia_ni->mft_no,
				(unsigned) le32_to_cpu(ia->index.allocated_size) + 0x18,
				(unsigned)index_block_size);
		*p_errno = EIO;
		goto close_err_out;
	}
	index_end = (u8*)&ia->index + le32_to_cpu(ia->index.index_length);
	if (index_end > (u8*)ia + index_block_size) {
		ntfs_log_error(dir_ni->vol->sb, "Size of index buffer (VCN 0x%llx) of directory inode "
				"0x%llx exceeds maximum size.\n",
				(long long)vcn, (unsigned long long)ia_ni->mft_no);
		*p_errno = EIO;
		goto close_err_out;
	}

	/* The first index entry. */
	ie = (INDEX_ENTRY*)((u8*)&ia->index +
			le32_to_cpu(ia->index.entries_offset));
	/*
	 * Iterate similar to above big loop but applied to index buffer, thus
	 * loop until we exceed valid memory (corruption case) or until we
	 * reach the last entry.
	 */
	for (;; ie = (INDEX_ENTRY*)((u8*)ie + le16_to_cpu(ie->length))) {
		/* Bounds check. */
		if ((u8*)ie < (u8*)ia || (u8*)ie +
				sizeof(INDEX_ENTRY_HEADER) > index_end ||
				(u8*)ie + le16_to_cpu(ie->key_length) >
				index_end) {
			ntfs_log_error(dir_ni->vol->sb, "Index entry out of bounds in directory "
				       "inode %lld.\n", 
				       (unsigned long long)ia_ni->mft_no);
			*p_errno = EIO;
			goto close_err_out;
		}
		/*
		 * The last entry cannot contain a name. It can however contain
		 * a pointer to a child node in the B+tree so we just break out.
		 */
		if (ie->ie_flags & INDEX_ENTRY_END)
			break;
		
		if (!le16_to_cpu(ie->length)) {
			*p_errno = EIO;
			ntfs_log_error(dir_ni->vol->sb, "Zero length index entry in inode %lld"
				       "\n", (unsigned long long)ia_ni->mft_no);
			goto close_err_out;
		}
		/*
		 * Not a perfect match, need to do full blown collation so we
		 * know which way in the B+tree we have to go.
		 */
		rc = ntfs_names_collate(uname, uname_len,
				(ntfschar*)&ie->key.file_name.file_name,
				ie->key.file_name.file_name_length, 1,
				IGNORE_CASE, vol->upcase, vol->upcase_len);
		/*
		 * If uname collates before the name of the current entry, there
		 * is definitely no such name in this index but we might need to
		 * descend into the B+tree so we just break out of the loop.
		 */
		if (rc == -1)
			break;
		/* The names are not equal, continue the search. */
		if (rc)
			continue;
		/*
		 * Names match with case insensitive comparison, now try the
		 * case sensitive comparison, which is required for proper
		 * collation.
		 */
		rc = ntfs_names_collate(uname, uname_len,
				(ntfschar*)&ie->key.file_name.file_name,
				ie->key.file_name.file_name_length, 1,
				CASE_SENSITIVE, vol->upcase, vol->upcase_len);
		if (rc == -1)
			break;
		if (rc)
			continue;
		
		mref = le64_to_cpu(ie->indexed_file);
		ntfs_free(ia);
		ntfs_attr_put_search_ctx(p_errno, ctx);
              eo = 0;
              ntfs_attr_inode_close(&eo,  ia_ni);
		return mref;
	}
	/*
	 * We have finished with this index buffer without success. Check for
	 * the presence of a child node.
	 */
	if (ie->ie_flags & INDEX_ENTRY_NODE) {
		if ((ia->index.ih_flags & NODE_MASK) == LEAF_NODE) {
			ntfs_log_error(dir_ni->vol->sb, "Index entry with child node found in a leaf "
					"node in directory inode %lld.\n",
					(unsigned long long)ia_ni->mft_no);
			*p_errno = EIO;
			goto close_err_out;
		}
		/* Child node present, descend into it. */
		vcn = sle64_to_cpup((sle64*)((u8*)ie + le16_to_cpu(ie->length) - 8));
		if (vcn >= 0)
			goto descend_into_child_node;
		ntfs_log_error(dir_ni->vol->sb, "Negative child node vcn in directory inode "
			       "0x%llx.\n", (unsigned long long)ia_ni->mft_no);
		*p_errno = EIO;
		goto close_err_out;
	}
	ntfs_free(ia);
	ntfs_attr_put_search_ctx(p_errno, ctx);
	/*
	 * No child node present, return error code ENOENT, unless we have got
	 * the mft reference of a matching name cached in mref in which case
	 * return mref.
	 */
       if(ia_ni)
           ntfs_attr_inode_close(&eo,  ia_ni);
       
	if (mref)
		return mref;
	ntfs_log_debug("Entry not found.\n");
	*p_errno = ENOENT;
	return -1;
put_err_out:
	eo = EIO;
	ntfs_log_debug("Corrupt directory. Aborting lookup.\n");
eo_put_err_out:
	ntfs_attr_put_search_ctx(p_errno, ctx);
       if(ia_ni)
           ntfs_attr_inode_close(p_errno, ia_ni);
	*p_errno = eo;
	return -1;
close_err_out:
	eo = *p_errno;
	ntfs_free(ia);
	goto eo_put_err_out;
}
/*
 * The little endian Unicode string ".." for ntfs_readdir().
 */
static const ntfschar dotdot[3] = { const_cpu_to_le16('.'),
				   const_cpu_to_le16('.'),
				   const_cpu_to_le16('\0') };

/*
 * union index_union -
 * More helpers for ntfs_readdir().
 */
typedef union {
	INDEX_ROOT *ir;
	INDEX_ALLOCATION *ia;
} index_union __attribute__((__transparent_union__));

/**
 * enum INDEX_TYPE -
 * More helpers for ntfs_readdir().
 */
typedef enum {
	INDEX_TYPE_ROOT,	/* index root */
	INDEX_TYPE_ALLOCATION,	/* index allocation */
} INDEX_TYPE;
 static int ntfs_fuse_filler(void *dirent, filldir_t filldir, 
		const ntfschar *name, const int name_len, const int name_type,
		const loff_t pos , const MFT_REF mref,
		const unsigned dt_type)
{
	char *filename = NULL;
	int ret = 0;

	if (name_type == FILE_NAME_DOS)
		return 0;
	
	if (ntfs_ucstombs(&ret, name, name_len, &filename, 0) < 0) {
		ntfs_log_perror("Skipping unrepresentable filename (inode %llu)",
				(unsigned long long)MREF(mref));
		return -1;
	}
    
       ret = 0;
       
	if (MREF(mref) == FILE_root || MREF(mref) >= FILE_first_user) 
       {

              unsigned int mode = 0;
		 
		if (dt_type == NTFS_DT_REG)
			mode = DT_REG;
		else if (dt_type == NTFS_DT_DIR)
			mode = DT_DIR;
		if (filldir(dirent, filename, strlen(filename),  pos, MREF(mref),  mode))
		{
		      ntfs_log_perror("failed to fill dir entry %s", filename);
		       ret = -1;
	       }
	}
	
	ntfs_free(filename);
	return ret;
}
/**
 * ntfs_filldir - ntfs specific filldir method
 * @dir_ni:	ntfs inode of current directory
 * @pos:	current position in directory
 * @ivcn_bits:	log(2) of index vcn size
 * @index_type:	specifies whether @iu is an index root or an index allocation
 * @iu:		index root or index block to which @ie belongs
 * @ie:		current index entry
 * @dirent:	context for filldir callback supplied by the caller
 * @filldir:	filldir callback supplied by the caller
 *
 * Pass information specifying the current directory entry @ie to the @filldir
 * callback.
 */
static int ntfs_filldir(ntfs_inode *dir_ni, s64 *pos, u8 ivcn_bits,
		const INDEX_TYPE index_type, index_union iu, INDEX_ENTRY *ie,
		void *dirent, filldir_t filldir)
{
	FILE_NAME_ATTR *fn = &ie->key.file_name;
	unsigned dt_type;

	ntfs_log_trace("Entering.\n");
	
	/* Advance the position even if going to skip the entry. */
	if (index_type == INDEX_TYPE_ALLOCATION)
		*pos = (u8*)ie - (u8*)iu.ia + (sle64_to_cpu(
				iu.ia->index_block_vcn) << ivcn_bits) +
				dir_ni->vol->mft_record_size;
	else /* if (index_type == INDEX_TYPE_ROOT) */
		*pos = (u8*)ie - (u8*)iu.ir;
	/* Skip root directory self reference entry. */
	if (MREF_LE(ie->indexed_file) == FILE_root)
		return 0;
	if (ie->key.file_name.file_attributes & FILE_ATTR_I30_INDEX_PRESENT)
		dt_type = NTFS_DT_DIR;
	else if (fn->file_attributes & FILE_ATTR_SYSTEM)
		dt_type = NTFS_DT_UNKNOWN;
	else
		dt_type = NTFS_DT_REG;
	return ntfs_fuse_filler(dirent,  filldir, fn->file_name, fn->file_name_length,
			fn->file_name_type, *pos,
			le64_to_cpu(ie->indexed_file), dt_type);
}

/**
 * ntfs_mft_get_parent_ref - find mft reference of parent directory of an inode
 * @ni:		ntfs inode whose parent directory to find
 *
 * Find the parent directory of the ntfs inode @ni. To do this, find the first
 * file name attribute in the mft record of @ni and return the parent mft
 * reference from that.
 *
 * Note this only makes sense for directories, since files can be hard linked
 * from multiple directories and there is no way for us to tell which one is
 * being looked for.
 *
 * Technically directories can have hard links, too, but we consider that as
 * illegal as Linux/UNIX do not support directory hard links.
 *
 * Return the mft reference of the parent directory on success or -1 on error
 * with errno set to the error code.
 */
MFT_REF ntfs_mft_get_parent_ref(int *p_errno, ntfs_inode *ni)
{
	MFT_REF mref;
	ntfs_attr_search_ctx *ctx;
	FILE_NAME_ATTR *fn;
	int eo;

	ntfs_log_trace("Entering.\n");
	
	if (!ni) {
		*p_errno = EINVAL;
		return ERR_MREF(-1);
	}
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
	ctx = ntfs_attr_get_search_ctx(p_errno, ni, NULL);
	if (!ctx)
		return ERR_MREF(-1);
	if (ntfs_attr_lookup(p_errno, AT_FILE_NAME, AT_UNNAMED, 0, 0, 0, NULL, 0, ctx)) {
		ntfs_log_error(ni->vol->sb, "No file name found in inode %lld\n", 
			       (unsigned long long)ni->mft_no);
		goto err_out;
	}
	if (ctx->attr->non_resident) {
		ntfs_log_error(ni->vol->sb, "File name attribute must be resident (inode "
			       "%lld)\n", (unsigned long long)ni->mft_no);
		goto io_err_out;
	}
	fn = (FILE_NAME_ATTR*)((u8*)ctx->attr +
			le16_to_cpu(ctx->attr->value_offset));
	if ((u8*)fn +	le32_to_cpu(ctx->attr->value_length) >
			(u8*)ctx->attr + le32_to_cpu(ctx->attr->length)) {
		ntfs_log_error(ni->vol->sb, "Corrupt file name attribute in inode %lld.\n",
			       (unsigned long long)ni->mft_no);
		goto io_err_out;
	}
	mref = le64_to_cpu(fn->parent_directory);
	ntfs_attr_put_search_ctx(p_errno, ctx);
	return mref;
io_err_out:
	*p_errno = EIO;
err_out:
	eo = *p_errno;
	ntfs_attr_put_search_ctx(p_errno, ctx);
	*p_errno = eo;
	return ERR_MREF(-1);
}
#if 1
static int ntfs_idx_cmp_uname(const ntfschar *name1, const u32 name1_len,
		const ntfschar *name2, const u32 name2_len,
		const ntfschar *upcase,
		const u32 upcase_len)
{
       int rc;

     rc = ntfs_names_collate(name1, name1_len,
				                  name2, name2_len, 1,
				                     IGNORE_CASE, upcase, upcase_len);
     if(rc)
           return rc;
      return ntfs_names_collate(name1, name1_len,
				                  name2, name2_len, 1,
				                     CASE_SENSITIVE, upcase, upcase_len);
}
static int ntfs_browser_dir_ia(int *p_errno, ntfs_inode *dir_ni, void *dirent,  ntfs_inode *ia_ni, ntfschar *pre_uname,
		 int *pre_uname_len, int skip_pre, VCN vcn, u8 index_vcn_size_bits, u32 index_block_size, s64 *pos, filldir_t filldir)
{

       INDEX_ALLOCATION *ia;
       s64 br;
       u8 *index_end;
       INDEX_ENTRY *ie;
       int rc;
       int res = 0;
       
	/* Allocate a buffer for the current index block. */
	ia = ntfs_malloc_nofs(p_errno, index_block_size);
	if (!ia) {
              *p_errno = ENOMEM;
		return -1;
	}

	/* Read the index block starting at vcn. */
	br = ntfs_inode_map_read(p_errno, VFS_I(ia_ni), vcn << index_vcn_size_bits, index_block_size, (u8*)ia);
	if (br != index_block_size) {
		if (br != -1)
			*p_errno = EIO;
		ntfs_log_perror("Failed to read vcn 0x%llx",
			       	(unsigned long long)vcn);
		goto close_err_out;
	}

	if (sle64_to_cpu(ia->index_block_vcn) != vcn) {
		ntfs_log_error(dir_ni->vol->sb, "Actual VCN (0x%llx) of index buffer is different "
				"from expected VCN (0x%llx).\n",
				(long long)sle64_to_cpu(ia->index_block_vcn),
				(long long)vcn);
		*p_errno = EIO;
		goto close_err_out;
	}
	if (le32_to_cpu(ia->index.allocated_size) + 0x18 != index_block_size) {
		ntfs_log_error(dir_ni->vol->sb, "Index buffer (VCN 0x%llx) of directory inode 0x%llx "
				"has a size (%u) differing from the directory "
				"specified size (%u).\n", (long long)vcn,
				(unsigned long long)ia_ni->mft_no,
				(unsigned) le32_to_cpu(ia->index.allocated_size) + 0x18,
				(unsigned)index_block_size);
		*p_errno = EIO;
		goto close_err_out;
	}
	index_end = (u8*)&ia->index + le32_to_cpu(ia->index.index_length);
	if (index_end > (u8*)ia + index_block_size) {
		ntfs_log_error(dir_ni->vol->sb, "Size of index buffer (VCN 0x%llx) of directory inode "
				"0x%llx exceeds maximum size.\n",
				(long long)vcn, (unsigned long long)ia_ni->mft_no);
		*p_errno = EIO;
		goto close_err_out;
	}

	/* The first index entry. */
	ie = (INDEX_ENTRY*)((u8*)&ia->index +
			le32_to_cpu(ia->index.entries_offset));
	/*
	 * Iterate similar to above big loop but applied to index buffer, thus
	 * loop until we exceed valid memory (corruption case) or until we
	 * reach the last entry.
	 */
	for (;; ie = (INDEX_ENTRY*)((u8*)ie + le16_to_cpu(ie->length))) {
		/* Bounds check. */
		if ((u8*)ie < (u8*)ia || (u8*)ie +
				sizeof(INDEX_ENTRY_HEADER) > index_end ||
				(u8*)ie + le16_to_cpu(ie->key_length) >
				index_end) {
			ntfs_log_error(dir_ni->vol->sb, "Index entry out of bounds in directory "
				       "inode %lld.\n", 
				       (unsigned long long)ia_ni->mft_no);
			*p_errno = EIO;
			goto close_err_out;
		}
		if (!le16_to_cpu(ie->length)) {
			*p_errno = EIO;
			ntfs_log_error(dir_ni->vol->sb, "Zero length index entry in inode %lld"
				       "\n", (unsigned long long)ia_ni->mft_no);
			goto close_err_out;
		}
		/*
		 * Not a perfect match, need to do full blown collation so we
		 * know which way in the B+tree we have to go.
		 */
	       if(skip_pre && !(ie->ie_flags & INDEX_ENTRY_END))
	       {
		        rc = ntfs_idx_cmp_uname(pre_uname, *pre_uname_len,
				              (ntfschar*)&ie->key.file_name.file_name,
				              ie->key.file_name.file_name_length, dir_ni->vol->upcase, dir_ni->vol->upcase_len);
		        if (rc > 0)
			       continue;
		        if (rc<0)
		        {	
                         if(ie->ie_flags & INDEX_ENTRY_NODE)
                         {
   		               if ((ia->index.ih_flags & NODE_MASK) == LEAF_NODE) {
			                ntfs_log_error(dir_ni->vol->sb, "Index entry with child node found in a leaf "
					                                                                           "node in directory inode %lld.\n",
					       (unsigned long long)ia_ni->mft_no);
			                *p_errno = EIO;
			                 goto close_err_out;
		               }
		                 /* Child node present, descend into it. */
		                vcn = sle64_to_cpup((sle64*)((u8*)ie + le16_to_cpu(ie->length) - 8));
		                if (vcn <0)
		                {
		                     ntfs_log_error(dir_ni->vol->sb, "Negative child node vcn in directory inode "
			              "0x%llx.\n", (unsigned long long)ia_ni->mft_no);
		                     *p_errno = EIO;
		                      goto close_err_out;
		                 }
                               res = ntfs_browser_dir_ia(p_errno, dir_ni, dirent,  ia_ni, pre_uname,
		                                          pre_uname_len, skip_pre, vcn, index_vcn_size_bits, index_block_size, pos, filldir);
                               if(res < 0)
		                       goto close_err_out;
                               else if(res > 0)
                                       goto done;
                         }  

                          rc = ntfs_filldir(dir_ni, pos, index_vcn_size_bits,
				                      INDEX_TYPE_ALLOCATION, ia, ie, dirent, filldir);
                          if (rc) 
                          {
                                   res = 1;
			             goto done;
		            }
                          memcpy(pre_uname, &ie->key.file_name.file_name, sizeof(ntfschar)*ie->key.file_name.file_name_length);
                           *pre_uname_len = ie->key.file_name.file_name_length;
		        }
                      skip_pre = 0;
                      continue;

	       }

              if(ie->ie_flags & INDEX_ENTRY_NODE)
              {
                        
 		          if ((ia->index.ih_flags & NODE_MASK) == LEAF_NODE) {
			                ntfs_log_error(dir_ni->vol->sb, "Index entry with child node found in a leaf "
					                                                                           "node in directory inode %lld.\n",
					       (unsigned long long)ia_ni->mft_no);
			                *p_errno = EIO;
			                 goto close_err_out;
		          }
		           /* Child node present, descend into it. */
		           vcn = sle64_to_cpup((sle64*)((u8*)ie + le16_to_cpu(ie->length) - 8));
		           if (vcn <0)
		           {
		                     ntfs_log_error(dir_ni->vol->sb, "Negative child node vcn in directory inode "
			              "0x%llx.\n", (unsigned long long)ia_ni->mft_no);
		                     *p_errno = EIO;
		                      goto close_err_out;
		          }
        
                         res = ntfs_browser_dir_ia(p_errno, dir_ni, dirent, ia_ni, pre_uname,
		                   pre_uname_len, skip_pre, vcn, index_vcn_size_bits, index_block_size, pos, filldir);
                         if(res < 0)
		                    goto close_err_out;
                         else if(res > 0)
                            goto done;
               }  
                if (ie->ie_flags & INDEX_ENTRY_END)
			   goto done;
		    /*
		        * Submit the directory entry to ntfs_filldir(), which will
		      * invoke the filldir() callback as appropriate.
		     */
		    rc = ntfs_filldir(dir_ni, pos, index_vcn_size_bits,
				INDEX_TYPE_ALLOCATION, ia, ie, dirent, filldir);
		    if (rc) {
                        res = 1;
			   goto done;
		    }
                  memcpy(pre_uname, &ie->key.file_name.file_name, sizeof(ntfschar)*ie->key.file_name.file_name_length);
                   *pre_uname_len = ie->key.file_name.file_name_length;

	}
 done:
       ntfs_free(ia);
	return res;
close_err_out:
	if(*p_errno == 0)
           *p_errno = EIO;        
	ntfs_free(ia);
	return -1;
}

static int ntfs_browser_dir_ir(int *p_errno, ntfs_inode *dir_ni,  void *dirent, ntfschar *pre_uname,
		int *pre_uname_len, int skip_pre, s64 *pos, filldir_t filldir)
{
	VCN vcn;
	ntfs_volume *vol = dir_ni->vol;
	ntfs_attr_search_ctx *ctx;
	INDEX_ROOT *ir;
	INDEX_ENTRY *ie;
	u8 *index_end;
	int eo, rc;
	u32 index_block_size, index_vcn_size;
	u8 index_vcn_size_bits;
       ntfs_inode *ia_ni = NULL;
       int res = 0;

	ntfs_log_trace("Entering\n");

	if (!dir_ni) {
		*p_errno = EINVAL;
		return -1;
	}
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(dir_ni));
	ctx = ntfs_attr_get_search_ctx(p_errno, dir_ni, NULL);
	if (!ctx)
		return -1;

	/* Find the index root attribute in the mft record. */
	if (ntfs_attr_lookup(p_errno, AT_INDEX_ROOT, NTFS_INDEX_I30, 4, CASE_SENSITIVE, 0, NULL,
			0, ctx)) {
		ntfs_log_perror("Index root attribute missing in directory inode "
				"%lld", (unsigned long long)dir_ni->mft_no);
		goto put_err_out;
	}
	/* Get to the index root value. */
	ir = (INDEX_ROOT*)((u8*)ctx->attr +
			le16_to_cpu(ctx->attr->value_offset));
	index_block_size = le32_to_cpu(ir->index_block_size);
	if (index_block_size < NTFS_BLOCK_SIZE ||
			index_block_size & (index_block_size - 1)) {
		ntfs_log_error(dir_ni->vol->sb, "Index block size %u is invalid.\n",
				(unsigned)index_block_size);
		goto put_err_out;
	}
	index_end = (u8*)&ir->index + le32_to_cpu(ir->index.index_length);
	/* The first index entry. */
	ie = (INDEX_ENTRY*)((u8*)&ir->index +
			le32_to_cpu(ir->index.entries_offset));
	/*
	 * Loop until we exceed valid memory (corruption case) or until we
	 * reach the last entry.
	 */
	ia_ni = ntfs_attr_inode_open(p_errno, dir_ni, AT_INDEX_ALLOCATION, NTFS_INDEX_I30, 4);
      /* Determine the size of a vcn in the directory index. */
	if (vol->cluster_size <= index_block_size) {
		              index_vcn_size = vol->cluster_size;
		             index_vcn_size_bits = vol->cluster_size_bits;
	} else {
		     index_vcn_size = vol->sector_size;
		     index_vcn_size_bits = vol->sector_size_bits;
	}
	for (;; ie = (INDEX_ENTRY*)((u8*)ie + le16_to_cpu(ie->length))) {
		/* Bounds checks. */
		if ((u8*)ie < (u8*)ctx->mrec || (u8*)ie +
				sizeof(INDEX_ENTRY_HEADER) > index_end ||
				(u8*)ie + le16_to_cpu(ie->key_length) >
				index_end) {
			ntfs_log_error(dir_ni->vol->sb, "Index entry out of bounds in inode %lld"
				       "\n", (unsigned long long)dir_ni->mft_no);
			goto put_err_out;
		}
		
		if (!le16_to_cpu(ie->length)) {
			ntfs_log_error(dir_ni->vol->sb, "Zero length index entry in inode %lld"
				       "\n", (unsigned long long)dir_ni->mft_no);
			goto put_err_out;
		}
		/*
		 * Not a perfect match, need to do full blown collation so we
		 * know which way in the B+tree we have to go.
		 */
		/*
		 * Not a perfect match, need to do full blown collation so we
		 * know which way in the B+tree we have to go.
		 */
	       if(skip_pre && !(ie->ie_flags & INDEX_ENTRY_END))
	       {
		        rc = ntfs_idx_cmp_uname(pre_uname, *pre_uname_len,
				              (ntfschar*)&ie->key.file_name.file_name,
				              ie->key.file_name.file_name_length, dir_ni->vol->upcase, dir_ni->vol->upcase_len);
		        if (rc > 0)
			       continue;
		        if (rc<0)
		        {

                          if(ie->ie_flags & INDEX_ENTRY_NODE)
                         {
                              if(!ia_ni)
                             {
                                	ntfs_log_error(dir_ni->vol->sb, "Negative child node vcn in directory inode "
			                       "0x%llx.\n", (unsigned long long)ia_ni->mft_no);
		                     *p_errno = EIO;
		                      goto put_err_out;
                             }

		                 /* Child node present, descend into it. */
		                vcn = sle64_to_cpup((sle64*)((u8*)ie + le16_to_cpu(ie->length) - 8));
		                if (vcn <0)
		                {
		                     ntfs_log_error(dir_ni->vol->sb, "Negative child node vcn in directory inode "
			              "0x%llx.\n", (unsigned long long)ia_ni->mft_no);
		                     *p_errno = EIO;
		                      goto put_err_out;
		                 }
                               res = ntfs_browser_dir_ia(p_errno, dir_ni, dirent,  ia_ni, pre_uname,
		                                          pre_uname_len, skip_pre, vcn, index_vcn_size_bits, index_block_size, pos,  filldir);
                               if(res < 0)
		                       goto put_err_out;
                               else if(res > 0)
                                       goto done;
      
                            }
                            rc = ntfs_filldir(dir_ni, pos, index_vcn_size_bits,
				                      INDEX_TYPE_ROOT, ir, ie, dirent, filldir);
                             if (rc) 
                             {
                                          res = 1;
			                    goto done;
		                }
                               memcpy(pre_uname, &ie->key.file_name.file_name, sizeof(ntfschar)*ie->key.file_name.file_name_length);
                                          *pre_uname_len = ie->key.file_name.file_name_length;
		          }
                        skip_pre = 0;
                        continue;

	       }
              if(ie->ie_flags & INDEX_ENTRY_NODE)
              {
                  if(!ia_ni)
                 {
                       ntfs_log_error(dir_ni->vol->sb, "Negative child node vcn in directory inode "
			                       "0x%llx.\n", (unsigned long long)ia_ni->mft_no);
		         *p_errno = EIO;
		          goto put_err_out;
                 }

	           /* Get the starting vcn of the index_block holding the child node. */
	          vcn = sle64_to_cpup((sle64*)((u8*)ie + le16_to_cpu(ie->length) - 8));
                 res = ntfs_browser_dir_ia(p_errno, dir_ni, dirent,  ia_ni, pre_uname,
		                                          pre_uname_len, skip_pre, vcn, index_vcn_size_bits, index_block_size, pos, filldir);
                 if(res < 0)
		           goto put_err_out;
                 else if(res > 0)
                         goto done;
              }

              if (ie->ie_flags & INDEX_ENTRY_END)
              {
                                    goto done;
              }
		rc = ntfs_filldir(dir_ni, pos, index_vcn_size_bits,
				INDEX_TYPE_ROOT, ir, ie, dirent, filldir);
		if (rc) 
              {
                     res = 1;
			goto done;
		}
              memcpy(pre_uname, &ie->key.file_name.file_name, sizeof(ntfschar)*ie->key.file_name.file_name_length);
                           *pre_uname_len = ie->key.file_name.file_name_length;

	}
done:
       ntfs_attr_put_search_ctx(p_errno, ctx);
        if(ia_ni)
           ntfs_attr_inode_close(p_errno, ia_ni);
	return res;
put_err_out:
	eo = EIO;
	ntfs_attr_put_search_ctx(p_errno, ctx);
       if(ia_ni)
           ntfs_attr_inode_close(p_errno, ia_ni);
	*p_errno = eo;
	return -1;
}
/**
 * ntfs_readdir - read the contents of an ntfs directory
 * @dir_ni:	ntfs inode of current directory
 * @pos:	current position in directory
 * @dirent:	context for filldir callback supplied by the caller
 * @filldir:	filldir callback supplied by the caller
 *
 * Parse the index root and the index blocks that are marked in use in the
 * index bitmap and hand each found directory entry to the @filldir callback
 * supplied by the caller.
 *
 * Return 0 on success or -1 on error with errno set to the error code.
 *
 * Note: Index blocks are parsed in ascending vcn order, from which follows
 * that the directory entries are not returned sorted.
 */

static int ntfs3g_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
	ntfs_volume *vol;
       loff_t pos = filp->f_pos;
       struct inode *vdir = filp->f_path.dentry->d_inode;
       ntfs_inode *dir_ni;
       int errcode = 0;
       ntfs3g_dir_private_info *info =  filp->private_data;
       int res = 0;
       MFT_RECORD *dir_mrec;
       int rc;
      


	ntfs_log_trace("Entering.\n");
	
	if (!vdir || !filldir) {
		return -EINVAL;
	}
    
       dir_ni = NTFS_I(vdir);

       if(dir_ni->mft_no == 28&& pos == 6808)
        {
           printk("trace here\n");
        }
       dir_mrec = map_mft_record(dir_ni);
       if(!dir_mrec)
       {
           return -EIO;
       }
	if (!(dir_mrec->flags & MFT_RECORD_IS_DIRECTORY)) {
              unmap_mft_record(dir_ni);
		return -ENOTDIR;
	}
       unmap_mft_record(dir_ni);
      
	vol = dir_ni->vol;

	ntfs_log_trace("Entering for inode %lld, *pos 0x%llx.\n",
			(unsigned long long)dir_ni->mft_no, (long long)pos);

	/* Open the index allocation attribute. */

	NTFS3G_BUG_TRACE(!dir_ni->map_na || dir_ni->map_na->type != AT_INDEX_ROOT);

       if(!info)
       {
              info = ntfs_malloc_nofs(&errcode, sizeof(ntfs3g_dir_private_info));
              if(!info)
                goto err_out;
              memset(info, 0, sizeof(*info));
              filp->private_data = info;
       }
	rc = 0;

	/* Emulate . and .. for all directories. */
	if (!pos) {
		rc = ntfs_fuse_filler(dirent,  filldir, dotdot, 1, FILE_NAME_POSIX, pos,
				MK_MREF(dir_ni->mft_no,
				le16_to_cpu(dir_mrec->sequence_number)),
				NTFS_DT_DIR);
		if (rc)
			goto done;
		++pos;
	}
	if (pos == 1) {
		MFT_REF parent_mref;
              errcode = 0;
		parent_mref = ntfs_mft_get_parent_ref(&errcode, dir_ni);
		if (parent_mref == ERR_MREF(-1)) {
			ntfs_log_perror("Parent directory not found");
			goto dir_err_out;
		}

		rc =ntfs_fuse_filler(dirent,  filldir, dotdot, 2, FILE_NAME_POSIX, pos,
				parent_mref, NTFS_DT_DIR);
		if (rc)
			goto done;
		++pos;
	}
       res = ntfs_browser_dir_ir(&errcode, dir_ni,  dirent, info->pre_uname,
		               &info->pre_uname_len, info->pre_uname_len, &pos, filldir);
       if(res < 0)
          goto err_out;

done:
	ntfs_log_debug("EOD, *pos 0x%llx, returning 0.\n", (long long)pos);
       filp->f_pos = pos;
	return 0;
dir_err_out:
	errcode = EIO;
err_out:
       if(!errcode)
           errcode = EIO;
       filp->f_pos = pos;
	return -errcode;
}
#else
/**
 * ntfs_readdir - read the contents of an ntfs directory
 * @dir_ni:	ntfs inode of current directory
 * @pos:	current position in directory
 * @dirent:	context for filldir callback supplied by the caller
 * @filldir:	filldir callback supplied by the caller
 *
 * Parse the index root and the index blocks that are marked in use in the
 * index bitmap and hand each found directory entry to the @filldir callback
 * supplied by the caller.
 *
 * Return 0 on success or -1 on error with errno set to the error code.
 *
 * Note: Index blocks are parsed in ascending vcn order, from which follows
 * that the directory entries are not returned sorted.
 */
static int ntfs3g_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
	s64 i_size, br, ia_pos, bmp_pos, ia_start;
	ntfs_volume *vol;
	ntfs_inode *bmp_ni = NULL, *ia_ni=NULL;
	ntfs_attr_search_ctx *ctx = NULL;
	u8 *index_end, *bmp = NULL;
	INDEX_ROOT *ir;
	INDEX_ENTRY *ie;
	INDEX_ALLOCATION *ia = NULL;
	int rc, ir_pos, bmp_buf_size, bmp_buf_pos, eo;
	u32 index_block_size, index_vcn_size;
	u8 index_block_size_bits, index_vcn_size_bits;
       loff_t pos = filp->f_pos;
       struct inode *vdir = filp->f_path.dentry->d_inode;
       ntfs_inode *dir_ni;
       int errcode = 0;
       MFT_RECORD *dir_mrec;


	ntfs_log_trace("Entering.\n");
	
	if (!vdir || !filldir) {
		return -EINVAL;
	}
    
       dir_ni = NTFS_I(vdir);

       dir_mrec = map_mft_record(dir_ni);
       if(!dir_mrec)
       {
           return -EIO;
       }
	if (!(dir_mrec->flags & MFT_RECORD_IS_DIRECTORY)) {
              unmap_mft_record(dir_ni);
		return -ENOTDIR;
	}
       unmap_mft_record(dir_ni);
      
	vol = dir_ni->vol;

	ntfs_log_trace("Entering for inode %lld, *pos 0x%llx.\n",
			(unsigned long long)dir_ni->mft_no, (long long)pos);

	/* Open the index allocation attribute. */

	NTFS3G_BUG_TRACE(!dir_ni->map_na || dir_ni->map_na->type != AT_INDEX_ROOT);

       
	rc = 0;

	/* Emulate . and .. for all directories. */
	if (!pos) {
		rc = ntfs_fuse_filler(dirent,  filldir, dotdot, 1, FILE_NAME_POSIX, pos,
				MK_MREF(dir_ni->mft_no,
				le16_to_cpu(dir_mrec->sequence_number)),
				NTFS_DT_DIR);
		if (rc)
			goto done;
		++pos;
	}
	if (pos == 1) {
		MFT_REF parent_mref;
              errcode = 0;
		parent_mref = ntfs_mft_get_parent_ref(&errcode, dir_ni);
		if (parent_mref == ERR_MREF(-1)) {
			ntfs_log_perror("Parent directory not found");
			goto dir_err_out;
		}

		rc =ntfs_fuse_filler(dirent,  filldir, dotdot, 2, FILE_NAME_POSIX, pos,
				parent_mref, NTFS_DT_DIR);
		if (rc)
			goto done;
		++pos;
	}

	ctx = ntfs_attr_get_search_ctx(&errcode, dir_ni, NULL);
	if (!ctx)
		goto err_out;

	/* Get the offset into the index root attribute. */
	ir_pos = (int)pos;
	/* Find the index root attribute in the mft record. */
       errcode  = 0;
	if (ntfs_attr_lookup(&errcode, AT_INDEX_ROOT, NTFS_INDEX_I30, 4, CASE_SENSITIVE, 0, NULL,
			0, ctx)) {
		ntfs_log_perror("Index root attribute missing in directory inode "
				"%lld", (unsigned long long)dir_ni->mft_no);
		goto dir_err_out;
	}
	/* Get to the index root value. */
	ir = (INDEX_ROOT*)((u8*)ctx->attr +
			le16_to_cpu(ctx->attr->value_offset));

	/* Determine the size of a vcn in the directory index. */
	index_block_size = le32_to_cpu(ir->index_block_size);
	if (index_block_size < NTFS_BLOCK_SIZE ||
			index_block_size & (index_block_size - 1)) {
		ntfs_log_error(vol->sb, "Index block size %u is invalid.\n",
				(unsigned)index_block_size);
		goto dir_err_out;
	}
	index_block_size_bits = ffs(index_block_size) - 1;
	if (vol->cluster_size <= index_block_size) {
		index_vcn_size = vol->cluster_size;
		index_vcn_size_bits = vol->cluster_size_bits;
	} else {
		index_vcn_size = vol->sector_size;
		index_vcn_size_bits = vol->sector_size_bits;
	}

	/* Are we jumping straight into the index allocation attribute? */
	if (pos >= vol->mft_record_size) {
		ntfs_attr_put_search_ctx(&errcode, ctx);
		ctx = NULL;
		goto skip_index_root;
	}

	index_end = (u8*)&ir->index + le32_to_cpu(ir->index.index_length);
	/* The first index entry. */
	ie = (INDEX_ENTRY*)((u8*)&ir->index +
			le32_to_cpu(ir->index.entries_offset));
	/*
	 * Loop until we exceed valid memory (corruption case) or until we
	 * reach the last entry or until filldir tells us it has had enough
	 * or signals an error (both covered by the rc test).
	 */
	for (;; ie = (INDEX_ENTRY*)((u8*)ie + le16_to_cpu(ie->length))) {
		ntfs_log_debug("In index root, offset 0x%x.\n", (u8*)ie - (u8*)ir);
		/* Bounds checks. */
		if ((u8*)ie < (u8*)ctx->mrec || (u8*)ie +
				sizeof(INDEX_ENTRY_HEADER) > index_end ||
				(u8*)ie + le16_to_cpu(ie->key_length) >
				index_end)
			goto dir_err_out;
		/* The last entry cannot contain a name. */
		if (ie->ie_flags & INDEX_ENTRY_END)
			break;
		
		if (!le16_to_cpu(ie->length))
			goto dir_err_out;
		
		/* Skip index root entry if continuing previous readdir. */
		if (ir_pos > (u8*)ie - (u8*)ir)
			continue;
		/*
		 * Submit the directory entry to ntfs_filldir(), which will
		 * invoke the filldir() callback as appropriate.
		 */
		rc = ntfs_filldir(dir_ni, &pos, index_vcn_size_bits,
				INDEX_TYPE_ROOT, ir, ie, dirent, filldir);
		if (rc) {
			ntfs_attr_put_search_ctx(&errcode, ctx);
			ctx = NULL;
			goto done;
		}
	}
	ntfs_attr_put_search_ctx(&errcode, ctx);
	ctx = NULL;

	/* Advance *pos to the beginning of the index allocation. */
	pos = vol->mft_record_size;

skip_index_root:


	/* Allocate a buffer for the current index block. */
      errcode = 0;
	ia = ntfs_malloc_nofs(&errcode, index_block_size);
	if (!ia)
		goto err_out;
    
       errcode = 0;
       
       ia_ni = ntfs_attr_inode_open(&errcode, dir_ni,  AT_INDEX_ALLOCATION, NTFS_INDEX_I30, 4);
	if (!ia_ni) {
		ntfs_log_perror("Failed to open index allocation attribute");
              if(errcode != ENOENT)
		     goto dir_err_out;
              goto done;
	}
       i_size = ia_ni->map_na->data_size;
	if (pos >= i_size + vol->mft_record_size)
		goto done;    
       NTFS3G_BUG_TRACE(!ia_ni->map_na);
       
       bmp_ni = ntfs_attr_inode_open(&errcode, dir_ni,  AT_BITMAP, NTFS_INDEX_I30, 4);
	if (!bmp_ni) {
		ntfs_log_perror("Failed to open index bitmap attribute");
		goto dir_err_out;
	}
       NTFS3G_BUG_TRACE(!bmp_ni->map_na);
       
	/* Get the offset into the index allocation attribute. */
	ia_pos = pos - vol->mft_record_size;

	bmp_pos = ia_pos >> index_block_size_bits;
	if (bmp_pos >> 3 >= bmp_ni->map_na->data_size) {
		ntfs_log_error(vol->sb, "Current index position exceeds index bitmap "
				"size.\n");
		goto dir_err_out;
	}

	bmp_buf_size = min(bmp_ni->map_na->data_size - (bmp_pos >> 3), (s64)4096);
	bmp = ntfs_malloc_nofs(&errcode, bmp_buf_size);
	if (!bmp)
		goto err_out;

	br = ntfs_inode_map_read(&errcode, VFS_I(bmp_ni), bmp_pos >> 3, bmp_buf_size, (u8*)bmp);
	if (br != bmp_buf_size) {
		if (br != -1)
			errcode = EIO;
		ntfs_log_perror("Failed to read from index bitmap attribute");
		goto err_out;
	}

	bmp_buf_pos = bmp_pos&7;
	/* If the index block is not in use find the next one that is. */
	while (!(bmp[bmp_buf_pos >> 3] & (1 << (bmp_buf_pos & 7)))) {
find_next_index_buffer:
		bmp_pos++;
		bmp_buf_pos++;
		/* If we have reached the end of the bitmap, we are done. */
		if (bmp_pos >> 3 >= bmp_ni->map_na->data_size)
			goto EOD;
		ia_pos = bmp_pos << index_block_size_bits;
		if (bmp_buf_pos >> 3 < bmp_buf_size)
			continue;
		/* Read next chunk from the index bitmap. */
		bmp_buf_pos = 0;
              NTFS3G_BUG_TRACE(bmp_pos&7);
		if ((bmp_pos >> 3) + bmp_buf_size > bmp_ni->map_na->data_size)
			bmp_buf_size = bmp_ni->map_na->data_size - (bmp_pos >> 3);
              br = ntfs_inode_map_read(&errcode, VFS_I(bmp_ni), bmp_pos >> 3, bmp_buf_size,  (u8*)bmp);
		//br = ntfs_attr_pread(&errcode, bmp_ni->map_na, bmp_pos >> 3, bmp_buf_size, bmp);
		if (br != bmp_buf_size) {
			if (br != -1)
				errcode = EIO;
			ntfs_log_perror("Failed to read from index bitmap attribute");
			goto err_out;
		}
	}

	ntfs_log_debug("Handling index block 0x%llx.\n", (long long)bmp_pos);

	/* Read the index block starting at bmp_pos. */
	br = ntfs_inode_map_read(&errcode, VFS_I(ia_ni), bmp_pos << index_block_size_bits, 
			index_block_size, (u8*)ia);
	if (br != index_block_size) {
		if (br != -1)
			errcode = EIO;
		ntfs_log_perror("Failed to read index block");
		goto err_out;
	}

	ia_start = ia_pos & ~(s64)(index_block_size - 1);
	if (sle64_to_cpu(ia->index_block_vcn) != ia_start >>
			index_vcn_size_bits) {
		ntfs_log_error(vol->sb, "Actual VCN (0x%llx) of index buffer is different "
				"from expected VCN (0x%llx) in inode 0x%llx.\n",
				(long long)sle64_to_cpu(ia->index_block_vcn),
				(long long)ia_start >> index_vcn_size_bits,
				(unsigned long long)dir_ni->mft_no);
		goto dir_err_out;
	}
	if (le32_to_cpu(ia->index.allocated_size) + 0x18 != index_block_size) {
		ntfs_log_error(vol->sb, "Index buffer (VCN 0x%llx) of directory inode %lld "
				"has a size (%u) differing from the directory "
				"specified size (%u).\n", (long long)ia_start >>
				index_vcn_size_bits,
				(unsigned long long)dir_ni->mft_no,
				(unsigned) le32_to_cpu(ia->index.allocated_size)
				+ 0x18, (unsigned)index_block_size);
		goto dir_err_out;
	}
	index_end = (u8*)&ia->index + le32_to_cpu(ia->index.index_length);
	if (index_end > (u8*)ia + index_block_size) {
		ntfs_log_error(vol->sb, "Size of index buffer (VCN 0x%llx) of directory inode "
				"%lld exceeds maximum size.\n",
				(long long)ia_start >> index_vcn_size_bits,
				(unsigned long long)dir_ni->mft_no);
		goto dir_err_out;
	}
	/* The first index entry. */
	ie = (INDEX_ENTRY*)((u8*)&ia->index +
			le32_to_cpu(ia->index.entries_offset));
	/*
	 * Loop until we exceed valid memory (corruption case) or until we
	 * reach the last entry or until ntfs_filldir tells us it has had
	 * enough or signals an error (both covered by the rc test).
	 */
	for (;; ie = (INDEX_ENTRY*)((u8*)ie + le16_to_cpu(ie->length))) {
		ntfs_log_debug("In index allocation, offset 0x%llx.\n",
				(long long)ia_start + ((u8*)ie - (u8*)ia));
		/* Bounds checks. */
		if ((u8*)ie < (u8*)ia || (u8*)ie +
				sizeof(INDEX_ENTRY_HEADER) > index_end ||
				(u8*)ie + le16_to_cpu(ie->key_length) >
				index_end) {
			ntfs_log_error(vol->sb, "Index entry out of bounds in directory inode "
				"%lld.\n", (unsigned long long)dir_ni->mft_no);
			goto dir_err_out;
		}
		/* The last entry cannot contain a name. */
		if (ie->ie_flags & INDEX_ENTRY_END)
			break;
		
		if (!le16_to_cpu(ie->length))
			goto dir_err_out;
		
		/* Skip index entry if continuing previous readdir. */
		if (ia_pos - ia_start > (u8*)ie - (u8*)ia)
			continue;
		/*
		 * Submit the directory entry to ntfs_filldir(), which will
		 * invoke the filldir() callback as appropriate.
		 */
		rc = ntfs_filldir(dir_ni, &pos, index_vcn_size_bits,
				INDEX_TYPE_ALLOCATION, ia, ie, dirent, filldir);
		if (rc)
			goto done;
	}
	goto find_next_index_buffer;
EOD:
	/* We are finished, set *pos to EOD. */
	pos = i_size + vol->mft_record_size;
done:
	ntfs_free(ia);
	ntfs_free(bmp);
	if (bmp_ni)
		ntfs_attr_inode_close(&errcode, bmp_ni);
	if (ia_ni)
		ntfs_attr_inode_close(&errcode, ia_ni);
	ntfs_log_debug("EOD, *pos 0x%llx, returning 0.\n", (long long)pos);
       filp->f_pos = pos;
	return 0;
dir_err_out:
	errcode = EIO;
err_out:
       pos = i_size + vol->mft_record_size;
       filp->f_pos = pos;
       
	eo = errcode;
       errcode = 0;
	ntfs_log_trace("failed.\n");
	if (ctx)
		ntfs_attr_put_search_ctx(&errcode, ctx);
	ntfs_free(ia);
	ntfs_free(bmp);
	if (bmp_ni)
	{
	       errcode = 0;
		ntfs_attr_inode_close(&errcode, bmp_ni);
	}
 	if (ia_ni)
	{
	       errcode = 0;
		ntfs_attr_inode_close(&errcode, ia_ni);
	}
	errcode = eo;
	return -1;
}
#endif

/**
 * __ntfs_create - create object on ntfs volume
 * @dir_ni:	ntfs inode for directory in which create new object
 * @name:	unicode name of new object
 * @name_len:	length of the name in unicode characters
 * @type:	type of the object to create
 * @dev:	major and minor device numbers (obtained from makedev())
 * @target:	target in unicode (only for symlinks)
 * @target_len:	length of target in unicode characters
 *
 * Internal, use ntfs_create{,_device,_symlink} wrappers instead.
 *
 * @type can be:
 *	S_IFREG		to create regular file
 *	S_IFDIR		to create directory
 *	S_IFBLK		to create block device
 *	S_IFCHR		to create character device
 *	S_IFLNK		to create symbolic link
 *	S_IFIFO		to create FIFO
 *	S_IFSOCK	to create socket
 * other values are invalid.
 *
 * @dev is used only if @type is S_IFBLK or S_IFCHR, in other cases its value
 * ignored.
 *
 * @target and @target_len are used only if @type is S_IFLNK, in other cases
 * their value ignored.
 *
 * Return opened ntfs inode that describes created object on success or NULL
 * on error with errno set to the error code.
 */
extern const struct inode_operations ntfs3g_file_inode_ops;
extern const struct file_operations ntfs3g_file_ops;
extern const struct address_space_operations ntfs3g_mst_aops;
extern const struct address_space_operations ntfs3g_aops;
extern const struct inode_operations ntfs3g_dir_inode_ops;
extern const struct file_operations ntfs3g_dir_ops;

static ntfs_inode *__ntfs_create(int *p_errno, ntfs_inode *dir_ni,
		ntfschar *name, u8 name_len, dev_t type, dev_t dev,
		ntfschar *target, int target_len)
{
	ntfs_inode *ni;
	int rollback_data = 0, rollback_sd = 0;
	FILE_NAME_ATTR *fn = NULL;
	STANDARD_INFORMATION *si = NULL;
	int err=0, fn_len, si_len;
       struct inode *tmp_ino;
       MFT_RECORD *ni_mrec;

	ntfs_log_trace("Entering.\n");
	
	/* Sanity checks. */
	if (!dir_ni || !name || !name_len) {
		ntfs_log_error(dir_ni->vol->sb, "Invalid arguments.\n");
		*p_errno = EINVAL;
		return NULL;
	}

       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(dir_ni));
	
	if (dir_ni->base_ni_attr.flags & FILE_ATTR_REPARSE_POINT) {
		*p_errno = EOPNOTSUPP;
		return NULL;
	}
	tmp_ino = new_inode(dir_ni->vol->sb);
	if (!tmp_ino) {
			ntfs_log_error(dir_ni->vol->sb, "Failed to load essential metadata.");
		*p_errno =  ENOMEM;
                return NULL;
	}

    
	/* Manually setup an ntfs_inode. */
	ni = NTFS_I(tmp_ino);

       memset(ni, 0, sizeof(*ni));
       ni->base_ni_attr.mft_page = NULL;
       ni->base_ni_attr.mft_page_refcnt = 0;
       ni->vol = dir_ni->vol;
       mutex_init(&ni->base_ni_attr.mrec_lock);
        tmp_ino->i_mode = S_IRWXUGO;
       
	if(ntfs_mft_record_alloc(p_errno, dir_ni->vol, ni, NULL))
	{
	    iput(tmp_ino);
           return NULL;
	}
       else
       {
    	     tmp_ino->i_ino = MREF_INO(ni->mft_no);
	     insert_inode_hash(tmp_ino);
       }
	/*
	 * Create STANDARD_INFORMATION attribute. Write STANDARD_INFORMATION
	 * version 1.2, windows will upgrade it to version 3 if needed.
	 */
	si_len = offsetof(STANDARD_INFORMATION, v1_end);
	si = ntfs_malloc_nofs(p_errno, si_len);
	if (!si) {
		err = *p_errno;
		goto err_out;
	}
	si->creation_time = utc2ntfs(ni->base_ni_attr.creation_time);
	si->last_data_change_time = utc2ntfs(ni->base_ni_attr.last_data_change_time);
	si->last_mft_change_time = utc2ntfs(ni->base_ni_attr.last_mft_change_time);
	si->last_access_time = utc2ntfs(ni->base_ni_attr.last_access_time);
	if (!S_ISREG(type) && !S_ISDIR(type)) {
		si->file_attributes = FILE_ATTR_SYSTEM;
		ni->base_ni_attr.flags = FILE_ATTR_SYSTEM;
	}
	/* Add STANDARD_INFORMATION to inode. */
	if (ntfs_attr_add(p_errno, ni, AT_STANDARD_INFORMATION, AT_UNNAMED, 0,
			(u8*)si, si_len)) {
		err = *p_errno;
		ntfs_log_error(dir_ni->vol->sb, "Failed to add STANDARD_INFORMATION "
				"attribute.\n");
		goto err_out;
	}
	
	if (ntfs_sd_add_everyone(p_errno, ni)) {
		err = *p_errno;
		goto err_out;
	}
	rollback_sd = 1;

	if (S_ISDIR(type)) {
		INDEX_ROOT *ir = NULL;
		INDEX_ENTRY *ie;
		int ir_len, index_len;

		/* Create INDEX_ROOT attribute. */
		index_len = sizeof(INDEX_HEADER) + sizeof(INDEX_ENTRY_HEADER);
		ir_len = offsetof(INDEX_ROOT, index) + index_len;
		ir = ntfs_malloc_nofs(p_errno, ir_len);
		if (!ir) {
			err = *p_errno;
			goto err_out;
		}
		ir->type = AT_FILE_NAME;
		ir->collation_rule = COLLATION_FILE_NAME;
		ir->index_block_size = cpu_to_le32(ni->vol->indx_record_size);
		if (ni->vol->cluster_size <= ni->vol->indx_record_size)
			ir->clusters_per_index_block =
					ni->vol->indx_record_size >>
					ni->vol->cluster_size_bits;
		else
			ir->clusters_per_index_block = 
					ni->vol->indx_record_size >>
					ni->vol->sector_size_bits;
		ir->index.entries_offset = cpu_to_le32(sizeof(INDEX_HEADER));
		ir->index.index_length = cpu_to_le32(index_len);
		ir->index.allocated_size = cpu_to_le32(index_len);
		ie = (INDEX_ENTRY*)((u8*)ir + sizeof(INDEX_ROOT));
		ie->length = cpu_to_le16(sizeof(INDEX_ENTRY_HEADER));
		ie->key_length = 0;
		ie->ie_flags = INDEX_ENTRY_END;
        
              tmp_ino->i_size = 0;
              tmp_ino->i_ino = MREF_LE(ni->mft_no);
	       tmp_ino->i_version = 1;
              tmp_ino->i_mode |= S_IFDIR;

		/* Add INDEX_ROOT attribute to inode. */
		if (ntfs_attr_add(p_errno, ni, AT_INDEX_ROOT, NTFS_INDEX_I30, 4,
				(u8*)ir, ir_len)) {
			err = *p_errno;
			ntfs_free(ir);
			ntfs_log_error(dir_ni->vol->sb, "Failed to add INDEX_ROOT attribute.\n");
			goto err_out;
		}
		ntfs_free(ir);
	} else {
		INTX_FILE *data;
		int data_len;

		switch (type) {
                     case 0:
                     case S_IFREG:
                        data = NULL;
			   data_len = 0;
                        break;
                     default:
                          BUG();
                       /*
			case S_IFBLK:
			case S_IFCHR:
				data_len = offsetof(INTX_FILE, device_end);
				data = ntfs_malloc_nofs(p_errno, data_len);
				if (!data) {
					err = *p_errno;
					goto err_out;
				}
				data->major = cpu_to_le64(MAJOR(dev));
				data->minor = cpu_to_le64(MINOR(dev));
				if (type == S_IFBLK)
					data->magic = INTX_BLOCK_DEVICE;
				if (type == S_IFCHR)
					data->magic = INTX_CHARACTER_DEVICE;
				break;
			case S_IFLNK:
				data_len = sizeof(INTX_FILE_TYPES) +
						target_len * sizeof(ntfschar);
				data = ntfs_malloc_nofs(p_errno, data_len);
				if (!data) {
					err = *p_errno;
					goto err_out;
				}
				data->magic = INTX_SYMBOLIC_LINK;
				memcpy(data->target, target,
						target_len * sizeof(ntfschar));
				break;
			case S_IFSOCK:
				data = NULL;
				data_len = 1;
				break;
                     */
		}
		/* Add DATA attribute to inode. */
		if (ntfs_attr_add(p_errno, ni, AT_DATA, AT_UNNAMED, 0, (u8*)data,
				data_len)) {
			err = *p_errno;
			ntfs_log_error(dir_ni->vol->sb, "Failed to add DATA attribute.\n");
			ntfs_free(data);
			goto err_out;
		}
		rollback_data = 1;
		ntfs_free(data);
 
              tmp_ino->i_size = 0;
              tmp_ino->i_ino = MREF_LE(ni->mft_no);
	       tmp_ino->i_version = 1;
              tmp_ino->i_mode |= S_IFREG;
	}
	/* Create FILE_NAME attribute. */
	fn_len = sizeof(FILE_NAME_ATTR) + name_len * sizeof(ntfschar);
	fn = ntfs_malloc_nofs(p_errno, fn_len);
	if (!fn) {
		err = *p_errno;
		goto err_out;
	}

       ni_mrec = map_mft_record(dir_ni);
       if(!ni_mrec)
       {
            *p_errno = err = EIO;
            ntfs_log_error(dir_ni->vol->sb, "failed to map mft record for file %lld.\n", dir_ni->mft_no);
            goto err_out;
       }
	fn->parent_directory = MK_LE_MREF(dir_ni->mft_no,
			le16_to_cpu(ni_mrec->sequence_number));
       unmap_mft_record(dir_ni);
	fn->file_name_length = name_len;
	fn->file_name_type = FILE_NAME_POSIX;
	if (S_ISDIR(type))
		fn->file_attributes = FILE_ATTR_I30_INDEX_PRESENT;
	if (!S_ISREG(type) && !S_ISDIR(type))
		fn->file_attributes = FILE_ATTR_SYSTEM;
	fn->creation_time = utc2ntfs(ni->base_ni_attr.creation_time);
	fn->last_data_change_time = utc2ntfs(ni->base_ni_attr.last_data_change_time);
	fn->last_mft_change_time = utc2ntfs(ni->base_ni_attr.last_mft_change_time);
	fn->last_access_time = utc2ntfs(ni->base_ni_attr.last_access_time);
	fn->data_size = cpu_to_sle64(ni->map_na?ni->map_na->data_size:0);
	fn->allocated_size = cpu_to_sle64(ni->map_na?ni->map_na->allocated_size:0);
	memcpy(fn->file_name, name, name_len * sizeof(ntfschar));
	/* Add FILE_NAME attribute to inode. */
	if (ntfs_attr_add(p_errno, ni, AT_FILE_NAME, AT_UNNAMED, 0, (u8*)fn, fn_len)) {
		err = *p_errno;
		ntfs_log_error(dir_ni->vol->sb, "Failed to add FILE_NAME attribute.\n");
		goto err_out;
	}
       ni_mrec = map_mft_record(ni);
       if(!ni_mrec)
       {
            *p_errno = err = EIO;
            ntfs_log_error(dir_ni->vol->sb, "failed to map mft record for file %lld.\n", ni->mft_no);
            goto err_out;
       }
	/* Add FILE_NAME attribute to index. */
	if (ntfs_index_add_filename(p_errno, dir_ni, fn, MK_MREF(ni->mft_no,
			le16_to_cpu(ni_mrec->sequence_number)))) {
		err = *p_errno;
              unmap_mft_record(ni);
		ntfs_log_error(dir_ni->vol->sb, "Failed to add entry to the index");
		goto err_out;
	}

	/* Set hard links count and directory flag. */
	ni_mrec->link_count = cpu_to_le16(1);
       tmp_ino->i_nlink = 1;
	if (S_ISDIR(type))
       {
		ni_mrec->flags |= MFT_RECORD_IS_DIRECTORY;
               ni->map_na = ntfs_attr_open(p_errno, ni, AT_INDEX_ROOT, NTFS_INDEX_I30, 4);
               tmp_ino->i_op = &ntfs3g_dir_inode_ops;
	        tmp_ino->i_fop = &ntfs3g_dir_ops;
               tmp_ino->i_mapping->a_ops = &ntfs3g_aops;
 
	}
       else
        {
              ni->map_na = ntfs_attr_open(p_errno, ni, AT_DATA, AT_UNNAMED, 0);	
               tmp_ino->i_op = &ntfs3g_file_inode_ops;
		 tmp_ino->i_fop = &ntfs3g_file_ops;
               tmp_ino->i_mapping->a_ops = &ntfs3g_aops;
        }

       ntfs_inode_mark_dirty(p_errno, ni);
       mark_mft_record_dirty(ni);
       unmap_mft_record(ni);

	/* Done! */
	ntfs_free(fn);
	ntfs_free(si);
	ntfs_log_trace("Done.\n");

	return ni;
err_out:
	ntfs_log_error(dir_ni->vol->sb, "Failed %d.\n", *p_errno);

	if (rollback_sd)
		ntfs_attr_remove(p_errno, ni, AT_SECURITY_DESCRIPTOR, AT_UNNAMED, 0);
	
	if (rollback_data)
		ntfs_attr_remove(p_errno, ni, AT_DATA, AT_UNNAMED, 0);
	/*
	 * Free extent MFT records (should not exist any with current
	 * ntfs_create implementation, but for any case if something will be
	 * changed in the future).
	 */
	while (ni->base_ni_attr.nr_extents)
		if (ntfs_mft_record_free(p_errno, ni->vol, *(ni->base_ni_attr.extent_nis))) {
			err = *p_errno;
			ntfs_log_error(dir_ni->vol->sb, "Failed to free extent MFT record.  "
					"Leaving inconsistent metadata.\n");
		}
	if (ntfs_mft_record_free(p_errno, ni->vol, ni))
		ntfs_log_error(dir_ni->vol->sb, "Failed to free MFT record.  "
				"Leaving inconsistent metadata. Run chkdsk.\n");
	ntfs_free(fn);
	ntfs_free(si);
	*p_errno = err;
	return NULL;
}

/**
 * Some wrappers around __ntfs_create() ...
 */

ntfs_inode *ntfs_create(int *p_errno, ntfs_inode *dir_ni, ntfschar *name, u8 name_len,
		dev_t type)
{
	if (type != S_IFREG && type != S_IFDIR && type != S_IFIFO &&
			type != S_IFSOCK) {
		ntfs_log_error(dir_ni->vol->sb, "Invalid arguments.\n");
		return NULL;
	}
	return __ntfs_create(p_errno, dir_ni, name, name_len, type, 0, NULL, 0);
}

ntfs_inode *ntfs_create_device(int *p_errno, ntfs_inode *dir_ni, ntfschar *name, u8 name_len,
		dev_t type, dev_t dev)
{
	if (type != S_IFCHR && type != S_IFBLK) {
		ntfs_log_error(dir_ni->vol->sb, "Invalid arguments.\n");
		return NULL;
	}
	return __ntfs_create(p_errno, dir_ni, name, name_len, type, dev, NULL, 0);
}

ntfs_inode *ntfs_create_symlink(int *p_errno, ntfs_inode *dir_ni, ntfschar *name, u8 name_len,
		ntfschar *target, int target_len)
{
	if (!target || !target_len) {
		ntfs_log_error(dir_ni->vol->sb, "%s: Invalid argument (%p, %d)\n", __FUNCTION__,
			       target, target_len);
		return NULL;
	}
	return __ntfs_create(p_errno, dir_ni, name, name_len, S_IFLNK, 0,
			target, target_len);
}

int ntfs_check_empty_dir(int *p_errno, ntfs_inode *ni)
{
	int ret = 0;
       MFT_RECORD *ni_mrec;

       ni_mrec = map_mft_record(ni);
       if(!ni_mrec)
       {
             *p_errno = EIO;
             return -1;
       }
	
	if (!(ni_mrec->flags & MFT_RECORD_IS_DIRECTORY))
	{
	       unmap_mft_record(ni);
		return 0;
	}
       unmap_mft_record(ni);

       NTFS3G_BUG_TRACE(ni->map_na->type != AT_INDEX_ROOT);

	
	/* Non-empty directory? */
	if ((ni->map_na->data_size != sizeof(INDEX_ROOT) + sizeof(INDEX_ENTRY_HEADER))){
		/* Both ENOTEMPTY and EEXIST are ok. We use the more common. */
		*p_errno = ENOTEMPTY;
		ntfs_log_debug("Directory is not empty\n");
		ret = -1;
	}
	
	return ret;
}

static int ntfs_check_unlinkable_dir(int *p_errno, ntfs_inode *ni, FILE_NAME_ATTR *fn)
{
	int link_count;
	int ret;
       MFT_RECORD *ni_mrec;
	
	ret = ntfs_check_empty_dir(p_errno, ni);
	if (!ret || *p_errno != ENOTEMPTY)
		return ret;
    
       ni_mrec = map_mft_record(ni);
       if(!ni_mrec)
       {
             *p_errno = EIO;
             return -1;
       }
       link_count = le16_to_cpu(ni_mrec->link_count);
       unmap_mft_record(ni);
	/* 
	 * Directory is non-empty, so we can unlink only if there is more than
	 * one "real" hard link, i.e. links aren't different DOS and WIN32 names
	 */
	if ((link_count == 1) || 
	    (link_count == 2 && fn->file_name_type == FILE_NAME_DOS)) {
		*p_errno = ENOTEMPTY;
		ntfs_log_debug("Non-empty directory without hard links\n");
		goto no_hardlink;
	}
	
	ret = 0;
no_hardlink:	
	return ret;
}

/**
 * ntfs_delete - delete file or directory from ntfs volume
 * @ni:		ntfs inode for object to delte
 * @dir_ni:	ntfs inode for directory in which delete object
 * @name:	unicode name of the object to delete
 * @name_len:	length of the name in unicode characters
 *
 * @ni is always closed after the call to this function (even if it failed),
 * user does not need to call ntfs_inode_close himself.
 *
 * Return 0 on success or -1 on error with errno set to the error code.
 */
int ntfs_delete(int *p_errno, ntfs_inode *ni, ntfs_inode *dir_ni, ntfschar *name, u8 name_len)
{
	ntfs_attr_search_ctx *actx = NULL;
	FILE_NAME_ATTR *fn = NULL;
	BOOL looking_for_dos_name = FALSE, looking_for_win32_name = FALSE;
	BOOL case_sensitive_match = TRUE;
	int err = 0;
       MFT_RECORD *ni_mrec;

	ntfs_log_trace("Entering.\n");
	
	if (!ni || !dir_ni || !name || !name_len) {
		ntfs_log_error(dir_ni->vol->sb, "Invalid arguments.\n");
		*p_errno = EINVAL;
		goto err_out;
	}
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
        NTFS3G_BUG_TRACE(NInoNTFSAttrNI(dir_ni));
       
	if (ni->base_ni_attr.nr_extents == -1)
		ni = ni->base_ni_attr.base_ni;
	if (dir_ni->base_ni_attr.nr_extents == -1)
		dir_ni = dir_ni->base_ni_attr.base_ni;
	/*
	 * Search for FILE_NAME attribute with such name. If it's in POSIX or
	 * WIN32_AND_DOS namespace, then simply remove it from index and inode.
	 * If filename in DOS or in WIN32 namespace, then remove DOS name first,
	 * only then remove WIN32 name.
	 */
	actx = ntfs_attr_get_search_ctx(p_errno, ni, NULL);
	if (!actx)
		goto err_out;
search:
	while (!ntfs_attr_lookup(p_errno, AT_FILE_NAME, AT_UNNAMED, 0, CASE_SENSITIVE,
			0, NULL, 0, actx)) {
		char *s;
		BOOL case_sensitive = IGNORE_CASE;

		*p_errno = 0;
		fn = (FILE_NAME_ATTR*)((u8*)actx->attr +
				le16_to_cpu(actx->attr->value_offset));
		s = ntfs_attr_name_get(p_errno, fn->file_name, fn->file_name_length);
		ntfs_log_trace("name: '%s'  type: %d  dos: %d  win32: %d  "
			       "case: %d\n", s, fn->file_name_type,
			       looking_for_dos_name, looking_for_win32_name,
			       case_sensitive_match);
		ntfs_attr_name_free(p_errno, &s);
		if (looking_for_dos_name) {
			if (fn->file_name_type == FILE_NAME_DOS)
				break;
			else
				continue;
		}
		if (looking_for_win32_name) {
			if  (fn->file_name_type == FILE_NAME_WIN32)
				break;
			else
				continue;
		}
		
		/* Ignore hard links from other directories */
		if (dir_ni->mft_no != MREF_LE(fn->parent_directory)) {
			ntfs_log_debug("MFT record numbers don't match "
				       "(%llu != %llu)\n", dir_ni->mft_no, 
				       MREF_LE(fn->parent_directory));
			continue;
		}
		     
		if (fn->file_name_type == FILE_NAME_POSIX || case_sensitive_match)
			case_sensitive = CASE_SENSITIVE;
		
		if (ntfs_names_are_equal(fn->file_name, fn->file_name_length,
					 name, name_len, case_sensitive, 
					 ni->vol->upcase, ni->vol->upcase_len)){
			
			if (fn->file_name_type == FILE_NAME_WIN32) {
				looking_for_dos_name = TRUE;
				ntfs_attr_reinit_search_ctx(p_errno, actx);
				continue;
			}
			if (fn->file_name_type == FILE_NAME_DOS)
				looking_for_dos_name = TRUE;
			break;
		}
	}
	if (*p_errno) {
		/*
		 * If case sensitive search failed, then try once again
		 * ignoring case.
		 */
		if (*p_errno == ENOENT && case_sensitive_match) {
			case_sensitive_match = FALSE;
			ntfs_attr_reinit_search_ctx(p_errno, actx);
			goto search;
		}
		goto err_out;
	}
	
	if (ntfs_check_unlinkable_dir(p_errno, ni, fn) < 0)
		goto err_out;
		
	if (ntfs_index_remove(p_errno, dir_ni, fn, le32_to_cpu(actx->attr->value_length)))
		goto err_out;
	
	if (ntfs_attr_record_rm(p_errno, actx))
		goto err_out;

       ni_mrec = map_mft_record(ni);
       if(!ni_mrec)
        {
                *p_errno = EIO;
                goto err_out;
        }
	ni_mrec->link_count = cpu_to_le16(le16_to_cpu(
			ni_mrec->link_count) - 1);
    
	ntfs_inode_mark_dirty(p_errno, ni);
       mark_mft_record_dirty(ni);
	if (looking_for_dos_name) {
		looking_for_dos_name = FALSE;
		looking_for_win32_name = TRUE;
		ntfs_attr_reinit_search_ctx(p_errno, actx);
              unmap_mft_record(ni);
		goto search;
	}
	/* TODO: Update object id, quota and securiry indexes if required. */
	/*
	 * If hard link count is not equal to zero then we are done. In other
	 * case there are no reference to this inode left, so we should free all
	 * non-resident attributes and mark all MFT record as not in use.
	 */
	if (ni_mrec->link_count) {
		ntfs_inode_update_times(p_errno, ni, NTFS_UPDATE_CTIME);
               unmap_mft_record(ni);
		goto ok;
	}
       unmap_mft_record(ni);
       
	ntfs_attr_reinit_search_ctx(p_errno, actx);
	while (!ntfs_attrs_walk(p_errno, actx)) {
		if (actx->attr->non_resident) {
			runlist *rl;

			rl = ntfs_mapping_pairs_decompress(p_errno, ni->vol, actx->attr,
					NULL);
			if (!rl) {
				err = *p_errno;
				ntfs_log_error(dir_ni->vol->sb, "Failed to decompress runlist.  "
						"Leaving inconsistent metadata.\n");
				continue;
			}
			if (ntfs_cluster_free_from_rl(p_errno, ni->vol, rl)) {
				err = *p_errno;
				ntfs_log_error(dir_ni->vol->sb, "Failed to free clusters.  "
						"Leaving inconsistent metadata.\n");
				continue;
			}
			ntfs_free(rl);
		}
	}
	if (*p_errno != ENOENT) {
		err = *p_errno;
		ntfs_log_error(dir_ni->vol->sb, "Attribute enumeration failed.  "
				"Probably leaving inconsistent metadata.\n");
	}
	/* All extents should be attached after attribute walk. */
	while (ni->base_ni_attr.nr_extents)
		if (ntfs_mft_record_free(p_errno, ni->vol, *(ni->base_ni_attr.extent_nis))) {
			err = *p_errno;
			ntfs_log_error(dir_ni->vol->sb, "Failed to free extent MFT record.  "
					"Leaving inconsistent metadata.\n");
		}
	if (ntfs_mft_record_free(p_errno, ni->vol, ni)) {
		err = *p_errno;
		ntfs_log_error(dir_ni->vol->sb, "Failed to free base MFT record.  "
				"Leaving inconsistent metadata.\n");
	}
	ni = NULL;
ok:	
	ntfs_inode_update_times(p_errno, dir_ni, NTFS_UPDATE_MCTIME);
out:
	if (actx)
		ntfs_attr_put_search_ctx(p_errno, actx);
	if (ntfs_big_inode_close(p_errno, ni) && !err)
		err = *p_errno;
	if (err) {
		*p_errno = err;
		ntfs_log_debug("Could not delete file: %s\n", strerror(errno));
		return -err;
	}
	ntfs_log_trace("Done.\n");
	return 0;
err_out:
	err = *p_errno;
	goto out;
}

/**
 * ntfs_link - create hard link for file or directory
 * @ni:		ntfs inode for object to create hard link
 * @dir_ni:	ntfs inode for directory in which new link should be placed
 * @name:	unicode name of the new link
 * @name_len:	length of the name in unicode characters
 *
 * NOTE: At present we allow creating hardlinks to directories, we use them
 * in a temporary state during rename. But it's defenitely bad idea to have
 * hard links to directories as a result of operation.
 * FIXME: Create internal  __ntfs_link that allows hard links to a directories
 * and external ntfs_link that do not. Write ntfs_rename that uses __ntfs_link.
 *
 * Return 0 on success or -1 on error with errno set to the error code.
 */
int ntfs_link(int *p_errno, ntfs_inode *ni, ntfs_inode *dir_ni, ntfschar *name, u8 name_len)
{
	FILE_NAME_ATTR *fn = NULL;
	int fn_len, err=0;
       MFT_RECORD *dir_mrec=NULL, *ni_mrec=NULL;

	ntfs_log_trace("Entering.\n");
	
	if (!ni || !dir_ni || !name || !name_len || 
			ni->mft_no == dir_ni->mft_no) {
		err = EINVAL;
		ntfs_log_perror("ntfs_link wrong arguments");
		goto err_out;
	}
	NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(dir_ni));
    
	if (ni->base_ni_attr.flags & FILE_ATTR_REPARSE_POINT) {
		err = EOPNOTSUPP;
		goto err_out;
	}
       ni_mrec = map_mft_record(ni);
       if(!ni_mrec)
       {
		err = EIO;
		goto err_out;
	}   
	
	/* Create FILE_NAME attribute. */
	fn_len = sizeof(FILE_NAME_ATTR) + name_len * sizeof(ntfschar);
	fn = ntfs_malloc_nofs(p_errno, fn_len);
	if (!fn) {
		err = *p_errno;
		goto err_out;
	}
       dir_mrec = map_mft_record(dir_ni);
       if(!dir_mrec)
       {
                err = EIO;
                goto err_out;
       }
	fn->parent_directory = MK_LE_MREF(dir_ni->mft_no,
			le16_to_cpu(dir_mrec->sequence_number));
       unmap_mft_record(dir_ni);
	fn->file_name_length = name_len;
	fn->file_name_type = FILE_NAME_POSIX;
	fn->file_attributes = ni->base_ni_attr.flags;
	if (ni_mrec->flags & MFT_RECORD_IS_DIRECTORY)
		fn->file_attributes |= FILE_ATTR_I30_INDEX_PRESENT;
	fn->allocated_size = cpu_to_sle64(ni->map_na?ni->map_na->allocated_size:0);
	fn->data_size = cpu_to_sle64(ni->map_na?ni->map_na->data_size:0);
	fn->creation_time = utc2ntfs(ni->base_ni_attr.creation_time);
	fn->last_data_change_time = utc2ntfs(ni->base_ni_attr.last_data_change_time);
	fn->last_mft_change_time = utc2ntfs(ni->base_ni_attr.last_mft_change_time);
	fn->last_access_time = utc2ntfs(ni->base_ni_attr.last_access_time);
	memcpy(fn->file_name, name, name_len * sizeof(ntfschar));
	/* Add FILE_NAME attribute to index. */
	if (ntfs_index_add_filename(p_errno, dir_ni, fn, MK_MREF(ni->mft_no,
			le16_to_cpu(ni_mrec->sequence_number)))) {
		err = *p_errno;
		ntfs_log_perror("Failed to add filename to the index");
		goto err_out;
	}
	/* Add FILE_NAME attribute to inode. */
	if (ntfs_attr_add(p_errno, ni, AT_FILE_NAME, AT_UNNAMED, 0, (u8*)fn, fn_len)) {
		ntfs_log_error(dir_ni->vol->sb, "Failed to add FILE_NAME attribute.\n");
		err = *p_errno;
		/* Try to remove just added attribute from index. */
		if (ntfs_index_remove(p_errno, dir_ni, fn, fn_len))
			goto rollback_failed;
		goto err_out;
	}
	/* Increment hard links count. */
	ni_mrec->link_count = cpu_to_le16(le16_to_cpu(
			ni_mrec->link_count) + 1);
       
	/* Done! */
	ntfs_inode_mark_dirty(p_errno, ni);
       mark_mft_record_dirty(ni);
	ntfs_free(fn); 
       if(ni_mrec)
             unmap_mft_record(ni);
	ntfs_log_trace("Done.\n");

	return 0;
rollback_failed:
	ntfs_log_error(dir_ni->vol->sb, "Rollback failed. Leaving inconsistent metadata.\n");
err_out:
	ntfs_free(fn);
	*p_errno = err;
       if(ni_mrec)
           unmap_mft_record(ni);
	return -1;
}
static int ntfs3g_dir_fsync(struct file *filp, struct dentry *dentry,
		int datasync)
{
	struct inode *vi = dentry->d_inode;
	int err=0, ret;

	ntfs_log_debug("Entering for inode 0x%lx.", vi->i_ino);
	NTFS3G_BUG_TRACE(!S_ISDIR(vi->i_mode));

       ret = -ntfs_inode_sync(&err, NTFS_I(vi));
	
	write_inode_now(vi, !datasync);
	err = sync_blockdev(vi->i_sb->s_bdev);
	if (unlikely(err && !ret))
		ret = err;
	if (likely(!ret))
		ntfs_log_debug("Done.");
	else
		ntfs_log_warning(vi->i_sb, "Failed to f%ssync inode 0x%lx.  Error "
				"%u.", datasync ? "data" : "", vi->i_ino, -ret);
	return ret;
}
static int ntfs3g_release_dir (struct inode * inode, struct file * filp)
{
       if (filp->private_data)
		ntfs_free(filp->private_data);
	return 0;
}

static int ntfs3g_ioctl_volume_label(struct inode *dir, unsigned long arg)
{
	struct super_block *sb = dir->i_sb;
    struct _ntfs_volume *vol = NTFS3G_SB(sb);//(struct ntfs_volume *)sb->s_fs_info;
    char * vn = vol->vol_name;
    char __user *v = (char __user *)arg;
    
    if(v == NULL)
        return -EFAULT;
    //APP must pay attention to the mem overwrite risk
    copy_to_user(v,vn, NTFS_VOLUME_NAME_LEN );
	return 0;
}


static int ntfs3g_ioctl (struct inode *inode, struct file *filp,
			 unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case NTFS_IOCTL_GET_VOLUME_LABEL :
		return ntfs3g_ioctl_volume_label(inode, arg);
        break;
	default:
		return -EFAULT;
	}

    return 0;
}


const struct file_operations ntfs3g_dir_ops = {
	.llseek		= generic_file_llseek,	/* Seek inside directory. */
	.read		= generic_read_dir,	/* Return -EISDIR. */
	.readdir	= ntfs3g_readdir,		/* Read directory contents. */
	.fsync		= ntfs3g_dir_fsync,	/* Sync a directory to disk. */
	.release   = ntfs3g_release_dir,
	/*.aio_fsync	= ,*/			/* Sync all outstanding async
						   i/o operations on a kiocb. */
	.ioctl	= ntfs3g_ioctl,			/* Perform function on the
						   mounted filesystem. */
	//.open		= ntfs_dir_open,	/* Open directory. */
};
