/*
 * file.c - NTFS kernel file operations.  Part of the Linux-NTFS project.
 *
 * Copyright (c) 2001-2007 Anton Altaparmakov
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/stddef.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>	/* For bdev_hardsect_size(). */
#include <linux/backing-dev.h>
#include <linux/buffer_head.h>
#include <linux/vfs.h>
#include <linux/moduleparam.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/mount.h>

#include "config.h"
#include "compat.h"
#include "attrib.h"
#include "attrlist.h"
#include "mft.h"
#include "debug.h"
#include "mst.h"
#include "volume.h"
#include "types.h"
#include "layout.h"
#include "inode.h"
#include "runlist.h"
#include "lcnalloc.h"
#include "dir.h"
#include "compress.h"
#include "bitmap.h"
#include "misc.h"


#define NTFS_IOCTL_GET_ATTRIBUTES	_IOR('r', 0x10, __u32)
/**
 * ntfs_file_open - called when an inode is about to be opened
 * @vi:		inode to be opened
 * @filp:	file structure describing the inode
 *
 * Limit file size to the page cache limit on architectures where unsigned long
 * is 32-bits. This is the most we can do for now without overflowing the page
 * cache page index. Doing it this way means we don't run into problems because
 * of existing too large files. It would be better to allow the user to read
 * the beginning of the file but I doubt very much anyone is going to hit this
 * check on a 32-bit architecture, so there is no point in adding the extra
 * complexity required to support this.
 *
 * On 64-bit architectures, the check is hopefully optimized away by the
 * compiler.
 *
 * After the check passes, just call generic_file_open() to do its work.
 */
static int ntfs3g_file_open(struct inode *vi, struct file *filp)
{
	if (sizeof(unsigned long) < 8) {
		if (i_size_read(vi) > MAX_LFS_FILESIZE)
			return -EOVERFLOW;
	}
	return generic_file_open(vi, filp);
}



/**
 * ntfs_file_fsync - sync a file to disk
 * @filp:	file to be synced
 * @dentry:	dentry describing the file to sync
 * @datasync:	if non-zero only flush user data and not metadata
 *
 * Data integrity sync of a file to disk.  Used for fsync, fdatasync, and msync
 * system calls.  This function is inspired by fs/buffer.c::file_fsync().
 *
 * If @datasync is false, write the mft record and all associated extent mft
 * records as well as the $DATA attribute and then sync the block device.
 *
 * If @datasync is true and the attribute is non-resident, we skip the writing
 * of the mft record and all associated extent mft records (this might still
 * happen due to the write_inode_now() call).
 *
 * Also, if @datasync is true, we do not wait on the inode to be written out
 * but we always wait on the page cache pages to be written out.
 *
 * Note: In the past @filp could be NULL so we ignore it as we don't need it
 * anyway.
 *
 * Locking: Caller must hold i_mutex on the inode.
 *
 * TODO: We should probably also write all attribute/index inodes associated
 * with this inode but since we have no simple way of getting to them we ignore
 * this problem for now.
 */
static int ntfs3g_file_fsync(struct file *filp, struct dentry *dentry,
		int datasync)
{
	struct inode *vi = dentry->d_inode;
	int ret = 0;
       int err = 0;


       ntfs_inode_sync(&err, NTFS_I(vi));
       ret = -err;     

	write_inode_now(vi, !datasync);
	/*
	 * NOTE: If we were to use mapping->private_list (see ext2 and
	 * fs/buffer.c) for dirty blocks then we could optimize the below to be
	 * sync_mapping_buffers(vi->i_mapping).
	 */
	err = sync_blockdev(vi->i_sb->s_bdev);
	if (unlikely(err && !ret))
		ret = err;
	if (likely(!ret))
		ntfs_log_debug("Done.");
	else
		ntfs_log_warning(vi->i_sb, "Failed to f%ssync inode 0x%lx.  Error "
				"%u.", datasync ? "data" : "", vi->i_ino, -ret);
	return ret;
}

void ntfs3g_truncate_vfs(struct inode *vi) {

       ntfs_inode *ni = NTFS_I(vi);
       int err = 0;
     
       if (!ni->map_na)
       {
             ntfs_log_perror("ntfs_truncate_vfs can not found attribute\n");
	       return;
       }
	ntfs_attr_truncate(&err, ni->map_na, vi->i_size);

}

int ntfs3g_setattr(struct dentry *dentry, struct iattr *attr)
{
	struct inode *vi = dentry->d_inode;
	int err = 0;
	unsigned int ia_valid = attr->ia_valid;
 
       
	err = inode_change_ok(vi, attr);
	if (err)
		goto out;
	/* We do not support NTFS ACLs yet. */
	if (ia_valid & (ATTR_UID | ATTR_GID | ATTR_MODE)) {
		ntfs_log_warning(vi->i_sb, "Changes in user/group/mode are not "
				"supported yet, ignoring.");
		err = -EOPNOTSUPP;
		goto out;
	}
	if (ia_valid & ATTR_SIZE ) {
		if (attr->ia_size != i_size_read(vi)) {
			ntfs_inode *ni = NTFS_I(vi);

                     if(S_ISDIR(vi->i_mode))
                     {
                        	  err = -EOPNOTSUPP;
		                goto out;
                     }
                     if(!ni->map_na)
                     {
                        	  err = -EOPNOTSUPP;
		                goto out;
                     }                       
			/*
			 * FIXME: For now we do not support resizing of
			 * compressed or encrypted files yet.
			 */
			if (NAttrCompressed(ni->map_na) || NAttrEncrypted(ni->map_na)) {
				ntfs_log_warning(vi->i_sb, "Changes in inode size "
						"are not supported yet for "
						"%s files, ignoring.",
						NAttrCompressed(na) ?
						"compressed" : "encrypted");
				err = -EOPNOTSUPP;
			} else
				err = vmtruncate(vi, attr->ia_size);
			if (err || ia_valid == ATTR_SIZE)
				goto out;
		} else {
			/*
			 * We skipped the truncate but must still update
			 * timestamps.
			 */
			ia_valid |= ATTR_MTIME | ATTR_CTIME;
		}
	}
	if (ia_valid & ATTR_ATIME)
		vi->i_atime = timespec_trunc(attr->ia_atime,
				vi->i_sb->s_time_gran);
	if (ia_valid & ATTR_MTIME)
		vi->i_mtime = timespec_trunc(attr->ia_mtime,
				vi->i_sb->s_time_gran);
	if (ia_valid & ATTR_CTIME)
		vi->i_ctime = timespec_trunc(attr->ia_ctime,
				vi->i_sb->s_time_gran);
	mark_inode_dirty(vi);
        err = 0;
out:
	return err;
}
int ntfs3g_getattr(struct vfsmount *mnt, struct dentry *dentry, struct kstat *stat)
{
	struct inode *inode = dentry->d_inode;
	generic_fillattr(inode, stat);

	stat->blksize = NTFS3G_SB(inode->i_sb)->cluster_size;
	return 0;
}


ssize_t ntfs3g_file_aio_write(struct kiocb *iocb, const struct iovec *iov,
		unsigned long nr_segs, loff_t pos)
{
	struct file *file = iocb->ki_filp;
	struct address_space *mapping = file->f_mapping;
	struct inode *vi =mapping->host;
	ntfs_inode *ni = NTFS_I(vi);
       int err = 0;
       s64 len = 0;
       unsigned long i;
       VCN vcn, target_vcn;
       runlist_element *rl;
       s64 ofs;

       NTFS3G_BUG_TRACE(NInoAttrMstProtected(ni));
   
       for( i=0; i<nr_segs; i++)
          len += iov[i].iov_len;

       if(ni->map_na->data_size < pos+len)
       {
		if (ntfs_attr_truncate(&err, ni->map_na,pos+len)) {
			ntfs_log_perror("Failed to enlarge attribute");
                     NTFS3G_BUG_TRACE(!err);
			return -EIO;
		}
              NTFS3G_BUG_TRACE(ni->map_na->data_size < pos+len);
       //}
              if(NAttrNonResident(ni->map_na))
             {
               ntfs_attr_search_ctx *ctx;
                  BOOL bRetry = FALSE;
                    ctx = ntfs_attr_get_search_ctx(&err, ni, NULL);
		      if (!ctx)
			       return -ENOMEM;
		     if (ntfs_attr_lookup(&err, ni->map_na->type, ni->map_na->name, ni->map_na->name_len, 0,
				0, NULL, 0, ctx))
		    {
			      ntfs_attr_put_search_ctx(&err, ctx);
                           return -EIO;
		     }

                   err = 0;
                   vcn = pos;
	            vcn >>= ni->vol->cluster_size_bits;
                   target_vcn = (pos+len+ni->vol->cluster_size-1);
                   target_vcn >>= ni->vol->cluster_size_bits;
                   ofs = pos - (vcn<<ni->vol->cluster_size_bits);

                   while(vcn < target_vcn)
                   {
	                 rl = ntfs_attr_find_vcn(&err, ni->map_na,  vcn);
	                 if(!rl)
	                 {
		                 ntfs_attr_put_search_ctx(&err, ctx);
                               return -EIO;
	                 }
                        NTFS3G_BUG_TRACE(rl->vcn > vcn || rl->vcn+rl->length<= vcn);
                        if(rl->lcn == LCN_HOLE)
                        {
                             VCN update_from = -1;

                             NTFS3G_BUG_TRACE(bRetry);
                   
                             err = 0;
                             //ofs is offset from run list beginning
                             ofs = pos - (rl->vcn << ni->vol->cluster_size_bits);
                             if (ntfs_attr_fill_hole(&err, ni->map_na,  len, &ofs, &rl, &update_from))
                             {
                                   ntfs_attr_put_search_ctx(&err, ctx);
                                    return  -ENOSPC;
                             }
                              if (update_from!=-1 &&
		                      ntfs_attr_update_mapping_pairs(&err, ni->map_na, 0 /*update_from*/))                  
                             {
                                    ntfs_attr_put_search_ctx(&err, ctx);
                                    return -EIO;
                              }
                              bRetry = TRUE;
                              continue; //try map same vcn again
                         } 
                         else
                         {
                               NTFS3G_BUG_TRACE(rl->lcn < LCN_HOLE);
                               vcn = rl->vcn+rl->length;
                               ofs = 0;
                               bRetry = FALSE;
                          }
                    }
                    ctx->attr->initialized_size = cpu_to_sle64(pos + len);
                    mark_mft_record_dirty(ctx->ntfs_ino);
                    ni->map_na->initialized_size = pos + len;
                    ntfs_attr_put_search_ctx(&err, ctx);
             }  
             else
             {
                 NTFS3G_BUG_TRACE(ni->map_na->data_size != ni->map_na->initialized_size);
             }
              
	   }

        return generic_file_aio_write(iocb, iov, nr_segs, pos);
}


static int ntfs3g_file_ioctl (struct inode *inode, struct file *filp,
			 unsigned int cmd, unsigned long arg)
{
    ntfs_inode *ni;
    unsigned long * flags = (unsigned long *)arg;
    ni = NTFS_I(inode);

	switch (cmd) {
     
	case NTFS_IOCTL_GET_ATTRIBUTES :
		//return ntfs3g_ioctl_volume_label(inode, arg);
		*flags = ni->base_ni_attr.flags;
		return ni->base_ni_attr.flags;
        break;
	default:
		return -EFAULT;
	}

    return 0;
}

const struct file_operations ntfs3g_file_ops = {
	.llseek		= generic_file_llseek,	 /* Seek inside file. */
	.read		= do_sync_read,		 /* Read from file. */
	.aio_read	= generic_file_aio_read, /* Async read from file. */

	.write		= do_sync_write,	 /* Write to file. */
	.aio_write	= ntfs3g_file_aio_write,	 /* Async write to file. */
	/*.release	= ,*/			 /* Last file is closed.  See
						    fs/ext2/file.c::
						    ext2_release_file() for
						    how to use this to discard
						    preallocated space for
						    write opened files. */
	.fsync		= ntfs3g_file_fsync,	 /* Sync a file to disk. */
	/*.aio_fsync	= ,*/			 /* Sync all outstanding async
						    i/o operations on a
						    kiocb. */
	.ioctl	= ntfs3g_file_ioctl,			 /* Perform function on the
						    mounted filesystem. */
	.mmap		= generic_file_mmap,	 /* Mmap file. */
	.open		= ntfs3g_file_open,	 /* Open file. */
	.splice_read	= generic_file_splice_read /* Zero-copy data send with
						    the data source being on
						    the ntfs partition.  We do
						    not need to care about the
						    data destination. */
	/*.sendpage	= ,*/			 /* Zero-copy data send with
						    the data destination being
						    on the ntfs partition.  We
						    do not need to care about
						    the data source. */
};

const struct inode_operations ntfs3g_file_inode_ops = {
	.truncate	= ntfs3g_truncate_vfs,
	.setattr	= ntfs3g_setattr,
	.getattr    = ntfs3g_getattr
};
