/**
 * inode.c - Inode handling code. Originated from the Linux-NTFS project.
 *
 * Copyright (c) 2002-2005 Anton Altaparmakov
 * Copyright (c) 2002-2008 Szabolcs Szakacsits
 * Copyright (c) 2004-2007 Yura Pakhuchiy
 * Copyright (c) 2004-2005 Richard Russon
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the NTFS-3G
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/stddef.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>	/* For bdev_hardsect_size(). */
#include <linux/backing-dev.h>
#include <linux/buffer_head.h>
#include <linux/vfs.h>
#include <linux/moduleparam.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/mount.h>

#include "config.h"

#include "compat.h"
#include "types.h"
#include "attrib.h"
#include "inode.h"
#include "debug.h"
#include "mft.h"
#include "attrlist.h"
#include "runlist.h"
#include "lcnalloc.h"
#include "index.h"
#include "dir.h"
#include "ntfstime.h"
#include "misc.h"
#include "aops.h"

int ntfs_ext_inode_close(int *p_errno, ntfs_inode *ni);

typedef struct 
{
   MFT_REF mft_no;
   ATTR_TYPES type;
   ntfschar *name;
   u32 name_len;
}inode_cmp_attr;

int ntfs_test_inode(struct inode *vi, void *data)
{
       ntfs_inode *ni = NTFS_I(vi);
       inode_cmp_attr *indicator = (inode_cmp_attr *)data;
       
	if (MREF_INO(vi->i_ino) != MREF_INO(indicator->mft_no))
		return 0;

       if(MREF_INO(ni->mft_no) != MREF_INO(indicator->mft_no))
       {
              BUG();
              return 0;
       }
       NTFS3G_BUG_TRACE(!ni->map_na);
       if(ni->map_na->type != indicator->type)
       {
           return 0;
       }

       if(ni->map_na->name_len != indicator->name_len)
       {
           return 0;
       }

       //should we compare lowest_vcn at here?
       
       if(ni->map_na->name == indicator->name)
       {
           return 1;
       }
       
 	if (memcmp(ni->map_na->name,indicator->name,indicator->name_len*sizeof(ntfschar)))
		return 0;
       return 1;
}

int ntfs_attr_of_vfs_inode(ATTR_TYPES type,ntfschar *name, const u32 name_len)
{
     if(type!=AT_DATA && type != AT_INDEX_ROOT)
           return 0;
     if(type == AT_DATA && name_len == 0)
           return 1;
     if(type==AT_INDEX_ROOT && name_len==4 && memcmp(NTFS_INDEX_I30, name, 4*sizeof(ntfschar))==0)
          return 1;
     return 0;
}
int ntfs_attr_of_attr_inode(ATTR_TYPES type,ntfschar *name, const u32 name_len)
{
     if(type == AT_ATTRIBUTE_LIST)
            return 0;

     if(type == AT_DATA && name_len == 0)
           return 0;
     if(type==AT_INDEX_ROOT && name_len==4 && memcmp(NTFS_INDEX_I30, name, 4*sizeof(ntfschar))==0)
          return 0;
     return 1;
}

static int ntfs_init_locked_inode(struct inode *vi,  void *data)
{
	ntfs_inode *ni = NTFS_I(vi);
        inode_cmp_attr *indicator = (inode_cmp_attr *)data;

	vi->i_ino = MREF_INO(indicator->mft_no);
       memset(ni, 0, sizeof(*ni));
       ni->mft_no = indicator->mft_no;

	return 0;
}



struct inode *ntfs3g_alloc_big_inode(struct super_block *sb)
{
	big_ntfs3g_inode *big_ni;

	ntfs_log_trace("Entering.");
	big_ni = kmem_cache_alloc(ntfs3g_big_inode_cache, GFP_NOFS);
	if (likely(big_ni != NULL)) {
		return VFS_I((&big_ni->ntfs_inode));
	}
	ntfs_log_error(sb, "Allocation of NTFS big inode structure failed.");
	return NULL;
}

void ntfs3g_destroy_big_inode(struct inode *vi)
{
	ntfs_inode *ni = NTFS_I(vi);
//       int err= 0;

	ntfs_log_trace("Entering.");      
#if 1
      if(!NInoNTFSAttrNI(ni))
      {
             NTFS3G_BUG_TRACE(ni->base_ni_attr.nr_extents);
             NTFS3G_BUG_TRACE(NInoAttrList(ni) && ni->base_ni_attr.attr_list);
             NTFS3G_BUG_TRACE(ni->base_ni_attr.mft_page_refcnt);
             NTFS3G_BUG_TRACE(ni->base_ni_attr.extent_nis);
             NTFS3G_BUG_TRACE(NInoDirty(ni) || NInoAttrListDirty(ni));
       }
       NTFS3G_BUG_TRACE(ni->map_na);
#else
       NTFS3G_BUG_TRACE(ni->nr_extents  != 0);
       NTFS3G_BUG_TRACE(ni->base_ni != NULL);
       NTFS3G_BUG_TRACE(NInoDirty(ni) || NInoAttrListDirty(ni));
       
	//NTFS3G_BUG_TRACE(ni->page);
	//if (!atomic_dec_and_test(&ni->count))
	//	BUG();
	if (NInoDirty(ni) || NInoAttrListDirty(ni)) {
		if (ntfs_inode_sync(&err, ni)) {
                     ntfs_log_perror("failed to sync big inode %08x\n", ni);
 		}
	}
    


	/* Is this a base inode with mapped extent inodes? */
	if (ni->nr_extents > 0) 
       {
		while (ni->nr_extents > 0) {
			if (ntfs_ext_inode_close(&err, ni->extent_nis[0])) {
				ntfs_log_perror("failed to sync ext inode %08x for big inode %08x\n", ni, ni->extent_nis[0]);
			}
		 }
       }
    
    	if (NInoAttrList(ni) && ni->attr_list)
		ntfs_free(ni->attr_list);
	ntfs_free(ni->mrec);
#endif
	kmem_cache_free(ntfs3g_big_inode_cache, NTFS_I(vi));
}


ntfs_inode *ntfs_inode_base(int *p_errno, ntfs_inode *ni)
{
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
	if (ni->base_ni_attr.nr_extents == -1)
		return ni->base_ni_attr.base_ni;
	return ni;
}

/**
 * ntfs_big_inode_close - close an ntfs inode and free all associated memory
 * @ni:		ntfs inode to close
 *
 * Make sure the ntfs inode @ni is clean.
 *
 * If the ntfs inode @ni is a base inode, close all associated extent inodes,
 * then deallocate all memory attached to it, and finally free the ntfs inode
 * structure itself.
 *
 * If it is an extent inode, we disconnect it from its base inode before we
 * destroy it.
 *
 * It is OK to pass NULL to this function, it is just noop in this case.
 *
 * Return 0 on success or -1 on error with errno set to the error code. On
 * error, @ni has not been freed. The user should attempt to handle the error
 * and call ntfs_inode_close() again. The following error codes are defined:
 *
 *	EBUSY	@ni and/or its attribute list runlist is/are dirty and the
 *		attempt to write it/them to disk failed.
 *	EINVAL	@ni is invalid (probably it is an extent inode).
 *	EIO	I/O error while trying to write inode to disk.
 */
int ntfs_big_inode_close(int *p_errno, ntfs_inode *ni)
{
	int err;
	
	if (!ni)
		return 0;

       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
       NTFS3G_BUG_TRACE(ni->base_ni_attr.nr_extents < 0);
       
	ntfs_log_enter("Entering for inode %lld\n", (long long)ni->mft_no);

	/* If we have dirty metadata, write it out. */
	if (NInoDirty(ni) || NInoAttrListDirty(ni)) {
		if ((err=ntfs_inode_sync(p_errno, ni))) {
			if (err != EIO)
				err = EBUSY;
			goto err_exit;
		}
	}
      #if 0  // do not sync extension inode because mft of extension inode already be marked as dirty before
	/* Is this a base inode with mapped extent inodes? */
	if (ni->base_ni_attr.nr_extents > 0) 
       {              
              int k;
		for(k=0; k<ni->base_ni_attr.nr_extents;k++) {

			if ((err=ntfs_inode_sync(p_errno, ni->base_ni_attr.extent_nis[k]))) {
				if (err != EIO)
					err = EBUSY;
				goto err_exit;
			}
               
		 }
       }
       #endif

	iput(VFS_I(ni));
	err = 0;
err_exit:
	ntfs_log_leave("\n");
	return err;
}
/**
 * __ntfs_inode_allocate - Create and initialise an NTFS inode object
 * @vol:
 *
 * Description...
 *
 * Returns:
 */
 ntfs_inode *ntfs_alloc_ext_inode(int *p_errno, ntfs_volume *vol)
{
	ntfs_inode *ni;

	ntfs_log_trace("Entering.");
	ni = kmem_cache_alloc(ntfs3g_ext_inode_cache, GFP_NOFS);
	if (likely(ni != NULL)) {
              memset(ni, 0, sizeof(*ni));
              ni->vol = vol;
              NInoClearNTFSAttrNI(ni);
              mutex_init(&ni->base_ni_attr.mrec_lock);
		return ni;
	}
       *p_errno = ENOMEM;
	ntfs_log_perror("Allocation of NTFS ext inode structure failed.");
	return NULL;
}
void ntfs_destroy_ext_inode(ntfs_inode *ni)
{
	//ntfs_inode *ni = NTFS_I(inode);
	int err = 0;

	ntfs_log_trace("Entering.");
	//NTFS3G_BUG_TRACE(ni->page);
	//if (!atomic_dec_and_test(&ni->count))
	//	BUG();
	if(ni == NULL)
           return;
    
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
       
       if(ni->map_na)
           ntfs_attr_close(&err, ni->map_na);
    
	if (NInoDirty(ni))
		ntfs_log_error(ni->vol->sb, "Releasing dirty ext inode %lld!\n", 
			       (long long)ni->mft_no);
	if (NInoAttrList(ni) && ni->base_ni_attr.attr_list)
		ntfs_free(ni->base_ni_attr.attr_list);
       if(ni->map_na)
       {
            err = 0;
            ntfs_attr_close(&err, ni->map_na);
       }
	kmem_cache_free(ntfs3g_ext_inode_cache, ni);
}


/**
 * ntfs_ext_inode_close - close an ntfs inode and free all associated memory
 * @ni:		ntfs inode to close
 *
 * Make sure the ntfs inode @ni is clean.
 *
 * If the ntfs inode @ni is a base inode, close all associated extent inodes,
 * then deallocate all memory attached to it, and finally free the ntfs inode
 * structure itself.
 *
 * If it is an extent inode, we disconnect it from its base inode before we
 * destroy it.
 *
 * It is OK to pass NULL to this function, it is just noop in this case.
 *
 * Return 0 on success or -1 on error with errno set to the error code. On
 * error, @ni has not been freed. The user should attempt to handle the error
 * and call ntfs_inode_close() again. The following error codes are defined:
 *
 *	EBUSY	@ni and/or its attribute list runlist is/are dirty and the
 *		attempt to write it/them to disk failed.
 *	EINVAL	@ni is invalid (probably it is an extent inode).
 *	EIO	I/O error while trying to write inode to disk.
 */
int ntfs_ext_inode_close(int *p_errno, ntfs_inode *ni)
{
	int err = -1;
	ntfs_inode **tmp_nis;
	ntfs_inode *base_ni;
	s32 i;
       
	if (!ni)
		return 0;
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
       
	ntfs_log_enter("Entering for inode %lld\n", (long long)ni->mft_no);
#if 0
	/* If we have dirty metadata, write it out. */
	if (NInoDirty(ni) || NInoAttrListDirty(ni)) {
		if ((err=ntfs_inode_sync(p_errno, ni))) {
			if (err != EIO)
				err = EBUSY;
			goto err_exit;
		}
	}
#endif
        NTFS3G_BUG_TRACE(ni->base_ni_attr.nr_extents >= 0 || ni->base_ni_attr.base_ni==NULL);
        NTFS3G_BUG_TRACE(NInoAttrList(ni));
	/* Is this a base inode with mapped extent inodes? */
		/*
		 * If the inode is an extent inode, disconnect it from the
		 * base inode before destroying it.
		 */
	base_ni = ni->base_ni_attr.base_ni;
	for (i = 0; i < base_ni->base_ni_attr.nr_extents; ++i) {
			tmp_nis = base_ni->base_ni_attr.extent_nis;
			if (tmp_nis[i] != ni)
				continue;
			/* Found it. Disconnect. */
			memmove(tmp_nis + i, tmp_nis + i + 1,
					(base_ni->base_ni_attr.nr_extents - i - 1) *
					sizeof(ntfs_inode *));
			/* Buffer should be for multiple of four extents. */
			if ((--base_ni->base_ni_attr.nr_extents) & 3) {
				i = -1;
				break;
			}
			/*
			 * ElectricFence is unhappy with realloc(x,0) as free(x)
			 * thus we explicitly separate these two cases.
			 */
			if (base_ni->base_ni_attr.nr_extents) {
				/* Resize the memory buffer. */
				tmp_nis = ntfs_realloc_nofs(p_errno, tmp_nis, base_ni->base_ni_attr.nr_extents *
						  sizeof(ntfs_inode *), base_ni->base_ni_attr.nr_extents *
						  sizeof(ntfs_inode *));
				/* Ignore errors, they don't really matter. */
				if (tmp_nis)
					base_ni->base_ni_attr.extent_nis = tmp_nis;
			} else if (tmp_nis)
			{
				ntfs_free(tmp_nis);
                             base_ni->base_ni_attr.extent_nis = NULL;
			}
			/* Allow for error checking. */
			i = -1;
			break;
	}

       NTFS3G_BUG_TRACE( i >= 0);
       
 	if (NInoDirty(ni))
		ntfs_log_perror("Releasing dirty inode %lld!\n", 
			       (long long)ni->mft_no);
	if (NInoAttrList(ni) && ni->base_ni_attr.attr_list)
		ntfs_free(ni->base_ni_attr.attr_list);
	ntfs_destroy_ext_inode(ni);
	err = 0;
	ntfs_log_leave("\n");
	return err;
}

/**
 * ntfs_inode_mark_dirty - set the inode (and its base inode if it exists) dirty
 * @ni:		ntfs inode to set dirty
 *
 * Set the inode @ni dirty so it is written out later (at the latest at
 * ntfs_inode_close() time). If @ni is an extent inode, set the base inode
 * dirty, too.
 *
 * This function cannot fail.
 */
void ntfs_inode_mark_dirty(int *p_errno, ntfs_inode *ni)
{
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
	NInoSetDirty(ni);
	if (ni->base_ni_attr.nr_extents == -1)
		NInoSetDirty(ni->base_ni_attr.base_ni);
}


/**
 * ntfs_big_inode_open - open an inode ready for access
 * @vol:	volume to get the inode from
 * @mref:	inode number / mft record number to open
 *
 * Allocate an ntfs_inode structure and initialize it for the given inode
 * specified by @mref. @mref specifies the inode number / mft record to read,
 * including the sequence number, which can be 0 if no sequence number checking
 * is to be performed.
 *
 * Then, allocate a buffer for the mft record, read the mft record from the
 * volume @vol, and attach it to the ntfs_inode structure (->mrec). The
 * mft record is mst deprotected and sanity checked for validity and we abort
 * if deprotection or checks fail.
 *
 * Finally, search for an attribute list attribute in the mft record and if one
 * is found, load the attribute list attribute value and attach it to the
 * ntfs_inode structure (->attr_list). Also set the NI_AttrList bit to indicate
 * this.
 *
 * Return a pointer to the ntfs_inode structure on success or NULL on error,
 * with errno set to the error code.
 */
extern const struct inode_operations ntfs3g_file_inode_ops;
extern const struct file_operations ntfs3g_file_ops;
extern const struct address_space_operations ntfs3g_mst_aops;
extern const struct address_space_operations ntfs3g_aops;
extern const struct inode_operations ntfs3g_dir_inode_ops;
extern const struct file_operations ntfs3g_dir_ops;
static int __ntfs_big_inode_open(int *p_errno, ntfs_volume *vol, ntfs_inode *ni, MFT_REF mref,
                                                                         ATTR_TYPES type, ntfschar *name, const u32 name_len)
{
	s64 l;
	ntfs_attr_search_ctx *ctx;
       STANDARD_INFORMATION *std_info;
       struct inode *vi = VFS_I(ni);
       MFT_RECORD *mrec;
       MFT_RECORD_FLAGS mftflags;

	ntfs_log_trace("Entering for inode %lld\n", MREF(mref));
	if (!vol) {
		*p_errno = EINVAL;
		  return -1;
	}       
       ni->base_ni_attr.mft_page = NULL;
       ni->base_ni_attr.mft_page_refcnt = 0;
       ni->mft_no = MREF(mref);
       ni->vol = vol;
       mutex_init(&ni->base_ni_attr.mrec_lock);
       NInoClearNTFSAttrNI(ni);

       mrec = map_mft_record(ni);
	if (mrec == NULL)
	{
	       *p_errno = EIO;
		goto err_out;
	}
	if (!(mrec->flags & MFT_RECORD_IN_USE)) {
		*p_errno = ENOENT;
              printk("error when __ntfs_big_inode_open %lld\n", MREF(mref));
              unmap_mft_record(ni);
		goto err_out;
	}      
       vi->i_mode = 0;
 	vi->i_generation = le16_to_cpu(mrec->sequence_number);
	vi->i_nlink = le16_to_cpu(mrec->link_count);
	/* Everyone gets all permissions. */
	vi->i_mode |= S_IRWXUGO;
	ctx = ntfs_attr_get_search_ctx(p_errno, ni, NULL); 
       mftflags = mrec->flags;
       unmap_mft_record(ni);
	if (!ctx)
		goto err_out;
	/* Receive some basic information about inode. */
	if (ntfs_attr_lookup(p_errno, AT_STANDARD_INFORMATION, AT_UNNAMED,
				0, CASE_SENSITIVE, 0, NULL, 0, ctx)) {
		ntfs_log_trace("Failed to receive STANDARD_INFORMATION "
				"attribute.\n");
		goto put_err_out;
	}
	std_info = (STANDARD_INFORMATION *)((u8 *)ctx->attr +
			le16_to_cpu(ctx->attr->value_offset));
	ni->base_ni_attr.flags = std_info->file_attributes;
	ni->base_ni_attr.creation_time = ntfs2utc(std_info->creation_time);
	ni->base_ni_attr.last_data_change_time = ntfs2utc(std_info->last_data_change_time);
	ni->base_ni_attr.last_mft_change_time = ntfs2utc(std_info->last_mft_change_time);
	ni->base_ni_attr.last_access_time = ntfs2utc(std_info->last_access_time);
	vi->i_mtime = ntfs2timespec(std_info->last_data_change_time);
	vi->i_ctime = ntfs2timespec(std_info->last_mft_change_time);
	vi->i_atime = ntfs2timespec(std_info->last_access_time);
    
       if(mftflags & MFT_RECORD_IS_DIRECTORY)
       {
		vi->i_mode |= S_IFDIR;
		/* Things break without this kludge! */
		if (vi->i_nlink > 1)
			vi->i_nlink = 1;
	} else {
		vi->i_mode |= S_IFREG;
		/* Apply the file permissions mask set in the mount options. */
	}
	/* Set attribute list information. */
	if (ntfs_attr_lookup(p_errno, AT_ATTRIBUTE_LIST, AT_UNNAMED, 0, 0, 0, NULL, 0,
			ctx)) {
		if (*p_errno != ENOENT)
			goto put_err_out;
		/* Attribute list attribute does not present. */
              *p_errno = 0;
		goto get_size;
	}
	NInoSetAttrList(ni);
	l = ntfs_get_attribute_value_length(p_errno, ctx->attr);
	if (!l)
		goto put_err_out;
	if (l > 0x40000) {
		*p_errno = EIO;
		goto put_err_out;
	}
	ni->base_ni_attr.attr_list_size = l;
	ni->base_ni_attr.attr_list = ntfs_malloc_nofs(p_errno, ni->base_ni_attr.attr_list_size);
	if (!ni->base_ni_attr.attr_list)
		goto put_err_out;
	l = ntfs_get_attribute_value(p_errno, vol, ctx->attr, ni->base_ni_attr.attr_list);
	if (!l)
		goto put_err_out;
	if (l != ni->base_ni_attr.attr_list_size) {
		*p_errno = EIO;
		goto put_err_out;
	}
get_size:
    if (S_ISDIR(vi->i_mode)) {
            if(type!=AT_INDEX_ROOT || name!=NTFS_INDEX_I30 || name_len!=4)
             {
			ntfs_log_perror("For folder, we only support open inode index_allocation & bitmap attribute\n", mref);
			goto put_err_out;
            }
 
	     ni->map_na = ntfs_attr_open(p_errno, ni, AT_INDEX_ROOT, NTFS_INDEX_I30, 4);
            if(!ni->map_na )
            {
                    *p_errno = EIO;
                    goto put_err_out;
            }
            vi->i_size = ni->map_na->data_size;
            vi->i_op = &ntfs3g_dir_inode_ops;
	     vi->i_fop = &ntfs3g_dir_ops;
    }
    else
    {
          NTFS3G_BUG_TRACE(type != AT_DATA);
          ni->map_na = ntfs_attr_open(p_errno, ni, type, name, name_len);
	   if (!ni->map_na) {
                       *p_errno = EIO;
			  goto put_err_out;
	    } else {
	              vi->i_size = ni->map_na->data_size;
            	       vi->i_op = &ntfs3g_file_inode_ops;
		       vi->i_fop = &ntfs3g_file_ops;
                     vi->i_size = ni->map_na->initialized_size;
	    }
           if(type==AT_DATA && ni->map_na->name==AT_UNNAMED && (mref==FILE_MFT||mref==FILE_MFTMirr))
           {
                  ni->fixup_block_size = vol->mft_record_size;
                  NInoSetAttrMstProtected(ni);
           }
             
     }
    
    if (NInoAttrMstProtected(ni))
	   vi->i_mapping->a_ops = &ntfs3g_mst_aops;
    else
           vi->i_mapping->a_ops = &ntfs3g_aops;
     ntfs_attr_put_search_ctx(p_errno, ctx);
     ntfs_log_trace("\n");
     return 0;

put_err_out:
	ntfs_attr_put_search_ctx(p_errno, ctx);
err_out:
       if(!*p_errno)
           *p_errno = EIO;
       vi->i_nlink = 0;
	return -1;
}

ntfs_inode *ntfs_big_inode_open(int *p_errno, ntfs_volume *vol, const MFT_REF mref,
              ATTR_TYPES type,ntfschar *name, const u32 name_len)
{
	ntfs_inode *ni = NULL;
       struct inode *tmp_ino;
       int err= 0;
       inode_cmp_attr indicator;

	ntfs_log_trace("Entering for inode %lld\n", MREF(mref));
	if (!vol) {
		*p_errno = EINVAL;
		 return NULL;
	}
    
      NTFS3G_BUG_TRACE(!ntfs_attr_of_vfs_inode(type, name, name_len));
      
      indicator.mft_no = mref;
      indicator.type = type;
      indicator.name = name;
      indicator.name_len = name_len;
      
      tmp_ino = iget5_locked(vol->sb, MREF_INO(mref),  ntfs_test_inode,
            ntfs_init_locked_inode, (void*)&indicator);
	if (unlikely(!tmp_ino))
	{
		*p_errno = ENOMEM;
               return NULL;
	 }
       ni = NTFS_I(tmp_ino);
       if ((tmp_ino->i_state & I_NEW)) {
		__ntfs_big_inode_open(&err, vol, ni, mref, type, name, name_len);
		 unlock_new_inode(tmp_ino);
	}

	if (unlikely((err))) {
		iput(tmp_ino);
              *p_errno = err;
		return NULL;
	}
       NTFS3G_BUG_TRACE(!ni->map_na);
	return ni;
}

static int __ntfs_attr_inode_open(int *p_errno, ntfs_inode *vfs_base_ni, ntfs_inode *ni,
                                                                         ATTR_TYPES type, ntfschar *name, const u32 name_len)
{
	ntfs_attr_search_ctx *ctx;
       STANDARD_INFORMATION *std_info;
       struct inode *vi = VFS_I(ni);
       MFT_RECORD *mrec;



       NTFS3G_BUG_TRACE(!vfs_base_ni);

       ntfs_log_trace("Entering for attr inode %lld\n", MREF(vfs_base_ni->mft_no));;
       
       ni->ntfs_mft_base_ni = vfs_base_ni;
       ni->mft_no = vfs_base_ni->mft_no;
       ni->vol = vfs_base_ni->vol;
       NInoSetNTFSAttrNI(ni);

       mrec = map_mft_record(vfs_base_ni);
	if (mrec == NULL)
		goto err_out;
	if (!(mrec->flags & MFT_RECORD_IN_USE)) {
		*p_errno = ENOENT;
		goto err_out;
	}      
 	vi->i_generation = le16_to_cpu(mrec->sequence_number);
	vi->i_nlink = le16_to_cpu(mrec->link_count);
	/* Everyone gets all permissions. */
	vi->i_mode |= S_IRWXUGO;
	ctx = ntfs_attr_get_search_ctx(p_errno, vfs_base_ni, NULL); 
       unmap_mft_record(vfs_base_ni);
	if (!ctx)
		goto err_out;
	/* Receive some basic information about inode. */
	if (ntfs_attr_lookup(p_errno, AT_STANDARD_INFORMATION, AT_UNNAMED,
				0, CASE_SENSITIVE, 0, NULL, 0, ctx)) {
		ntfs_log_trace("Failed to receive STANDARD_INFORMATION "
				"attribute.\n");
		goto put_err_out;
	}
	std_info = (STANDARD_INFORMATION *)((u8 *)ctx->attr +
			le16_to_cpu(ctx->attr->value_offset));
	ni->base_ni_attr.flags = std_info->file_attributes;
	ni->base_ni_attr.creation_time = ntfs2utc(std_info->creation_time);
	ni->base_ni_attr.last_data_change_time = ntfs2utc(std_info->last_data_change_time);
	ni->base_ni_attr.last_mft_change_time = ntfs2utc(std_info->last_mft_change_time);
	ni->base_ni_attr.last_access_time = ntfs2utc(std_info->last_access_time);
	vi->i_mtime = ntfs2timespec(std_info->last_data_change_time);
	vi->i_ctime = ntfs2timespec(std_info->last_mft_change_time);
	vi->i_atime = ntfs2timespec(std_info->last_access_time);
    
       vi->i_mode = VFS_I(vfs_base_ni)->i_mode & ~S_IFMT;

       NTFS3G_BUG_TRACE(!ntfs_attr_of_attr_inode(type, name, name_len));

        if(type ==AT_INDEX_ALLOCATION )
        {
              NTFS3G_BUG_TRACE(vfs_base_ni->map_na->type!=AT_INDEX_ROOT);
              if (ntfs_attr_lookup(p_errno, AT_INDEX_ROOT, NTFS_INDEX_I30, 4, CASE_SENSITIVE, 0, NULL,
			0, ctx)) {
			ntfs_log_perror("Failed to open index root for big inode for %lld!\n", mref);
                     *p_errno = EIO;
			goto put_err_out;
             }
             else
             {     
                     INDEX_ROOT *ir;
                     u32 index_block_size;
                     if(ctx->attr->non_resident)
                     {
                          *p_errno = EIO;
	                    ntfs_log_perror("Index root attribute is non residual for mref %lld\n", mref);
                           goto put_err_out;
                     }
                	ir = (INDEX_ROOT*)((u8*)ctx->attr + le16_to_cpu(ctx->attr->value_offset));
                    	index_block_size = le32_to_cpu(ir->index_block_size);
	              if (index_block_size < NTFS_BLOCK_SIZE ||
			       index_block_size & (index_block_size - 1)) {
			        *p_errno = EIO;
		                ntfs_log_error(ni->vol->sb, "Index block size %u is invalid.\n",   (unsigned)index_block_size);
		               goto put_err_out;
	              }
                     /*during ntfs_ia_add(), ir_index flag may still be SMALL_INDEX when attr_inode open for AT_INDEX_ALLOCATION, so we
                     do not check it here
	              if (!(ir->index.ih_flags & LARGE_INDEX)) {
                             *p_errno = ENOENT;
                             goto put_err_out;
                                 
	              } else */{
	                       ni->map_na = ntfs_attr_open(p_errno, ni, 
                                                           (type==AT_UNUSED)?AT_INDEX_ALLOCATION:type, NTFS_INDEX_I30, 4);
                              if(!ni->map_na )
                              {
                                   *p_errno = ENOENT;
                                   goto put_err_out;
                              }
                              vi->i_size = ni->map_na->data_size;
                              ni->fixup_block_size = index_block_size;
                              NInoSetAttrMstProtected(ni);
                              vi->i_op = NULL;
		                vi->i_fop = NULL;
                              vi->i_mapping->a_ops = &ntfs3g_mst_aops;
                     }
		} 
        }
        else
        {
            ni->map_na = ntfs_attr_open(p_errno, ni, type, name, name_len);
	     if (!ni->map_na) {
                       *p_errno = ENOENT;
			  goto put_err_out;
	      } else {
	              vi->i_size = ni->map_na->data_size;
            	       vi->i_op = NULL;
		       vi->i_fop = NULL;
                     vi->i_size = ni->map_na->initialized_size;
                     vi->i_mapping->a_ops = &ntfs3g_aops;
	     }
        }
     ntfs_attr_put_search_ctx(p_errno, ctx);
     ntfs_log_trace("\n");
     return 0;

put_err_out:
	ntfs_attr_put_search_ctx(p_errno, ctx);
err_out:
       if(!*p_errno)
           *p_errno = EIO;
       vi->i_nlink = 0;
	return -1;
}

ntfs_inode *ntfs_attr_inode_open(int *p_errno, ntfs_inode *vfs_base_ni,
              ATTR_TYPES type,ntfschar *name, const u32 name_len)
{
	ntfs_inode *ni = NULL;
       struct inode *tmp_ino;
       int err= 0;
       inode_cmp_attr indicator;
       

	ntfs_log_trace("Entering for inode %lld\n", MREF(mref));
	if (!vfs_base_ni) {
		*p_errno = EINVAL;
		 return NULL;
	}
      indicator.mft_no = vfs_base_ni->mft_no;
      indicator.type = type;
      indicator.name = name;
      indicator.name_len = name_len;

      NTFS3G_BUG_TRACE(ntfs_test_inode(VFS_I(vfs_base_ni), &indicator));
      
      tmp_ino = iget5_locked(vfs_base_ni->vol->sb, vfs_base_ni->mft_no,  ntfs_test_inode,
            ntfs_init_locked_inode, (void*)&indicator);
	if (unlikely(!tmp_ino))
	{
		*p_errno = ENOMEM;
               return NULL;
	 }
       ni = NTFS_I(tmp_ino);
       if ((tmp_ino->i_state & I_NEW)) {
		__ntfs_attr_inode_open(&err, vfs_base_ni, ni,  type, name, name_len);
		 unlock_new_inode(tmp_ino);
	}
	if (unlikely((err))) {
		iput(tmp_ino);
              *p_errno = err;
		return NULL;
	}
       NTFS3G_BUG_TRACE(!ni->map_na);
	return ni;
}

int ntfs_attr_inode_close(int *p_errno, ntfs_inode *ni)
{
	
	if (!ni)
		return 0;
       NTFS3G_BUG_TRACE(!NInoNTFSAttrNI(ni));
	ntfs_log_enter("Entering for inode %lld\n", (long long)ni->mft_no);
	iput(VFS_I(ni));
	ntfs_log_leave("\n");
	return 0;
}

int ntfs3g_write_inode(struct inode *vi, int sync)
{
    ntfs_inode *ni = NTFS_I(vi);
    int err = 0;
    
    ntfs_log_trace("Entering.");
    
    return ntfs_inode_sync(&err, ni);;
}

/**
 * ntfs_extent_inode_open - load an extent inode and attach it to its base
 * @base_ni:	base ntfs inode
 * @mref:	mft reference of the extent inode to load (in little endian)
 *
 * First check if the extent inode @mref is already attached to the base ntfs
 * inode @base_ni, and if so, return a pointer to the attached extent inode.
 *
 * If the extent inode is not already attached to the base inode, allocate an
 * ntfs_inode structure and initialize it for the given inode @mref. @mref
 * specifies the inode number / mft record to read, including the sequence
 * number, which can be 0 if no sequence number checking is to be performed.
 *
 * Then, allocate a buffer for the mft record, read the mft record from the
 * volume @base_ni->vol, and attach it to the ntfs_inode structure (->mrec).
 * The mft record is mst deprotected and sanity checked for validity and we
 * abort if deprotection or checks fail.
 *
 * Finally attach the ntfs inode to its base inode @base_ni and return a
 * pointer to the ntfs_inode structure on success or NULL on error, with errno
 * set to the error code.
 *
 * Note, extent inodes are never closed directly. They are automatically
 * disposed off by the closing of the base inode.
 */
ntfs_inode *ntfs_extent_inode_open(int *p_errno, ntfs_inode *base_ni, const MFT_REF mref)
{
	u64 mft_no = MREF_LE(mref);
	ntfs_inode *ni = NULL;
	ntfs_inode **extent_nis;
	int i;
       MFT_RECORD *ni_mrec;

	if (!base_ni) {
		*p_errno = EINVAL;
		ntfs_log_perror("%s", __FUNCTION__);
		return NULL;
	}
	NTFS3G_BUG_TRACE(NInoNTFSAttrNI(base_ni));
	ntfs_log_enter("Opening extent inode %lld (base mft record %lld).\n",
			(unsigned long long)mft_no,
			(unsigned long long)base_ni->mft_no);
	
	/* Is the extent inode already open and attached to the base inode? */
	if (base_ni->base_ni_attr.nr_extents > 0) {
		extent_nis = base_ni->base_ni_attr.extent_nis;
		for (i = 0; i < base_ni->base_ni_attr.nr_extents; i++) {
			u16 seq_no;

			ni = extent_nis[i];
			if (mft_no != ni->mft_no)
				continue;
                     if((ni_mrec=map_mft_record(ni))== NULL)
                     {
                     	ntfs_log_perror("failed to map mft record for %lld", ni->mft_no);
                            goto out;
                     }
			/* Verify the sequence number if given. */
			seq_no = MSEQNO_LE(mref);
			if (seq_no && seq_no != le16_to_cpu(
					ni_mrec->sequence_number)) {
				*p_errno = EIO;
				ntfs_log_perror("Found stale extent mft "
					"reference mft=%lld",
					(long long)ni->mft_no);
                            unmap_mft_record(ni);
				goto out;
			}
                      unmap_mft_record(ni);
			goto out;
		}
	}
	/* Wasn't there, we need to load the extent inode. */
	ni = ntfs_alloc_ext_inode(p_errno, base_ni->vol);
	if (!ni)
		goto out;

	ni->mft_no = mft_no;
	ni->base_ni_attr.nr_extents = -1;
	ni->base_ni_attr.base_ni = base_ni;
	/* Attach extent inode to base inode, reallocating memory if needed. */
	if (!(base_ni->base_ni_attr.nr_extents & 3)) {
		i = (base_ni->base_ni_attr.nr_extents + 4) * sizeof(ntfs_inode *);

		extent_nis = ntfs_malloc_nofs(p_errno, i);
		if (!extent_nis)
			goto err_out;
		if (base_ni->base_ni_attr.nr_extents) {
			memcpy(extent_nis, base_ni->base_ni_attr.extent_nis,
					i - 4 * sizeof(ntfs_inode *));
			ntfs_free(base_ni->base_ni_attr.extent_nis);
		}
		base_ni->base_ni_attr.extent_nis = extent_nis;
	}
	base_ni->base_ni_attr.extent_nis[base_ni->base_ni_attr.nr_extents++] = ni;
out:
	ntfs_log_leave("\n");
	return ni;
err_out:
	ntfs_destroy_ext_inode(ni);
	ni = NULL;
	goto out;
}

/**
 * ntfs_inode_attach_all_extents - attach all extents for target inode
 * @ni:		opened ntfs inode for which perform attach
 *
 * Return 0 on success and -1 on error with errno set to the error code.
 */
int ntfs_inode_attach_all_extents(int *p_errno, ntfs_inode *ni)
{
	ATTR_LIST_ENTRY *ale;
	u64 prev_attached = 0;

	if (!ni) {
		ntfs_log_trace("Invalid arguments.\n");
		*p_errno = EINVAL;
		return -1;
	}
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
       
	if (ni->base_ni_attr.nr_extents == -1)
		ni = ni->base_ni_attr.base_ni;

	ntfs_log_trace("Entering for inode 0x%llx.\n", (long long) ni->mft_no);

	/* Inode haven't got attribute list, thus nothing to attach. */
	if (!NInoAttrList(ni))
		return 0;

	if (!ni->base_ni_attr.attr_list) {
		ntfs_log_trace("Corrupt in-memory struct.\n");
		*p_errno = EINVAL;
		return -1;
	}

	/* Walk through attribute list and attach all extents. */
	*p_errno = 0;
	ale = (ATTR_LIST_ENTRY *)ni->base_ni_attr.attr_list;
	while ((u8*)ale < ni->base_ni_attr.attr_list + ni->base_ni_attr.attr_list_size) {
		if (ni->mft_no != MREF_LE(ale->mft_reference) &&
				prev_attached != MREF_LE(ale->mft_reference)) {
			if (!ntfs_extent_inode_open(p_errno, ni, ale->mft_reference)) {
				ntfs_log_trace("Couldn't attach extent inode.\n");
				return -1;
			}
			prev_attached = MREF_LE(ale->mft_reference);
		}
		ale = (ATTR_LIST_ENTRY *)((u8*)ale + le16_to_cpu(ale->length));
	}
	return 0;
}

/**
 * ntfs_inode_sync_standard_information - update standard information attribute
 * @ni:		ntfs inode to update standard information
 *
 * Return 0 on success or -1 on error with errno set to the error code.
 */

static int ntfs_inode_sync_standard_information(int *p_errno, ntfs_inode *ni)
{
	ntfs_attr_search_ctx *ctx;
	STANDARD_INFORMATION *std_info;
	int err;
       int stdinfo_changed = 0;

	ntfs_log_trace("Entering for inode 0x%llx.\n", (long long) ni->mft_no);
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
	ctx = ntfs_attr_get_search_ctx(p_errno, ni, NULL);
	if (!ctx)
		return -1;
	if (ntfs_attr_lookup(p_errno, AT_STANDARD_INFORMATION, AT_UNNAMED,
				0, CASE_SENSITIVE, 0, NULL, 0, ctx)) {
		err = *p_errno;
		ntfs_log_trace("Failed to receive STANDARD_INFORMATION "
				"attribute.\n");
		ntfs_attr_put_search_ctx(p_errno, ctx);
		*p_errno = err;
		return -1;
	}
	std_info = (STANDARD_INFORMATION *)((u8 *)ctx->attr +
			le16_to_cpu(ctx->attr->value_offset));

       if(std_info->file_attributes != ni->base_ni_attr.flags)
       {
            std_info->file_attributes = ni->base_ni_attr.flags;
            stdinfo_changed = 1;
       }
       if(ntfs2utc(std_info->creation_time) != ni->base_ni_attr.creation_time)
       {
            std_info->creation_time = utc2ntfs(ni->base_ni_attr.creation_time);
            stdinfo_changed = 1;
       }
       if(ntfs2utc(std_info->last_data_change_time) != ni->base_ni_attr.last_data_change_time)
       {
            std_info->last_data_change_time = utc2ntfs(ni->base_ni_attr.last_data_change_time);
            stdinfo_changed = 1;
       }
       if(ntfs2utc(std_info->last_mft_change_time) != ni->base_ni_attr.last_mft_change_time)
       {
            std_info->last_mft_change_time = utc2ntfs(ni->base_ni_attr.last_mft_change_time);
            stdinfo_changed = 1;
       }
       if(ntfs2utc(std_info->last_access_time) != ni->base_ni_attr.last_access_time)
       {
            std_info->last_access_time = utc2ntfs(ni->base_ni_attr.last_access_time);
            stdinfo_changed = 1;
       }

       if(stdinfo_changed)
       {
	    ntfs_inode_mark_dirty(p_errno, ctx->ntfs_ino);
           mark_mft_record_dirty(ctx->ntfs_ino);
       }
	ntfs_attr_put_search_ctx(p_errno, ctx);
	return 0;
}

/**
 * ntfs_inode_sync_file_name - update FILE_NAME attributes
 * @ni:		ntfs inode to update FILE_NAME attributes
 *
 * Update all FILE_NAME attributes for inode @ni in the index.
 *
 * Return 0 on success or -1 on error with errno set to the error code.
 */
static int ntfs_inode_sync_file_name(int *p_errno, ntfs_inode *ni)
{
	ntfs_attr_search_ctx *ctx = NULL;
	ntfs_index_context *ictx;
	ntfs_inode *index_ni;
	FILE_NAME_ATTR *fn;
	int err = 0;
    
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
       
	ntfs_log_trace("Entering for inode %lld\n", (long long)ni->mft_no);

	ctx = ntfs_attr_get_search_ctx(p_errno, ni, NULL);
	if (!ctx) {
		err = *p_errno;
		goto err_out;
	}
	/* Walk through all FILE_NAME attributes and update them. */
	while (!ntfs_attr_lookup(p_errno, AT_FILE_NAME, NULL, 0, 0, 0, NULL, 0, ctx)) {
		fn = (FILE_NAME_ATTR *)((u8 *)ctx->attr +
				le16_to_cpu(ctx->attr->value_offset));
		if (MREF_LE(fn->parent_directory) == ni->mft_no) {
			/*
			 * WARNING: We cheat here and obtain 2 attribute
			 * search contexts for one inode (first we obtained
			 * above, second will be obtained inside
			 * ntfs_index_lookup), it's acceptable for library,
			 * but will deadlock in the kernel.
			 */
			index_ni = ni;
		} else
			index_ni = ntfs_big_inode_open(p_errno, ni->vol, 
					le64_to_cpu(fn->parent_directory), AT_INDEX_ROOT, NTFS_INDEX_I30, 4);
		if (!index_ni) {
			if (!err)
				err = *p_errno;
			ntfs_log_perror("Failed to open inode %lld with index",
				(long long)le64_to_cpu(fn->parent_directory));
			continue;
		}
		ictx = ntfs_index_ctx_get(p_errno, index_ni, NTFS_INDEX_I30, 4);
		if (!ictx) {
			if (!err)
				err = *p_errno;
			ntfs_log_perror("Failed to get index ctx, inode %lld",
					(long long)index_ni->mft_no);
			if (ni != index_ni && ntfs_big_inode_close(p_errno, index_ni) && !err)
				err = *p_errno;
			continue;
		}
		if (ntfs_index_lookup(p_errno, fn, sizeof(FILE_NAME_ATTR), ictx)) {
			if (!err) {
				if (*p_errno == ENOENT)
					err = EIO;
				else
					err = *p_errno;
			}
			ntfs_log_perror("Index lookup failed, inode %lld",
					(long long)index_ni->mft_no);
			ntfs_index_ctx_put(p_errno, ictx);
			if (ni != index_ni && ntfs_big_inode_close(p_errno, index_ni) && !err)
				err = *p_errno;
			continue;
		}
		/* Update flags and file size. */
		fn = (FILE_NAME_ATTR *)ictx->data;
		fn->file_attributes =
				(fn->file_attributes & ~FILE_ATTR_VALID_FLAGS) |
				(ni->base_ni_attr.flags & FILE_ATTR_VALID_FLAGS);
             if (ni->map_na && ni->map_na->type == AT_DATA && ni->map_na->name == AT_UNNAMED) 
            {
		        if (NAttrSparse(ni->map_na))
			       fn->allocated_size = ni->map_na->compressed_size;
                      else
                            fn->allocated_size = ni->map_na->compressed_size;
             }
             else
             {
		       fn->allocated_size = cpu_to_sle64(ni->map_na?ni->map_na->allocated_size:0);
              }
		fn->data_size = cpu_to_sle64(ni->map_na?ni->map_na->data_size:0);
		fn->creation_time = utc2ntfs(ni->base_ni_attr.creation_time);
		fn->last_data_change_time = utc2ntfs(ni->base_ni_attr.last_data_change_time);
		fn->last_mft_change_time = utc2ntfs(ni->base_ni_attr.last_mft_change_time);
		fn->last_access_time = utc2ntfs(ni->base_ni_attr.last_access_time);
		ntfs_index_entry_mark_dirty(p_errno, ictx);
		ntfs_index_ctx_put(p_errno, ictx);
		if ((ni != index_ni) && ntfs_big_inode_close(p_errno, index_ni) && !err)
			err = *p_errno;
	}
	/* Check for real error occurred. */
	if (*p_errno != ENOENT) {
		err = *p_errno;
		ntfs_log_perror("Attribute lookup failed, inode %lld",
				(long long)ni->mft_no);
		goto err_out;
	}
	ntfs_attr_put_search_ctx(p_errno, ctx);
	if (err) {
		*p_errno = err;
		return -1;
	}
	return 0;
err_out:
	if (ctx)
		ntfs_attr_put_search_ctx(p_errno, ctx);
	*p_errno = err;
	return -1;
}

/**
 * ntfs_inode_sync - write the inode (and its dirty extents) to disk
 * @ni:		ntfs inode to write
 *
 * Write the inode @ni to disk as well as its dirty extent inodes if such
 * exist and @ni is a base inode. If @ni is an extent inode, only @ni is
 * written completely disregarding its base inode and any other extent inodes.
 *
 * For a base inode with dirty extent inodes if any writes fail for whatever
 * reason, the failing inode is skipped and the sync process is continued. At
 * the end the error condition that brought about the failure is returned. Thus
 * the smallest amount of data loss possible occurs.
 *
 * Return 0 on success or -1 on error with errno set to the error code.
 * The following error codes are defined:
 *	EINVAL	- Invalid arguments were passed to the function.
 *	EBUSY	- Inode and/or one of its extents is busy, try again later.
 *	EIO	- I/O error while writing the inode (or one of its extents).
 */
int ntfs_inode_sync(int *p_errno, ntfs_inode *ni)
{
	int ret = 0;
	int err = 0;
       MFT_RECORD *ni_mrec;

       if(!ni->map_na)
        {
            printk("No Addr inode when writing\n");
    		NInoClearDirty(ni);
    		return 0;
        }
        
       
       if(NInoNTFSAttrNI(ni))
        {
               NTFS3G_BUG_TRACE(!ni->ntfs_mft_base_ni);
               return ntfs_inode_sync(p_errno, ni->ntfs_mft_base_ni);
        }
       NTFS3G_BUG_TRACE(!ni->map_na);
       
	if (!ni) {
		*p_errno = EINVAL;
		ntfs_log_error(ni->vol->sb,"Failed to sync NULL inode\n");
		return -1;
	}

	ntfs_log_enter("Entering for inode %lld\n", (long long)ni->mft_no);
       if((ni_mrec=map_mft_record(ni))== NULL)
        {
               ntfs_log_perror(ni->vol->sb, "failed to map mft record for %lld", ni->mft_no);
               *p_errno = EIO;
               return -1;
        }
	/* Update STANDARD_INFORMATION. */
	if ((ni_mrec->flags & MFT_RECORD_IN_USE) && ni->base_ni_attr.nr_extents != -1 &&
			ntfs_inode_sync_standard_information(p_errno, ni)) {
		if (!err || *p_errno == EIO) {
			err = *p_errno;
			if (err != EIO)
				err = EBUSY;
		}
		ntfs_log_perror("Failed to sync standard info (inode %lld)",
				(long long)ni->mft_no);
	}

	/* Update FILE_NAME's in the index. */
       if(!(ni_mrec->flags & MFT_RECORD_IN_USE))
             NInoFileNameTestAndClearDirty(ni);
	else if ( ni->base_ni_attr.nr_extents != -1 &&
			NInoFileNameTestAndClearDirty(ni) &&
			ntfs_inode_sync_file_name(p_errno, ni)) {
		if (!err || *p_errno == EIO) {
			err = *p_errno;
			if (err != EIO)
				err = EBUSY;
		}
		ntfs_log_perror("Failed to sync FILE_NAME (inode %lld)",
				(long long)ni->mft_no);
		NInoFileNameSetDirty(ni);
	}

	/* Write out attribute list from cache to disk. */
        if(!(ni_mrec->flags & MFT_RECORD_IN_USE)) 
               NInoAttrListTestAndClearDirty(ni);
	 else if (ni->base_ni_attr.nr_extents != -1 &&
			NInoAttrList(ni) && NInoAttrListTestAndClearDirty(ni)) {
		ntfs_attr *na;

		na = ntfs_attr_open(p_errno, ni, AT_ATTRIBUTE_LIST, AT_UNNAMED, 0);
		if (!na) {
			if (!err || *p_errno == EIO) {
				err = *p_errno;
				if (err != EIO)
					err = EBUSY;
				ntfs_log_perror("Attribute list sync failed "
						"(open, inode %lld)",
						(long long)ni->mft_no);
			}
			NInoAttrListSetDirty(ni);
			goto sync_inode;
		} 
		
		if (na->data_size == ni->base_ni_attr.attr_list_size) {
			if (ntfs_attr_pwrite(p_errno, na, 0, ni->base_ni_attr.attr_list_size,
				        ni->base_ni_attr.attr_list) != ni->base_ni_attr.attr_list_size) {
				if (!err || *p_errno == EIO) {
					err = *p_errno;
					if (err != EIO)
						err = EBUSY;
					ntfs_log_perror("Attribute list sync "
						"failed (write, inode %lld)",
						(long long)ni->mft_no);
				}
				NInoAttrListSetDirty(ni);
			}
		} else {
			err = EIO;
			ntfs_log_error("Attribute list sync failed (bad size, "
				       "inode %lld)\n", (long long)ni->mft_no);
			NInoAttrListSetDirty(ni);
		}
		ntfs_attr_close(p_errno, na);
	}
	
sync_inode:
	/* Write this inode out to the $MFT (and $MFTMirr if applicable). */
	if (NInoTestAndClearDirty(ni)) {
      #if 0 //@@track
		if (ntfs_mft_record_write(p_errno, ni->vol, ni->mft_no, ni_mrec)) {
			if (!err || *p_errno == EIO) {
				err = *p_errno;
				if (err != EIO)
					err = EBUSY;
			}
			NInoSetDirty(ni);
			ntfs_log_perror("MFT record sync failed, inode %lld",
					(long long)ni->mft_no);
		}
      #endif
	}

	/* If this is a base inode with extents write all dirty extents, too. */
	if (ni->base_ni_attr.nr_extents > 0) {
		s32 i;

		for (i = 0; i < ni->base_ni_attr.nr_extents; ++i) {
			ntfs_inode *eni;
                     //MFT_RECORD *eni_mrec;

			eni = ni->base_ni_attr.extent_nis[i];
			if (!NInoTestAndClearDirty(eni))
				continue;      
       #if 0 //@@track
			if((eni_mrec=map_mft_record(eni))== NULL)
                     {
                           ntfs_log_perror("failed to map mft record for %lld", ni->mft_no);
                             *p_errno = EIO;
                     }   
    
			if (!eni_mrec || ntfs_mft_record_write(p_errno, eni->vol, eni->mft_no, 
						  eni_mrec)) {
				if (!err || *p_errno == EIO) {
					err = *p_errno;
					if (err != EIO)
						err = EBUSY;
				}
				NInoSetDirty(eni);
				ntfs_log_perror("Extent MFT record sync failed,"
						" inode %lld/%lld",
						(long long)ni->mft_no,
						(long long)eni->mft_no);   
              
			}  
                     if(eni_mrec)
                         unmap_mft_record(eni);
        #endif
                
		}
	}

	if (err) {
		*p_errno = err;
		ret = -1;
	}
	unmap_mft_record(ni);
	ntfs_log_leave("\n");
	return ret;
}

/**
 * ntfs_inode_add_attrlist - add attribute list to inode and fill it
 * @ni: opened ntfs inode to which add attribute list
 *
 * Return 0 on success or -1 on error with errno set to the error code.
 * The following error codes are defined:
 *	EINVAL	- Invalid arguments were passed to the function.
 *	EEXIST	- Attribute list already exist.
 *	EIO	- Input/Ouput error occurred.
 *	ENOMEM	- Not enough memory to perform add.
 */
int ntfs_inode_add_attrlist(int *p_errno, ntfs_inode *ni)
{
	int err;
	ntfs_attr_search_ctx *ctx;
	u8 *al = NULL, *aln;
	int al_len = 0;
	ATTR_LIST_ENTRY *ale = NULL;
	ntfs_attr *na;
       MFT_RECORD *ni_mrec;
       
	if (!ni) {
		*p_errno = EINVAL;
		ntfs_log_perror("%s", __FUNCTION__);
		return -1;
	}   
    
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
    
       ni_mrec = map_mft_record(ni);
       if(ni_mrec == NULL)
      {
		*p_errno = EIO;
		ntfs_log_perror("%s", __FUNCTION__);
		return -1;
	} 

	ntfs_log_trace("inode %llu\n", (unsigned long long) ni->mft_no);

	if (NInoAttrList(ni) || ni->base_ni_attr.nr_extents) {
		*p_errno = EEXIST;
		ntfs_log_perror("Inode already has attribute list");
		return -1;
	}

	/* Form attribute list. */
	ctx = ntfs_attr_get_search_ctx(p_errno, ni, NULL);
	if (!ctx) {
		err = *p_errno;
		goto err_out;
	}
	/* Walk through all attributes. */
	while (!ntfs_attr_lookup(p_errno, AT_UNUSED, NULL, 0, 0, 0, NULL, 0, ctx)) {
		
		int ale_size;
		
		if (ctx->attr->type == AT_ATTRIBUTE_LIST) {
			err = EIO;
			ntfs_log_perror("Attribute list already present");
			goto put_err_out;
		}
		
		ale_size = (sizeof(ATTR_LIST_ENTRY) + sizeof(ntfschar) *
					ctx->attr->name_length + 7) & ~7;
		al_len += ale_size;
		
		aln = ntfs_realloc_nofs(p_errno, al, al_len, al_len-ale_size);
		if (!aln) {
			err = *p_errno;
			ntfs_log_perror("Failed to realloc %d bytes", al_len);
			goto put_err_out;
		}
		ale = (ATTR_LIST_ENTRY *)(aln + ((u8 *)ale - al));
		al = aln;
		
		memset(ale, 0, ale_size);
		
		/* Add attribute to attribute list. */
		ale->type = ctx->attr->type;
		ale->length = cpu_to_le16((sizeof(ATTR_LIST_ENTRY) +
			sizeof(ntfschar) * ctx->attr->name_length + 7) & ~7);
		ale->name_length = ctx->attr->name_length;
		ale->name_offset = (u8 *)ale->name - (u8 *)ale;
		if (ctx->attr->non_resident)
			ale->lowest_vcn = ctx->attr->lowest_vcn;
		else
			ale->lowest_vcn = 0;
		ale->mft_reference = MK_LE_MREF(ni->mft_no,
			le16_to_cpu(ni_mrec->sequence_number));
		ale->instance = ctx->attr->instance;
		memcpy(ale->name, (u8 *)ctx->attr +
				le16_to_cpu(ctx->attr->name_offset),
				ctx->attr->name_length * sizeof(ntfschar));
		ale = (ATTR_LIST_ENTRY *)(al + al_len);
	}
	/* Check for real error occurred. */
	if (*p_errno != ENOENT) {
		err = *p_errno;
		ntfs_log_perror("%s: Attribute lookup failed, inode %lld",
				__FUNCTION__, (long long)ni->mft_no);
		goto put_err_out;
	}

	/* Set in-memory attribute list. */
	ni->base_ni_attr.attr_list = al;
	ni->base_ni_attr.attr_list_size = al_len;
	NInoSetAttrList(ni);
	NInoAttrListSetDirty(ni);

	/* Free space if there is not enough it for $ATTRIBUTE_LIST. */
	if (le32_to_cpu(ni_mrec->bytes_allocated) -
			le32_to_cpu(ni_mrec->bytes_in_use) <
			offsetof(ATTR_RECORD, resident_end)) {
		if (ntfs_inode_free_space(p_errno, ni,
				offsetof(ATTR_RECORD, resident_end))) {
			/* Failed to free space. */
			err = *p_errno;
			ntfs_log_perror("Failed to free space for attrlist");
			goto rollback;
		}
	}

	/* Add $ATTRIBUTE_LIST to mft record. */
	if (ntfs_resident_attr_record_add(p_errno, ni,
				AT_ATTRIBUTE_LIST, NULL, 0, NULL, 0, 0) < 0) {
		err = *p_errno;
		ntfs_log_perror("Couldn't add $ATTRIBUTE_LIST to MFT");
		goto rollback;
	}

	/* Resize it. */
	na = ntfs_attr_open(p_errno, ni, AT_ATTRIBUTE_LIST, AT_UNNAMED, 0);
	if (!na) {
		err = *p_errno;
		ntfs_log_perror("Failed to open just added $ATTRIBUTE_LIST");
		goto remove_attrlist_record;
	}
	if (ntfs_attr_truncate(p_errno, na, al_len)) {
		err = *p_errno;
		ntfs_log_perror("Failed to resize just added $ATTRIBUTE_LIST");
		ntfs_attr_close(p_errno, na);
		goto remove_attrlist_record;;
	}
	
	ntfs_attr_put_search_ctx(p_errno, ctx);
	ntfs_attr_close(p_errno, na);
       unmap_mft_record(ni);
	return 0;

remove_attrlist_record:
	/* Prevent ntfs_attr_recorm_rm from freeing attribute list. */
	ni->base_ni_attr.attr_list = NULL;
	NInoClearAttrList(ni);
	/* Remove $ATTRIBUTE_LIST record. */
	ntfs_attr_reinit_search_ctx(p_errno, ctx);
	if (!ntfs_attr_lookup(p_errno, AT_ATTRIBUTE_LIST, NULL, 0,
				CASE_SENSITIVE, 0, NULL, 0, ctx)) {
		if (ntfs_attr_record_rm(p_errno, ctx))
			ntfs_log_perror("Rollback failed to remove attrlist");
	} else
		ntfs_log_perror("Rollback failed to find attrlist");
	/* Setup back in-memory runlist. */
	ni->base_ni_attr.attr_list = al;
	ni->base_ni_attr.attr_list_size = al_len;
	NInoSetAttrList(ni);
rollback:
	/*
	 * Scan attribute list for attributes that placed not in the base MFT
	 * record and move them to it.
	 */
	ntfs_attr_reinit_search_ctx(p_errno, ctx);
	ale = (ATTR_LIST_ENTRY*)al;
	while ((u8*)ale < al + al_len) {
		if (MREF_LE(ale->mft_reference) != ni->mft_no) {
			if (!ntfs_attr_lookup(p_errno, ale->type, ale->name,
						ale->name_length,
						CASE_SENSITIVE,
						sle64_to_cpu(ale->lowest_vcn),
						NULL, 0, ctx)) {
				if (ntfs_attr_record_move_to(p_errno, ctx, ni))
					ntfs_log_perror("Rollback failed to "
							"move attribute");
			} else
				ntfs_log_perror("Rollback failed to find attr");
			ntfs_attr_reinit_search_ctx(p_errno, ctx);
		}
		ale = (ATTR_LIST_ENTRY*)((u8*)ale + le16_to_cpu(ale->length));
	}
	/* Remove in-memory attribute list. */
	ni->base_ni_attr.attr_list = NULL;
	ni->base_ni_attr.attr_list_size = 0;
	NInoClearAttrList(ni);
	NInoAttrListClearDirty(ni);
put_err_out:
	ntfs_attr_put_search_ctx(p_errno, ctx);
err_out:
	ntfs_free(al);
	*p_errno = err;
       unmap_mft_record(ni);
	return -1;
}

/**
 * ntfs_inode_free_space - free space in the MFT record of inode
 * @ni:		ntfs inode in which MFT record free space
 * @size:	amount of space needed to free
 *
 * Return 0 on success or -1 on error with errno set to the error code.
 */
int ntfs_inode_free_space(int *p_errno, ntfs_inode *ni, int size)
{
	ntfs_attr_search_ctx *ctx;
	int freed, err;
       MFT_RECORD *ni_mrec;

	if (!ni || size < 0) {
		*p_errno = EINVAL;
		ntfs_log_perror("%s: ni=%p size=%d", __FUNCTION__, ni, size);
		return -1;
	}
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
	ntfs_log_trace("Entering for inode %lld, size %d\n",
		       (unsigned long long)ni->mft_no, size);
    
       ni_mrec = map_mft_record(ni);
       if(ni_mrec == NULL)
      {
		*p_errno = EIO;
		ntfs_log_perror("%s", __FUNCTION__);
		return -1;
	} 
       
	freed = (le32_to_cpu(ni_mrec->bytes_allocated) -
				le32_to_cpu(ni_mrec->bytes_in_use));

       unmap_mft_record(ni);

	if (size <= freed)
		return 0;

	ctx = ntfs_attr_get_search_ctx(p_errno, ni, NULL);
	if (!ctx)
		return -1;

	/*
	 * Chkdsk complain if $STANDARD_INFORMATION is not in the base MFT
	 * record. FIXME: I'm not sure in this, need to recheck. For now simply
	 * do not move $STANDARD_INFORMATION at all.
	 *
	 * Also we can't move $ATTRIBUTE_LIST from base MFT_RECORD, so position
	 * search context on first attribute after $STANDARD_INFORMATION and
	 * $ATTRIBUTE_LIST.
	 *
	 * Why we reposition instead of simply skip this attributes during
	 * enumeration? Because in case we have got only in-memory attribute
	 * list ntfs_attr_lookup will fail when it will try to find
	 * $ATTRIBUTE_LIST.
	 */
	if (ntfs_attr_lookup(p_errno, AT_FILE_NAME, NULL, 0, CASE_SENSITIVE, 0, NULL,
				0, ctx)) {
		if (*p_errno != ENOENT) {
			err = *p_errno;
			ntfs_log_perror("%s: attr lookup failed #2", __FUNCTION__);
			goto put_err_out;
		}
		if (ctx->attr->type == AT_END) {
			err = ENOSPC;
			goto put_err_out;
		}
	}

	while (1) {
		int record_size;

		/*
		 * Check whether attribute is from different MFT record. If so,
		 * find next, because we don't need such.
		 */
		while (ctx->ntfs_ino->mft_no != ni->mft_no) {
retry:			
			if (ntfs_attr_lookup(p_errno, AT_UNUSED, NULL, 0, CASE_SENSITIVE,
						0, NULL, 0, ctx)) {
				err = *p_errno;
				if (*p_errno != ENOENT) {
					ntfs_log_perror("Attr lookup failed #2");
				} else
					err = ENOSPC;
				goto put_err_out;
			}
		}

		if (ntfs_inode_base(p_errno,ctx->ntfs_ino)->mft_no == FILE_MFT && 
		    ctx->attr->type == AT_DATA)
			goto retry;

		record_size = le32_to_cpu(ctx->attr->length);

		/* Move away attribute. */
		if (ntfs_attr_record_move_away(p_errno, ctx, 0)) {
			err = *p_errno;
			ntfs_log_perror("Failed to move out attribute #2");
			break;
		}
		freed += record_size;

		/* Check whether we done. */
		if (size <= freed) {
			ntfs_attr_put_search_ctx(p_errno, ctx);
			return 0;
		}

		/*
		 * Reposition to first attribute after $STANDARD_INFORMATION and
		 * $ATTRIBUTE_LIST (see comments upwards).
		 */
		ntfs_attr_reinit_search_ctx(p_errno, ctx);
		if (ntfs_attr_lookup(p_errno, AT_FILE_NAME, NULL, 0, CASE_SENSITIVE, 0,
				NULL, 0, ctx)) {
			if (*p_errno != ENOENT) {
				err = *p_errno;
				ntfs_log_perror("Attr lookup #2 failed");
				break;
			}
			if (ctx->attr->type == AT_END) {
				err = ENOSPC;
				break;
			}
		}
	}
put_err_out:
	ntfs_attr_put_search_ctx(p_errno, ctx);
	if (err == ENOSPC)
		ntfs_log_trace("No attributes left that can be moved out.\n");
	*p_errno = err;
	return -1;
}

/**
 * ntfs_inode_update_times - update selected time fields for ntfs inode
 * @ni:		ntfs inode for which update time fields
 * @mask:	select which time fields should be updated
 *
 * This function updates time fields to current time. Fields to update are
 * selected using @mask (see enum @ntfs_time_update_flags for posssible values).
 */
void ntfs_inode_update_times(int *p_errno, ntfs_inode *ni, ntfs_time_update_flags mask)
{
	time_t now;

       if(ni && NInoNTFSAttrNI(ni))
        {
            NTFS3G_BUG_TRACE(!ni->ntfs_mft_base_ni);
            ni = ni->ntfs_mft_base_ni;
        }

	if (!ni) {
		ntfs_log_error("%s(): Invalid arguments.\n", __FUNCTION__);
		return;
	}

	if ((ni->mft_no < FILE_first_user && ni->mft_no != FILE_root) ||
			NVolReadOnly(ni->vol) || !mask)
		return;

	 now = get_seconds();
	if (mask & NTFS_UPDATE_ATIME)
		ni->base_ni_attr.last_access_time = now;
	if (mask & NTFS_UPDATE_MTIME)
		ni->base_ni_attr.last_data_change_time = now;
	if (mask & NTFS_UPDATE_CTIME)
		ni->base_ni_attr.last_mft_change_time = now;
	
	NInoFileNameSetDirty(ni);
	NInoSetDirty(ni);
}

/**
 * ntfs_inode_badclus_bad - check for $Badclus:$Bad data attribute
 * @mft_no:		mft record number where @attr is present
 * @attr:		attribute record used to check for the $Bad attribute
 *
 * Check if the mft record given by @mft_no and @attr contains the bad sector
 * list. Please note that mft record numbers describing $Badclus extent inodes
 * will not match the current $Badclus:$Bad check.
 * 
 * On success return 1 if the file is $Badclus:$Bad, otherwise return 0.
 * On error return -1 with errno set to the error code.
 */
int ntfs_inode_badclus_bad(int *p_errno, u64 mft_no, ATTR_RECORD *attr)
{
	int len, ret = 0;
	ntfschar *ustr;

	if (!attr) {
		ntfs_log_error(ni->vol->sb, "Invalid argument.\n");
		*p_errno = EINVAL;
		return -1;
	}
	
	if (mft_no != FILE_BadClus)
	       	return 0;

	if (attr->type != AT_DATA)
	       	return 0;

	if ((ustr = ntfs_str2ucs(p_errno, "$Bad", &len)) == NULL) {
		ntfs_log_perror("Couldn't convert '$Bad' to Unicode");
		return -1;
	}

	if (ustr && ntfs_names_are_equal(ustr, len,
			(ntfschar *)((u8 *)attr + le16_to_cpu(attr->name_offset)),
			attr->name_length, 0, NULL, 0))
		ret = 1;

	ntfs_ucsfree(ustr);

	return ret;
}
s64 ntfs_inode_map_read(int *p_errno, struct inode *vi, s64 pos, s64 count, u8 *b)
{
       u32 offset = 0;
       s64 t;
	u32 br, total = 0;

    	u32 index;
	struct address_space *mapping;
       struct page *page;
       u8 *kaddr;
       s64 isize;
       ntfs_inode *ni;

	if (!vi  || !b || pos < 0 || count < 0) {
		*p_errno = EINVAL;
		ntfs_log_perror("%s:   b=%p  pos=%lld  count=%lld",
				__FUNCTION__,  b, (long long)pos,
				(long long)count);
		return -1;
	}
       ni = NTFS_I(vi);
       isize =  i_size_read(vi);
       if(pos > isize)
      {
		*p_errno = EINVAL;
		ntfs_log_perror("%s:   b=%p  pos=%lld  count=%lld",
				__FUNCTION__,  b, (long long)pos,
				(long long)count);
		return -1;
	}
       if(pos+count >  isize)
           count =  isize-pos;
    
      	mapping = vi->i_mapping;

       index = (u32)(pos/PAGE_CACHE_SIZE);
       
       while(count)
        {
		/* Update @index and get the next page. */
		page = ntfs_map_page(mapping, index);
		if (IS_ERR(page))
			break;
              kaddr = page_address(page);
              t = pos;
              offset = do_div(t, PAGE_CACHE_SIZE);
              br = PAGE_CACHE_SIZE-offset;
              if(br > count)
                  br = count;
             //to avoid race condition with mst write page, we lock buffer here for mst protected page
             if(NInoAttrMstProtected(ni))
             {
                    if(ntfs_lock_page_buffers(ni->vol, page, offset, br))
                    {
                           //err
                            ntfs_unmap_page(page);
                            break;
                    }
             }
              memcpy(b, kaddr+offset, br);    
              if(NInoAttrMstProtected(ni))
                    ntfs_unlock_page_buffers(ni->vol, page, offset, br);
              total += br;
              b += br;
              pos +=br;
              count -= br;
              ntfs_unmap_page(page);
              index++;
	}
       if(count && total == 0)
       {
              *p_errno = EIO;
              return -1;
       }
	return total;
}
static inline int ntfs_expand_na_space(int *p_errno, struct inode *vi,  s64 pos, s64 count,
    s64*fill_zero_start, s64*fill_zero_length)
{
       s64 old_data_size;
       ntfs_inode *vfs_base_ni;
       ntfs_attr_search_ctx *ctx = NULL;
       ntfs_attr *map_na;
       map_na = NTFS_I(vi)->map_na;

       NTFS3G_BUG_TRACE(!map_na);

       NTFS3G_BUG_TRACE(NTFS_I(vi) != map_na->ni);
       
       *fill_zero_length = 0;
       *fill_zero_start = 0;
       
       NTFS3G_BUG_TRACE(map_na->initialized_size > map_na->data_size);
       NTFS3G_BUG_TRACE(!map_na->ni);

       if(pos+count <= map_na->initialized_size)
             return 0;
	if (NAttrEncrypted(map_na) && NAttrNonResident(map_na)) {
		*p_errno = EACCES;
		return -1;
	}
	/* If this is a compressed attribute it needs special treatment. */
	if (NAttrCompressed(map_na)) {
		// TODO: Implement writing compressed attributes! (AIA)
		// return ntfs_attr_pwrite_compressed(ntfs_attr *na,
		//		const s64 pos, s64 count, void *b);
		*p_errno = EOPNOTSUPP;
		return -1;
	}
	old_data_size = map_na->data_size;
	if (pos + count > map_na->data_size) {
		if (ntfs_attr_truncate(p_errno, map_na, pos + count)) {
			ntfs_log_perror("Failed to enlarge attribute");
			return -1;
		}
	}


	if (!NAttrNonResident(map_na)) {
              NTFS3G_BUG_TRACE(map_na->initialized_size!=map_na->data_size);
              if(likely(vi->i_size < map_na->data_size))
                     i_size_write(vi, map_na->data_size);
               return 0;
	}
    
       NTFS3G_BUG_TRACE(vi->i_size != map_na->initialized_size);
       
       if(NInoNTFSAttrNI(map_na->ni))
       {
            NTFS3G_BUG_TRACE(!map_na->ni->ntfs_mft_base_ni);
            vfs_base_ni = map_na->ni->ntfs_mft_base_ni;
            NTFS3G_BUG_TRACE(NInoNTFSAttrNI(vfs_base_ni));
       }
       else
            vfs_base_ni = map_na->ni;
	/* Handle writes beyond initialized_size. */
	if (pos + count > map_na->initialized_size) {
		/* Set initialized_size to @pos + @count. */
		ctx = ntfs_attr_get_search_ctx(p_errno, vfs_base_ni, NULL);
		if (!ctx)
		       return -1;
		if (ntfs_attr_lookup(p_errno, map_na->type, map_na->name, map_na->name_len, 0,
				0, NULL, 0, ctx))
		{
			ntfs_attr_put_search_ctx(p_errno, ctx);
                     return -1;
		}
		
		/* If write starts beyond initialized_size, zero the gap. */
		if (pos > map_na->initialized_size)
		{
		        *fill_zero_start = map_na->initialized_size;
                      *fill_zero_length = pos - map_na->initialized_size;
		}
			
		ctx->attr->initialized_size = cpu_to_sle64(pos + count);
              mark_mft_record_dirty(ctx->ntfs_ino);
		map_na->initialized_size = pos + count; 
		ntfs_attr_put_search_ctx(p_errno, ctx);
              NTFS3G_BUG_TRACE(vi->i_size >= map_na->initialized_size);
              i_size_write(vi, map_na->initialized_size);
	}
    return 0;
}
s64 ntfs_inode_map_write(int *p_errno, struct inode *vi, s64 pos, s64 count, u8 *b)
{
       u32 offset = 0;
       s64 t;
	u32 br, total = 0;
    	u32 index;
	struct address_space *mapping;
       struct page *page;
       u8 *kaddr;
       s64 isize;
       s64 fill_zero_start, fill_zero_length;
       ntfs_attr *map_na;
       unsigned bh_size = vi->i_sb->s_blocksize;
       ntfs_inode *ni;

	if (!vi  || !b || pos < 0 || count < 0) {
		*p_errno = EINVAL;
		ntfs_log_perror("%s:   b=%p  pos=%lld  count=%lld",
				__FUNCTION__,  b, (long long)pos,
				(long long)count);
		return -1;
	}
       ni = NTFS_I(vi);

       NTFS3G_BUG_TRACE(ni == ni->vol->mft_dat_ni);
	  
       if(ntfs_expand_na_space(p_errno, vi,  pos, count,
                                           &fill_zero_start, &fill_zero_length))
       {
       	ntfs_log_perror("%s:  explan inode for write b=%p  pos=%lld  count=%lld failed",
				__FUNCTION__,  b, (long long)pos,
				(long long)count);
             return -1;
       }
       map_na = NTFS_I(vi)->map_na;
       
       mapping = vi->i_mapping;
      
       while(fill_zero_length)
       {
              index = (u32)(fill_zero_start/PAGE_CACHE_SIZE);
		/* Update @index and get the next page. */
		page = ntfs_map_page(mapping, index);
		if (IS_ERR(page))
			break;
              kaddr = page_address(page);
              t = fill_zero_start;
              offset = do_div(t, PAGE_CACHE_SIZE);
              br = PAGE_CACHE_SIZE-offset;
              if(br > fill_zero_length)
                  br = fill_zero_length;
              memset(kaddr+offset, 0, br);
		lock_page(page);
              set_page_dirty(page);
              if (unlikely(page_has_buffers(page))) 
              {
                       struct buffer_head *bh, *head;
                        unsigned int  bh_ofs;
		          bh = head = page_buffers(page);
	                 NTFS3G_BUG_TRACE(!bh);
	                 do {
		                bh_ofs = bh_offset(bh);
		               if (bh_ofs + bh_size <= offset)
			             continue;
		             if (unlikely(bh_ofs >= offset+br))
			             break;
		             set_buffer_dirty(bh);
	                 } while ((bh = bh->b_this_page) != head);
              }	
		unlock_page(page);
              fill_zero_start +=br;
              fill_zero_length -= br;

              ntfs_unmap_page(page);
	}

       isize =  i_size_read(vi);
       NTFS3G_BUG_TRACE(pos+count>map_na->data_size);
        
       index = (u32)(pos/PAGE_CACHE_SIZE);
       while(count)
        {
		/* Update @index and get the next page. */
		page = ntfs_map_page(mapping, index);
		if (IS_ERR(page))
			break;

              kaddr = page_address(page);
              t = pos;
              offset = do_div(t, PAGE_CACHE_SIZE);
              br = PAGE_CACHE_SIZE-offset;
              if(br > count)
                  br = count;
             //to avoid race condition with mst write page, we lock buffer here for mst protected page
             if(NInoAttrMstProtected(ni))
             {
                    if(ntfs_lock_page_buffers(ni->vol, page, offset, br))
                    {
                           //err
                            ntfs_unmap_page(page);
                            break;
                    }
             }
              memcpy(kaddr+offset, b, br);
                           //to avoid race condition with mst write page, we lock buffer here for mst protected page
             if(NInoAttrMstProtected(ni))
                    ntfs_unlock_page_buffers(ni->vol, page, offset, br);
              total += br;
              b += br;
              pos +=br;
              count -= br;
		lock_page(page);
              set_page_dirty(page);
              if (likely(page_has_buffers(page))) 
              {
                       struct buffer_head *bh, *head;
                        unsigned int  bh_ofs;
		          bh = head = page_buffers(page);
	                 NTFS3G_BUG_TRACE(!bh);
	                 do {
		                bh_ofs = bh_offset(bh);
		               if (bh_ofs + bh_size <= offset)
			             continue;
		             if (unlikely(bh_ofs >= offset+br))
			             break;
		             set_buffer_dirty(bh);
	                 } while ((bh = bh->b_this_page) != head);
              }	
		unlock_page(page);
              ntfs_unmap_page(page);
              index++;
	}
        
       if(count && total == 0)
        {
           *p_errno = EIO;
           return -1;
        }
	return total;
}
void ntfs_sync_inode_size(ntfs_inode *ni, ntfs_attr *map_na)
{ 
      struct inode *vi;
      NTFS3G_BUG_TRACE(!NInoNTFSAttrNI(ni) && ni->base_ni_attr.nr_extents < 0);
      NTFS3G_BUG_TRACE(ni->map_na != map_na);

      vi = VFS_I(ni);
      if(vi->i_size != (loff_t)map_na->initialized_size)
                   i_size_write(vi, (loff_t)map_na->initialized_size);
}
