/*
 * inode.h - Defines for NTFS inode handling. Originated from the Linux-NTFS project.
 *
 * Copyright (c) 2001-2004 Anton Altaparmakov
 * Copyright (c) 2004-2007 Yura Pakhuchiy
 * Copyright (c) 2004-2005 Richard Russon
 * Copyright (c) 2006-2008 Szabolcs Szakacsits
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the NTFS-3G
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _NTFS_INODE_H
#define _NTFS_INODE_H

/* Forward declaration */
typedef struct _ntfs_inode ntfs_inode;

#include <asm/atomic.h>

#include <linux/fs.h>
#include <linux/list.h>
#include <linux/mm.h>
#include <linux/mutex.h>
#include <linux/seq_file.h>

#include "types.h"
#include "layout.h"
#include "support.h"
#include "volume.h"
#include "attrib.h"

/**
 * enum ntfs_inode_state_bits -
 *
 * Defined bits for the state field in the ntfs_inode structure.
 * (f) = files only, (d) = directories only
 */
typedef enum {
	NI_Dirty,		/* 1: Mft record needs to be written to disk. */

	/* The NI_AttrList* tests only make sense for base inodes. */
	NI_AttrList,		/* 1: Mft record contains an attribute list. */
	NI_AttrListDirty,	/* 1: Attribute list needs to be written to the
				      mft record and then to disk. */
	NI_FileNameDirty,	/* 1: FILE_NAME attributes need to be updated
				      in the index. */
} ntfs_inode_state_bits;

#define  test_nino_flag(ni, flag)	   ntfs3g_test_bit(NI_##flag, (NInoNTFSAttrNI(ni)?ni->ntfs_mft_base_ni:ni)->base_ni_attr.vfs_baseni_state)
#define   set_nino_flag(ni, flag)	    ntfs3g_set_bit(NI_##flag, (NInoNTFSAttrNI(ni)?ni->ntfs_mft_base_ni:ni)->base_ni_attr.vfs_baseni_state)
#define clear_nino_flag(ni, flag)	  ntfs3g_clear_bit(NI_##flag, (NInoNTFSAttrNI(ni)?ni->ntfs_mft_base_ni:ni)->base_ni_attr.vfs_baseni_state)

#define test_and_set_nino_flag(ni, flag)	\
				   ntfs3g_test_and_set_bit(NI_##flag, (ni)->base_ni_attr.vfs_baseni_state)
#define test_and_clear_nino_flag(ni, flag)	\
				 ntfs3g_test_and_clear_bit(NI_##flag, (ni)->base_ni_attr.vfs_baseni_state)

#define NInoDirty(ni)				  test_nino_flag(ni, Dirty)
#define NInoSetDirty(ni)			   set_nino_flag(ni, Dirty)
#define NInoClearDirty(ni)			 clear_nino_flag(ni, Dirty)
#define NInoTestAndSetDirty(ni)		  test_and_set_nino_flag(ni, Dirty)
#define NInoTestAndClearDirty(ni)	test_and_clear_nino_flag(ni, Dirty)

#define NInoAttrList(ni)			  test_nino_flag(ni, AttrList)
#define NInoSetAttrList(ni)			   set_nino_flag(ni, AttrList)
#define NInoClearAttrList(ni)			 clear_nino_flag(ni, AttrList)


#define  test_nino_al_flag(ni, flag)	 test_nino_flag(ni, AttrList##flag)
#define   set_nino_al_flag(ni, flag)	  set_nino_flag(ni, AttrList##flag)
#define clear_nino_al_flag(ni, flag)	clear_nino_flag(ni, AttrList##flag)

#define test_and_set_nino_al_flag(ni, flag)	\
				 test_and_set_nino_flag(ni, AttrList##flag)
#define test_and_clear_nino_al_flag(ni, flag)	\
			       test_and_clear_nino_flag(ni, AttrList##flag)

#define NInoAttrListDirty(ni)			    test_nino_al_flag(ni, Dirty)
#define NInoAttrListSetDirty(ni)		     set_nino_al_flag(ni, Dirty)
#define NInoAttrListClearDirty(ni)		   clear_nino_al_flag(ni, Dirty)
#define NInoAttrListTestAndSetDirty(ni)	    test_and_set_nino_al_flag(ni, Dirty)
#define NInoAttrListTestAndClearDirty(ni) test_and_clear_nino_al_flag(ni, Dirty)

#define NInoFileNameDirty(ni)                 test_nino_flag(ni, FileNameDirty)
#define NInoFileNameSetDirty(ni)               set_nino_flag(ni, FileNameDirty)
#define NInoFileNameClearDirty(ni)           clear_nino_flag(ni, FileNameDirty)
#define NInoFileNameTestAndSetDirty(ni)		\
				      test_and_set_nino_flag(ni, FileNameDirty)
#define NInoFileNameTestAndClearDirty(ni)	\
				    test_and_clear_nino_flag(ni, FileNameDirty)



/**
 * enum ntfs_inode_ext_state_bits -
 *
 * Defined bits for the ext_state field in the ntfs_inode structure.
 * (f) = files only, (d) = directories only
 */
typedef enum {
	NI_EXT_NTFS_ATTR_NI,		/*is NTFS base vfs inode*/
     	NI_EXT_NTFS_MST_PROTECTED	/* 1: NI is mst fixup protected, need to pre/post fix up when sync disk. */
} ntfs_inode_ext_state_bits;

#define  test_ni_ext_flag(ni, flag)	   ntfs3g_test_bit(NI_EXT_##flag, (ni)->ext_state)
#define   set_ni_ext_flag(ni, flag)	   ntfs3g_set_bit(NI_EXT_##flag,  (ni)->ext_state)
#define clear_ni_ext_flag(ni, flag)	   ntfs3g_clear_bit(NI_EXT_##flag, (ni)->ext_state)

#define NInoNTFSAttrNI(ni)			 test_ni_ext_flag(ni, NTFS_ATTR_NI)
#define NInoSetNTFSAttrNI(ni)		 set_ni_ext_flag(ni, NTFS_ATTR_NI)
#define NInoClearNTFSAttrNI(ni)	        clear_ni_ext_flag(ni, NTFS_ATTR_NI)
#define NInoAttrMstProtected(ni)			 test_ni_ext_flag(ni, NTFS_MST_PROTECTED)
#define NInoSetAttrMstProtected(ni)		 set_ni_ext_flag(ni, NTFS_MST_PROTECTED)
#define NInoClearAttrMstProtected(ni)	        clear_ni_ext_flag(ni, NTFS_MST_PROTECTED)

/**
 * struct _ntfs_inode - The NTFS in-memory inode structure.
 *
 * It is just used as an extension to the fields already provided in the VFS
 * inode.
 */
 typedef struct
{
	unsigned long vfs_baseni_state;	/* NTFS specific flags describing this inode.
				   See ntfs_inode_state_bits above. */
	FILE_ATTR_FLAGS flags;	/* Flags describing the file.
				   (Copy from STANDARD_INFORMATION) */
	/*
	 * Attribute list support (for use by the attribute lookup functions).
	 * Setup during ntfs_open_inode() for all inodes with attribute lists.
	 * Only valid if NI_AttrList is set in state.
	 */
	u32 attr_list_size;	/* Length of attribute list value in bytes. */
	u8 *attr_list;		/* Attribute list value itself. */
	/* Below fields are always valid. */
	s32 nr_extents;		/* For a base mft record, the number of
				   attached extent inodes (0 if none), for
				   extent records this is -1. */
	union {		/* This union is only used if nr_extents != 0. */
		ntfs_inode **extent_nis;/* For nr_extents > 0, array of the
					   ntfs inodes of the extent mft
					   records belonging to this base
					   inode which have been loaded. */
		ntfs_inode *base_ni;	/* For nr_extents == -1, the ntfs
					   inode of the base mft record. */
	};
	/*
	 * These four fields are copy of relevant fields from
	 * STANDARD_INFORMATION attribute and used to sync it and FILE_NAME
	 * attribute in the index.
	 */
	time_t creation_time;
	time_t last_data_change_time;
	time_t last_mft_change_time;
	time_t last_access_time;

       struct page *mft_page;
       u32 mft_page_ofs;
       struct mutex mrec_lock;
       int   mft_page_refcnt;
       int   mft_record_modified;//flag for sync to mirror mft
}VFS_base_ni_attr;
struct _ntfs_inode {
        
	u64 mft_no;		/* Inode / mft record number. */
	ntfs_volume *vol;	/* Pointer to the ntfs volume of this inode. */
       u32  ext_state;

	/* Below fields are valid only for base inode. */

	/*
	 * These two fields are used to sync filename index and guaranteed to be
	 * correct, however value in index itself maybe wrong (windows itself
	 * do not update them properly).
	 */
	u32 fixup_block_size; //ntfs fixup block size which is used in mst, index file & logfile
	ntfs_attr *map_na;
    
       union 
       {
           VFS_base_ni_attr base_ni_attr;
           ntfs_inode *ntfs_mft_base_ni;  //ntfs big or extenal inode which this attribute read from
       };
};

/*
 * The full structure containing a ntfs_inode and a vfs struct inode. Used for
 * all real and fake inodes but not for extent inodes which lack the vfs struct
 * inode.
 */
typedef struct {
	ntfs_inode ntfs_inode;
	struct inode vfs_inode;		/* The vfs inode structure. */
} big_ntfs3g_inode;

/**
 * NTFS_I - return the ntfs inode given a vfs inode
 * @inode:	VFS inode
 *
 * NTFS_I() returns the ntfs inode associated with the VFS @inode.
 */
static inline ntfs_inode *NTFS_I(struct inode *inode)
{
	return (ntfs_inode *)list_entry(inode, big_ntfs3g_inode, vfs_inode);
}

static inline struct inode *VFS_I(ntfs_inode *ni)
{
	return &((big_ntfs3g_inode *)ni)->vfs_inode;
}


typedef enum {
	NTFS_UPDATE_ATIME = 1 << 0,
	NTFS_UPDATE_MTIME = 1 << 1,
	NTFS_UPDATE_CTIME = 1 << 2,
} ntfs_time_update_flags;

#define NTFS_UPDATE_MCTIME  (NTFS_UPDATE_MTIME | NTFS_UPDATE_CTIME)
#define NTFS_UPDATE_AMCTIME (NTFS_UPDATE_ATIME | NTFS_UPDATE_MCTIME)


extern ntfs_inode *ntfs_inode_base(int *p_errno, ntfs_inode *ni);


extern ntfs_inode *ntfs_big_inode_open(int *p_errno, ntfs_volume *vol, const MFT_REF mref,
                          ATTR_TYPES type,ntfschar *name, const u32 name_len);

extern int ntfs_big_inode_close(int *p_errno, ntfs_inode *ni);

extern int ntfs_ext_inode_close(int *p_errno, ntfs_inode *ni);

extern ntfs_inode *ntfs_extent_inode_open(int *p_errno, ntfs_inode *base_ni, const MFT_REF mref);

extern int ntfs_inode_attach_all_extents(int *p_errno, ntfs_inode *ni);

extern void ntfs_inode_mark_dirty(int *p_errno, ntfs_inode *ni);

extern void ntfs_inode_update_times(int *p_errno, ntfs_inode *ni, ntfs_time_update_flags mask);

extern int ntfs_inode_sync(int *p_errno, ntfs_inode *ni);

extern int ntfs_inode_add_attrlist(int *p_errno, ntfs_inode *ni);

extern int ntfs_inode_free_space(int *p_errno, ntfs_inode *ni, int size);

extern int ntfs_inode_badclus_bad(int *p_errno, u64 mft_no, ATTR_RECORD *a);

extern struct inode *ntfs3g_alloc_big_inode(struct super_block *sb);
extern void ntfs3g_destroy_big_inode(struct inode *vi);
extern ntfs_inode *ntfs_alloc_ext_inode(int *p_errno, ntfs_volume *vol);
extern void ntfs_destroy_ext_inode(ntfs_inode *ni);
int ntfs3g_write_inode(struct inode *vi, int sync);
int ntfs_test_inode(struct inode *vi, void *data);
extern s64 ntfs_inode_map_read(int *p_errno, struct inode *vi, s64 pos, s64 count, u8 *b);
extern s64 ntfs_inode_map_write(int *p_errno, struct inode *vi, s64 pos, s64 count, u8 *b);
extern ntfs_inode *ntfs_attr_inode_open(int *p_errno, ntfs_inode *vfs_base_ni,
              ATTR_TYPES type,ntfschar *name, const u32 name_len);
extern int ntfs_attr_inode_close(int *p_errno, ntfs_inode *ni);
extern int ntfs_attr_of_vfs_inode(ATTR_TYPES type,ntfschar *name, const u32 name_len);
extern int ntfs_attr_of_attr_inode(ATTR_TYPES type,ntfschar *name, const u32 name_len);
extern void ntfs_sync_inode_size(ntfs_inode *ni, ntfs_attr *map_na);
#endif /* defined _NTFS_INODE_H */
