#ifndef _NTFS_MISC_H_
#define _NTFS_MISC_H_

#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/highmem.h>

#define TRACE_MEM_LEAK 0

#if TRACE_MEM_LEAK
extern void ntfs_add_track_ptr(void *addr, unsigned size);
extern void ntfs_rm_track_ptr(void *addr);
#else
#define ntfs_add_track_ptr(addr, size) do{}while(0)
#define ntfs_rm_track_ptr(addr) do{}while(0)
#endif

/**
 * __ntfs_malloc - allocate memory in multiples of pages
 * @size:	number of bytes to allocate
 * @gfp_mask:	extra flags for the allocator
 *
 * Internal function.  You probably want ntfs_malloc_nofs()...
 *
 * Allocates @size bytes of memory, rounded up to multiples of PAGE_SIZE and
 * returns a pointer to the allocated memory.
 *
 * If there was insufficient memory to complete the request, return NULL.
 * Depending on @gfp_mask the allocation may be guaranteed to succeed.
 */
static inline void *__ntfs_malloc(int *p_errno, unsigned long size, gfp_t gfp_mask)
{
	if (likely(size <= PAGE_SIZE)) {
		BUG_ON(!size);
		/* kmalloc() has per-CPU caches so is faster for now. */
		return kmalloc(PAGE_SIZE, gfp_mask & ~__GFP_HIGHMEM);
		/* return (void *)__get_free_page(gfp_mask); */
	}
	if (likely(size >> PAGE_SHIFT < num_physpages))
		return __vmalloc(size, gfp_mask, PAGE_KERNEL);
	return NULL;
}

/**
 * ntfs_malloc_nofs - allocate memory in multiples of pages
 * @size:	number of bytes to allocate
 *
 * Allocates @size bytes of memory, rounded up to multiples of PAGE_SIZE and
 * returns a pointer to the allocated memory.
 *
 * If there was insufficient memory to complete the request, return NULL.
 */
static inline void *ntfs_malloc_nofs(int *p_errno, unsigned long size)
{
	void *ptr = __ntfs_malloc(p_errno, size, GFP_NOFS | __GFP_HIGHMEM);
       if(likely(ptr))
        {
           ntfs_add_track_ptr(ptr, size);
           memset(ptr, 0, size);
        }
       else
          *p_errno = ENOMEM;
       return ptr;
}

/**
 * ntfs_malloc_nofs_nofail - allocate memory in multiples of pages
 * @size:	number of bytes to allocate
 *
 * Allocates @size bytes of memory, rounded up to multiples of PAGE_SIZE and
 * returns a pointer to the allocated memory.
 *
 * This function guarantees that the allocation will succeed.  It will sleep
 * for as long as it takes to complete the allocation.
 *
 * If there was insufficient memory to complete the request, return NULL.
 */
static inline void *ntfs_malloc_nofs_nofail(int *p_errno, unsigned long size)
{
	void *ptr = __ntfs_malloc(p_errno, size, GFP_NOFS | __GFP_HIGHMEM | __GFP_NOFAIL);
       if(likely(ptr))
        {
          ntfs_add_track_ptr(ptr, size);
          memset(ptr, 0, size);
        }
       else
          *p_errno = ENOMEM;
       return ptr;
}

static inline void ntfs_free(void *addr)
{  
       if(unlikely(addr == NULL))
               return;
       ntfs_rm_track_ptr(addr);
	if (!is_vmalloc_addr(addr)) {
		kfree(addr);
		/* free_page((unsigned long)addr); */
		return;
	}
	vfree(addr);
}

static inline void *ntfs_realloc_nofs(int *p_errno, void *addr, unsigned long new_size, unsigned long cpy_size)
{
       void* pnew_addr;

        if(new_size == 0)
        {
            ntfs_free(addr);
            return NULL;
        }
        
        pnew_addr = ntfs_malloc_nofs(p_errno, new_size);

       if(pnew_addr == NULL)
            return NULL;
       if(addr)
        {
           cpy_size = min(cpy_size, new_size);
           if(cpy_size)
                memcpy(pnew_addr, addr, cpy_size);
           ntfs_free(addr);
        }
        return pnew_addr;       
}


#endif /* _NTFS_MISC_H_ */

