/*
 * namei.c - NTFS kernel directory inode operations. Part of the Linux-NTFS
 *	     project.
 *
 * Copyright (c) 2001-2006 Anton Altaparmakov
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <linux/stddef.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>	/* For bdev_hardsect_size(). */
#include <linux/backing-dev.h>
#include <linux/buffer_head.h>
#include <linux/vfs.h>
#include <linux/moduleparam.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/mount.h>
#include <linux/exportfs.h>

#include "config.h"
#include "compat.h"
#include "attrib.h"
#include "attrlist.h"
#include "mft.h"
#include "debug.h"
#include "mst.h"
#include "volume.h"
#include "types.h"
#include "layout.h"
#include "inode.h"
#include "runlist.h"
#include "lcnalloc.h"
#include "dir.h"
#include "compress.h"
#include "bitmap.h"
#include "misc.h"

/**
 * ntfs_lookup - find the inode represented by a dentry in a directory inode
 * @dir_ino:	directory inode in which to look for the inode
 * @dent:	dentry representing the inode to look for
 * @nd:		lookup nameidata
 *
 * In short, ntfs_lookup() looks for the inode represented by the dentry @dent
 * in the directory inode @dir_ino and if found attaches the inode to the
 * dentry @dent.
 *
 * In more detail, the dentry @dent specifies which inode to look for by
 * supplying the name of the inode in @dent->d_name.name. ntfs_lookup()
 * converts the name to Unicode and walks the contents of the directory inode
 * @dir_ino looking for the converted Unicode name. If the name is found in the
 * directory, the corresponding inode is loaded by calling ntfs_iget() on its
 * inode number and the inode is associated with the dentry @dent via a call to
 * d_splice_alias().
 *
 * If the name is not found in the directory, a NULL inode is inserted into the
 * dentry @dent via a call to d_add(). The dentry is then termed a negative
 * dentry.
 *
 * Only if an actual error occurs, do we return an error via ERR_PTR().
 *
 * In order to handle the case insensitivity issues of NTFS with regards to the
 * dcache and the dcache requiring only one dentry per directory, we deal with
 * dentry aliases that only differ in case in ->ntfs_lookup() while maintaining
 * a case sensitive dcache. This means that we get the full benefit of dcache
 * speed when the file/directory is looked up with the same case as returned by
 * ->ntfs_readdir() but that a lookup for any other case (or for the short file
 * name) will not find anything in dcache and will enter ->ntfs_lookup()
 * instead, where we search the directory for a fully matching file name
 * (including case) and if that is not found, we search for a file name that
 * matches with different case and if that has non-POSIX semantics we return
 * that. We actually do only one search (case sensitive) and keep tabs on
 * whether we have found a case insensitive match in the process.
 *
 * To simplify matters for us, we do not treat the short vs long filenames as
 * two hard links but instead if the lookup matches a short filename, we
 * return the dentry for the corresponding long filename instead.
 *
 * There are three cases we need to distinguish here:
 *
 * 1) @dent perfectly matches (i.e. including case) a directory entry with a
 *    file name in the WIN32 or POSIX namespaces. In this case
 *    ntfs_lookup_inode_by_name() will return with name set to NULL and we
 *    just d_splice_alias() @dent.
 * 2) @dent matches (not including case) a directory entry with a file name in
 *    the WIN32 namespace. In this case ntfs_lookup_inode_by_name() will return
 *    with name set to point to a kmalloc()ed ntfs_name structure containing
 *    the properly cased little endian Unicode name. We convert the name to the
 *    current NLS code page, search if a dentry with this name already exists
 *    and if so return that instead of @dent.  At this point things are
 *    complicated by the possibility of 'disconnected' dentries due to NFS
 *    which we deal with appropriately (see the code comments).  The VFS will
 *    then destroy the old @dent and use the one we returned.  If a dentry is
 *    not found, we allocate a new one, d_splice_alias() it, and return it as
 *    above.
 * 3) @dent matches either perfectly or not (i.e. we don't care about case) a
 *    directory entry with a file name in the DOS namespace. In this case
 *    ntfs_lookup_inode_by_name() will return with name set to point to a
 *    kmalloc()ed ntfs_name structure containing the mft reference (cpu endian)
 *    of the inode. We use the mft reference to read the inode and to find the
 *    file name in the WIN32 namespace corresponding to the matched short file
 *    name. We then convert the name to the current NLS code page, and proceed
 *    searching for a dentry with this name, etc, as in case 2), above.
 *
 * Locking: Caller must hold i_mutex on the directory.
 */
static struct dentry *vfs_ntfs3g_lookup(struct inode *dir_ino, struct dentry *dent,
		struct nameidata *nd)
{
	ntfs_volume *vol = NTFS3G_SB(dir_ino->i_sb);
	struct inode *dent_inode;
	ntfschar *unicode = NULL;
	MFT_REF mref;
	unsigned long dent_ino;
	int uname_len;
       int err = 0;
       ntfs_inode *ni;
       MFT_RECORD mrec;
        

	ntfs_log_debug("Looking up %s in directory inode 0x%lx.",
			dent->d_name.name, dir_ino->i_ino);
	/* Convert the name of the dentry to Unicode. */
       uname_len = ntfs_mbstoucs(&err, dent->d_name.name, dent->d_name.len, &unicode);
	if (uname_len < 0) {
		if (uname_len != -ENAMETOOLONG)
			ntfs_log_error(vol->sb, "Failed to convert name to "
					"Unicode.");
		return ERR_PTR(uname_len);
	}
    //variable 'err' in ntfs_inode_lookup_by_name alway positive
	mref = ntfs_inode_lookup_by_name(&err, NTFS_I(dir_ino), unicode, uname_len);
       ntfs_free(unicode);
//	kmem_cache_free(ntfs_name_cache, uname);
	if (mref != -1) {
		dent_ino = MREF(mref);
		ntfs_log_debug("Found inode 0x%lx. Calling ntfs_iget.", dent_ino);
              err = 0;
              if(ntfs_inode_map_read(&err, VFS_I(vol->mft_dat_ni), MREF_INO(mref) << vol->mft_record_size_bits,
                                             sizeof(mrec), (u8*)&mrec) != sizeof(mrec))
             {
                 dent_inode = ERR_PTR(-err);
             }
             else
             {
                if(mrec.flags & MFT_RECORD_IS_DIRECTORY)
                {
		    ni = ntfs_big_inode_open(&err, vol, mref, AT_INDEX_ROOT, NTFS_INDEX_I30, 4);
                 }
                else
                {
                  ni = ntfs_big_inode_open(&err, vol, mref, AT_DATA, AT_UNNAMED, 0);
                 }
                 if(ni == NULL)
                       dent_inode = ERR_PTR(-err);
                 else
                 {
                       dent_inode = VFS_I(ni);
                       if(mrec.flags & MFT_RECORD_IS_DIRECTORY)
                               dent_inode->i_mode |= S_IFDIR;
                       else
                              dent_inode->i_mode |= S_IFREG;
                  }
              }
          
		if (likely(!IS_ERR(dent_inode))) {
			/* Consistency check. */
                     #ifdef NTFS3G_FIXME
			if (is_bad_inode(dent_inode) ||MSEQNO(mref) ==
 					MSEQNO(NTFS_I(dent_inode)->mft_no) ||
					dent_ino == FILE_MFT)
			#endif
                      {
				/* Perfect WIN32/POSIX match. -- Case 1. */

				return d_splice_alias(dent_inode, dent);

			}
			ntfs_log_error(vol->sb, "Found stale reference to inode "
					"0x%lx (reference sequence number = "
					"0x%x"
					"returning -EIO. Run chkdsk.",
					dent_ino, MSEQNO(mref));
			iput(dent_inode);
			dent_inode = ERR_PTR(-EIO);
		} else
			ntfs_log_error(vol->sb, "ntfs_iget(0x%lx) failed with "
					"error code %li.", dent_ino,
					PTR_ERR(dent_inode));
		/* Return the error code. */
		return (struct dentry *)dent_inode;
	}
	/* It is guaranteed that @name is no longer allocated at this point. */
	if (err == ENOENT) {
		ntfs_log_debug("Entry was not found, adding negative dentry.");
		/* The dcache will handle negative entries. */
		d_add(dent, NULL);
		ntfs_log_debug("Done.");
		return NULL;
	}
	ntfs_log_error(vol->sb, "ntfs_lookup_ino_by_name() failed with error "
			"code %i.", -err);
	return ERR_PTR(-err);  //err alway positive. infact we must return a negative err 

}

/**
 * ntfs_get_parent - find the dentry of the parent of a given directory dentry
 * @child_dent:		dentry of the directory whose parent directory to find
 *
 * Find the dentry for the parent directory of the directory specified by the
 * dentry @child_dent.  This function is called from
 * fs/exportfs/expfs.c::find_exported_dentry() which in turn is called from the
 * default ->decode_fh() which is export_decode_fh() in the same file.
 *
 * The code is based on the ext3 ->get_parent() implementation found in
 * fs/ext3/namei.c::ext3_get_parent().
 *
 * Note: ntfs_get_parent() is called with @child_dent->d_inode->i_mutex down.
 *
 * Return the dentry of the parent directory on success or the error code on
 * error (IS_ERR() is true).
 */
static struct dentry *ntfs3g_get_parent(struct dentry *child_dent)
{
	struct inode *vi = child_dent->d_inode;
	ntfs_inode *ni = NTFS_I(vi);
	unsigned long parent_ino;
	int err = 0;
    
    	ntfs_log_enter("Entering for inode 0x%lx.", vi->i_ino);
	parent_ino = ntfs_mft_get_parent_ref(&err, ni);
	if (parent_ino == ERR_MREF(-1)) {
		ntfs_log_perror("Parent directory not found");
		ERR_PTR(-err);
	}
       err = 0;
       ni = ntfs_big_inode_open(&err, ni->vol, parent_ino, AT_INDEX_ROOT, NTFS_INDEX_I30, 4);
       if(ni == NULL)
           return ERR_PTR(-err);
	return d_obtain_alias(VFS_I(ni));
}

static struct inode *ntfs3g_nfs_get_inode(struct super_block *sb,
		u64 ino, u32 generation)
{
	struct inode *vi;
       int err = 0;
       ntfs_inode *ni;

	ni = ntfs_big_inode_open(&err, NTFS3G_SB(sb), ino, AT_INDEX_ROOT, NTFS_INDEX_I30, 4);
       if(ni == NULL)
           ni = ntfs_big_inode_open(&err, NTFS3G_SB(sb), ino, AT_DATA, AT_UNNAMED, 0);
       if(ni == NULL)
            vi = ERR_PTR(-err);
       else
            vi = VFS_I(ni);
	if (!IS_ERR(vi)) {
		if (is_bad_inode(vi) || vi->i_generation != generation) {
			iput(vi);
			vi = ERR_PTR(-ESTALE);
		}
	}

	return vi;
}

static struct dentry *ntfs3g_fh_to_dentry(struct super_block *sb, struct fid *fid,
		int fh_len, int fh_type)
{
	return generic_fh_to_dentry(sb, fid, fh_len, fh_type,
				    ntfs3g_nfs_get_inode);
}

static struct dentry *ntfs3g_fh_to_parent(struct super_block *sb, struct fid *fid,
		int fh_len, int fh_type)
{
	return generic_fh_to_parent(sb, fid, fh_len, fh_type,
				    ntfs3g_nfs_get_inode);
}
static int vfs_ntfs3g_create(struct inode *dir, struct dentry *dentry, int mode,
		       struct nameidata *nd)
{
	struct super_block *sb = dir->i_sb;
	struct inode *vi;
	struct timespec ts;
	int err = 0;
       ntfs_inode *ni;
       ntfschar *unicode = NULL;
       int uname_len;
       
       uname_len = ntfs_mbstoucs(&err, dentry->d_name.name, dentry->d_name.len, &unicode);
	if (uname_len < 0) {
		if (uname_len != -ENAMETOOLONG)
			ntfs_log_error(sb, "Failed to convert name to "
					"Unicode.");
		return -ENOMEM;
	}
    
	lock_super(sb);
    
     
       
	ts = CURRENT_TIME_SEC;
       ni = ntfs_create(&err, NTFS_I(dir), unicode, uname_len, S_IFREG);
       ntfs_free(unicode);
       if(ni == NULL)
       {
              err = -err;
		goto out;
       }
       err = 0;
       vi = VFS_I(ni);
	vi->i_version++;
	vi->i_mtime = vi->i_atime = vi->i_ctime = ts;
	/* timestamp is already written, so mark_inode_dirty() is unneeded. */
	dentry->d_time = dentry->d_parent->d_inode->i_version;
	d_instantiate(dentry, vi);
out:
	unlock_super(sb);
	return err;
}


static int vfs_ntfs3g_unlink(struct inode *dir, struct dentry *dentry)
{
	struct inode *vi = dentry->d_inode;
	struct super_block *sb = dir->i_sb;
	int err = 0;
       ntfs_inode *ni;
       ntfschar *unicode = NULL;
       int uname_len;

       ni = NTFS_I(vi);

       uname_len = ntfs_mbstoucs(&err, dentry->d_name.name, dentry->d_name.len, &unicode);
	if (uname_len < 0) {
		if (uname_len != -ENAMETOOLONG)
			ntfs_log_error(vol->sb, "Failed to convert name to "
					"Unicode.");
		return -ENOMEM;
	}

	lock_super(sb); 
        __iget(vi);

	if (ntfs_delete(&err, ni, NTFS_I(dir), unicode, uname_len))
       {
              iput(vi);
              ntfs_free(unicode);
		err = -err;
		goto out;
       }
       err = 0;
	clear_nlink(vi);
	vi->i_mtime = vi->i_atime = CURRENT_TIME_SEC;
out:
       ntfs_free(unicode);
	unlock_super(sb);

	return err;
}


static unsigned char bad_chars[] = "*?<>|\"\n";
//static unsigned char bad_if_strict[] = "+=,; ";

/***** Formats an Legal file name. Rejects invalid names. */
static int ntfs_format_name(const unsigned char *name, int len)
	/*
	 * name is the proposed name, len is its length, 
	 * allows dots at the beginning of name (for hidden files)
	 */
{
	int walk;
	unsigned char c;

	c = 0;
	for (walk = 0; walk < len; walk++) {
		c = *name++;
		if (strchr(bad_chars, c)){
			return -EINVAL;
		}
//		if (strchr(bad_if_strict, c))
//			return -EINVAL;
	}

	return 0;
}

static int vfs_ntfs3g_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
	struct super_block *sb = dir->i_sb;
	struct inode *vi;
	struct timespec ts;
	int err = 0;
       ntfs_inode *ni;
       ntfschar *unicode = NULL;
       int uname_len;
 	err = ntfs_format_name(dentry->d_name.name, dentry->d_name.len);
	if (err)
		return err;//goto out;
    
       uname_len = ntfs_mbstoucs(&err, dentry->d_name.name, dentry->d_name.len, &unicode);
	if (uname_len < 0) {
		if (uname_len != -ENAMETOOLONG)
			ntfs_log_error(vol->sb, "Failed to convert name to "
					"Unicode.");
		return -ENOMEM;
	}
    
	lock_super(sb);
    
     
       
	ts = CURRENT_TIME_SEC;
       ni = ntfs_create(&err, NTFS_I(dir), unicode, uname_len, S_IFDIR);
       ntfs_free(unicode);
       if(ni == NULL)
       {
              err = -err;
		goto out;
       }
       err = 0;
       vi = VFS_I(ni);
	vi->i_version++;
       vi->i_nlink = 2;
	vi->i_mtime = vi->i_atime = vi->i_ctime = ts;
	/* timestamp is already written, so mark_inode_dirty() is unneeded. */
	dentry->d_time = dentry->d_parent->d_inode->i_version;
	d_instantiate(dentry, vi);
out:
	unlock_super(sb);
	return err;
}

static int vfs_ntfs3g_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct inode *vi = dentry->d_inode;
	struct super_block *sb = dir->i_sb;
	int err = 0;
       ntfs_inode *ni;
       ntfschar *unicode = NULL;
       int uname_len;

       ni = NTFS_I(vi);
       
       uname_len = ntfs_mbstoucs(&err, dentry->d_name.name, dentry->d_name.len, &unicode);
	if (uname_len < 0) {
		if (uname_len != -ENAMETOOLONG)
			ntfs_log_error(vol->sb, "Failed to convert name to "
					"Unicode.");
		return -ENOMEM;
	}

	lock_super(sb);

       __iget(vi);

	if (ntfs_delete(&err, ni, NTFS_I(dir), unicode, uname_len))
       {
              ntfs_free(unicode);
              iput(vi);
		err = -err;
		goto out;
       }
       err = 0;
	clear_nlink(vi);
	vi->i_mtime = vi->i_atime = CURRENT_TIME_SEC;
out:
       ntfs_free(unicode);
	unlock_super(sb);

	return err;
}
     
static int vfs_ntfs3g_rename(struct inode *old_dir, struct dentry *old_dentry,
		       struct inode *new_dir, struct dentry *new_dentry)
{
	struct inode *old_inode, *new_inode;
	struct timespec ts;
	int err = 0;
       int is_dir;
	struct super_block *sb = old_dir->i_sb;
       ntfschar *unicode_new = NULL;
       ntfschar *unicode_old = NULL;
       int new_name_len;
       int old_name_len;

       new_name_len = ntfs_mbstoucs(&err, new_dentry->d_name.name, new_dentry->d_name.len, &unicode_new);
	if (new_name_len < 0) {
		if (new_name_len != -ENAMETOOLONG)
			ntfs_log_error(vol->sb, "Failed to convert name to "
					"Unicode.");
		err = ENOMEM;
              goto error_ret_unlock;
	}
       old_name_len = ntfs_mbstoucs(&err, old_dentry->d_name.name, old_dentry->d_name.len, &unicode_old);
	if (old_name_len < 0) {
		if (old_name_len != -ENAMETOOLONG)
			ntfs_log_error(vol->sb, "Failed to convert name to "
					"Unicode.");
		err = ENOMEM;
              goto error_ret_unlock;
	}
    
	old_inode = old_dentry->d_inode;
	new_inode = new_dentry->d_inode;
	lock_super(sb);

	is_dir = S_ISDIR(old_inode->i_mode);

	ts = CURRENT_TIME_SEC;
	if (new_inode) {
		if (is_dir) {
			if (ntfs_check_empty_dir(&err, NTFS_I(new_inode)))
				goto out;
		}
              __iget(new_inode);
		if(ntfs_delete(&err, NTFS_I(new_inode), NTFS_I(new_dir), unicode_new, new_name_len))
		{
		     iput(new_inode);
                   goto error_ret;
		 }
	}
       
       if(ntfs_link(&err, NTFS_I(old_inode), NTFS_I(new_dir), unicode_new, new_name_len))
           goto error_ret;
       
       __iget(old_inode);
       if(ntfs_delete(&err, NTFS_I(old_inode), NTFS_I(old_dir), unicode_old, old_name_len))
       {
             ntfs_delete(&err, NTFS_I(old_inode), NTFS_I(new_dir), unicode_new, new_name_len);
            {
                    iput(old_inode);
                    goto error_ret;
             }
       }
       err = 0;
	new_dir->i_version++;
out:
      if(unicode_new)
           ntfs_free(unicode_new);
      if(unicode_old)
           ntfs_free(unicode_old);
      unlock_super(sb);
	return err;
error_ret:
       unlock_super(sb);
error_ret_unlock:
       err = -EIO;
	goto out;
}
/**
 * Export operations allowing NFS exporting of mounted NTFS partitions.
 *
 * We use the default ->encode_fh() for now.  Note that they
 * use 32 bits to store the inode number which is an unsigned long so on 64-bit
 * architectures is usually 64 bits so it would all fail horribly on huge
 * volumes.  I guess we need to define our own encode and decode fh functions
 * that store 64-bit inode numbers at some point but for now we will ignore the
 * problem...
 *
 * We also use the default ->get_name() helper (used by ->decode_fh() via
 * fs/exportfs/expfs.c::find_exported_dentry()) as that is completely fs
 * independent.
 *
 * The default ->get_parent() just returns -EACCES so we have to provide our
 * own and the default ->get_dentry() is incompatible with NTFS due to not
 * allowing the inode number 0 which is used in NTFS for the system file $MFT
 * and due to using iget() whereas NTFS needs ntfs_iget().
 */
const struct export_operations ntfs3g_export_ops = {
	.get_parent	= ntfs3g_get_parent,	/* Find the parent of a given
						   directory. */
	.fh_to_dentry	= ntfs3g_fh_to_dentry,
	.fh_to_parent	= ntfs3g_fh_to_parent,
};
extern int ntfs3g_setattr(struct dentry *dentry, struct iattr *attr);
extern int ntfs3g_getattr(struct vfsmount *mnt, struct dentry *dentry, struct kstat *stat);
const struct inode_operations ntfs3g_dir_inode_ops = {
	.create		= vfs_ntfs3g_create,
	.lookup		= vfs_ntfs3g_lookup,
	.unlink		= vfs_ntfs3g_unlink,
	.mkdir		= vfs_ntfs3g_mkdir,
	.rmdir		= vfs_ntfs3g_rmdir,
	.rename		= vfs_ntfs3g_rename,
	.setattr	= ntfs3g_setattr,
	.getattr	= ntfs3g_getattr,
};
