/**
 * ntfs-4g - Fourth Generation NTFS Driver from NTFS-3G
 *
 * Copyright (c) 2005-2007 Yura Pakhuchiy
 * Copyright (c) 2005 Yuval Fledel
 * Copyright (c) 2006-2009 Szabolcs Szakacsits
 * Copyright (c) 2007-2009 Jean-Pierre Andre
 * Copyright (c) 2009 Erik Larsson
 *
 * This file is originated from the Linux-NTFS3G project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the NTFS-4G
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/stddef.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>	/* For bdev_hardsect_size(). */
#include <linux/backing-dev.h>
#include <linux/buffer_head.h>
#include <linux/vfs.h>
#include <linux/moduleparam.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/mount.h>


#include "config.h"
#include "debug.h"
#include "attrib.h"
#include "volume.h"
#include "mft.h"
#include "logfile.h"
#include "misc.h"
#include "compress.h"
#include "dir.h"
#include "aops.h"

#define NTFS3G_SKIP_MSRECOVER 1
#define NTFS3G_SKIP_CONSISTENCY_CHK 1

/* Driver wide mutex. */
DEFINE_MUTEX(ntfs3g_lock);

/**
 * ntfs_volume_alloc - Create an NTFS volume object and initialise it
 *
 * Description...
 *
 * Returns:
 */
ntfs_volume *ntfs_volume_alloc(void)
{
       ntfs_volume* vol;
	vol = kmalloc(sizeof(ntfs_volume), GFP_NOFS);
       if(vol)
           memset(vol, 0, sizeof(*vol));
       return vol;
}
void ntfs_volume_free(ntfs_volume *vol)
{ 
       if(vol)
	     kfree(vol);
}

void ntfs_attr_free(int *p_errno, ntfs_attr **na)
{
	if (na && *na) {
		ntfs_attr_close(p_errno, *na);
		*na = NULL;
	}
}

static s64 ntfs_get_nr_free_mft_records(int *p_errno, ntfs_volume *vol)
{
	ntfs_inode *bmp_ni = vol->mft_bmp_ni;
	s64 nr_free = ntfs_attr_get_free_bits(p_errno, VFS_I(bmp_ni));

	if (nr_free >= 0)
		nr_free += (bmp_ni->map_na->allocated_size - bmp_ni->map_na->data_size) << 3;
	return nr_free;
}

static int __ntfs_volume_release(ntfs_volume *v)
{
	int errcode = 0;
       int err = 0;

	if (v->vol_ni)
       {  
              ntfs_big_inode_close(&err, v->vol_ni);
              errcode |= err;
              err = 0;
              v->vol_ni = NULL;
	}
	/* 
	 * FIXME: Inodes must be synced before closing
	 * attributes, otherwise unmount could fail.
	 */
	if (v->lcnbmp_ni)
       {  
              ntfs_big_inode_close(&err, v->lcnbmp_ni);
              errcode |= err;
              err = 0;
              v->lcnbmp_ni = NULL;
	}

 	if (v->mft_bmp_ni)
       {  
              ntfs_attr_inode_close(&err, v->mft_bmp_ni);
              errcode |= err;
              err = 0;
              v->mft_bmp_ni = NULL;
	}  
       if(v->mft_dat_ni)
       {
           ntfs_inode_sync(&err, v->mft_dat_ni); 
           err = 0;
      }
       if (v->mftmirr_ni)
       {  
              NTFS3G_BUG_TRACE(!v->mft_dat_ni);
              ntfs_big_inode_close(&err, v->mftmirr_ni);
              errcode |= err;
              err = 0;
              v->mftmirr_ni = NULL;
	}	
       
  	if (v->mft_dat_ni)
       {  
              ntfs_big_inode_close(&err, v->mft_dat_ni);
              errcode |= err;
              err = 0;
              v->mft_dat_ni = NULL;
	}	 

	ntfs_free(v->vol_name);
	ntfs_free(v->upcase);
	ntfs_free(v->attrdef);
	ntfs_free(v);

	errcode = err;
	return errcode ? -1 : 0;
}

 #if !NTFS3G_SKIP_CONSISTENCY_CHK
/**
 * ntfs_volume_check_logfile - check logfile on target volume
 * @vol:	volume on which to check logfile
 *
 * Return 0 on success and -1 on error with errno set error code.
 */
static int ntfs_volume_check_logfile(int *p_errno, ntfs_volume *vol)
{
	ntfs_inode *ni;
	RESTART_PAGE_HEADER *rp = NULL;
	int err = 0;

	ni = ntfs_big_inode_open(p_errno, vol, FILE_LogFile,AT_DATA, AT_UNNAMED, 0);
	if (!ni) {
		ntfs_log_perror("Failed to open inode FILE_LogFile");
		*p_errno = EIO;
		return -1;
	}
	
	if (!ntfs_check_logfile(p_errno, ni, &rp) || !ntfs_is_logfile_clean(p_errno, ni, rp))
		err = EOPNOTSUPP;
	ntfs_free(rp);
    
	if (ntfs_big_inode_close(p_errno, ni))
		ntfs_error_set(&err, *p_errno);
	if (err) {
		*p_errno = err;
		return -1;
	}
	return 0;
}
#endif
/**
 * ntfs_put_super - called by the vfs to unmount a volume
 * @sb:		vfs superblock of volume to unmount
 *
 * ntfs_put_super() is called by the VFS (from fs/super.c::do_umount()) when
 * the volume is being unmounted (umount system call has been invoked) and it
 * releases all inodes and memory belonging to the NTFS specific part of the
 * super block.
 */
static void ntfs3g_put_super(struct super_block *sb)
{
	ntfs_volume *vol = NTFS3G_SB(sb);

	ntfs_log_trace("Entering.");
#if 0
	if (sb_has_dirty_inodes(sb)) {
		const char *s1, *s2;

		mutex_lock(&vol->mft_ino->i_mutex);
		truncate_inode_pages(vol->mft_ino->i_mapping, 0);
		mutex_unlock(&vol->mft_ino->i_mutex);
		write_inode_now(vol->mft_ino, 1);
		if (sb_has_dirty_inodes(sb)) {
			static const char *_s1 = "inodes";
			static const char *_s2 = "";
			s1 = _s1;
			s2 = _s2;
		} else {
			static const char *_s1 = "mft pages";
			static const char *_s2 = "They have been thrown "
					"away.  ";
			s1 = _s1;
			s2 = _s2;
		}
		ntfs3g_error(sb, "Dirty %s found at umount time.  %sYou should "
				"run chkdsk.  Please email "
				"linux-ntfs-dev@lists.sourceforge.net and say "
				"that you saw this message.  Thank you.", s1, s2);
	    }
#endif
       __ntfs_volume_release(vol);

}

/**
 * ntfs_statfs - return information about mounted NTFS volume
 * @dentry:	dentry from mounted volume
 * @sfs:	statfs structure in which to return the information
 *
 * Return information about the mounted NTFS volume @dentry in the statfs structure
 * pointed to by @sfs (this is initialized with zeros before ntfs_statfs is
 * called). We interpret the values to be correct of the moment in time at
 * which we are called. Most values are variable otherwise and this isn't just
 * the free values but the totals as well. For example we can increase the
 * total number of file nodes if we run out and we can keep doing this until
 * there is no more space on the volume left at all.
 *
 * Called from vfs_statfs which is used to handle the statfs, fstatfs, and
 * ustat system calls.
 *
 * Return 0 on success or -errno on error.
 */
static int ntfs3g_statfs(struct dentry *dentry, struct kstatfs *sfs)
{
	struct super_block *sb = dentry->d_sb;
	ntfs_volume *vol = NTFS3G_SB(sb);
	s64 size;
	int delta_bits;

	if (!vol)
		return -ENODEV;
    
	sfs->f_type   = NTFS3G_SB_MAGIC;
	/* 
	 * File system block size. Used to calculate used/free space by df.
	 * Incorrectly documented as "optimal transfer block size". 
	 */
	sfs->f_bsize = PAGE_CACHE_SIZE;
	
	/* Fundamental file system block size, used as the unit. */
	sfs->f_frsize = vol->cluster_size;
	
	/*
	 * Total number of blocks on file system in units of f_frsize.
	 * Since inodes are also stored in blocks ($MFT is a file) hence
	 * this is the number of clusters on the volume.
	 */
	sfs->f_blocks = vol->nr_clusters<<vol->cluster_size_bits>>PAGE_CACHE_SHIFT;
	
	/* Free blocks available for all and for non-privileged processes. */
	size = vol->free_clusters<<vol->cluster_size_bits>>PAGE_CACHE_SHIFT;
	if (size < 0)
		size = 0;
	sfs->f_bavail = sfs->f_bfree = size;
	
	/* Free inodes on the free space */
	delta_bits = vol->cluster_size_bits - vol->mft_record_size_bits;
	if (delta_bits >= 0)
		size <<= delta_bits;
	else
		size >>= -delta_bits;
	
	/* Number of inodes at this point in time. */
	sfs->f_files = (vol->mft_bmp_ni->map_na->allocated_size << 3) + size;
	
	/* Free inodes available for all and for non-privileged processes. */
	size += vol->free_mft_records;
	if (size < 0)
		size = 0;
	sfs->f_ffree = size;
	
	/* Maximum length of filenames. */
	sfs->f_namelen = NTFS_MAX_NAME_LEN;
	return 0;
}

/**
 * ntfs_clear_big_inode - clean up the ntfs specific part of an inode
 * @vi:		vfs inode pending annihilation
 *
 * When the VFS is going to remove an inode from memory, ntfs_clear_big_inode()
 * is called, which deallocates all memory belonging to the NTFS specific part
 * of the inode and returns.
 *
 * If the MFT record is dirty, we commit it before doing anything else.
 */
void ntfs3g_clear_big_inode(struct inode *vi)
{
	ntfs_inode *ni;
       int errcode = 0;
       s32 i;
	
	if (!vi)
		return;
      ni = NTFS_I(vi);
      
     if(NInoNTFSAttrNI(ni))
     {
         ntfs_attr_close(&errcode, ni->map_na);
         ni->map_na = NULL;
         return;
     }

      NTFS3G_BUG_TRACE(ni->base_ni_attr.nr_extents < 0);
      
	ntfs_log_trace("Entering for inode %lld\n", (long long)ni->mft_no);

	if (NInoDirty(ni) || NInoAttrListDirty(ni)) {
		if (ntfs_inode_sync(&errcode, ni)) {
                     ntfs_log_perror("failed to sync big inode %08x, but we must forgive"
                        "sync inode data onto disk since this is the last chance we could do sync inode data \n", ni);
                     NInoAttrListClearDirty(ni);
                     NInoClearDirty(ni);
                     
 		}
	}
    
      for( i=0; i<ni->base_ni_attr.nr_extents; i++)
        {
	     /* If we have dirty metadata, write it out. */
	     if (NInoDirty(ni->base_ni_attr.extent_nis[i]) || NInoAttrListDirty(ni->base_ni_attr.extent_nis[i]))
            {
                  NTFS3G_BUG_TRACE(NInoAttrList(ni->base_ni_attr.extent_nis[i]));
                  NTFS3G_BUG_TRACE(NInoAttrListDirty(ni->base_ni_attr.extent_nis[i]));
                  #if 0
		    if (ntfs_inode_sync(&errcode, ni->base_ni_attr.extent_nis[i])) 
                  {
                        ntfs_log_perror("failed to sync big inode %08x, but we must forgive"
                                  "sync inode data onto disk since this is the last chance we could do sync inode data \n", ni);
		    }
                   #endif
                  if (NInoAttrList(ni->base_ni_attr.extent_nis[i]) && ni->base_ni_attr.extent_nis[i]->base_ni_attr.attr_list)
                  {
                        NInoClearAttrList(ni->base_ni_attr.extent_nis[i]);
		          ntfs_free(ni->base_ni_attr.extent_nis[i]->base_ni_attr.attr_list);
                        ni->base_ni_attr.extent_nis[i]->base_ni_attr.attr_list = NULL;
                   }
                   ntfs_destroy_ext_inode(ni->base_ni_attr.extent_nis[i]);
		}
	}
        if(ni->base_ni_attr.extent_nis)
        {
                NTFS3G_BUG_TRACE(ni->base_ni_attr.nr_extents  == 0);
                ntfs_free(ni->base_ni_attr.extent_nis);
                ni->base_ni_attr.extent_nis = NULL;
                ni->base_ni_attr.nr_extents = 0;
                
        }
        else
        {
               NTFS3G_BUG_TRACE(ni->base_ni_attr.nr_extents !=0 );
        }
        if (NInoAttrList(ni) && ni->base_ni_attr.attr_list)
        {
                
		  ntfs_free(ni->base_ni_attr.attr_list);
                ni->base_ni_attr.attr_list = NULL;
                NInoClearAttrList(ni);;
        }
        NTFS3G_BUG_TRACE(ni->base_ni_attr.mft_page_refcnt);
        ntfs_attr_close(&errcode, ni->map_na);
        ni->map_na = NULL;
}


/**
 * ntfs_show_options - show mount options in /proc/mounts
 * @sf:		seq_file in which to write our mount options
 * @mnt:	vfs mount whose mount options to display
 *
 * Called by the VFS once for each mounted ntfs volume when someone reads
 * /proc/mounts in order to display the NTFS specific mount options of each
 * mount. The mount options of the vfs mount @mnt are written to the seq file
 * @sf and success is returned.
 */
int ntfs3g_show_options(struct seq_file *sf, struct vfsmount *mnt)
{
	ntfs_volume *vol = NTFS3G_SB(mnt->mnt_sb);

	//seq_printf(sf, ",uid=%i", vol->uid);
	//seq_printf(sf, ",gid=%i", vol->gid);
	if (vol->fmask == vol->dmask)
		seq_printf(sf, ",umask=0%o", vol->fmask);
	else {
		seq_printf(sf, ",fmask=0%o", vol->fmask);
		seq_printf(sf, ",dmask=0%o", vol->dmask);
	}
	if (NVolCaseSensitive(vol))
		seq_printf(sf, ",case_sensitive");

	seq_printf(sf, ",mft_zone_multiplier=%i", vol->mft_zone_multiplier);
	return 0;
}

/**
 * The complete super operations.
 */
static const struct super_operations ntfs3g_sops = {
	.alloc_inode	= ntfs3g_alloc_big_inode,	  /* VFS: Allocate new inode. */
	.destroy_inode	= ntfs3g_destroy_big_inode, /* VFS: Deallocate inode. */
	//.dirty_inode	= NULL,			/* VFS: Called from
	//					   __mark_inode_dirty(). */
	.write_inode	= ntfs3g_write_inode,	/* VFS: Write dirty inode to
						   disk. */
	//.drop_inode	= NULL,			/* VFS: Called just after the
	//					   inode reference count has
	//					   been decreased to zero.
	//					   NOTE: The inode lock is
	//					   held. See fs/inode.c::
	//					   generic_drop_inode(). */
	//.delete_inode	= NULL,			/* VFS: Delete inode from disk.
	//					   Called when i_count becomes
	//					   0 and i_nlink is also 0. */
	//.write_super	= NULL,			/* Flush dirty super block to
	//					   disk. */
	//.sync_fs	= NULL,			/* ? */
	//.write_super_lockfs	= NULL,		/* ? */
	//.unlockfs	= NULL,			/* ? */
	.put_super	= ntfs3g_put_super,	/* Syscall: umount. */
	.statfs		= ntfs3g_statfs,		/* Syscall: statfs */
	//.remount_fs	= ntfs3g_remount,		/* Syscall: mount -o remount. */
	.clear_inode	= ntfs3g_clear_big_inode,	/* VFS: Called when an inode is
						   removed from memory. */
	//.umount_begin	= NULL,			/* Forced umount. */
	.show_options	= ntfs3g_show_options,	/* Show mount options in
						   proc. */
};


extern const struct export_operations ntfs3g_export_ops;



int  ntfs3g_fill_super(struct super_block *sb, void *opt, const int silent)
{
	s64 l;
	ntfs_volume *vol;
 #if !NTFS3G_SKIP_CONSISTENCY_CHK
	u8 *m = NULL, *m2 = NULL;
 #endif
	ntfs_attr_search_ctx *ctx = NULL;
	ntfs_inode *ni;
	ATTR_RECORD *a;
	VOLUME_INFORMATION *vinf;
	ntfschar *vname;
	int j;
	u32 u;
       unsigned long flags = MS_IGNORE_HIBERFILE;
       int blocksize;
       int err = 0;
       ntfs_inode *root_ni = NULL;		/* The VFS inode of the root*/
 #if !NTFS3G_SKIP_CONSISTENCY_CHK
       int i;
 #endif
    
#ifdef FIXME_NTFS3G
	/* Important to get the mount options dealt with now. */
	if (!parse_options(&flags, (char*)opt))
		goto err_out_now;
#endif

	lockdep_off();
	ntfs_log_trace("Entering.");

	/* Allocate a new ntfs_volume and place it in sb->s_fs_info. */
	sb->s_fs_info = (void*)ntfs_volume_alloc();
	vol = NTFS3G_SB(sb);
	if (!vol) {
		if (!silent)
			ntfs_log_perror("Allocation of NTFS volume structure "
					"failed. Aborting mount...");
		lockdep_on();
		return -ENOMEM;
	}
      sb->s_op = &ntfs3g_sops;
	/* Initialize ntfs_volume structure. */
	*vol = (ntfs_volume) {
		.sb = sb,
		/*
		 * Default is group and other don't have any access to files or
		 * directories while owner has full access. Further, files by
		 * default are not executable but directories are of course
		 * browseable.
		 */
		.fmask = 0177,
		.dmask = 0077,
	};
	mutex_init(&vol->mftbmp_lock);
	mutex_init(&vol->lcnbmp_lock);

	unlock_kernel();
    
    	/* We support sector sizes up to the PAGE_CACHE_SIZE. */
	if (bdev_hardsect_size(sb->s_bdev) > PAGE_CACHE_SIZE) {
		if (!silent)
			ntfs_log_perror("Device has unsupported sector size "
					"(%i).  The maximum supported sector "
					"size on this architecture is %lu "
					"bytes.",
					bdev_hardsect_size(sb->s_bdev),
					PAGE_CACHE_SIZE);
		goto io_error_exit;
	}
	/*
	 * Setup the device access block size to NTFS_BLOCK_SIZE or the hard
	 * sector size, whichever is bigger.
	 */
	blocksize = sb_min_blocksize(sb, NTFS_BLOCK_SIZE);
	if (blocksize < NTFS_BLOCK_SIZE) {
		if (!silent)
			ntfs_log_perror("Unable to set device block size.");
		goto error_exit;
	}
	NTFS3G_BUG_TRACE(blocksize != sb->s_blocksize);
	ntfs_log_trace("Set device block size to %i bytes (block size bits %i).",
			blocksize, sb->s_blocksize_bits);
	/* Determine the size of the device in units of block_size bytes. */
	if (!i_size_read(sb->s_bdev->bd_inode)) {
		if (!silent)
			ntfs_log_perror("Unable to determine device size.");
		goto io_error_exit;
	}
	vol->nr_blocks = i_size_read(sb->s_bdev->bd_inode) >>
			sb->s_blocksize_bits;
    
	if (ntfs_volume_startup(&err, sb, flags))
	{
		goto error_exit;
	}
 #if !NTFS3G_SKIP_CONSISTENCY_CHK
	/* Load data from $MFT and $MFTMirr and compare the contents. */
	m  = ntfs_malloc_nofs(&err, vol->mftmirr_size << vol->mft_record_size_bits);
	m2 = ntfs_malloc_nofs(&err, vol->mftmirr_size << vol->mft_record_size_bits);
	if (!m || !m2)
		goto error_exit;

	l = ntfs_inode_map_read(&err, VFS_I(vol->mft_dat_ni),  0, vol->mftmirr_size << vol->mft_record_size_bits,m);
	if (l != vol->mftmirr_size << vol->mft_record_size_bits) {
		if (l == -1)
			ntfs_log_perror("Failed to read $MFT");
		else {
			ntfs_log_perror("Failed to read $MFT, unexpected length "
				       "(%lld != %d).\n", (long long)l,
				       vol->mftmirr_size);
			err = EIO;
		}
		goto error_exit;
	}
	l = ntfs_inode_map_read(&err, VFS_I(vol->mftmirr_ni),  0, vol->mftmirr_size << vol->mft_record_size_bits,m2);
	if (l != vol->mftmirr_size << vol->mft_record_size_bits) {
		if (l == -1) {
			ntfs_log_perror("Failed to read $MFTMirr");
			goto error_exit;
		}
		vol->mftmirr_size = l>> vol->mft_record_size_bits;
	}
	ntfs_log_debug("Comparing $MFTMirr to $MFT...\n");
	for (i = 0; i < vol->mftmirr_size; ++i) {
		MFT_RECORD *mrec, *mrec2;
		const char *ESTR[12] = { "$MFT", "$MFTMirr", "$LogFile",
			"$Volume", "$AttrDef", "root directory", "$Bitmap",
			"$Boot", "$BadClus", "$Secure", "$UpCase", "$Extend" };
		const char *s;

		if (i < 12)
			s = ESTR[i];
		else if (i < 16)
			s = "system file";
		else
			s = "mft record";

		mrec = (MFT_RECORD*)(m + i * vol->mft_record_size);
		if (mrec->flags & MFT_RECORD_IN_USE) {
			if (ntfs_is_baad_recordp(mrec)) {
				ntfs_log_perror("$MFT error: Incomplete multi "
					       "sector transfer detected in "
					       "'%s'.\n", s);
				goto io_error_exit;
			}
			if (!ntfs_is_mft_recordp(mrec)) {
				ntfs_log_perror("$MFT error: Invalid mft "
						"record for '%s'.\n", s);
				goto io_error_exit;
			}
		}
		mrec2 = (MFT_RECORD*)(m2 + i * vol->mft_record_size);
		if (mrec2->flags & MFT_RECORD_IN_USE) {
			if (ntfs_is_baad_recordp(mrec2)) {
				ntfs_log_perror("$MFTMirr error: Incomplete "
						"multi sector transfer "
						"detected in '%s'.\n", s);
				goto io_error_exit;
			}
			if (!ntfs_is_mft_recordp(mrec2)) {
				ntfs_log_perror("$MFTMirr error: Invalid mft "
						"record for '%s'.\n", s);
				goto io_error_exit;
			}
		}
		if (memcmp(mrec, mrec2, ntfs_mft_record_get_data_size(mrec))) {
			ntfs_log_perror("$MFTMirr does not match $MFT (record "
				       "%d).\n", i);
			goto io_error_exit;
		}
	}

	ntfs_free(m2);
	ntfs_free(m);
	m = m2 = NULL;
#endif
	/* Now load the bitmap from $Bitmap. */
	ntfs_log_debug("Loading $Bitmap...\n");
	vol->lcnbmp_ni = ntfs_big_inode_open(&err, vol, FILE_Bitmap, AT_DATA, AT_UNNAMED, 0);
	if (!vol->lcnbmp_ni) {
		ntfs_log_perror("Failed to open inode FILE_Bitmap");
		goto error_exit;
	}


	/* Now load the upcase table from $UpCase. */
	ntfs_log_debug("Loading $UpCase...\n");
	ni = ntfs_big_inode_open(&err, vol, FILE_UpCase, AT_DATA, AT_UNNAMED, 0);
	if (!ni) {
		ntfs_log_perror("Failed to open inode FILE_UpCase");
		goto error_exit;
	}

	if (ni->map_na->data_size & ~0x1ffffffffULL) {
		ntfs_log_perror("Error: Upcase table is too big (max 32-bit "
				"allowed).\n");
		err = EINVAL;
		goto error_exit;
	}
	if (vol->upcase_len != ni->map_na->data_size >> 1) {
		vol->upcase_len = ni->map_na->data_size >> 1;
		/* Throw away default table. */
		ntfs_free(vol->upcase);
		vol->upcase = ntfs_malloc_nofs(&err, ni->map_na->data_size);
		if (!vol->upcase)
			goto error_exit;
	}
	/* Read in the $DATA attribute value into the buffer. */
       l = ntfs_inode_map_read(&err, VFS_I(ni),  0, ni->map_na->data_size, (u8*)vol->upcase);
	if (l != ni->map_na->data_size) {
		ntfs_log_perror("Failed to read $UpCase, unexpected length "
			       "(%lld != %lld).\n", (long long)l,
			       (long long)na->data_size);
		err = EIO;
		goto error_exit;
	}
	/* Done with the $UpCase mft record. */
	if ((err=ntfs_big_inode_close(&err, ni))) {
		ntfs_log_perror("Failed to close $UpCase");
		goto error_exit;
	}

	/*
	 * Now load $Volume and set the version information and flags in the
	 * vol structure accordingly.
	 */
	ntfs_log_debug("Loading $Volume...\n");
	vol->vol_ni = ntfs_big_inode_open(&err, vol, FILE_Volume, AT_DATA,  AT_UNNAMED, 0);
	if (!vol->vol_ni) {
		ntfs_log_perror("Failed to open inode FILE_Volume");
		goto error_exit;
	}
	/* Get a search context for the $Volume/$VOLUME_INFORMATION lookup. */
	ctx = ntfs_attr_get_search_ctx(&err, vol->vol_ni, NULL);
	if (!ctx)
		goto error_exit;

	/* Find the $VOLUME_INFORMATION attribute. */
	if (ntfs_attr_lookup(&err, AT_VOLUME_INFORMATION, AT_UNNAMED, 0, 0, 0, NULL,
			0, ctx)) {
		ntfs_log_perror("$VOLUME_INFORMATION attribute not found in "
				"$Volume");
		goto error_exit;
	}
	a = ctx->attr;
	/* Has to be resident. */
	if (a->non_resident) {
		ntfs_log_perror("Attribute $VOLUME_INFORMATION must be "
			       "resident but it isn't.\n");
		err = EIO;
		goto error_exit;
	}
	/* Get a pointer to the value of the attribute. */
	vinf = (VOLUME_INFORMATION*)(le16_to_cpu(a->value_offset) + (char*)a);
	/* Sanity checks. */
	if ((char*)vinf + le32_to_cpu(a->value_length) > (char*)ctx->mrec +
			le32_to_cpu(ctx->mrec->bytes_in_use) ||
			le16_to_cpu(a->value_offset) + le32_to_cpu(
			a->value_length) > le32_to_cpu(a->length)) {
		ntfs_log_perror("$VOLUME_INFORMATION in $Volume is corrupt.\n");
		err = EIO;
		goto error_exit;
	}
	/* Setup vol from the volume information attribute value. */
	vol->major_ver = vinf->major_ver;
	vol->minor_ver = vinf->minor_ver;
	/* Do not use le16_to_cpu() macro here as our VOLUME_FLAGS are
	   defined using cpu_to_le16() macro and hence are consistent. */
	vol->flags = vinf->flags;
	/*
	 * Reinitialize the search context for the $Volume/$VOLUME_NAME lookup.
	 */
	ntfs_attr_reinit_search_ctx(&err, ctx);
	if (ntfs_attr_lookup(&err, AT_VOLUME_NAME, AT_UNNAMED, 0, 0, 0, NULL, 0,
			ctx)) {
		if (err != ENOENT) {
			ntfs_log_perror("Failed to lookup of $VOLUME_NAME in "
					"$Volume failed");
			goto error_exit;
		}
		/*
		 * Attribute not present.  This has been seen in the field.
		 * Treat this the same way as if the attribute was present but
		 * had zero length.
		 */
		vol->vol_name = ntfs_malloc_nofs(&err, 1);
		if (!vol->vol_name)
			goto error_exit;
		vol->vol_name[0] = '\0';
	} else {
		a = ctx->attr;
		/* Has to be resident. */
		if (a->non_resident) {
			ntfs_log_perror("$VOLUME_NAME must be resident.\n");
			err = EIO;
			goto error_exit;
		}
		/* Get a pointer to the value of the attribute. */
		vname = (ntfschar*)(le16_to_cpu(a->value_offset) + (char*)a);
		u = le32_to_cpu(a->value_length) / 2;
		/*
		 * Convert Unicode volume name to current locale multibyte
		 * format.
		 */
		vol->vol_name = NULL;
		if (ntfs_ucstombs(&err, vname, u, &vol->vol_name, 0) == -1) {
			ntfs_log_perror("Volume name could not be converted "
					"to current locale");
			ntfs_log_debug("Forcing name into ASCII by replacing "
				"non-ASCII characters with underscores.\n");
			vol->vol_name = ntfs_malloc_nofs(&err, u + 1);
			if (!vol->vol_name)
				goto error_exit;
			
			for (j = 0; j < (s32)u; j++) {
				ntfschar uc = le16_to_cpu(vname[j]);
				if (uc > 0xff)
					uc = (ntfschar)'_';
				vol->vol_name[j] = (char)uc;
			}
			vol->vol_name[u] = '\0';
		}
	}
	ntfs_attr_put_search_ctx(&err, ctx);
	ctx = NULL;
	/* Now load the attribute definitions from $AttrDef. */
	ntfs_log_debug("Loading $AttrDef...\n");
	ni = ntfs_big_inode_open(&err, vol, FILE_AttrDef, AT_DATA,  AT_UNNAMED, 0);
	if (!ni) {
		ntfs_log_perror("Failed to open $AttrDef");
		goto error_exit;
	}

	/* Check we don't overflow 32-bits. */
	if (ni->map_na->data_size > 0xffffffffLL) {
		ntfs_log_perror("Attribute definition table is too big (max "
			       "32-bit allowed).\n");
		err = EINVAL;
		goto error_exit;
	}
	vol->attrdef_len = ni->map_na->data_size;
	vol->attrdef = ntfs_malloc_nofs(&err, ni->map_na->data_size);
	if (!vol->attrdef)
		goto error_exit;
	/* Read in the $DATA attribute value into the buffer. */
      l = ntfs_inode_map_read(&err, VFS_I(ni),  0, ni->map_na->data_size, (u8*)vol->attrdef);
	if (l != ni->map_na->data_size) {
		ntfs_log_perror("Failed to read $AttrDef, unexpected length "
			       "(%lld != %lld).\n", (long long)l,
			       (long long)na->data_size);
		err = EIO;
		goto error_exit;
	}
	/* Done with the $AttrDef mft record. */
	if ((err=ntfs_big_inode_close(&err, ni))) {
		ntfs_log_perror("Failed to close $AttrDef");
		goto error_exit;
	}

	root_ni = ntfs_big_inode_open(&err, vol, FILE_root, AT_INDEX_ROOT, NTFS_INDEX_I30, 4);
	if (!root_ni) {
		ntfs_log_perror("Failed to open FILE_root");
		goto error_exit;
	}
 #if !NTFS3G_SKIP_CONSISTENCY_CHK
	/*
	 * Check for dirty logfile and hibernated Windows.
	 * We care only about read-write mounts.
	 */
	if (!(flags & MS_RDONLY)) {
		if (!(flags & MS_IGNORE_HIBERFILE) && 
		    ntfs_volume_check_hiberfile(&err, vol, 1, root_ni))
			goto error_exit;
              err = 0;
		if (ntfs_volume_check_logfile(&err, vol) ) {
            #if NTFS3G_SKIP_MSRECOVER
                   printk("The file system wasn't safely closed on Windows. Skip fixing in ntfs3g...\n");
            #else
                    	if (!(flags & MS_RECOVER))
				goto error_exit;
			ntfs_log_perror("The file system wasn't safely "
				      "closed on Windows. Fixing.\n");
			if ((err = ntfs_logfile_reset(&err, vol)))
				goto error_exit;    
            #endif
		}
        
	}
#endif
	vol->free_clusters = ntfs_attr_get_free_bits(&err, VFS_I(vol->lcnbmp_ni));
	if (vol->free_clusters < 0) {
		ntfs_log_perror("Failed to read NTFS $Bitmap");
		goto error_exit;
	}

	vol->free_mft_records = ntfs_get_nr_free_mft_records(&err, vol);
	if (vol->free_mft_records < 0) {
		ntfs_log_perror("Failed to calculate free MFT records");
		goto error_exit;
	}

    
#ifdef FIXME_NTFS3G 
	if (ctx->hiberfile && (err=ntfs_volume_check_hiberfile(&err, vol, 0, root_ni))) {
		if (err != EPERM)
			goto error_exit;
		if (ntfs_fuse_rm("/hiberfil.sys"))
			goto error_exit;
	}
#endif

	if (!(sb->s_root = d_alloc_root(VFS_I(root_ni)))) {
		/* We increment i_count simulating an ntfs_iget(). */
		ntfs_log_perror("Failed to open create root dentry");
		goto error_exit;
	}

       sb->s_magic = NTFS3G_SB_MAGIC;
	/*
	 * Ntfs allows 63 bits for the file size, i.e. correct would be:
	 *	sb->s_maxbytes = ~0ULL >> 1;
	 * But the kernel uses a long as the page cache page index which on
	 * 32-bit architectures is only 32-bits. MAX_LFS_FILESIZE is kernel
	 * defined to the maximum the page cache page index can cope with
	 * without overflowing the index or to 2^63 - 1, whichever is smaller.
	 */
	sb->s_maxbytes = MAX_LFS_FILESIZE;
	/* Ntfs measures time in 100ns intervals. */
	sb->s_time_gran = 100;
       sb->s_export_op = &ntfs3g_export_ops;
	lock_kernel();
	lockdep_on();
	return 0;
io_error_exit:
	err = EIO;
error_exit:	
      err = 0;
       if(root_ni)
           ntfs_big_inode_close(&err, root_ni);
       lock_kernel();	
       sb->s_fs_info = NULL;
       ntfs_log_perror("Failed, returning -EINVAL.");
	lockdep_on();
    
	if (ctx)
		ntfs_attr_put_search_ctx(&err, ctx);

#if !NTFS3G_SKIP_CONSISTENCY_CHK
       if(m)
	    ntfs_free(m);
       if(m2)
	    ntfs_free(m2);
#endif
	__ntfs_volume_release(vol);
	return -EINVAL;
}

/*
 * Slab caches to optimize allocations and deallocations of attribute search
 * contexts and index contexts, respectively.
 */
struct kmem_cache *ntfs3g_big_inode_cache;
struct kmem_cache *ntfs3g_ext_inode_cache;
struct kmem_cache *ntfs3g_search_ctx_cache;
struct kmem_cache *ntfs3g_bio_data_cache;

static int ntfs3g_get_sb(struct file_system_type *fs_type,
	int flags, const char *dev_name, void *data, struct vfsmount *mnt)
{
	return get_sb_bdev(fs_type, flags, dev_name, data, ntfs3g_fill_super,
			   mnt);
}

static struct file_system_type ntfs3g_fs_type = {
	.owner		= THIS_MODULE,
	.name		= "ntfs3g",
	.get_sb		= ntfs3g_get_sb,
	.kill_sb	= kill_block_super,
	.fs_flags	= FS_REQUIRES_DEV,
};

/* Stable names for the slab caches. */
//static const char ntfs3g_index_ctx_cache_name[] = "ntfs3g_index_ctx_cache";
//static const char ntfs3g_attr_ctx_cache_name[] = "ntfs3g_attr_ctx_cache";
static const char ntfs3g_bio_data_cache_name[] = "ntfs3g_io_data_cache";
static const char ntfs3g_search_ctx_cache_name[] = "ntfs3g_search_ctx_cache";
static const char ntfs3g_big_inode_cache_name[] = "ntfs3g_big_inode_cache";
static const char ntfs3g_ext_inode_cache_name[] = "ntfs3g_ext_inode_cache";
//static const char ntfs3g_big_inode_cache_name[] = "ntfs3g_big_inode_cache";
/* Init once constructor for the inode slab cache. */
static void ntfs3g_big_inode_init_once(void *foo)
{
	ntfs_inode *ni = (ntfs_inode *)foo;

	inode_init_once(VFS_I(ni));
}

static int __init init_ntfs3g_fs(void)
{
	int err = 0;

	/* This may be ugly but it results in pretty output so who cares. (-8 */
	ntfs_log_trace(KERN_INFO "NTFS-3G driver " NTFS3G_VERSION " [Flags: R/W"
#ifdef DEBUG
			" DEBUG"
#endif
#ifdef MODULE
			" MODULE"
#endif
			"].\n");

	ntfs_log_trace("Debug messages are enabled.");

    

	ntfs3g_search_ctx_cache = kmem_cache_create(ntfs3g_search_ctx_cache_name,
			                                  sizeof(ntfs_attr_search_ctx), 0, SLAB_HWCACHE_ALIGN, NULL);
	if (!ntfs3g_search_ctx_cache) {
		ntfs_log_perror(KERN_CRIT "NTFS-4G: Failed to create %s!\n",
				ntfs3g_search_ctx_cache_name);
		goto search_ctx_err_out;
	}

	ntfs3g_big_inode_cache = kmem_cache_create(ntfs3g_big_inode_cache_name,
			sizeof(big_ntfs3g_inode), 0,
			SLAB_RECLAIM_ACCOUNT|SLAB_MEM_SPREAD, ntfs3g_big_inode_init_once);
	if (!ntfs3g_big_inode_cache) {
		ntfs_log_perror(KERN_CRIT "NTFS-4G: Failed to create %s!\n",
				ntfs3g_big_inode_cache_name);
		goto inode_big_err_out;
	}
	ntfs3g_ext_inode_cache = kmem_cache_create(ntfs3g_ext_inode_cache_name,
			sizeof(ntfs_inode), 0,
			SLAB_RECLAIM_ACCOUNT|SLAB_MEM_SPREAD, NULL);
	if (!ntfs3g_ext_inode_cache) {
		ntfs_log_perror(KERN_CRIT "NTFS-4G: Failed to create %s!\n",
				ntfs3g_ext_inode_cache_name);
		goto inode_ext_err_out;
	}
    
	ntfs3g_bio_data_cache = kmem_cache_create(ntfs3g_bio_data_cache_name,
			                                  sizeof(struct submit_rw_io_data), 0, SLAB_HWCACHE_ALIGN, NULL);
	if (!ntfs3g_bio_data_cache) {
		ntfs_log_perror(KERN_CRIT "NTFS-4G: Failed to create %s!\n",
				ntfs3g_bio_data_cache_name);
		goto submit_io_data_err_out;
	}

    
       if(alloc_compression_buffers())
            goto decom_buf_err_out;
	err = register_filesystem(&ntfs3g_fs_type);
	if (!err) {
		ntfs_log_trace("NTFS-4G driver registered successfully.");
		return 0; /* Success! */
	}
	ntfs_log_perror(KERN_CRIT "NTFS-4G: Failed to register NTFS filesystem driver!\n");

decom_buf_err_out:    
      kmem_cache_destroy(ntfs3g_bio_data_cache);
submit_io_data_err_out:
       kmem_cache_destroy(ntfs3g_ext_inode_cache);
inode_ext_err_out:
       kmem_cache_destroy(ntfs3g_big_inode_cache);
inode_big_err_out:
	kmem_cache_destroy(ntfs3g_search_ctx_cache);
search_ctx_err_out:
	if (!err) {
		ntfs_log_perror(KERN_CRIT "NTFS-3G: Aborting NTFS-3G filesystem driver "
				"registration...\n");
		err = -ENOMEM;
	}
	return err;
}

static void __exit exit_ntfs3g_fs(void)
{
	ntfs_log_trace("Unregistering NTFS-4G driver.");

	unregister_filesystem(&ntfs3g_fs_type);
	kmem_cache_destroy(ntfs3g_big_inode_cache);
	kmem_cache_destroy(ntfs3g_ext_inode_cache);
	kmem_cache_destroy(ntfs3g_search_ctx_cache);
       free_compression_buffers();
	/* Unregister the ntfs sysctls. */
	//ntfs_sysctl(0);
}
s64 ntfs_pread(struct super_block *sb, const s64 pos, s64 count, void *b)
{
      u32 offset = (u32)(pos&(block_size(sb->s_bdev)-1));
      u32 block_index = (u32)(pos>>blksize_bits(block_size(sb->s_bdev)));
      s64 rd_cnt = 0;
      u8 *dst = (u8*)b;

       while(rd_cnt < count)
        {
           struct buffer_head *bh = sb_bread(sb, block_index);
	     if (bh == NULL) {
		    return rd_cnt;
	     }
            block_index++;
            if(count-rd_cnt>=bh->b_size-offset)
            {
               memcpy(dst, bh->b_data+offset, bh->b_size-offset);
               rd_cnt += bh->b_size-offset; 
               dst += bh->b_size-offset;
               offset = 0;
            }
            else
            {
               memcpy(dst, bh->b_data+offset, count-rd_cnt);
               rd_cnt = count;
            }
            brelse(bh);
        }
        return rd_cnt;
        
}
s64 ntfs_pwrite(struct super_block *sb, const s64 pos, s64 count,
		const void *b)
{
      u32 offset = (u32)(pos&(block_size(sb->s_bdev)-1));
      u32 block_index = (u32)(pos>>blksize_bits(block_size(sb->s_bdev)));
      s64 wr_cnt = 0;
      u8 *src = (u8*)b;
       while(wr_cnt < count)
        {
           struct buffer_head *bh;
           bh = sb_getblk(sb, block_index);
	    if (bh == NULL) {
		    return wr_cnt;
	    }
           block_index++;

           if(offset || count-wr_cnt<bh->b_size-offset)
           {
                //write doesn't cover the whole block, need read-back
                lock_buffer(bh);
	         if (buffer_uptodate(bh)) {
		         unlock_buffer(bh);
	         } else {
		         get_bh(bh);
		         bh->b_end_io = end_buffer_read_sync;
		         submit_bh(READ, bh);
		         wait_on_buffer(bh);
		         if (!buffer_uptodate(bh))
		         {   
		               brelse(bh);
                             BUG();
			        return wr_cnt;
		         }
		  }
	    }
              
            if(count-wr_cnt>=bh->b_size-offset)
            {
                    memcpy(bh->b_data+offset, src, bh->b_size-offset);
                    wr_cnt += bh->b_size-offset; 
                    src += bh->b_size-offset;
                    offset = 0;
            }
            else
            {
                    memcpy(bh->b_data+offset, src, count-wr_cnt);
                    wr_cnt = count;
            }
	      lock_buffer(bh);
	      bh->b_end_io = end_buffer_write_sync;
	      get_bh(bh);
             submit_bh(WRITE, bh);
             wait_on_buffer(bh);
             brelse(bh);
        }
        return wr_cnt;
}

MODULE_AUTHOR("tommy sun <mingyong.sun@gmail.com>");
MODULE_DESCRIPTION("NTFS-3G kernel driver - Ported from NTFS-3G usermode implement to avoid FUSE serious performance issue");
MODULE_VERSION(NTFS3G_VERSION);
MODULE_LICENSE("GPL");

module_init(init_ntfs3g_fs)
module_exit(exit_ntfs3g_fs)
