/**
 * volume.c - NTFS volume handling code. Originated from the Linux-NTFS project.
 *
 * Copyright (c) 2000-2006 Anton Altaparmakov
 * Copyright (c) 2002-2009 Szabolcs Szakacsits
 * Copyright (c) 2004-2005 Richard Russon
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the NTFS-3G
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/stddef.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>	/* For bdev_hardsect_size(). */
#include <linux/backing-dev.h>
#include <linux/buffer_head.h>
#include <linux/vfs.h>
#include <linux/moduleparam.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>

#include "config.h"
#include "volume.h"
#include "attrib.h"
#include "mft.h"
#include "bootsect.h"
#include "debug.h"
#include "inode.h"
#include "runlist.h"
#include "dir.h"
#include "inode.h"
#include "logfile.h"
#include "mst.h"
#include "misc.h"

const char *ntfs_home = 
"Ntfs-3g news, support and information:  http://ntfs-3g.org\n";

static const char *invalid_ntfs_msg =
"The device '%s' doesn't seem to have a valid NTFS.\n"
"Maybe the wrong device is used? Or the whole disk instead of a\n"
"partition (e.g. /dev/sda, not /dev/sda1)? Or the other way around?\n";

static const char *corrupt_volume_msg =
"NTFS is either inconsistent, or there is a hardware fault, or it's a\n"
"SoftRAID/FakeRAID hardware. In the first case run chkdsk /f on Windows\n"
"then reboot into Windows twice. The usage of the /f parameter is very\n"
"important! If the device is a SoftRAID/FakeRAID then first activate\n"
"it and mount a different device under the /dev/mapper/ directory, (e.g.\n"
"/dev/mapper/nvidia_eahaabcc1). Please see the 'dmraid' documentation\n"
"for more details.\n";

static const char *hibernated_volume_msg =
"The NTFS partition is hibernated. Please resume and shutdown Windows\n"
"properly, or mount the volume read-only with the 'ro' mount option, or\n"
"mount the volume read-write with the 'remove_hiberfile' mount option.\n"
"For example type on the command line:\n"
"\n"
"            mount -t ntfs-3g -o remove_hiberfile %s %s\n"
"\n";

static const char *unclean_journal_msg =
"Write access is denied because the disk wasn't safely powered\n"
"off and the 'norecover' mount option was specified.\n";

static const char *opened_volume_msg =
"Mount is denied because the NTFS volume is already exclusively opened.\n"
"The volume may be already mounted, or another software may use it which\n"
"could be identified for example by the help of the 'fuser' command.\n";

static const char *fakeraid_msg =
"Either the device is missing or it's powered down, or you have\n"
"SoftRAID hardware and must use an activated, different device under\n" 
"/dev/mapper/, (e.g. /dev/mapper/nvidia_eahaabcc1) to mount NTFS.\n"
"Please see the 'dmraid' documentation for help.\n";

static const char *access_denied_msg =
"Please check '%s' and the ntfs-3g binary permissions,\n"
"and the mounting user ID. More explanation is provided at\n"
"http://ntfs-3g.org/support.html#unprivileged\n";


static void ntfs_attr_setup_flag(int *p_errno, ntfs_inode *ni)
{
	STANDARD_INFORMATION *si;
       NTFS3G_BUG_TRACE(NInoNTFSAttrNI(ni));
	si = ntfs_attr_readall(p_errno, ni, AT_STANDARD_INFORMATION, AT_UNNAMED, 0, NULL);
	if (si) {
		ni->base_ni_attr.flags = si->file_attributes;
		ntfs_free(si);
	}
}

/**
 * ntfs_mft_load - load the $MFT and setup the ntfs volume with it
 * @vol:	ntfs volume whose $MFT to load
 *
 * Load $MFT from @vol and setup @vol with it. After calling this function the
 * volume @vol is ready for use by all read access functions provided by the
 * ntfs library.
 *
 * Return 0 on success and -1 on error with errno set to the error code.
 */
extern const struct address_space_operations ntfs3g_mst_aops;
static int ntfs_mft_load(int *p_errno, ntfs_volume *vol)
{
	VCN next_vcn, last_vcn, highest_vcn;
	s64 l;
	ntfs_attr_search_ctx *ctx = NULL;
	ATTR_RECORD *a;
	int eo;
       u32 remain_size;
       sector_t mst_block;
       u8 *pTmp;
       struct inode *tmp_ino;
       MFT_RECORD *mft_mrec;
       
       struct super_block *sb = (struct super_block *)vol->sb;

	tmp_ino = new_inode(sb);
	if (!tmp_ino) {
			printk("Failed to load essential metadata.");
		*p_errno =  ENOMEM;
                return -1;
	}
	tmp_ino->i_ino = FILE_MFT;
       tmp_ino->i_mode = S_IFREG;
	insert_inode_hash(tmp_ino);

       mft_mrec = NULL;
    
	/* Manually setup an ntfs_inode. */
	vol->mft_dat_ni = NTFS_I(tmp_ino);

	mft_mrec = (MFT_RECORD*)ntfs_malloc_nofs(p_errno, vol->mft_record_size);
	if (!vol->mft_dat_ni || !mft_mrec) {
		printk("Error allocating memory for $MFT");
		goto error_exit;
	}
	vol->mft_dat_ni->mft_no = FILE_MFT;
       memset(vol->mft_dat_ni, 0, sizeof(ntfs_inode));
       vol->mft_dat_ni->vol = vol;
	/*now read mst sectors*/
       remain_size  = vol->mft_record_size;
       mst_block = (vol->mft_lcn << vol->cluster_size_bits)>>blksize_bits(block_size(sb->s_bdev));
       pTmp = (u8 *)mft_mrec;
       while(remain_size)
        {
           struct buffer_head *bh = sb_bread(sb, mst_block);
	     if (bh == NULL) {
		    *p_errno  = EIO;
		    goto error_exit;
  
	     }
            mst_block++;
            if(remain_size>=bh->b_size)
            {
               memcpy(pTmp, bh->b_data, bh->b_size);
               remain_size -= bh->b_size;
               pTmp += bh->b_size;
            }
            else
            {
               memcpy(pTmp, bh->b_data, remain_size);
               pTmp += remain_size;
               remain_size =0;
            }
            brelse(bh);
            
        }
        NInoClearNTFSAttrNI(vol->mft_dat_ni);
	/* Apply the mst fixups. */
	if (ntfs_mst_post_read_fixup(p_errno, (NTFS_RECORD*)mft_mrec, vol->mft_record_size)) {
		/* FIXME: Try to use the $MFTMirr now. */
		printk("MST fixup failed. $MFT is corrupt.");
		goto error_exit;
	}
	if (ntfs_mft_record_check(p_errno, vol, 0, mft_mrec))
	{
		goto error_exit;
	}
	
	ctx = ntfs_attr_get_search_ctx(p_errno, vol->mft_dat_ni, mft_mrec);
	if (!ctx)
	{
		goto error_exit;
	}

	/* Find the $ATTRIBUTE_LIST attribute in $MFT if present. */
	if (ntfs_attr_lookup(p_errno, AT_ATTRIBUTE_LIST, AT_UNNAMED, 0, 0, 0, NULL, 0,
			ctx)) {
		if (*p_errno != ENOENT) {
			printk("$MFT has corrupt attribute list.\n");
			goto io_error_exit;
		}
		goto mft_has_no_attr_list;
	}
	NInoSetAttrList(vol->mft_dat_ni);
	l = ntfs_get_attribute_value_length(p_errno, ctx->attr);
	if (l <= 0 || l > 0x40000) {
		printk("$MFT/$ATTR_LIST invalid length (%lld).\n",
			       (long long)l);
              * p_errno = EIO;
		goto io_error_exit;
	}
    
     
       
	vol->mft_dat_ni->base_ni_attr.attr_list_size = l;
	vol->mft_dat_ni->base_ni_attr.attr_list = ntfs_malloc_nofs(p_errno, l);
	if (!vol->mft_dat_ni->base_ni_attr.attr_list)
	{
	       *p_errno = ENOMEM;
		goto error_exit;
	 }
	
	l = ntfs_get_attribute_value(p_errno, vol, ctx->attr, vol->mft_dat_ni->base_ni_attr.attr_list);
	if (!l) {
		printk("Failed to get value of $MFT/$ATTR_LIST.\n");
		goto io_error_exit;
	}
	if (l != vol->mft_dat_ni->base_ni_attr.attr_list_size) {
		printk("Partial read of $MFT/$ATTR_LIST (%lld != "
			       "%u).\n", (long long)l,
			       vol->mft_dat_ni->base_ni_attr.attr_list_size);
		goto io_error_exit;
	}

mft_has_no_attr_list:

	/* Read all extents from the $DATA attribute in $MFT. */
	ntfs_attr_reinit_search_ctx(p_errno, ctx);
	highest_vcn = next_vcn = 0;
	a = NULL;
       vol->mft_dat_ni->map_na = ntfs_malloc_nofs(p_errno, sizeof(ntfs_attr));
      // vol->mft_dat_ni->map_na = ntfs_direct_attr_open(p_errno, vol->mft_dat_ni, mft_mrec, AT_DATA, AT_UNNAMED, 0);

       if(!vol->mft_dat_ni->map_na)
           goto error_exit;
       //memset(vol->mft_dat_ni->map_na, 0, sizeof(ntfs_attr));
       vol->mft_dat_ni->map_na->rl=NULL;
       vol->mft_dat_ni->map_na->ni = vol->mft_dat_ni;
       vol->mft_dat_ni->map_na->type = AT_DATA;
       vol->mft_dat_ni->map_na->name = AT_UNNAMED;
       vol->mft_dat_ni->map_na->name_len = 0;
       mutex_init(&vol->mft_dat_ni->base_ni_attr.mrec_lock);
	while (!ntfs_attr_lookup(p_errno, AT_DATA, AT_UNNAMED, 0, 0, next_vcn, NULL, 0,
			ctx)) {
		runlist_element *nrl;

		a = ctx->attr;
		/* $MFT must be non-resident. */
		if (!a->non_resident) {
			printk("$MFT must be non-resident.\n");
			goto io_error_exit;
		}
		/* $MFT must be uncompressed and unencrypted. */
		if (a->flags & ATTR_COMPRESSION_MASK ||
				a->flags & ATTR_IS_ENCRYPTED) {
			printk("$MFT must be uncompressed and "
				       "unencrypted.\n");
			goto io_error_exit;
		}

		nrl = ntfs_mapping_pairs_decompress(p_errno, vol, a, vol->mft_dat_ni->map_na->rl);
		if (!nrl) {
			printk("ntfs_mapping_pairs_decompress() failed");
			goto error_exit;
		}
              //printk("mft data attribute run list mapped, from now on mft map mechanism can work.\n");
        	ntfs_attr_init(p_errno, vol->mft_dat_ni->map_na, TRUE, a->flags & ATTR_IS_COMPRESSED,
				a->flags & ATTR_IS_ENCRYPTED,
				a->flags & ATTR_IS_SPARSE,
				sle64_to_cpu(a->allocated_size),
				sle64_to_cpu(a->data_size),
				sle64_to_cpu(a->initialized_size),
				0, 0);
		vol->mft_dat_ni->map_na->rl = nrl;

		/* Get the lowest vcn for the next extent. */
		highest_vcn = sle64_to_cpu(a->highest_vcn);
		next_vcn = highest_vcn + 1;

		/* Only one extent or error, which we catch below. */
		if (next_vcn <= 0)
			break;

		/* Avoid endless loops due to corruption. */
		if (next_vcn < sle64_to_cpu(a->lowest_vcn)) {
			printk("$MFT has corrupt attribute list.\n");
			goto io_error_exit;
		}
	}
	if (!a) {
		printk("$MFT/$DATA attribute not found.\n");
              *p_errno = EIO;
		goto io_error_exit;
	}
    	last_vcn = vol->mft_dat_ni->map_na->allocated_size >> vol->cluster_size_bits;
	if (highest_vcn && highest_vcn != last_vcn - 1) {
		printk("Failed to load runlist for $MFT/$DATA.\n");
		printk("highest_vcn = 0x%llx, last_vcn - 1 = 0x%llx\n",
			       (long long)highest_vcn, (long long)last_vcn - 1);
              *p_errno = EIO;
		goto io_error_exit;
	}
	/* Done with the $Mft mft record. */
	ntfs_attr_put_search_ctx(p_errno, ctx);
       vol->mft_dat_ni->fixup_block_size = vol->mft_record_size;
       NInoSetAttrMstProtected(vol->mft_dat_ni);
       VFS_I(vol->mft_dat_ni)->i_mapping->a_ops = &ntfs3g_mst_aops;
       VFS_I(vol->mft_dat_ni)->i_size = vol->mft_dat_ni->map_na->initialized_size;

       ntfs_attr_setup_flag(p_errno, vol->mft_dat_ni);
       
	ctx = NULL;

	/* We now have a fully setup ntfs inode for $MFT in vol->mft_ni. */
    
	/*
	 * The volume is now setup so we can use all read access functions.
	 */
	vol->mft_bmp_ni = ntfs_attr_inode_open(p_errno, vol->mft_dat_ni,  AT_BITMAP, AT_UNNAMED, 0);
	if (!vol->mft_bmp_ni) {
		printk("Failed to open $MFT/$BITMAP");
              *p_errno = EIO;
		goto error_exit;
	}
       vol->mft_dat_ni->fixup_block_size = vol->mft_record_size;
       NInoSetAttrMstProtected(vol->mft_dat_ni);
       if(mft_mrec)
            ntfs_free(mft_mrec);
	return 0;
io_error_exit:
	*p_errno = EIO;
error_exit:
       eo = *p_errno;
	if (ctx)
		ntfs_attr_put_search_ctx(p_errno, ctx);
      iput(tmp_ino);
	if (vol->mft_dat_ni && VFS_I(vol->mft_dat_ni)!=tmp_ino) {
		iput(VFS_I(vol->mft_dat_ni));
	}
    	vol->mft_dat_ni = NULL;
       *p_errno = eo;
       if(mft_mrec)
            ntfs_free(mft_mrec);
	return -1;
}

/**
 * ntfs_mftmirr_load - load the $MFTMirr and setup the ntfs volume with it
 * @vol:	ntfs volume whose $MFTMirr to load
 *
 * Load $MFTMirr from @vol and setup @vol with it. After calling this function
 * the volume @vol is ready for use by all write access functions provided by
 * the ntfs library (assuming ntfs_mft_load() has been called successfully
 * beforehand).
 *
 * Return 0 on success and -1 on error with errno set to the error code.
 */
static int ntfs_mftmirr_load(int *p_errno, ntfs_volume *vol)
{

	vol->mftmirr_ni = ntfs_big_inode_open(p_errno, vol, FILE_MFTMirr, AT_DATA, AT_UNNAMED, 0);
	if (!vol->mftmirr_ni) {
		printk("Failed to open inode $MFTMirr");
		return -1;
	}	
	return 0;
}

/**
 * ntfs_volume_startup - allocate and setup an ntfs volume
 * @dev:	device to open
 * @flags:	optional mount flags
 *
 * Load, verify, and parse bootsector; load and setup $MFT and $MFTMirr. After
 * calling this function, the volume is setup sufficiently to call all read
 * and write access functions provided by the library.
 *
 * Return the allocated volume structure on success and NULL on error with
 * errno set to the error code.
 */
int ntfs_volume_startup(int *p_errno, struct super_block *sb, unsigned long flags)
{
	LCN mft_zone_size, mft_lcn;
	ntfs_volume *vol;
	NTFS_BOOT_SECTOR *bs = NULL;
       int err = 0;
       struct buffer_head *bh;

	if (!sb || !sb->s_bdev) {
		return EINVAL;
	}
	
	/* get the volume structure. */
       vol = NTFS3G_SB(sb);
	if (!vol)
	 {    
	       *p_errno = EINVAL;
		goto error_exit;
	 }
	
	vol->upcase_len = 65536;
	vol->upcase = ntfs_malloc_nofs(p_errno, vol->upcase_len * sizeof(ntfschar));
	if (!vol->upcase)
	{
	       err = ENOMEM;
		goto error_exit;
	}
	
	ntfs_upcase_table_build(vol->upcase,
			vol->upcase_len * sizeof(ntfschar));
	
	if (flags & MS_RDONLY)
		NVolSetReadOnly(vol);
	
	/* Now read the bootsector. */
       bh = sb_bread(sb, 0);
	if (bh == NULL) {
		err = EIO;
		goto error_exit;
	}
       bs = (NTFS_BOOT_SECTOR *)bh->b_data;
	if (!ntfs_boot_sector_is_ntfs(bs)) {
		err = EINVAL;
              brelse(bh);
		goto error_exit;
	}
	if (ntfs_boot_sector_parse(p_errno, vol, bs) < 0)
	{
	       err = EINVAL;
	       brelse(bh);
		goto error_exit;
	}
	
	brelse(bh);

#ifdef FIXME_NTFS3G
	/* Now set the device block size to the sector size. */
	if (ntfs_device_block_size_set(vol->dev, vol->sector_size))
		ntfs_log_debug("Failed to set the device block size to the "
				"sector size.  This may affect performance "
				"but should be harmless otherwise.  Error: "
				"%d\n", err);
#endif
	
	/* We now initialize the cluster allocator. */
	mft_zone_size = vol->nr_clusters >> 3;      /* 12.5% */

	/* Setup the mft zone. */
	vol->mft_zone_start = vol->mft_zone_pos = vol->mft_lcn;
	ntfs_log_debug("mft_zone_pos = 0x%llx\n", (long long)vol->mft_zone_pos);

	/*
	 * Calculate the mft_lcn for an unmodified NTFS volume (see mkntfs
	 * source) and if the actual mft_lcn is in the expected place or even
	 * further to the front of the volume, extend the mft_zone to cover the
	 * beginning of the volume as well. This is in order to protect the
	 * area reserved for the mft bitmap as well within the mft_zone itself.
	 * On non-standard volumes we don't protect it as the overhead would be
	 * higher than the speed increase we would get by doing it.
	 */
	mft_lcn = (8192 + 2 * vol->cluster_size - 1) / vol->cluster_size;
	if (mft_lcn * vol->cluster_size < 16 * 1024)
		mft_lcn = (16 * 1024 + vol->cluster_size - 1) /
				vol->cluster_size;
	if (vol->mft_zone_start <= mft_lcn)
		vol->mft_zone_start = 0;
	ntfs_log_debug("mft_zone_start = 0x%llx\n", (long long)vol->mft_zone_start);

	/*
	 * Need to cap the mft zone on non-standard volumes so that it does
	 * not point outside the boundaries of the volume. We do this by
	 * halving the zone size until we are inside the volume.
	 */
	vol->mft_zone_end = vol->mft_lcn + mft_zone_size;
	while (vol->mft_zone_end >= vol->nr_clusters) {
		mft_zone_size >>= 1;
		vol->mft_zone_end = vol->mft_lcn + mft_zone_size;
	}
	ntfs_log_debug("mft_zone_end = 0x%llx\n", (long long)vol->mft_zone_end);

	/*
	 * Set the current position within each data zone to the start of the
	 * respective zone.
	 */
	vol->data1_zone_pos = vol->mft_zone_end;
	ntfs_log_debug("data1_zone_pos = 0x%llx\n", vol->data1_zone_pos);
	vol->data2_zone_pos = 0;
	ntfs_log_debug("data2_zone_pos = 0x%llx\n", vol->data2_zone_pos);

	/* Set the mft data allocation position to mft record 24. */
	vol->mft_data_pos = 24;

	/*
	 * The cluster allocator is now fully operational.
	 */

	/* Need to setup $MFT so we can use the library read functions. */
	if (ntfs_mft_load(p_errno, vol)) {
		printk("Failed to load $MFT");
              err = EINVAL;
		goto error_exit;
	}

	/* Need to setup $MFTMirr so we can use the write functions, too. */
	if (ntfs_mftmirr_load(p_errno, vol) ) {
		printk("Failed to load $MFTMirr");
              err = EINVAL;
		goto error_exit;
	}
	return 0;
error_exit:
	return err;
}


/**
 * ntfs_hiberfile_open - Find and open '/hiberfil.sys'
 * @vol:    An ntfs volume obtained from ntfs_mount
 *
 * Return:  inode  Success, hiberfil.sys is valid
 *	    NULL   hiberfil.sys doesn't exist or some other error occurred
 */
/**
 * ntfs_hiberfile_open - Find and open '/hiberfil.sys'
 * @vol:    An ntfs volume obtained from ntfs_mount
 *
 * Return:  inode  Success, hiberfil.sys is valid
 *	    NULL   hiberfil.sys doesn't exist or some other error occurred
 */
static ntfs_inode *ntfs_hiberfile_open(int *p_errno, ntfs_volume *vol, ntfs_inode *ni_root)
{
	u64 inode;
	ntfs_inode *ni_hibr = NULL;
	ntfschar   *unicode = NULL;
	int unicode_len;
	const char *hiberfile = "hiberfil.sys";

	if (!vol) {
		*p_errno = EINVAL;
		return NULL;
	}

	unicode_len = ntfs_mbstoucs(p_errno, hiberfile, strlen(hiberfile), &unicode);
	if (unicode_len < 0) {
		ntfs_log_perror("Couldn't convert 'hiberfil.sys' to Unicode");
		goto out;
	}

	inode = ntfs_inode_lookup_by_name(p_errno, ni_root, unicode, unicode_len);
	if (inode == (u64)-1) {
		ntfs_log_debug("Couldn't find file '%s'.\n", hiberfile);
		goto out;
	}

	inode = MREF(inode);
	ni_hibr = ntfs_big_inode_open(p_errno, vol, inode, AT_DATA, AT_UNNAMED, 0);
	if (!ni_hibr) {
		ntfs_log_debug("Couldn't open inode %lld.\n", (long long)inode);
		goto out;
	}
out:
	ntfs_free(unicode);
	return ni_hibr;
}


#define NTFS_HIBERFILE_HEADER_SIZE	4096

/**
 * ntfs_volume_check_hiberfile - check hiberfil.sys whether Windows is
 *                               hibernated on the target volume
 * @vol:    volume on which to check hiberfil.sys
 *
 * Return:  0 if Windows isn't hibernated for sure
 *         -1 otherwise and errno is set to the appropriate value
 */
int ntfs_volume_check_hiberfile(int *p_errno, ntfs_volume *vol, int verbose, ntfs_inode *ni_root)
{
	ntfs_inode *ni;
	int bytes_read, err;
	char *buf = NULL;

	ni = ntfs_hiberfile_open(p_errno, vol, ni_root);
	if (!ni) {
		if (*p_errno == ENOENT)
			return 0;
		return -1;
	}

	buf = ntfs_malloc_nofs(p_errno, NTFS_HIBERFILE_HEADER_SIZE);
	if (!buf)
		goto out;
    
       bytes_read = ntfs_inode_map_read(p_errno, VFS_I(ni), 0 , NTFS_HIBERFILE_HEADER_SIZE,  (u8*)buf);
	//bytes_read = ntfs_attr_pread(p_errno, na, 0, NTFS_HIBERFILE_HEADER_SIZE, buf);
	if (bytes_read < 0) {
		ntfs_log_perror("Failed to read hiberfil.sys");
		goto out;
	}
	if (bytes_read < NTFS_HIBERFILE_HEADER_SIZE) {
		if (verbose)
			ntfs_log_error(ni->sb, "Hibernated non-system partition, "
				       "refused to mount.\n");
		*p_errno = EPERM;
		goto out;
	}
	if (memcmp(buf, "hibr", 4) == 0) {
		if (verbose)
			ntfs_log_error(ni->sb, "Windows is hibernated, refused to mount.\n");
		*p_errno = EPERM;
		goto out;
	}
        /* All right, all header bytes are zero */
	*p_errno = 0;
out:
	ntfs_free(buf);
	err = *p_errno;
	if (ntfs_big_inode_close(p_errno, ni))
		ntfs_error_set(&err, *p_errno);
	*p_errno = err;
	return *p_errno ? -1 : 0;
}



/**
 * ntfs_version_is_supported - check if NTFS version is supported.
 * @vol:	ntfs volume whose version we're interested in.
 *
 * The function checks if the NTFS volume version is known or not.
 * Version 1.1 and 1.2 are used by Windows NT3.x and NT4.
 * Version 2.x is used by Windows 2000 Betas.
 * Version 3.0 is used by Windows 2000.
 * Version 3.1 is used by Windows XP, Windows Server 2003 and Longhorn.
 *
 * Return 0 if NTFS version is supported otherwise -1 with errno set.
 *
 * The following error codes are defined:
 *	EOPNOTSUPP - Unknown NTFS version
 *	EINVAL	   - Invalid argument
 */
int ntfs_version_is_supported(ntfs_volume *vol)
{
	u8 major, minor;

	if (!vol) {
		return EINVAL;
	}

	major = vol->major_ver;
	minor = vol->minor_ver;

	if (NTFS_V1_1(major, minor) || NTFS_V1_2(major, minor))
		return 0;

	if (NTFS_V2_X(major, minor))
		return 0;

	if (NTFS_V3_0(major, minor) || NTFS_V3_1(major, minor))
		return 0;

	return EOPNOTSUPP;
}

/**
 * ntfs_logfile_reset - "empty" $LogFile data attribute value
 * @vol:	ntfs volume whose $LogFile we intend to reset.
 *
 * Fill the value of the $LogFile data attribute, i.e. the contents of
 * the file, with 0xff's, thus marking the journal as empty.
 *
 * FIXME(?): We might need to zero the LSN field of every single mft
 * record as well. (But, first try without doing that and see what
 * happens, since chkdsk might pickup the pieces and do it for us...)
 *
 * On success return 0.
 *
 * On error return -1 with errno set to the error code.
 */
int ntfs_logfile_reset(int *p_errno, ntfs_volume *vol)
{
	ntfs_inode *ni;
	int eo;

	if (!vol) {
		*p_errno = EINVAL;
		return -1;
	}

	ni = ntfs_big_inode_open(p_errno, vol, FILE_LogFile, AT_DATA, AT_UNNAMED, 0);
	if (!ni) {
		ntfs_log_perror(vol->sb, "Failed to open inode FILE_LogFile");
		return -1;
	}

	if (ntfs_empty_logfile(p_errno, ni)) {
		eo = *p_errno;
		goto error_exit;
	}
	
	return ntfs_big_inode_close(p_errno, ni);

error_exit:
	ntfs_big_inode_close(p_errno, ni);
	*p_errno = eo;
	return -1;
}
/**
 * ntfs_volume_write_flags - set the flags of an ntfs volume
 * @vol:	ntfs volume where we set the volume flags
 * @flags:	new flags
 *
 * Set the on-disk volume flags in the mft record of $Volume and
 * on volume @vol to @flags.
 *
 * Return 0 if successful and -1 if not with errno set to the error code.
 */
int ntfs_volume_write_flags(int *p_errno, ntfs_volume *vol, const u16 flags)
{
	ATTR_RECORD *a;
	VOLUME_INFORMATION *c;
	ntfs_attr_search_ctx *ctx;
	int ret = -1;	/* failure */

	if (!vol || !vol->vol_ni) {
		*p_errno = EINVAL;
		return -1;
	}
	/* Get a pointer to the volume information attribute. */
	ctx = ntfs_attr_get_search_ctx(p_errno, vol->vol_ni, NULL);
	if (!ctx)
		return -1;

	if (ntfs_attr_lookup(p_errno, AT_VOLUME_INFORMATION, AT_UNNAMED, 0, 0, 0, NULL,
			0, ctx)) {
		ntfs_log_error(vol->sb, "Attribute $VOLUME_INFORMATION was not found "
			       "in $Volume!\n");
		goto err_out;
	}
	a = ctx->attr;
	/* Sanity check. */
	if (a->non_resident) {
		ntfs_log_error(vol->sb, "Attribute $VOLUME_INFORMATION must be resident "
			       "but it isn't.\n");
		*p_errno = EIO;
		goto err_out;
	}
	/* Get a pointer to the value of the attribute. */
	c = (VOLUME_INFORMATION*)(le16_to_cpu(a->value_offset) + (char*)a);
	/* Sanity checks. */
	if ((char*)c + le32_to_cpu(a->value_length) > (char*)ctx->mrec +
			le32_to_cpu(ctx->mrec->bytes_in_use) ||
			le16_to_cpu(a->value_offset) +
			le32_to_cpu(a->value_length) > le32_to_cpu(a->length)) {
		ntfs_log_error(vol->sb, "Attribute $VOLUME_INFORMATION in $Volume is "
			       "corrupt!\n");
		*p_errno = EIO;
		goto err_out;
	}
	/* Set the volume flags. */
	vol->flags = c->flags = flags & VOLUME_FLAGS_MASK;
	/* Write them to disk. */
	ntfs_inode_mark_dirty(p_errno, vol->vol_ni);
       mark_mft_record_dirty(vol->vol_ni);
	if (ntfs_inode_sync(p_errno, vol->vol_ni))
		goto err_out;

	ret = 0; /* success */
err_out:
	ntfs_attr_put_search_ctx(p_errno, ctx);
	return ret;
}

int ntfs_volume_error(int err)
{
	int ret;

	switch (err) {
		case 0:
			ret = NTFS_VOLUME_OK;
			break;
		case EINVAL:
			ret = NTFS_VOLUME_NOT_NTFS;
			break;
		case EIO:
			ret = NTFS_VOLUME_CORRUPT;
			break;
		case EPERM:
			ret = NTFS_VOLUME_HIBERNATED;
			break;
		case EOPNOTSUPP:
			ret = NTFS_VOLUME_UNCLEAN_UNMOUNT;
			break;
		case EBUSY:
			ret = NTFS_VOLUME_LOCKED;
			break;
		case ENXIO:
			ret = NTFS_VOLUME_RAID;
			break;
		case EACCES:
			ret = NTFS_VOLUME_NO_PRIVILEGE;
			break;
		default:
			ret = NTFS_VOLUME_UNKNOWN_REASON;
			break;
	}
	return ret;
}


void ntfs_mount_error(const char *volume, const char *mntpoint, int err)
{
	switch (err) {
		case NTFS_VOLUME_NOT_NTFS:
			printk(invalid_ntfs_msg, volume);
			break;
		case NTFS_VOLUME_CORRUPT:
			printk("%s", corrupt_volume_msg);
			break;
		case NTFS_VOLUME_HIBERNATED:
			printk(hibernated_volume_msg, volume, mntpoint);
			break;
		case NTFS_VOLUME_UNCLEAN_UNMOUNT:
			printk("%s", unclean_journal_msg);
			break;
		case NTFS_VOLUME_LOCKED:
			printk("%s", opened_volume_msg);
			break;
		case NTFS_VOLUME_RAID:
			printk("%s", fakeraid_msg);
			break;
		case NTFS_VOLUME_NO_PRIVILEGE:
			printk(access_denied_msg, volume);
			break;
	}
}

